// **********************************************************************
//
// Copyright (c) 2000
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/OCI_IIOP.h>
#include <OB/IIOPConfig.h>

#include <OB/BiDirIIOPInit.h>
#include <OB/BiDirUtil.h>
#include <OB/OCI_BiDir_IIOP.h>

#include <OCI_BiDir_impl.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;

void
OBBiDir::IIOP::install(CORBA::ORB_ptr orb, const char* peer)
{
    Object_var obj;

    //
    // Get the connector factory registry
    //
    OCI::ConFactoryRegistry_var conFactoryRegistry;
    obj = orb -> resolve_initial_references("OCIConFactoryRegistry");
    conFactoryRegistry = OCI::ConFactoryRegistry::_narrow(obj);

    //
    // Get the IIOP connector factory for use as our "delegate"
    //
    OCI::ConFactory_var conFactoryDelegate;
    try
    {
        conFactoryDelegate =
            conFactoryRegistry -> get_factory(OCI::IIOP::TAG_IIOP);
    }
    catch(const OCI::NoSuchFactory&)
    {
        throw INITIALIZE();
    }

    //
    // Get the acceptor factory registry
    //
    OCI::AccFactoryRegistry_var accFactoryRegistry;
    obj = orb -> resolve_initial_references("OCIAccFactoryRegistry");
    accFactoryRegistry = OCI::AccFactoryRegistry::_narrow(obj);

    //
    // Get the IIOP acceptor factory for use as our "delegate"
    //
    OCI::AccFactory_var accFactoryDelegate;
    try
    {
        accFactoryDelegate =
            accFactoryRegistry -> get_factory(OCI::IIOP::TAG_IIOP);
    }
    catch(const OCI::NoSuchFactory&)
    {
        throw INITIALIZE();
    }

    //
    // Install factories
    //
    OBBiDir::installFactories(orb, OCI::BiDir::IIOP::TAG_BIDIR_IIOP,
                              BiDirIOP::TAG_BIDIR_IIOP, peer,
                              conFactoryDelegate, accFactoryDelegate);
}

//
// Create a new POAManager with support for bi-directional IIOP
//
PortableServer::POAManager_ptr
OBBiDir::IIOP::createPOAManager(const char* name, CORBA::ORB_ptr orb,
                                int& argc, char** argv,
                                OB::Properties_ptr properties)
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);

    //
    // Get the ORB's properties if none were provided
    //
    OB::Properties_var propDummy;
    if(is_nil(properties))
    {
        propDummy = oborb -> properties();
        properties = propDummy.in();
    }

    //
    // Translate command-line options into properties
    //
    OBBiDir::parseArgs(orb, argc, argv, properties);

    //
    // Validate the properties
    //
    OBBiDir::validateProperties(orb, properties);

    //
    // Get the properties
    //
    const char* peer = properties -> getProperty("ooc.bidir.peer");
    const char* mode = properties -> getProperty("ooc.bidir.mode");
    if(mode == 0)
        mode = "both";

    //
    // Install the connector and acceptor factories
    //
    install(orb, peer);

    //
    // Get the acceptor factory registry
    //
    OCI::AccFactoryRegistry_var accFactoryRegistry;
    Object_var obj =
        orb -> resolve_initial_references("OCIAccFactoryRegistry");
    accFactoryRegistry = OCI::AccFactoryRegistry::_narrow(obj);

    //
    // Get the bi-directional IIOP acceptor factory
    //
    OCI::AccFactory_var accFactory;
    try
    {
        accFactory = accFactoryRegistry -> get_factory(
            OCI::BiDir::IIOP::TAG_BIDIR_IIOP);
    }
    catch(const OCI::NoSuchFactory&)
    {
        assert(false);
    }

    OB::Logger_var logger = oborb -> logger();
    OBPortableServer::POAManagerFactory::AcceptorConfigSeq config;
    if(strcmp(mode, "client") == 0)
    {
        //
        // mode == "client", so only create a callback acceptor
        //
        config.length(1);
        config[0].id = OCI::BiDir::IIOP::TAG_BIDIR_IIOP;
        config[0].params.length(1);
        config[0].params[0].name = string_dup("callback");
        config[0].params[0].value <<= Any::from_boolean(true);
    }
    else if(strcmp(mode, "server") == 0)
    {
        //
        // mode == "server", so don't create a callback acceptor
        //

        //
        // Get the IIOP acceptor configuration
        //
        OB::IIOPConfig iiopConfig(properties, logger);
        iiopConfig.parseAll(name);
        OCI::ParamSeq_var params = iiopConfig.getParams();

        config.length(1);
        config[0].id = OCI::BiDir::IIOP::TAG_BIDIR_IIOP;
        config[0].params.length(1);
        config[0].params[0].name = string_dup("params");
        config[0].params[0].value <<= params.in();
    }
    else if(strcmp(mode, "both") == 0)
    {
        //
        // mode == "both", so create both kinds of acceptor
        //
        config.length(2);
        config[0].id = OCI::BiDir::IIOP::TAG_BIDIR_IIOP;
        config[0].params.length(1);
        config[0].params[0].name = string_dup("callback");
        config[0].params[0].value <<= Any::from_boolean(true);

        //
        // Get the IIOP acceptor configuration
        //
        OB::IIOPConfig iiopConfig(properties, logger);
        iiopConfig.parseAll(name);
        OCI::ParamSeq_var params = iiopConfig.getParams();

        config[1].id = OCI::BiDir::IIOP::TAG_BIDIR_IIOP;
        config[1].params.length(1);
        config[1].params[0].name = string_dup("params");
        config[1].params[0].value <<= params.in();
    }
    else
    {
        assert(false);
        return PortableServer::POAManager::_nil(); // to satisfy the compiler
    }

    //
    // Resolve POAManagerFactory
    //
    OBPortableServer::POAManagerFactory_var factory;
    obj = orb -> resolve_initial_references("POAManagerFactory");
    factory = OBPortableServer::POAManagerFactory::_narrow(obj);

    //
    // Create the POAManager with the name provided by the caller
    //
    PortableServer::POAManager_var manager;
    try
    {
        manager = factory -> create_poa_manager_with_config(name, config);
    }
    catch(const OBPortableServer::POAManagerFactory::POAManagerAlreadyExists&)
    {
        String_var err = string_dup("POAManager `");
        err += name;
        err += "': already exists";
        logger -> error(err);
        throw INITIALIZE();
    }
    catch(const OCI::NoSuchFactory&)
    {
        assert(false);
    }
    catch(const OCI::InvalidParam&)
    {
        assert(false);
    }

    return manager._retn();
}
