// **********************************************************************
//
// Copyright (c) 2000
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/BiDirUtil.h>

#include <stdlib.h>

#include <OCI_BiDir_impl.h>

using namespace CORBA;
using namespace OB;

void
OBBiDir::installFactories(CORBA::ORB_ptr orb, OCI::ProtocolId id,
                          OCI::ProfileId tag, const char* peer,
                          OCI::ConFactory_ptr conFactoryDelegate,
                          OCI::AccFactory_ptr accFactoryDelegate)
{
    Object_var obj;

    //
    // Get the connector factory registry
    //
    OCI::ConFactoryRegistry_var conFactoryRegistry;
    obj = orb -> resolve_initial_references("OCIConFactoryRegistry");
    conFactoryRegistry = OCI::ConFactoryRegistry::_narrow(obj);

    //
    // First check if the bi-directional connector factory has
    // already been installed
    //
    OCI::BiDir::ConFactory_impl* conFactoryImpl = 0;
    try
    {
        OCI::ConFactory_var f = conFactoryRegistry -> get_factory(id);
        conFactoryImpl = dynamic_cast<OCI::BiDir::ConFactory_impl*>(f.in());
        assert(conFactoryImpl != 0);
    }
    catch(const OCI::NoSuchFactory&)
    {
        // ignore
    }

    //
    // Install connector factory, if necessary
    //
    if(conFactoryImpl == 0)
    {
        //
        // Install the bi-directional connector factory
        //
        conFactoryImpl = new OCI::BiDir::ConFactory_impl(id, tag, peer,
                                                         conFactoryDelegate);
        OCI::ConFactory_var conFactory = conFactoryImpl;
        try
        {
            conFactoryRegistry -> add_factory(conFactoryImpl);
        }
        catch(const OCI::FactoryAlreadyExists&)
        {
            // ignore
        }
    }

    //
    // Get the acceptor factory registry
    //
    OCI::AccFactoryRegistry_var accFactoryRegistry;
    obj = orb -> resolve_initial_references("OCIAccFactoryRegistry");
    accFactoryRegistry = OCI::AccFactoryRegistry::_narrow(obj);

    //
    // First check if the bi-directional acceptor factory has
    // already been installed
    //
    OCI::BiDir::AccFactory_impl* accFactoryImpl = 0;
    try
    {
        OCI::AccFactory_var f = accFactoryRegistry -> get_factory(id);
        accFactoryImpl = dynamic_cast<OCI::BiDir::AccFactory_impl*>(f.in());
        assert(accFactoryImpl != 0);
    }
    catch(const OCI::NoSuchFactory&)
    {
        // ignore
    }

    //
    // Install acceptor factory, if necessary
    //
    if(accFactoryImpl == 0)
    {
        //
        // Create the bi-directional acceptor factory
        //
        OCI::BiDir::TransportManager_var transportManager =
            conFactoryImpl -> transportManager();
        accFactoryImpl = new OCI::BiDir::AccFactory_impl(
            transportManager, accFactoryDelegate);
        OCI::AccFactory_var accFactory = accFactoryImpl;
        try
        {
            accFactoryRegistry -> add_factory(accFactoryImpl);
        }
        catch(const OCI::FactoryAlreadyExists&)
        {
            // ignore
        }
    }
}

void
OBBiDir::parseArgs(CORBA::ORB_ptr orb, int& ac, char** av,
                   Properties_ptr properties)
{
    assert(!is_nil(properties));

    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    Logger_var logger = oborb -> logger();

    //
    // Process each argument. Turn each argument into an appropriate
    // property.
    //
    int i = 0;
    while(i < ac)
    {
        const char* option = av[i];
        if(strcmp(option, "-BIDIRpeer") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("BiDir: argument expected for "
                                "-BIDIRpeer");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.bidir.peer", av[i + 1]);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];

            ac -= 2;
        }
        else
            i++;
    }
}

void
OBBiDir::validateProperties(CORBA::ORB_ptr orb, Properties_ptr properties)
{
    //
    // Process each property
    //
    String_var peer;
    String_var mode = string_dup("both");

    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    Logger_var logger = oborb -> logger();

    Properties::KeySeq keys = properties -> getKeys("ooc.bidir.");
    for(ULong j = 0 ; j < keys.length() ; j++)
    {
        const char* key = keys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.bidir.peer") == 0)
        {
            if(strlen(value) == 0)
            {
                logger -> error("BiDir: invalid value for "
                                "ooc.bidir.peer");
                throw INITIALIZE();
            }
            peer = value;
        }
        else if(strcmp(key, "ooc.bidir.mode") == 0)
        {
            if(strcmp(value, "client") != 0 && strcmp(value, "server") != 0 &&
               strcmp(value, "both") != 0)
            {
                String_var err = string_dup("BiDir: invalid value for "
                                "ooc.bidir.mode: ");
                err += value;
                logger -> error(err);
                throw INITIALIZE();
            }
            mode = value;
        }
        else
        {
            String_var err = string_dup("BiDir: unknown property `");
            err += key;
            err += "'";
            logger -> warning(err);
        }
    }

    //
    // Validate peer ID
    //
    if(peer.in() == 0)
    {
        logger -> error("BiDir: no peer value specified");
        throw INITIALIZE();
    }
}
