// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/CosEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <PipeSegment.h>
#include <AdminBase.h>
#include <EC_ConsumerAdmin_impl.h>
#include <EC_SupplierAdmin_impl.h>
#include <SetDefaultProperties.h>

#include <EC_EventChannel_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// EC_EventChannel_impl constructor and destructor
// ----------------------------------------------------------------------

EC_EventChannel_impl::EC_EventChannel_impl(
    ORB_ptr orb,
    PortableServer::POA_ptr poa,
    OB::EventLogger_ptr logger,
    Reaper* reaper,
    const char* myId)
    : poa_(PortableServer::POA::_duplicate(poa)),
      logger_(OB::EventLogger::_duplicate(logger)),
      myId_(myId),
      destroy_(false)
{
    myPipe_ = new PipeSegment();

    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	CORBA::Object_var(orb -> resolve_initial_references("RootPOA")));

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("creating EventChannel(");
	msg += myId_;
	msg += ')';
	logger_ -> trace("event", msg);
    }

    consumerAdmin_ =
	new EC_ConsumerAdmin_impl(orb, rootPOA, logger, reaper, myPipe_);
    supplierAdmin_ =
	new EC_SupplierAdmin_impl(orb, rootPOA, logger, reaper, myPipe_);

    SetDefaultProperties::setDefaults(orb, logger);
}

EC_EventChannel_impl::~EC_EventChannel_impl()
{
    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("destroying EventChannel(");
	msg += myId_;
	msg += ')';
	logger_ -> trace("event", msg);
    }

    consumerAdmin_ -> _remove_ref();
    supplierAdmin_ -> _remove_ref();
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

CosEventChannelAdmin::ConsumerAdmin_ptr
EC_EventChannel_impl::for_consumers()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return consumerAdmin_ -> _this();
}

CosEventChannelAdmin::SupplierAdmin_ptr
EC_EventChannel_impl::for_suppliers()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return supplierAdmin_ -> _this();
}

void
EC_EventChannel_impl::destroy()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    destroy_ = true;

    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());

    myPipe_ -> destroy();

    consumerAdmin_ -> disconnect();
    supplierAdmin_ -> disconnect();
}

PortableServer::POA_ptr
EC_EventChannel_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

// ----------------------------------------------------------------------
// Internal methods
// ----------------------------------------------------------------------

void
EC_EventChannel_impl::discardRequests()
{
    JTCSynchronized sync(*this);

    if(destroy_)
	return;

    myPipe_ -> destroy();

    consumerAdmin_ -> discardRequests();
    supplierAdmin_ -> discardRequests();
}

bool
EC_EventChannel_impl::destroyed()
{
    JTCSynchronized sync(*this);

    return destroy_;
}
