// **********************************************************************
//
// Copyright (c) 2000 
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/OBTypedEventChannelFactory.h>
#include <OB/CosTypedEventChannelAdmin.h>
#include <OB/CosTypedEventComm_skel.h>

#include <TestSuiteUtilities.h>

#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#include <Echo_skel.h>

#if defined(HAVE_STRSTREAM)
#   include <strstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Global variable used for tracing.
//
int TraceLevel = 0;

//
// Use a strstream object to capture the output.  Before returning, 
// direct the contents of this stream to standard output.  This method 
// of directing information to standard outout is used to ensure only 
// one process is directing informtation at a time.  Otherwise a 
// deadlock may occur.
//
OB_STD(strstream) tout;


// ----------------------------------------------------------------------
// Echo_impl implementation
// ----------------------------------------------------------------------
class Echo_impl : virtual public POA_TypedEcho,
                  virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_; // The ORB
    PortableServer::POA_var poa_; // My POA.

public:
    
    Echo_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa)
        : orb_(CORBA::ORB::_duplicate(orb)),
	  poa_(PortableServer::POA::_duplicate(poa))
    {
    }
    
    //
    // IDL to C++ mapping
    //
    void
    echo_string(const char* str)
	throw(CORBA::SystemException)
    {
        tout << str << OB_ENDL;
    }

    void
    disconnect_push_consumer()
	throw(CORBA::SystemException)
    {
        tout << "Disconnect";
        PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
        poa_ -> deactivate_object(oid.in());

	orb_ -> shutdown(false);
    }

    void
    push(const CORBA::Any& data)
	throw(CORBA::SystemException)
    {
    }
    
    PortableServer::POA_ptr
    _default_POA()
    {
        return PortableServer::POA::_duplicate(poa_);
    }    
};

static
void Usage()
{
    OB_CERR << "typedpusheventclient <proxy-ior> [--disconnect]\n"
            "   --disconnect:  Call connect on the proxy and disconnect\n";
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    try
    {
        //
        // Resolve Root POA.
        //
        PortableServer::POA_var rootPOA =
            PortableServer::POA::_narrow(CORBA::Object_var(
                orb -> resolve_initial_references("RootPOA")));

        //
        // Get a reference to the POA manager.
        //
        PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

        //
        // Create a persistent POA.
        //
        CORBA::PolicyList pl(2);
        pl.length(2);
        pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
        pl[1] = rootPOA -> create_id_assignment_policy(
            PortableServer::USER_ID);

        PortableServer::POA_var consumerPOA =
            rootPOA -> create_POA("consumer", manager, pl);

        if(argc > 3)
        {
            Usage();
            return EXIT_FAILURE;
        }

	//
	// Just connect and exit.
	//
	bool disconnect = false;

	//
	// Find whether to append timestamps or not and make sure that no 
	// "new" arguments are given to ClientCommon.
	//
	for(int arg = 0 ; arg < argc ; arg++)
	{
	    if(strcmp(argv[arg],"--disconnect") == 0)
	    {
		disconnect=true;
		for(int j = arg ; j < argc-1 ; j++)
		    argv[j] = argv[j+1];
		argc--;
	    }
	}
	PrintStatusMsg(TraceLevel, "TypedPushEventClient", 2, "Starting");
	
	//
	// Get the proxy object.
	//
	CosEventChannelAdmin::ProxyPushSupplier_var proxy;
	CORBA::Object_var obj = orb -> string_to_object(argv[1]);
	proxy = CosEventChannelAdmin::ProxyPushSupplier::_narrow(obj);

	if(CORBA::is_nil(proxy))
	{
	    OB_CERR << "Proxy is nil in TypedPushEventClient" << OB_ENDL;
	    OB_COUT << "ERROR" << OB_ENDL;
	    return EXIT_FAILURE;
	}

	//
	// Create the implementation object.
	//
	Echo_impl* impl = new Echo_impl(orb, consumerPOA);

        //
        // This will cause the servant to be destroyed when `dummy'
        // goes out of scope.
        //
        PortableServer::ServantBase_var dummy = impl;

        //
        // Connect the consumer to the POA with the same name each
        // time.
        //
        PortableServer::ObjectId_var oid =
            PortableServer::string_to_ObjectId("DefaultConsumer");
        consumerPOA -> activate_object_with_id(oid.in(), impl);
        CosEventComm::PushConsumer_var consumer = impl -> _this();

	//
	// Connect the proxy. If we're already connected we'll get an
	// exception. This should only happen if we're reusing the
	// same proxy-object -- this this case the connected
	// SequencePushConsumer must be the same servant.
	//
	try
	{
	    proxy -> connect_push_consumer(consumer);
	    PrintStatusMsg(TraceLevel, "TypedPushEventClient", 2, 
			   "Created and connected to Echo");
	}
	catch(const CosEventChannelAdmin::AlreadyConnected&)
	{
	}

	manager -> activate();
	
	if(disconnect)
	{
	    return EXIT_SUCCESS;
	}
	tout << OB_STD(ends);
	OB_COUT << tout.str() << OB_ENDL;
    }
    catch(const CORBA::SystemException& ex)
    {	
	OB_CERR << "System exception in TypedPushEventClient" << OB_ENDL;
        OB_CERR << ex << OB_ENDL;
	return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Determine the test suite trace level.
	//
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0; i < keys.length(); ++i) 
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
                break;
            }
	    else
	    {
		OB_CERR << "Warning: Unknown property - " << key << OB_ENDL;
	    }
        }

        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_CERR << ex << OB_ENDL;
        OB_COUT << "ERROR" << OB_ENDL;
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
