// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_IIOP_CONFIG_H
#define OB_IIOP_CONFIG_H

#include <OB/Basic.h>
#include <OB/Template.h>
#include <OB/Any.h>
#include <OB/Object.h>
#include <OB/OCITypes.h>
#include <OB/Properties_fwd.h>
#include <OB/Logger_fwd.h>

namespace OB
{

//
// IIOPConfig converts configuration properties into parameters that
// can be used to create an OCI::Acceptor using an OCI::AccFactory.
// Each invocation of a parse() method may replace parameters created
// by a previous parse() method, therefore callers should invoke
// these methods in the desired order of precedence.
//
class IIOPConfig
{
    //
    // Hide copy-constructor and assignment operator
    //
    IIOPConfig(const IIOPConfig&);
    void operator=(const IIOPConfig&);

    //
    // The properties to be parsed
    //
    Properties_var props_;

    //
    // The Logger
    //
    Logger_var logger_;

    //
    // The parsed parameters
    //
    OCI::ParamSeq params_;

    OCI::Param* findParam(const char*);
    void parseHost(const char*);
    void parsePort(const char*);
    void parseBacklog(const char*);
    void parseBind(const char*);
    void parseNumeric(const char*);
    void parseMultiProfile(const char*);

public:

    IIOPConfig(Properties_ptr, Logger_ptr);

    //
    // Get the parameters for all properties that have been parsed
    //
    OCI::ParamSeq* getParams();

    //
    // Parse the following deprecated properties:
    //
    // ooc.orb.oa.host
    // ooc.orb.oa.numeric
    //
    void parseDeprecatedDefaults();

    //
    // Parse the following deprecated properties:
    //
    // ooc.orb.poamanager.<name>.host
    // ooc.orb.poamanager.<name>.numeric
    // ooc.orb.poamanager.<name>.port
    //
    void parseDeprecatedForName(const char*);

    //
    // Parse the following deprecated property:
    //
    // ooc.orb.oa.port
    //
    // This property generally only applies to the RootPOAManager.
    //
    void parseDeprecatedDefaultPort();

    //
    // Parse the following property:
    //
    // ooc.iiop.port
    //
    // This property generally only applies to the RootPOAManager.
    //
    void parseDefaultPort();

    //
    // Parse the following property:
    //
    // ooc.iiop.bind
    //
    // This property generally only applies to the RootPOAManager.
    //
    void parseDefaultBind();

    //
    // Parse the following properties:
    //
    // ooc.iiop.backlog
    // ooc.iiop.host
    // ooc.iiop.multi_profile
    // ooc.iiop.numeric
    //
    void parseDefaults();

    //
    // Parse the following properties:
    //
    // ooc.iiop.acceptor.<name>.backlog
    // ooc.iiop.acceptor.<name>.bind
    // ooc.iiop.acceptor.<name>.host
    // ooc.iiop.acceptor.<name>.multi_profile
    // ooc.iiop.acceptor.<name>.numeric
    // ooc.iiop.acceptor.<name>.port
    //
    void parseForName(const char*);

    //
    // Parse all properties in the standard order of precedence
    //
    void parseAll(const char*);
};

} // End of namespace OB

#endif
