// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_OBJECT_H
#define OB_OBJECT_H

#include <OB/Object_fwd.h>
#include <OB/StubImpl_fwd.h>
#include <OB/MarshalStubImpl_fwd.h>
#include <OB/IFR2_fwd.h>
#include <OB/Context_fwd.h>
#include <OB/NamedValue_fwd.h>
#include <OB/DII_fwd.h>
#include <OB/Policy_fwd.h>
#include <OB/OCI_fwd.h>
#include <OB/AbstractBase_fwd.h>
#include <OB/OBPOAManager_fwd.h>
#include <OB/ORBInstance_fwd.h>
#include <OB/RefCounted_fwd.h>

#include <OB/Any.h>
#include <OB/TypeCodeConst.h>
#include <OB/OCITypes.h>

namespace CORBA
{

//
// PolicyList
//
class OBUnique_PolicyList;
typedef OB::ObjSeq< Policy, OBUnique_PolicyList > PolicyList;

//
// PolicyType
//
typedef ULong PolicyType;
typedef ULong_out PolicyType_out;

//
// PolicyTypeSeq
//
class OBUnique_PolicyTypeSeq;
typedef OB::FixSeq< CORBA::ULong, OBUnique_PolicyTypeSeq > PolicyTypeSeq;

//
// The SetOverrideType
//
enum SetOverrideType
{
    SET_OVERRIDE,
    ADD_OVERRIDE
};

} // End of namespace CORBA

namespace CORBA
{

//
// The Object class
//
class Object : public OB::RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Object(const Object&);
    void operator=(const Object&);

    //
    // The IDs for CORBA::Object (only for consistency, the list must
    // be empty)
    //
    static const char* ids_[];

    //
    // Mutex for CORBA::Object members. Deriving from a mutex would be
    // to intrusive, as locality-constrained classes derive directly
    // from this class, and thus would inherit this mutex as well.
    //
#ifdef HAVE_JTC
    JTCMutex mutex_;
#endif

    //
    // The ORBInstance object
    //
    OB::ORBInstance_var orbInstance_;
    
    //
    // The IOR and the original IOR
    //
    OB::RefCountIOR_var IOR_;
    OB::RefCountIOR_var origIOR_;

    //
    // The list of policies
    //
    OB::RefCountPolicyList_var policies_;

    //
    // The stub implementation
    //
    OB::StubImplBase_var stubImpl_;

    //
    // Check whether it's safe to retry
    //
    void checkRetry(bool, const CORBA::SystemException&);

protected:

    //
    // Management operations for the stub implementation
    //
    OB::StubImplBase_ptr _OB_getStubImpl();
    OB::MarshalStubImpl_ptr _OB_getMarshalStubImpl();

    //
    // Create a new marshal stub implementation. Derived stubs
    // overwrite this method to return a more derived marshal stub
    // implementation.
    //
    virtual OB::MarshalStubImpl_ptr _OB_createMarshalStubImpl();

    //
    // Constructor is protected
    //
    Object();

    //
    // OB::ObjectFactory may create Objects
    //
    friend class ::OB::ObjectFactory;

    //
    // The DII needs access to _OB_getMarshalStubImpl()
    //
    friend class Request;

public:

    virtual ~Object();

    //
    // Standard IDL to C++ Mapping
    //
    static inline Object_ptr _duplicate(Object_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Object_ptr _nil()
    { return 0; }

    InterfaceDef_ptr _get_interface();
    Boolean _is_a(const char*);
    Boolean _non_existent();

    Boolean _is_equivalent(Object_ptr);
    ULong _hash(ULong);

    static Object_ptr _narrow(AbstractBase_ptr);

    void _create_request(Context_ptr, const char*, NVList_ptr, NamedValue_ptr,
                         Request_out, Flags);

    void _create_request(Context_ptr, const char*, NVList_ptr, NamedValue_ptr,
                         ExceptionList_ptr, ContextList_ptr, Request_out,
			 Flags);

    Request_ptr _request(const char* operation);

    Policy_ptr _get_policy(PolicyType);
    Object_ptr _set_policy_overrides(const PolicyList&, SetOverrideType);
    
    //
    // Policies
    // From draft messaging specification
    //
    Policy_ptr _get_effective_policy_override(PolicyType);
    PolicyList* _get_policy_overrides(const PolicyTypeSeq&);
    Boolean _validate_policies(const PolicyList&);

    //
    // Additional ORBacus-specific functions
    //
    OCI::ConnectorInfo_ptr _get_oci_connector_info();
    OCI::TransportInfo_ptr _get_oci_transport_info();

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    virtual const char** _OB_ids() const;
    static const char** _OB_staticIds();

    void _OB_setup(OB::ORBInstance_ptr, OB::RefCountIOR_ptr,
		   OB::RefCountPolicyList_ptr);
    void _OB_copyFrom(Object_ptr);

    OB::ORBInstance_ptr _OB_ORBInstance() const;
    OB::RefCountIOR_ptr _OB_IOR() const;
    OB::RefCountIOR_ptr _OB_origIOR() const;
    void _OB_marshalOrigIOR(OB::OutputStreamImpl*) const;
    OB::RefCountPolicyList_ptr _OB_policies() const;
    Policy_ptr _OB_getPolicy(PolicyType policyType) const;
    
    void _OB_handleException(const OB::ExceptionBase&, bool&);

    virtual bool _OB_isLC() const { return false; }

    virtual AbstractBase_ptr _OB_toAbstractBase();

    friend class OB::InputStream;
    friend class OB::OutputStream;
};

//
// release and is_nil
//
inline void
release(Object_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(Object_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

namespace OBCORBA
{

class ORB_impl; // For the friend declaration

} // End of namespace CORBA

namespace OB
{
    
//
// The ObjectFactory class
//
class ObjectFactory : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    ObjectFactory(const ObjectFactory&);
    void operator=(const ObjectFactory&);

    bool destroy_; // True if destroy() was called
    ORBInstance_var orbInstance_; // The ORBInstance object
    CORBA::PolicyManager_var policyManager_; // The PolicyManager object

    ObjectFactory();
    ~ObjectFactory();
    friend class ::OBCORBA::ORB_impl; // ORB_impl creates ObjectFactory

public:

    static inline ObjectFactory_ptr _duplicate(ObjectFactory_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ObjectFactory_ptr _nil()
    { return 0; }

    //
    // Destroy the ObjectFactory
    //
    void destroy();

    //
    // Set the ORBInstance object
    //
    void setORBInstance(ORBInstance_ptr);

    //
    // Set the PolicyManager object
    //
    void setPolicyManager(CORBA::PolicyManager_ptr);

    //
    // Create a new CORBA::Object
    //
    CORBA::Object_ptr createObject(OB::RefCountIOR_ptr);

    //
    // Convert an IOR or URL to an object reference
    //
    CORBA::Object_ptr stringToObject(const char*);

    //
    // Get the list of policies new objects are created with
    //
    OB::RefCountPolicyList_ptr policies();
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::ObjectFactory_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
