// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <InternalDomain_impl.h>
#include <ServerFactory_impl.h>
#include <ActiveStateFactory_impl.h>
#include <ServerStarter.h>
#include <OADEndpointManager_impl.h>
#include <OADDatabase.h>

using namespace IMR;

// ----------------------------------------------------------------------
// InternalDomain_impl constructor and destructor
// ----------------------------------------------------------------------

InternalDomain_impl::InternalDomain_impl(CORBA::ORB_ptr orb,
                                         PortableServer::POA_ptr poa,
                                         PortableServer::POA_ptr fwd,
			                 OADDatabase* oaddb,
                                         ServerDatabase* db,
					 OADEndpointManager_impl* endpoint,
			                 IMROAD::Logger_ptr logger,
					 bool administrative,
					 ServerStarter* serverStarter)
    : oaddb_(oaddb),
      poa_(PortableServer::POA::_duplicate(poa)),
      endpoint_(endpoint),
      administrative_(administrative)
{
    activeStateFactory_ = new ActiveStateFactory_impl(db, oaddb, orb, fwd, poa,
    						      logger, serverStarter);
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("ActiveStateFactory");
    poa -> activate_object_with_id(oid, activeStateFactory_);

    serverFactory_ = new ServerFactory_impl(db, oaddb, orb, poa, logger,
    					    administrative, serverStarter);
    oid = PortableServer::string_to_ObjectId("ServerFactory");
    poa -> activate_object_with_id(oid, serverFactory_);

    serverStarter -> setActiveStateFactory(activeStateFactory_);
}

InternalDomain_impl::~InternalDomain_impl()
{
    delete serverFactory_;
    delete activeStateFactory_;
}

// ----------------------------------------------------------------------
// InternalDomain_impl public member functions
// ----------------------------------------------------------------------

//
// Implementation for IMR::Domain
//
ActiveStateFactory_ptr
InternalDomain_impl::get_active_state_factory()
    throw(CORBA::SystemException)
{
    return activeStateFactory_ -> _this();
}

void
InternalDomain_impl::registerServer(const char* name,
                                        const char* exec,
                                        const char* host)
    throw(ServerAlreadyRegistered,
	  CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    Server_var server = serverFactory_ -> create_server_record(name);

    server -> exec(exec);
    server -> host(host);
}

//
// Implementation for IMR::Domain
//
ServerFactory_ptr
InternalDomain_impl::get_server_factory()
    throw(CORBA::SystemException)
{
    return serverFactory_ -> _this();
}

void
InternalDomain_impl::create_oad_record(const char* host)
    throw(IMR::OADAlreadyExists,
	  CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    JTCSynchronized sync(*oaddb_);

    IMRDB_4_0_2::OADRecord rec;
    if(oaddb_ -> queryRecordByHost(host, rec))
        throw OADAlreadyExists();

    rec.host = CORBA::string_dup(host);
    rec.controller = OAD::ProcessController::_nil();
    rec.manager = OAD::ProcessEndpointManager::_nil();
    rec.up = false;
    
    oaddb_ -> addRecord(rec);
}

void
InternalDomain_impl::remove_oad_record(const char* host)
    throw(NoSuchOAD,
          OADRunning,
	  CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    JTCSynchronized sync(*oaddb_);

    IMRDB_4_0_2::OADRecord rec;
    if(!oaddb_ -> queryRecordByHost(host, rec))
        throw NoSuchOAD(host);
    if(rec.up)
        throw OADRunning(host);

    oaddb_ -> removeRecord(host);
}

OADInfo*
InternalDomain_impl::get_oad_record(const char* host)
    throw(IMR::NoSuchOAD,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*oaddb_);

    IMRDB_4_0_2::OADRecord rec;
    if(!oaddb_ -> queryRecordByHost(host, rec))
        throw NoSuchOAD(host);

    OADInfo_var oadInfo = new OADInfo();
    oadInfo -> host = CORBA::string_dup(rec.host);
    oadInfo -> up = rec.up;
    
    return oadInfo._retn();
}

OADInfoSeq*
InternalDomain_impl::list_oads()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*oaddb_);

    IMRDB_4_0_2::OADRecordSeq_var oads = oaddb_ -> getAllRecords();

    OADInfoSeq_var oadInfos = new OADInfoSeq();
    oadInfos -> length(oads -> length());
    for(CORBA::ULong i = 0; i < oads -> length(); i++)
    {
	oadInfos[i].host = CORBA::string_dup(oads[i].host);
	oadInfos[i].up = oads[i].up;
    }
    
    return oadInfos._retn();
}

//
// Implementation for IMR::InternalDomain
//
OADEndpointManager_ptr
InternalDomain_impl::get_endpoint_manager()
    throw(CORBA::SystemException)
{
    return endpoint_ -> _this();
}


PortableServer::POA_ptr
InternalDomain_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}
