// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/IMR.h>
#include <OB/IIOP.h>
#include <OB/ObjectKey.h>
#include <OB/RefCounted.h>

#include <IMRLogger.h>
#include <IMRInternal.h>
#include <LocationForwarder.h>

using namespace IMR;

// ----------------------------------------------------------------------
// LocationForwarder private member implementation
// ----------------------------------------------------------------------

IOP::IOR*
LocationForwarder::getIOR(const char* name,
			  const IMR::POAName& poa)
{
    //
    // First get the appropriate server object.
    //
    IMR::InternalServer_var server;
    try
    {
	IMR::ActiveState_var s = factory_ -> get_active_state(name);
	server = IMR::InternalServer::_narrow(s);    
	assert(!CORBA::is_nil(server));
    }
    catch(const IMR::NoSuchServer&)
    {
	throw CORBA::OBJECT_NOT_EXIST();
    }

    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("contact for: server: ");
	msg += name;
	msg += " poa: ";
	for(CORBA::ULong i = 0; i < poa.length(); i++)
	{
	    msg += poa[i];
	    if(i != poa.length() -1)
		msg += '/';
	}
	logger_ -> trace("IMR", msg);
    }

    while(true)
    {
	IMR::ServerStatus status;
	IMR::POAStatus poaStatus;
	server -> status(poa, status, poaStatus);

	//
	// If the server is running and the poa status is ACTIVE or
	// HOLDING then return the IOR.
	//
	if(status == IMR::RUNNING && poaStatus != IMR::DISCARDING)
	    break;

	switch(status)
	{
	case IMR::FORKED:
	case IMR::STARTING:
	case IMR::STOPPING:
	    //
	    // Have we been waiting too long for a state
	    // transition? If so then kill the server.
	    //
	    if(server -> timeout())
	    {
		try
		{
		    server -> kill();
		}
		catch(const IMR::NoSuchOAD&)
		{
    		    if(logger_ -> traceProcessControl() >= 1)
		    {
		        CORBA::String_var msg = name;
		        msg += ": kill: NoSuchOAD";
		        logger_ -> trace("IMR", msg);
		    }
		    throw CORBA::TRANSIENT();
		}
	    }
	    break;

	case IMR::STOPPED:
	    //
	    // Ask the OAD to exec this server.
	    //
	    try
	    {
		server -> fork();

		//
		// Artificially set the status to FORKED to avoid the
		// spin. TODO: Return the new status from fork?
		//
		status = IMR::FORKED;
	    }
	    catch(const IMR::NoSuchOAD&)
	    {
    		if(logger_ -> traceProcessControl() >= 1)
		{
		    CORBA::String_var msg = name;
		    msg += ": fork: NoSuchOAD";
		    logger_ -> trace("IMR", msg);
		}
		throw CORBA::TRANSIENT();
	    }
	    break;

	case IMR::RUNNING:
	    //
	    // POA must be in DISCARDING or ACTIVE state. Handled below.
	    //
	    break;

	default:
	    abort();
	}

	//
	// If the server status is not RUNNING then wait for a state
	// transition. If the state or poa status has changed since
	// initial call to status or poa_status wait() will return
	// immediately.
	//
	if(status != IMR::RUNNING || poaStatus == IMR::DISCARDING)
	    server -> wait(status, poa, poaStatus);
	else if(poaStatus == IMR::INACTIVE)
	    throw CORBA::TRANSIENT();
    }

    return server -> ior_for(poa);
}

// ----------------------------------------------------------------------
// LocationForwarder constructor/destructor
// ----------------------------------------------------------------------

LocationForwarder::LocationForwarder(CORBA::ORB_ptr orb,
				     IMROAD::Logger_ptr logger,
				     IMR::ActiveStateFactory_ptr factory)
    : logger_(IMROAD::Logger::_duplicate(logger)),
      factory_(IMR::ActiveStateFactory::_duplicate(factory))
{
    OB::ORBInstance_var instance = OB::GetORBInstance(orb);
    objectFactory_ = instance -> getObjectFactory();
}

LocationForwarder::~LocationForwarder()
{
}

// ----------------------------------------------------------------------
// LocationForwarder public member implementation
// ----------------------------------------------------------------------

PortableServer::Servant
LocationForwarder::preinvoke(const PortableServer::ObjectId& oid,
			     PortableServer::POA_ptr poa,
			     const char*,
			     PortableServer::ServantLocator::Cookie&)
    throw(PortableServer::ForwardRequest,
	  CORBA::SystemException)
{
    //
    // Since the object-id is the object-key of the target servant
    // the key should be in ORBacus format -- parse it. However,
    // we must first place this object-id into an octet sequence.
    //
    OCI::ObjectKey oseq(oid.length());
    oseq.length(oid.length());
    ::memcpy(oseq.get_buffer(), oid.get_buffer(), oid.length());

    //
    // If the object isn't ORBacus format then throw an
    // OBJECT_NOT_EXIST exception
    //
    OB::ObjectKeyData data;
    if(!OB::ParseObjectKey(oseq, data))
	throw CORBA::OBJECT_NOT_EXIST();

    //
    // Locate the server, and the POA in the database.
    // Create a valid object key from the data in the object-id
    // and stuff this in the IIOP portion of the poa's IOR.
    //
    IOP::IOR_var ior = getIOR(data.poaID[0], data.poaID);

    //
    // Find the IIOP portion
    //
    CORBA::ULong i = 0;
    while(i < ior -> profiles.length())
	if(ior -> profiles[i].tag == IOP::TAG_INTERNET_IOP)
	    break;

    //
    // No IIOP Profile exists. Since this is an ORBacus internally
    // generated IOR and all IMR produced IORs are IIOP then this
    // is an internal error
    //
    assert(i < ior -> profiles.length());
	    
    //
    // Extract the IIOP Profile Body
    //
    const CORBA::Octet* coct = ior -> profiles[i].profile_data.get_buffer();
    CORBA::ULong len = ior -> profiles[i].profile_data.length();
    OCI::BufferImpl buf(len);
    memcpy(buf.data(), coct, len);
    OB::InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_1 p;
    ::IIOP::ProfileBody_1_1::_OB_unmarshal(p, &in);
	    
    //
    // Put in the correct object-key data
    //
    p.object_key.length(oid.length());
    ::memcpy(p.object_key.get_buffer(), oid.get_buffer(), oid.length());
	    
    //
    // Remarshal data and put back in the profile
    //
    OCI::BufferImpl buf2;
    OB::OutputStreamImpl out(&buf2);
    out._OB_writeEndian();
    p._OB_marshal(&out);
	    
    ior -> profiles[i].profile_data.length(out._OB_pos());
    memcpy(ior -> profiles[i].profile_data.get_buffer(),
	   buf2.data(), out._OB_pos());
	    
    //
    // Create the object and throw a ForwardRequest exeception
    //
    OB::RefCountIOR_var refIOR = new OB::RefCountIOR(ior);
    CORBA::Object_var obj = objectFactory_ -> createObject(refIOR);
    throw PortableServer::ForwardRequest(obj);
    return 0; // Some compilers need this
}

void
LocationForwarder::postinvoke(const PortableServer::ObjectId& oid,
			      PortableServer::POA_ptr poa,
			      const char*,
			      PortableServer::ServantLocator::Cookie,
			      PortableServer::Servant)
    throw(CORBA::SystemException)
{
}
