// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Net.h>
#include <OB/ObjectKey.h>
#include <OB/Properties.h>

#include <ctype.h>

void
Usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [-H IMRHOST] [-P IMRPORT] "
            << "server-name object-id poa1/poa2/.../poan\n";
}

static inline char
EncodeHex(int i)
{
    assert(i >= 0 && i < 16);
    if(i >= 0 && i <= 9)
	return i + '0';
    return (i - 10) + 'A';
}

//
// Escape characters according to URL encoding scheme except for the
// characters listed in 4.6.3
//
char*
AddURLEscapes(const OCI::ObjectKey& seq)
{
    CORBA::String_var result = CORBA::string_alloc(seq.length());
    *result.inout() = '\0';

    //
    // As per RFC 2396.
    //
    for(CORBA::ULong i = 0; i < seq.length(); i++)
    {
        unsigned char s = (unsigned char)seq[i];
	if(isalpha(s) || isdigit(s))
	    result += s;
	else
	{
	    switch(s)
	    {
	    case ';': case '/': case '?': case ':':
	    case '@': case '&': case '=': case  '+':
	    case '$': case ',': case '-': case '_':
	    case '.': case '!': case '~': case '*':
	    case '\'':case '(': case ')':
		result += s;
		break;

	    default:
		result += '%';
		result += EncodeHex((s >> 4) & 0xF);
		result += EncodeHex(s & 0xF);
	    }
	}
    }

    return result._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);

    OB::Properties_var properties = oborb -> properties();

    //
    // Check whether to use numeric hostname
    //
    bool numeric = false; 
    const char* value = properties -> getProperty("ooc.iiop.numeric");
    if(value != 0)
        numeric = (strcmp(value, "true") == 0);


    //
    // Parse command line arguments
    // 
    CORBA::String_var host = CORBA::string_dup("");
    CORBA::String_var port = CORBA::string_dup("");
    int arg;
    for(arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    {
        if(strcmp(argv[arg], "-H") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
	        Usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    host = CORBA::string_dup(argv[arg]);
	}
	else if(strcmp(argv[arg], "-P") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
	        Usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    port = CORBA::string_dup(argv[arg]);
	}
	else
	{
	    OB_CERR << argv[0] << ": unknown option `" 
	    	    << argv[arg] << "'\n" << OB_ENDL;
	    Usage(argv[0]);
	    return EXIT_FAILURE;
	}
    }
    
    if(argc - arg != 3)
    {
	Usage(argv[0]);
	return EXIT_FAILURE;
    }

    const char* serverName = argv[arg++];
    const char* id = argv[arg++];

    PortableServer::ObjectId_var oid = PortableServer::string_to_ObjectId(id);

    //
    // Create an object-key to represent the object-id, and POA
    // provided on the command line
    //
    OB::ObjectKeyData data;

    data.persistent = true;
    data.createTime = 0;
    data.oid = oid;

    CORBA::ULong poaIDLength = data.poaID.length();
    data.poaID.length(poaIDLength + 1);
    data.poaID[poaIDLength] = CORBA::string_dup(serverName);

    const char* p = strtok(argv[arg], "/");
    while(p)
    {
        int len = strcspn(p, "/");
	char* s = CORBA::string_alloc(len);
	strncpy(s, p, len);
	s[len] = '\0';

        CORBA::ULong poaIDLength = data.poaID.length();
        data.poaID.length(poaIDLength + 1);
        data.poaID[poaIDLength] = s;

	p = strtok(0, "/");
    }
    
    //
    // Create the object-key data, and copy this into the ObjectId
    // data
    //
    OCI::ObjectKey_var oct = OB::CreateObjectKey(data);
    oid -> length(oct -> length());
    ::memcpy(oid -> get_buffer(), oct -> get_buffer(), oid -> length());

    //
    // Next we have to create an object-key that points to the forward
    // POA in the IMR
    //
    data.persistent = true;
    data.createTime = 0;
    
    //
    // The output from the first object-key is the object-id for this
    // object-key
    //
    data.oid = oid;

    data.poaID.length(2);
    data.poaID[0] = CORBA::string_dup("_RootPOA");
    data.poaID[1] = CORBA::string_dup("forward");

    //
    // Create the object-key data for this
    //
    oct = OB::CreateObjectKey(data);

    //
    // Convert this object-key into the URL escaped form
    //
    CORBA::String_var str = AddURLEscapes(oct.in());

    //
    // Grab the port (ooc.imr.forward_port) if not already set.
    //
    if(strlen(port) == 0)
    {
        value = properties -> getProperty("ooc.imr.forward_port");
        if(value != 0)
            port = value;
        else
            port = CORBA::string_dup("9998"); // Default port is 9998
    }

    if(strlen(host) == 0)
        host = OB::GetCanonicalHostname(numeric);

    //
    // Output an corbaloc style URL that refers to this object
    //
    OB_COUT << "corbaloc::" << host << ":" << port << "/" << str << OB_ENDL;

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
