// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Client.h>
#include <OB/Except.h>
#include <OB/OCI.h>
#include <OB/ORBInstance.h>
#include <OB/OBPolicies.h>
#include <OB/RefCounted.h>
#include <OB/Properties.h>

#include <OB/function.h>
#include <OB/algorithms.h>

#include <CodeSetUtil.h>
#include <CollocatedClient.h>
#include <GIOPClient.h>

#include <POAManager_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Client_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Client constructor and destructor
// ----------------------------------------------------------------------

Client::Client(Client::ConcModel concModel, const CodeConverters& conv)
    : usage_(0),
      codeConverters_(conv),
      concModel_(concModel)
{
}

Client::~Client()
{
}

// ----------------------------------------------------------------------
// ClientManager constructor and destructor
// ----------------------------------------------------------------------

ClientManager::ClientManager(Client::ConcModel concModel)
    : destroy_(false),
      concModel_(concModel)
{
}

ClientManager::~ClientManager()
{
    assert(destroy_);
    assert(allClients_.empty());
    assert(reusableClients_.empty());
}

// ----------------------------------------------------------------------
// ClientManager public member implementation
// ----------------------------------------------------------------------

void
ClientManager::destroy()
{
    OB_SYNCHRONIZED

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;

    //
    // Set the destroy flag
    //
    destroy_ = true;

    //
    // Destroy all clients
    //
    for(OB_STL::list<Client_var>::iterator iter = allClients_.begin();
        iter != allClients_.end();
	iter++)
    {
        (*iter) -> destroy();
    }

    //
    // Reset internal data
    //
    orbInstance_ = ORBInstance::_nil();
    allClients_.clear();
    reusableClients_.clear();
}

void
ClientManager::setORBInstance(ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);

    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.client_timeout");
    if(value)
	defaultACMTimeout_ = atoi(value);
    else
	defaultACMTimeout_ = 0;
}

OB_STL::list<ClientProfilePair_var>*
ClientManager::getClientProfilePairs(const IOP::IOR& ior,
                                     const PolicyList& policies)
{
    assert_nca(ior.type_id.in() != 0, NCANullString);

    if(::strlen(ior.type_id) == 0 && ior.profiles.length() == 0)
	throw INV_OBJREF(); // Can't create Client for a nil object

    OB_SYNCHRONIZED

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    ULong i, j;

    //
    // Get the active connection management timeout
    //
    ULong acmTimeout = defaultACMTimeout_;
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == ACM_TIMEOUT_POLICY_ID)
        {
            ACMTimeoutPolicy_var policy =
		ACMTimeoutPolicy::_narrow(policies[i]);
            assert(!is_nil(policy));
            acmTimeout = policy -> value();
	    break;
        }
    }

    OB_STL::list<ClientProfilePair_var>* pairs =
        new OB_STL::list<ClientProfilePair_var>;
    try
    {
        //
        // Find out whether private clients are requested
        //
        OB::ConnectionReusePolicy_var reusePolicy;
        for(j = 0; j < policies.length(); j++)
            if(policies[j] -> policy_type() == CONNECTION_REUSE_POLICY_ID)
                reusePolicy = ConnectionReusePolicy::_narrow(policies[j]);
        bool privateClients;
        if(!is_nil(reusePolicy) && reusePolicy -> value() == false)
            privateClients = true;
        else
            privateClients = false;

        //
        // First try to create CollocatedClients
        //
        OBPortableServer::POAManagerFactory_var pmFactory =
            orbInstance_ -> getPOAManagerFactory();
        OBPortableServer::POAManagerSeq_var managers =
            pmFactory -> get_poa_managers();
        
        for(i = 0 ; i < managers -> length() ; i++)
        {
            try
            {
                OCI::AcceptorSeq_var acceptors =
                    managers[i] -> get_acceptors();
                bool local = false;
                for(j = 0 ; j < acceptors -> length() && !local ; j++)
                {
                    OCI::ProfileInfoSeq_var localProfileInfos =
                        acceptors[j] -> get_local_profiles(ior);
                    if(localProfileInfos -> length() > 0)
                        local = true;
                }

                if(local)
                {
                    //
                    // Retrieve the CollocatedServer from the POAManager
                    //
                    OBPortableServer::POAManager_impl* manager =
                        dynamic_cast<OBPortableServer::POAManager_impl*>(
                            managers[i].in());
                    CollocatedServer_var collocatedServer =
                        manager -> _OB_getCollocatedServer();
                    
                    //
                    // Create a new CollocatedClient and add the new client to
                    // both the list of all clients, and the list of clients
                    // that is returned
                    //
                    CodeConverters conv;
                    conv.inputCharConverter =
			LocalCodeConverter::instanceChar();
                    conv.outputCharConverter = conv.inputCharConverter;
                    conv.wcharConverter =
			LocalCodeConverter::instanceWchar();
                    Client_var client = new CollocatedClient(collocatedServer,
                                                             concModel_, conv);

                    allClients_.push_back(client);

                    OCI::ProfileInfoSeq_var profileInfos =
                        client -> getUsableProfiles(ior, policies);
                    for(j = 0 ; j < profileInfos -> length() ; j++)
                    {
                        ClientProfilePair_var pair = new ClientProfilePair();
                        pair -> client = client;
                        pair -> profile =
                            new RefCountProfileInfo(profileInfos[j]);
                        pairs -> push_back(pair);
                    }

                    //
                    // TODO: Introduce reusable CollocatedClients?
                    //
                }
            }
            catch(const OBPortableServer::POAManager::AdapterInactive&)
            {
                // Ignore
            }
        }

        //
        // If no private clients are requested, add all existing reusable
        // clients, which are usable for the given IOR and policies
        //
        if(!privateClients)
        {
            for(OB_STL::list<Client_var>::iterator iter =
                reusableClients_.begin() ;
                iter != reusableClients_.end() ;
                iter++)
            {
                OCI::ProfileInfoSeq_var profileInfos =
                    (*iter) -> getUsableProfiles(ior, policies);
                for(j = 0 ; j < profileInfos -> length() ; j++)
                {
                    ClientProfilePair_var pair = new ClientProfilePair();
                    pair -> client = (*iter);
                    pair -> profile = new RefCountProfileInfo(profileInfos[j]);
                    pairs -> push_back(pair);
                }
            }
        }

        //
        // Finally, create new GIOPClients for all connectors we can get
        //
        OCI::ConFactoryRegistry_var conFactoryRegistry =
            orbInstance_ -> getConFactoryRegistry();
        OCI::ConFactorySeq_var factories =
            conFactoryRegistry -> get_factories();
        for(i = 0 ; i < factories -> length() ; i++)
        {
            OCI::ConnectorSeq_var connectors =
                factories[i] -> create_connectors(ior, policies);
            for(j = 0 ; j < connectors -> length() ; j++)
            {
                //
                // Get all usable profiles
                //
                OCI::ProfileInfoSeq_var profileInfos =
                    connectors[j] -> get_usable_profiles(ior, policies);
                assert(profileInfos -> length() >= 1);

                //
                // Create a new GIOPClient for each usable profile,
                // and set the concurrency model, code converters, and
                // active connection management timeout. Filter out
                // clients that are equivalent to other clients we
                // already have.
                //
                for(ULong k = 0 ; k < profileInfos -> length() ; k++)
                {
                    CodeConverters conv;
                    CodeSetUtil::getCodeConverters(profileInfos[k], conv);
                    Client_var client =
                        new GIOPClient(orbInstance_, connectors[j],
                                       concModel_, conv, acmTimeout);

                    if(pairs -> size() > 0)
                    {
                        OB_STL::list<ClientProfilePair_var>::iterator iter;
                        
                        for(iter = pairs -> begin() ;
                            iter != pairs -> end() ;
                            iter++)
                        {
                            if((*iter) -> client -> equal(client))
                                break;
                        }
                        
                        if(iter != pairs -> end())
                        {
                            client -> destroy();
                            continue;
                        }
                    }

                    //
                    // Add the new client to the list of all clients
                    //
                    allClients_.push_back(client);

                    //
                    // Add client/profile pairs
                    //
                    OCI::ProfileInfoSeq_var clientProfileInfos =
                        client -> getUsableProfiles(ior, policies);
                    for(ULong l = 0 ; l < clientProfileInfos -> length() ; l++)
                    {
                        ClientProfilePair_var pair = new ClientProfilePair();
                        pair -> client = client;
                        pair -> profile =
                            new RefCountProfileInfo(clientProfileInfos[l]);
                        pairs -> push_back(pair);
                    }

                    //
                    // If no private clients have been requested, add it also
                    // to the list of existing reusable clients
                    //
                    if(!privateClients) 
                        reusableClients_.push_back(client);
                }
            }
        }

        //
        // If we still don't have any client/profile pairs, throw a
        // NO_RESOURCES exception
        //
        if(pairs -> size() == 0)
            throw NO_RESOURCES();

        //
        // Increment the usage count on all clients
        //
        OB_STL::list<ClientProfilePair_var>::iterator iter;
        for(iter = pairs -> begin() ; iter != pairs -> end() ; iter++)
            (*iter) -> client -> incUsage();
    }
    catch(...)
    {
        delete pairs;
        throw;
    }

    return pairs;
}

void
ClientManager::releaseClient(Client_ptr client)
{
    OB_SYNCHRONIZED

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	return;

    //
    // TODO:
    //
    // We need to examine the entire shutdown sequence again since the
    // GIOPClientWorkers and what-not get destroyed after the ORB has
    // been marked as destroyed
    //
    // if(destroy_)
    //     throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    bool inUse = client -> decUsage();

    if(!inUse)
    {
	OB_STL::list<Client_var>::iterator iter;

	for(iter = reusableClients_.begin();
	    iter != reusableClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		reusableClients_.erase(iter);
		break;
	    }
	}
	
	for(iter = allClients_.begin();
	    iter != allClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		client -> destroy();
		allClients_.erase(iter);
		return;
	    }
	}
    }
}

Boolean
ClientManager::equivalent(const IOP::IOR& ior1, const IOP::IOR& ior2)
{
    OCI::ConFactoryRegistry_var conFactoryRegistry =
        orbInstance_ -> getConFactoryRegistry();

    OCI::ConFactorySeq_var factories =
        conFactoryRegistry -> get_factories();
    for(ULong i = 0 ; i < factories -> length() ; i++)
        if(!factories[i] -> equivalent(ior1, ior2))
            return false;

    return true;
}

ULong
ClientManager::hash(const IOP::IOR& ior, ULong maximum)
{
    OCI::ConFactoryRegistry_var conFactoryRegistry =
        orbInstance_ -> getConFactoryRegistry();

    OCI::ConFactorySeq_var factories =
        conFactoryRegistry -> get_factories();
    ULong hash = 0;
    for(ULong i = 0 ; i < factories -> length() ; i++)
        hash ^= factories[i] -> hash(ior, maximum);

    return hash % (maximum + 1);
}
