// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <CodeSetDatabase.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// CodeConverterBase static members
// ----------------------------------------------------------------------

const ULong CodeConverterBase::CONVERT_NONE = 0;
const ULong CodeConverterBase::CONVERT_FROM = 1;
const ULong CodeConverterBase::CONVERT_TO   = 2;
const ULong CodeConverterBase::CONVERT_BOTH = 3;

// ----------------------------------------------------------------------
// CodeConverterBase constructor
// ----------------------------------------------------------------------

CodeConverterBase::CodeConverterBase(
    const CodeSetDatabase::CodeSetInfo* fromSet,
    const CodeSetDatabase::CodeSetInfo* toSet)
    : fromSet_(fromSet), toSet_(toSet)
{
}

// ----------------------------------------------------------------------
// CodeConverter constructor
// ----------------------------------------------------------------------

CodeConverter::CodeConverter(
    const CodeSetDatabase::CodeSetInfo* fromSet,
    const CodeSetDatabase::CodeSetInfo* toSet)
    : CodeConverterBase(fromSet, toSet),
      fromMap_(0), toMap_(0)
{
    CodeSetDatabase* db = CodeSetDatabase::instance();

    if(fromSet)
    {
	//
	// Narrow characters
	//
	if(fromSet -> max_bytes == 1)
	{
	    if(fromSet -> rgy_value != CodeSetDatabase::ISOLATIN1)
	    {
		fromMap_ = db -> getCharMapInfo(fromSet);
	    }
	}
	//
	// Wide characters
	//
	else
	{
	    if(fromSet -> rgy_value != CodeSetDatabase::UTF16)
	    {
		fromMap_ = db -> getCharMapInfo(fromSet);
	    }
	}
    }

    if(toSet)
    {
	//
	// Narrow characters
	//
	if(toSet -> max_bytes == 1)
	{
	    if(toSet -> rgy_value != CodeSetDatabase::ISOLATIN1)
	    {
		toMap_ = db -> getCharMapInfo(toSet);
	    }
	}
	//
	// Wide characters
	//
	else
	{
	    if(toSet -> rgy_value != CodeSetDatabase::UTF16)
	    {
		toMap_ = db -> getCharMapInfo(toSet);
	    }
	}
    }

    if(fromMap_ != 0 && toMap_ != 0)
        conversionType_ = CONVERT_BOTH;
    else if(fromMap_ != 0)
        conversionType_ = CONVERT_FROM;
    else if(toMap_ != 0)
        conversionType_ = CONVERT_TO;
    else
        conversionType_ = CONVERT_NONE;
}

// ----------------------------------------------------------------------
// Public member implementations
// ----------------------------------------------------------------------

Char
CodeConverter::convertCharFrom(Char v)
{
    Char value = v;

    if(value >= (Char)fromMap_ -> lower_bound &&
       value <= (Char)fromMap_ -> upper_bound)
    {
        value = (Char)fromMap_ -> map_values[value - fromMap_ -> lower_bound];
    }

    return value;
}

Char
CodeConverter::convertCharTo(Char v)
{
    Char value = v;

    for(ULong i = 0 ; i <= toMap_ -> upper_bound - toMap_ -> lower_bound ; i++)
    {
        if(toMap_ -> map_values[i] == value)
        {
            value = (Char)(toMap_ -> lower_bound + i);
            break;
        }
    }

    return value;
}

Char
CodeConverter::convertCharBoth(Char v)
{
    Char value = v;

    if(value >= (Char)fromMap_ -> lower_bound &&
       value <= (Char)fromMap_ -> upper_bound)
    {
        value = (Char)fromMap_ -> map_values[value - fromMap_ -> lower_bound];
    }

    for(ULong i = 0 ; i <= toMap_ -> upper_bound - toMap_ -> lower_bound ; i++)
    {
        if(toMap_ -> map_values[i] == value)
        {
            value = (Char)(toMap_ -> lower_bound + i);
            break;
        }
    }

    return value;
}

// ----------------------------------------------------------------------
// Read/write/count wchar
// ----------------------------------------------------------------------

ULong
CodeConverter::count_wchar(Octet*& in)
{
    ULong len;

    //
    // UTF-8 is a special case
    //
    if(toSet_ -> rgy_value == CodeSetDatabase::UTF8)
    {
	Octet first = *in;

	if((first & 0x80) == 0)
	{
	    len = 1;
	}
	else if((first & 0xe0) == 0xc0)
	{
	    len = 2;
	}
	else if((first & 0xf0) == 0xe0)
	{
	    len = 3;
	}
	else if((first & 0xf8) == 0xf0)
	{
	    len = 4;
	}
	else if((first & 0xfc) == 0xf8)
	{
	    len = 5;
	}
	else if((first & 0xfe) == 0xfc)
	{
	    len = 6;
	}
	else
	    throw MARSHAL(MinorReadWStringOverflow, COMPLETED_NO);
    }
    else
	len = toSet_ -> max_bytes;

    return len;
}

ULong
CodeConverter::count_wchar(WChar value)
{
    ULong len;

    //
    // UTF-8 is a special case
    //
    if(toSet_ -> rgy_value == CodeSetDatabase::UTF8)
    {
	if(value < 0x80)
	    len = 1;
	else if(value < 0x7ff)
	    len = 2;
	else if(value < 0xffff)
	    len = 3;
	else if(value < 0x1fffff)
	    len = 4;
	else if(value < 0x3ffffff)
	    len = 5;
	else
	    len = 6;
    }
    else
    {
	len = toSet_ -> max_bytes;
    }

    return len;
}

WChar
CodeConverter::read_wchar(Octet*& in, ULong len)
    throw(DATA_CONVERSION)
{
    if(toSet_ -> rgy_value == CodeSetDatabase::UTF8)
	return CodeSetUtil::utf8ToUnicode(in);
    else
    {
        //
        // Skip most significant bytes as long as they are zero
        //
        while(len > 1 && !*in)
        {
            in++;
            len--;
        }

	//
	// Remaining bytes must not exceed the local wide character size
	// and must not exceed the maximum number of bytes for the
	// code set used
	//
	if(!len || len > sizeof(WChar) || len > fromSet_ -> max_bytes)
	    throw DATA_CONVERSION();

	WChar v;

	switch(len)
	{

	case 1:
	{
	    v = *in++;
	    break;
	}

	case 2:
	{
	    switch(toSet_ -> max_bytes)
	    {
	    case 2:
	    case 4:
	    {
		v = *in++ << 8;
		v |= *in++;
		break;
	    }

	    default:
		throw DATA_CONVERSION();
	    }

	    break;
	}

	case 3:
	{
	    switch(toSet_ -> max_bytes)
	    {
	    case 3:
	    case 4:
	    {
		v = *in++ << 16;
		v |= *in++ << 8;
		v |= *in++;
		break;
	    }

	    default:
		throw DATA_CONVERSION();
	    }
	    
	    break;
	}

	case 4:
	{
	    switch(toSet_ -> max_bytes)
	    {
	    case 4:
	    {
		v = *in++ << 24;
		v |= *in++ << 16;
		v |= *in++ << 8;
		v |= *in++;
		break;
	    }

	    default:
		throw DATA_CONVERSION();
	    }

	    break;
	}

	default:
	    throw DATA_CONVERSION();
	}

	return v;
    }
}

void
CodeConverter::write_wchar(Octet*& out, WChar v)
    throw(DATA_CONVERSION)
{
    if(toSet_ -> rgy_value == CodeSetDatabase::UTF8)
	CodeSetUtil::unicodeToUtf8(out, v);
    else
    {
	ULong len = count_wchar(v);

	//
	// Write valid bytes
	//
	WChar value = v;
	while(len)
	{
	    out[--len] = (Octet)value;
	    value >>= 8;
	}
    }
}

WChar
CodeConverter::convert_wchar_to(WChar v)
    throw(DATA_CONVERSION)
{
    WChar value = v;

    if(value)
    {
	//
	// Convert to intermediate Unicode character first
	//
	if(fromMap_)
	{
	    if(value >= (WChar)fromMap_ -> lower_bound &&
	       value <= (WChar)fromMap_ -> upper_bound)
	    {
		value =
		    fromMap_ -> map_values[value - fromMap_ -> lower_bound];
	    }
	}

	//
	// Now convert to final character from Unicode character
	//
	if(toMap_)
	{
	    for(ULong i = 0 ;
		i <= toMap_ -> upper_bound - toMap_ -> lower_bound; i++)
	    {
		if(toMap_ -> map_values[i] == value)
		{
		    value = (WChar)(toMap_ -> lower_bound + i);
		    break;
		}
	    }
	}
    }

    return value;
}

WChar
CodeConverter::convert_wchar_from(WChar v)
    throw(DATA_CONVERSION)
{
    WChar value = v;

    if(value)
    {
	//
	// Convert to intermediate Unicode character first
	//
	if(toMap_)
	{
	    if(value >= (WChar)toMap_ -> lower_bound &&
	       value <= (WChar)toMap_ -> upper_bound)
	    {
		value = toMap_ -> map_values[value - toMap_ -> lower_bound];
	    }
	}

	//
	// Now convert to final character from Unicode character
	//
	if(fromMap_ && value)
	{
	    for(ULong i = 0 ;
		i <= fromMap_ -> upper_bound - fromMap_ -> lower_bound; i++)
	    {
		if(fromMap_ -> map_values[i] == value)
		{
		    value = (WChar)(fromMap_ -> lower_bound + i);
		    break;
		}
	    }
	}
    }

    return value;
}

// ----------------------------------------------------------------------
// Local read/write wchar (no conversion necessary)
// ----------------------------------------------------------------------

#ifdef HAVE_JTC
JTCMutex LocalCodeConverter::instanceCharMutex_;
JTCMutex LocalCodeConverter::instanceWcharMutex_;
#endif
OB::LocalCodeConverter* LocalCodeConverter::instanceChar_ = 0;
OB::LocalCodeConverter* LocalCodeConverter::instanceWchar_ = 0;

namespace OB
{

class LocalCodeConverterDestroyer
{
public:

    ~LocalCodeConverterDestroyer()
    {
	delete LocalCodeConverter::instanceChar_;
	LocalCodeConverter::instanceChar_ = 0;
	delete LocalCodeConverter::instanceWchar_;
	LocalCodeConverter::instanceWchar_ = 0;
    }
};

} // End of namespace OB

static LocalCodeConverterDestroyer destroyer;

LocalCodeConverter::LocalCodeConverter(const CodeSetDatabase::CodeSetInfo* cs,
				       const CodeSetDatabase::CodeSetInfo* wcs)
    : CodeConverterBase(cs, wcs)
{
}

// ----------------------------------------------------------------------
// Public member implementations
// ----------------------------------------------------------------------

LocalCodeConverter*
LocalCodeConverter::instanceChar()
{
    //
    // Double checked locking
    //
    if(instanceChar_ == 0)
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(instanceCharMutex_);
#endif
	if(instanceChar_ == 0)
	{
	    instanceChar_ = new LocalCodeConverter(
		CodeSetDatabase::instance() -> getCodeSetInfo(
		    CodeSetUtil::native_cs()),
		CodeSetDatabase::instance() -> getCodeSetInfo(
		    CodeSetUtil::native_cs()));
	}
    }

    return instanceChar_;
}

LocalCodeConverter*
LocalCodeConverter::instanceWchar()
{
    //
    // Double checked locking
    //
    if(instanceWchar_ == 0)
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(instanceWcharMutex_);
#endif
	if(instanceWchar_ == 0)
	{
	    instanceWchar_ = new LocalCodeConverter(
		CodeSetDatabase::instance() -> getCodeSetInfo(
		    CodeSetUtil::native_wcs()),
		CodeSetDatabase::instance() -> getCodeSetInfo(
		    CodeSetUtil::native_wcs()));
	}
    }

    return instanceWchar_;
}

Char
LocalCodeConverter::convertCharFrom(Char value)
{
    return value;
}

Char
LocalCodeConverter::convertCharTo(Char value)
{
    return value;
}

Char
LocalCodeConverter::convertCharBoth(Char value)
{
    return value;
}

WChar
LocalCodeConverter::read_wchar(Octet*& in, ULong len)
    throw(DATA_CONVERSION)
{
    WChar val = 0;
    for(ULong i = 0 ; i < len ; i++)
    {
        val <<= 8;
        val += *in++;
    }

    return val;
}

void
LocalCodeConverter::write_wchar(Octet*& out, WChar v)
    throw(DATA_CONVERSION)
{
    ULong len = toSet_ -> max_bytes;
    for(ULong i = len ; i > 0 ; i--)
    {
	out[i - 1] = (Octet)v;
	v >>= 8;
    }
}
