// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_CODE_CONVERTER_H
#define OB_CODE_CONVERTER_H

#include <OB/Basic.h>

#include <CodeSetDatabase.h>

namespace OB
{

class CodeConverterBase
{

protected:

    //
    // Source and destination code set
    //
    const CodeSetDatabase::CodeSetInfo* fromSet_;
    const CodeSetDatabase::CodeSetInfo* toSet_;

public:

    CodeConverterBase(const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverterBase() { };

    //
    // Conversion types
    //
    static const CORBA::ULong CONVERT_NONE;
    static const CORBA::ULong CONVERT_FROM;
    static const CORBA::ULong CONVERT_TO;
    static const CORBA::ULong CONVERT_BOTH;

    //
    // Get source/destination code set
    //
    const CodeSetDatabase::CodeSetInfo* getFrom() const { return fromSet_; }
    const CodeSetDatabase::CodeSetInfo* getTo() const { return toSet_; }

    //
    // Get conversion type
    //
    virtual CORBA::ULong getConversionType() const = 0;

    //
    // Convert character from/to
    //
    virtual CORBA::Char convertCharFrom(CORBA::Char) = 0;
    virtual CORBA::Char convertCharTo(CORBA::Char) = 0;
    virtual CORBA::Char convertCharBoth(CORBA::Char) = 0;

    //
    // Read/write/count wide character
    //
    virtual CORBA::WChar read_wchar(CORBA::Octet*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::ULong count_wchar(CORBA::Octet*&) = 0;
    virtual CORBA::ULong count_wchar(CORBA::WChar) = 0;

    //
    // Convert wide character
    //
    virtual CORBA::WChar convert_wchar_to(CORBA::WChar)
        throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::WChar convert_wchar_from(CORBA::WChar)
        throw(CORBA::DATA_CONVERSION) = 0;
};


class CodeConverter : public CodeConverterBase
{
    //
    // Unicode character mapping tables
    //
    const CodeSetDatabase::CharMapInfo* fromMap_;
    const CodeSetDatabase::CharMapInfo* toMap_;

    //
    // Conversion type
    //
    CORBA::ULong conversionType_;

public:

    CodeConverter(const CodeSetDatabase::CodeSetInfo*,
		  const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverter() { };

    //
    // Get conversion type
    //
    virtual CORBA::ULong getConversionType() const { return conversionType_; }

    //
    // Convert narrow character from/to
    //
    virtual CORBA::Char convertCharFrom(CORBA::Char);
    virtual CORBA::Char convertCharTo(CORBA::Char);
    virtual CORBA::Char convertCharBoth(CORBA::Char);

    //
    // Read/write/count wide character
    //
    CORBA::WChar read_wchar(CORBA::Octet*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::Octet*&);
    CORBA::ULong count_wchar(CORBA::WChar);

    //
    // Convert via Unicode
    //
    CORBA::WChar convert_wchar_to(CORBA::WChar)
        throw(CORBA::DATA_CONVERSION);
    CORBA::Char convert_char(const CodeSetDatabase::CharMapInfo*,
			     const CodeSetDatabase::CharMapInfo*,
			     CORBA::Char);
    CORBA::WChar convert_wchar_from(CORBA::WChar)
        throw(CORBA::DATA_CONVERSION);
};


class LocalCodeConverter : public CodeConverterBase
{
    //
    // The LocalCodeConverter singleton
    //
#ifdef HAVE_JTC
    static JTCMutex instanceCharMutex_;
    static JTCMutex instanceWcharMutex_;
#endif
    static OB_IMPORT LocalCodeConverter* instanceChar_;
    static OB_IMPORT LocalCodeConverter* instanceWchar_;
    friend class LocalCodeConverterDestroyer;

public:

    LocalCodeConverter(const CodeSetDatabase::CodeSetInfo*,
		       const CodeSetDatabase::CodeSetInfo*);
    virtual ~LocalCodeConverter() { };

    //
    // Get singleton instances
    //
    static LocalCodeConverter* instanceChar();
    static LocalCodeConverter* instanceWchar();

    //
    // Get conversion type
    //
    virtual CORBA::ULong getConversionType() const { return CONVERT_NONE; }

    //
    // Convert narrow character from/to
    //
    virtual CORBA::Char convertCharFrom(CORBA::Char);
    virtual CORBA::Char convertCharTo(CORBA::Char);
    virtual CORBA::Char convertCharBoth(CORBA::Char);

    //
    // Read/write/count wide character
    //
    CORBA::WChar read_wchar(CORBA::Octet*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::Octet*&) { return toSet_ -> max_bytes; }
    CORBA::ULong count_wchar(CORBA::WChar) { return toSet_ -> max_bytes; }

    //
    // Convert wide character
    //
    CORBA::WChar convert_wchar_to(CORBA::WChar v)
	throw(CORBA::DATA_CONVERSION) { return v; }
    CORBA::WChar convert_wchar_from(CORBA::WChar v)
	throw(CORBA::DATA_CONVERSION) { return v; }
};

} // End of namespace OB

#endif
