// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <CodeSetDatabase.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

#include <locale.h>

using namespace CORBA;
using namespace OB;

#ifdef HAVE_JTC
JTCMutex CodeSetDatabase::instanceMutex_;
#endif
CodeSetDatabase* CodeSetDatabase::instance_ = 0;

namespace OB
{

class CodeSetDatabaseDestroyer
{
public:

    ~CodeSetDatabaseDestroyer()
    {

	delete CodeSetDatabase::instance_;
	CodeSetDatabase::instance_ = 0;
    }
};

} // End of namespace OB

static CodeSetDatabaseDestroyer destroyer;

const ULong CodeSetDatabase::ISOLATIN1 = 0x00010001;
const ULong CodeSetDatabase::ISOLATIN2 = 0x00010002;
const ULong CodeSetDatabase::ISOLATIN3 = 0x00010003;
const ULong CodeSetDatabase::ISOLATIN4 = 0x00010004;
const ULong CodeSetDatabase::ISOLATIN5 = 0x00010005;
const ULong CodeSetDatabase::ISOLATIN7 = 0x00010007;
const ULong CodeSetDatabase::ISOLATIN9 = 0x00010009;
const ULong CodeSetDatabase::PCS = 0x00010020;
const ULong CodeSetDatabase::UTF8 = 0x05010001;
const ULong CodeSetDatabase::UTF16 = 0x00010109;
const ULong CodeSetDatabase::UCS2 = 0x00010100;

// ----------------------------------------------------------------------
// CodeSetDatabase constructor and destructor
// ----------------------------------------------------------------------

CodeSetDatabase::CodeSetDatabase()
{
}

CodeSetDatabase::~CodeSetDatabase()
{
    for(ULong i = 0 ; i < codeConverters_.length() ; i++)
	delete codeConverters_[i];
    codeConverters_.length(0);
}

// ----------------------------------------------------------------------
// CodeSetDatabase public member implementation
// ----------------------------------------------------------------------

CodeSetDatabase*
CodeSetDatabase::instance()
{
    //
    // Double-checked locking
    //
    if(instance_ == 0)
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(instanceMutex_);
#endif
	if(instance_ == 0)
	{
	    instance_ = new CodeSetDatabase();

	    //
	    // Add locale specific char codesets
	    //
	    const char* loc = setlocale(LC_ALL, 0);

	    if(strcmp(loc, "C") == 0 || strcmp(loc, "POSIX"))
	    {
		CodeSetUtil::addCharCodeSet(ISOLATIN1);
	    }
	    else
	    {
		//
		// West European (C, POSIX, Germany, England, France,
		// Netherlands, Portugal)
		//
		if(strncmp(loc, "de", 2) == 0 || strncmp(loc, "en", 2) == 0 ||
		   strncmp(loc, "fr", 2) == 0 || strncmp(loc, "nl", 2) == 0 ||
		   strncmp(loc, "pt", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN1);
		}
		//
		// North European (Denmark, Finland, Island, Norway, Sweden)
		//
		else if(strncmp(loc, "da", 2) == 0 ||
			strncmp(loc, "fi", 2) == 0 ||
			strncmp(loc, "is", 2) == 0 ||
			strncmp(loc, "no", 2) == 0 ||
			strncmp(loc, "sv", 2))
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN4);
		}
		//
		// South European (Italy)
		//
		else if(strncmp(loc, "it", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN3);
		}
		//
		// East European (Czek, Hungary, Poland, Slovakia)
		//
		else if(strncmp(loc, "cs", 2) == 0 ||
			strncmp(loc, "hu", 2) == 0 ||
			strncmp(loc, "pl", 2) == 0 ||
			strncmp(loc, "sk", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN2);
		}
		//
		// Greek (Greece)
		//
		else if(strncmp(loc, "el", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN7);
		}
		//
		// Cyrillic (Russia)
		//
		else if(strncmp(loc, "ru", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN5);
		}
		//
		// Turkish (Turkey)
		//
		else if(strncmp(loc, "tr", 2) == 0)
		{
		    CodeSetUtil::addCharCodeSet(ISOLATIN9);
		}
	    }

	    //
	    // Always supported
	    //
	    CodeSetUtil::addCharCodeSet(PCS);

	    //
	    // Add the supported wchar codesets in the preferred order
	    //
	    CodeSetUtil::addWcharCodeSet(UTF16);
	    CodeSetUtil::addWcharCodeSet(UCS2);

	    //
	    // Always supported, but only as transmission codeset
	    //
//	    CodeSetUtil::addWcharCodeSet(UTF8);
	}
    }

    return instance_;
}

//
// This is the only method that acccesses mutable data thus needs
// synchronization
//
CodeConverterBase*
CodeSetDatabase::getConverter(ULong to, ULong from)
{
    //
    // Check for legal code sets
    //
    if(!to || !from)
	return 0;

    //
    // Optimization: Use local converter for identical 8 bit char codesets
    //
    if(to == from)
    {
	const CodeSetInfo* codeSetInfo = getCodeSetInfo(to);
	if(codeSetInfo->max_bytes == 1)
	    return LocalCodeConverter::instanceChar();
    }

    OB_SYNCHRONIZED

    //
    // Check if the right converter is already available
    //
    for(ULong i = 0 ; i < codeConverters_.length() ; i++)
    {
	if(codeConverters_[i] -> getTo() -> rgy_value == to &&
	   codeConverters_[i] -> getFrom() -> rgy_value == from)
	    return codeConverters_[i];
    }

    //
    // Create new converter and add it to the converter list
    //
    CodeConverterBase* converter =
	new CodeConverter(getCodeSetInfo(from), getCodeSetInfo(to));
    CORBA::ULong codeConvertersLength = codeConverters_.length();
    codeConverters_.length(codeConvertersLength + 1);
    codeConverters_[codeConvertersLength] = converter;

    return converter;
}

const CodeSetDatabase::CodeSetInfo*
CodeSetDatabase::getCodeSetInfo(ULong rgy_value)
{
    //
    // Check if code set id is listed in registry
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
	if(codeSetInfoArray_[i].rgy_value == rgy_value)
	   return &codeSetInfoArray_[i];

    return 0;
}

ULong
CodeSetDatabase::determineTCS(const CONV_FRAME::CodeSetComponent& clientCS,
			      const CONV_FRAME::CodeSetComponent& serverCS,
			      ULong fallback)
{
    //
    // Check if native code sets are valid
    //
    if(!clientCS.native_code_set || !serverCS.native_code_set)
	return 0;

    //
    // Check if the native code sets are identical
    // In case they are no conversion is required
    //
    if(clientCS.native_code_set == serverCS.native_code_set)
	return serverCS.native_code_set;

    //
    // Check if client can convert
    //
    if(checkCodeSetId(clientCS, serverCS.native_code_set))
	return serverCS.native_code_set;

    //
    // Check if server can convert
    //
    if(checkCodeSetId(serverCS, clientCS.native_code_set))
	return clientCS.native_code_set;

    //
    // Check for common code set that can be used for transmission
    // The server supported code sets have preference
    //
    for(ULong i = 0 ; i < serverCS.conversion_code_sets.length() ; i++)
    {
	if(checkCodeSetId(clientCS, serverCS.conversion_code_sets[i]))
	    return serverCS.conversion_code_sets[i];
    }

    //
    // Check compatibility by using the OSF registry,
    // use fallback code set if compatible
    //
    if(isCompatible(clientCS.native_code_set, serverCS.native_code_set))
	return fallback;

    throw CODESET_INCOMPATIBLE();
    return 0; // Some compilers need this
}

// ----------------------------------------------------------------------
// Check code set compatibility
// ----------------------------------------------------------------------

bool
CodeSetDatabase::isCompatible(ULong id1, ULong id2)
{
    const CodeSetInfo* cs1 = getCodeSetInfo(id1);
    const CodeSetInfo* cs2 = getCodeSetInfo(id2);

    if(!cs1 || !cs2)
	return false;

    for(ULong i = 0 ; i < cs1 -> char_values_size ; i++)
    {
	for(ULong j = 0 ; j < cs2 -> char_values_size ; j++)
	{
	    //
	    // In order to be compatible in OFS terms two code sets must have
	    // a common character set
	    //
	    if(cs1 -> char_values[i] == cs2 -> char_values[j])
		return true;
	}
    }

    return false;
}

bool
CodeSetDatabase::checkCodeSetId(const CONV_FRAME::CodeSetComponent& cs,
				ULong id)
{
    for(ULong i = 0 ; i < cs.conversion_code_sets.length() ; i++)
    {
	if(cs.conversion_code_sets[i] == id)
	    return true;
    }

    //
    // ID not found
    //
    return false;
}

ULong
CodeSetDatabase::nameToId(const char* name)
{
    assert(name != 0);

    //
    // Check if codeset name is listed in registry
    // Return first match so that shortcuts can be used
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
    {
	if(strstr(codeSetInfoArray_[i].description, name) != 0)
	    return codeSetInfoArray_[i].rgy_value;
    }

    //
    // Name not found
    //
    return 0;
}

const char*
CodeSetDatabase::idToName(ULong rgy_value)
{
    //
    // Check if codeset ID is listed in registry
    //
    for(ULong i = 0 ; i < codeSetInfoArraySize_ ; i++)
	if(codeSetInfoArray_[i].rgy_value == rgy_value)
	    return codeSetInfoArray_[i].description;

    //
    // ID not found
    //
    return 0;
}

const CodeSetDatabase::CharMapInfo*
CodeSetDatabase::getCharMapInfo(const CodeSetInfo* codeSetInfo)
{
    const char* name = "";

    switch(codeSetInfo -> rgy_value)
    {
    case ISOLATIN1:
	name = "ISO 8859-1 (1987) to Unicode";
	break;

    case ISOLATIN2:
	name = "ISO 8859-2 (1987) to Unicode";
	break;

    case ISOLATIN3:
	name = "ISO 8859-3 (1988) to Unicode";
	break;

    case ISOLATIN4:
	name = "ISO 8859-4 (1988) to Unicode";
	break;

    case ISOLATIN5:
	name = "ISO 8859-5 (1988) to Unicode";
	break;

    case ISOLATIN7:
	name = "ISO 8859-7 (1987) to Unicode";
	break;

    case ISOLATIN9:
	name = "ISO 8859-9 (1989) to Unicode";
	break;

    case PCS:
	name = "PCS to Unicode";
	break;

    default:
	break;
    }

    for(ULong i = 0 ; i < charMapInfoArraySize_ ; i++)
    {
	if(strcmp(charMapInfoArray_[i].name, name) == 0)
	    return &charMapInfoArray_[i];
    }

    return 0;
}
