// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Except.h>
#include <OB/PIIOR.h>
#include <OB/PIManager.h>
#include <OB/Stream.h>

#include <CodeSetDatabase.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;

FixSeq< ULong, int > CodeSetUtil::supportedCharCodeSets_;
FixSeq< ULong, int > CodeSetUtil::supportedWcharCodeSets_;

ULong CodeSetUtil::native_cs_ = CodeSetDatabase::ISOLATIN1;
ULong CodeSetUtil::native_wcs_ = CodeSetDatabase::UTF16;
ULong CodeSetUtil::default_wcs_ = 0;

// ----------------------------------------------------------------------
// Get and set native code sets
// ----------------------------------------------------------------------

ULong
CodeSetUtil::native_cs()
{
    return native_cs_;
}

void
CodeSetUtil::native_cs(ULong id)
{
    native_cs_ = id;
}

ULong
CodeSetUtil::native_wcs()
{
    return native_wcs_;
}

void
CodeSetUtil::native_wcs(ULong id)
{
    native_wcs_ = id;
}

void
CodeSetUtil::default_wcs(ULong id)
{
    default_wcs_ = id;
}

// ----------------------------------------------------------------------
// Add code set
// ----------------------------------------------------------------------

void
CodeSetUtil::addCharCodeSet(ULong id)
{
    CORBA::ULong supportedCharCodeSetsLength =
	supportedCharCodeSets_.length();
    supportedCharCodeSets_.length(supportedCharCodeSetsLength + 1);
    supportedCharCodeSets_[supportedCharCodeSetsLength] = id;
}

void
CodeSetUtil::addWcharCodeSet(ULong id)
{
    CORBA::ULong supportedCharCodeSetsLength =
	supportedCharCodeSets_.length();
    supportedCharCodeSets_.length(supportedCharCodeSetsLength + 1);
    supportedCharCodeSets_[supportedCharCodeSetsLength] = id;

    CORBA::ULong supportedWcharCodeSetsLength =
	supportedWcharCodeSets_.length();
    supportedWcharCodeSets_.length(supportedWcharCodeSetsLength + 1);
    supportedWcharCodeSets_[supportedWcharCodeSetsLength] = id;
}

// ----------------------------------------------------------------------
// Create code set component and info for current code set settings
// ----------------------------------------------------------------------

CONV_FRAME::CodeSetComponentInfo*
CodeSetUtil::createCodeSetComponentInfo()
{
    CONV_FRAME::CodeSetComponent_var c =
	createCharCodeSetComponent(native_cs_);
    CONV_FRAME::CodeSetComponent_var wc =
	createWcharCodeSetComponent(native_wcs_);

    CONV_FRAME::CodeSetComponentInfo* codeSetComponentInfo =
	new CONV_FRAME::CodeSetComponentInfo();
    codeSetComponentInfo -> ForCharData = c;
    codeSetComponentInfo -> ForWcharData = wc;

    return codeSetComponentInfo;
}

CONV_FRAME::CodeSetComponent*
CodeSetUtil::createCharCodeSetComponent(ULong id)
{
    CodeSetDatabase::instance();

    CONV_FRAME::CodeSetComponent* codeSetComponent =
	new CONV_FRAME::CodeSetComponent();

    codeSetComponent -> native_code_set = id;

    //
    // Add conversion code sets, filter native code set
    //
    for(ULong i = 0 ; i < supportedCharCodeSets_.length() ; i++)
    {
	if(supportedCharCodeSets_[i] != id) {
            OB::FixSeq<CONV_FRAME::CodeSetId,
                       CONV_FRAME::CodeSetComponent::OBUnique_conversion_code_sets>&
                code_sets = codeSetComponent -> conversion_code_sets;

	    CORBA::ULong conversion_code_sets_length = code_sets.length();
	    code_sets.length(conversion_code_sets_length + 1);
	    code_sets[conversion_code_sets_length] = supportedCharCodeSets_[i];
        }
    }

    return codeSetComponent;
}

CONV_FRAME::CodeSetComponent*
CodeSetUtil::createWcharCodeSetComponent(ULong id)
{
    CodeSetDatabase::instance();

    CONV_FRAME::CodeSetComponent* codeSetComponent =
	new CONV_FRAME::CodeSetComponent();

    codeSetComponent -> native_code_set = id;

    //
    // Add conversion code sets, filter native code set
    //
    for(ULong i = 0 ; i < supportedWcharCodeSets_.length() ; i++)
    {
	if(supportedWcharCodeSets_[i] != id) {
            OB::FixSeq<CONV_FRAME::CodeSetId,
                       CONV_FRAME::CodeSetComponent::OBUnique_conversion_code_sets>&
                code_sets = codeSetComponent -> conversion_code_sets;

	    CORBA::ULong conversion_code_sets_length = code_sets.length();
	    code_sets.length(conversion_code_sets_length + 1);
	    code_sets[conversion_code_sets_length] = supportedWcharCodeSets_[i];
        }
    }

    return codeSetComponent;
}

bool
CodeSetUtil::getCodeSetInfoFromComponents(
    const OCI::ProfileInfo& profileInfo,
    CONV_FRAME::CodeSetComponentInfo& info)
{
    //
    // IIOP 1.0 has no code set information
    //
    if(profileInfo.major == 1 && profileInfo.minor == 0)
	return false;
    //
    // IIOP 1.1
    //
    for(ULong i = 0 ; i < profileInfo.components.length() ; i++)
    {
	if(checkForCodeSetInfo(profileInfo.components[i], info))
	    return true;
    }

    return false;
}

// ----------------------------------------------------------------------
// Get code converter from ProfileInfo and/or IOR
// ----------------------------------------------------------------------

void
CodeSetUtil::getCodeConverters(const OCI::ProfileInfo& profileInfo,
			       OB::CodeConverters& conv)
{
    //
    // Set code set defaults: ISO 8859-1 for char, no default for wchar
    // (13.7.2.4)
    //
    CONV_FRAME::CodeSetComponentInfo serverInfo;
    serverInfo.ForCharData.native_code_set = CodeSetDatabase::ISOLATIN1;
    serverInfo.ForWcharData.native_code_set = default_wcs_;

    //
    // Try to get code set information from IIOP profile
    //
    getCodeSetInfoFromComponents(profileInfo, serverInfo);

    CodeSetDatabase* db = CodeSetDatabase::instance();

    //
    // Set up char code converter
    //
    CONV_FRAME::CodeSetComponent_var client_cs =
	createCharCodeSetComponent(native_cs());
    ULong tcs_c = db -> determineTCS(client_cs, serverInfo.ForCharData,
				     CodeSetDatabase::UTF8);
    conv.inputCharConverter = db -> getConverter(native_cs(), tcs_c);
    conv.outputCharConverter = db -> getConverter(tcs_c, native_cs());

    //
    // Set up wchar code converter
    //
    CONV_FRAME::CodeSetComponent_var client_wcs =
	createWcharCodeSetComponent(native_wcs());
    ULong tcs_wc = db -> determineTCS(client_wcs, serverInfo.ForWcharData,
				      CodeSetDatabase::UTF16);
    conv.wcharConverter = db -> getConverter(tcs_wc, native_wcs());
}

// ----------------------------------------------------------------------
// Check for code set information in a tagged component
// ----------------------------------------------------------------------

bool
CodeSetUtil::checkForCodeSetInfo(const IOP::TaggedComponent& comp,
                                 CONV_FRAME::CodeSetComponentInfo& info)
{
    if(comp.tag == IOP::TAG_CODE_SETS)
    {
	const Octet* coct = comp.component_data.get_buffer();
	ULong len = comp.component_data.length();
        OCI::BufferImpl buf((Octet*)coct, len);
	InputStreamImpl in(&buf, 0, false);
	in._OB_readEndian();
	CONV_FRAME::CodeSetComponentInfo::_OB_unmarshal(info, &in);
	return true;
    }

    return false;
}

// ----------------------------------------------------------------------
// Extract code set context from service context
// ----------------------------------------------------------------------

void
CodeSetUtil::extractCodeSetContext(const IOP::ServiceContext& context,
				   CONV_FRAME::CodeSetContext& ctx)
{
    const Octet* coct = context.context_data.get_buffer();
    ULong len = context.context_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    CONV_FRAME::CodeSetContext::_OB_unmarshal(ctx, &in);
}

// ----------------------------------------------------------------------
// Convert between UTF-8 and Unicode
// ----------------------------------------------------------------------

void
CodeSetUtil::unicodeToUtf8(Octet*& out, WChar value)
    throw(DATA_CONVERSION)
{
    ULong len;
    WChar first;

    //
    // Direct mapping (7 bits) for characters < 0x80
    //
    if(value < 0x80)
    {
	first = value;
	len = 1;
    }
    else if(value < 0x7ff)
    {
        // 5 free bits (%110xxxxx | %vvvvv)
        first = 0xc0;
	len = 2;
    }
    else if(value < 0xffff)
    {
        // 4 free bits (%1110xxxx | %vvvv)
        first = 0xe0;
	len = 3;
    }
    else if(value < 0x1fffff)
    {
        // 3 free bits (%11110xxx | %vvv)
        first = 0xf0;
	len = 4;
    }
    else if(value < 0x3ffffff)
    {
        // 2 free bits (%111110xx | %vv)
        first = 0xf8;
	len = 5;
    }
    else if(value < 0x7fffffff)
    {
        // 1 free bit (%1111110x | %v)
        first = 0xfc;
	len = 6;
    }
    else
	throw DATA_CONVERSION();

    //
    // Write continuation bytes (6 free bits each)
    //
    while(len > 1)
    {
	out[--len] = 0x80 | (value & 0x3f);
	value >>= 6;
    }

    //
    // Write the rest
    //
    first |= value;
    *out = (Octet)first;
}

WChar
CodeSetUtil::utf8ToUnicode(Octet*& in)
    throw(DATA_CONVERSION)
{
    Octet first = *in++;
 
    //
    // Direct mapping for characters < 0x80
    //
    if((first & 0x80) == 0)
	return first;

    ULong rem;
    WChar value;

    if((first & 0xe0) == 0xc0)
    {
	// 5 free bits
	rem = 1;
	value = first & 0x1f;
    }
    else if((first & 0xf0) == 0xe0)
    {
	// 4 free bits
	rem = 2;
	value = first & 0x0f;
    }
    else if((first & 0xf8) == 0xf0)
    {
	// 3 free bits;
	rem = 3;
	value = first & 0x07;
    }
    else if((first & 0xfc) == 0xf8)
    {
	// 2 free bits
	rem = 4;
	value = first & 0x03;
    }
    else if((first & 0xfe) == 0xfc)
    {
	// 1 free bit
	rem = 5;
	value = first & 0x01;
    }
    else
	throw DATA_CONVERSION();

    //
    // Check for 16 bit overflow
    //
    if(sizeof(WChar) < 4 && rem >= 3)
	throw DATA_CONVERSION();

    //
    // Get continuation bytes (6 free bits)
    //
    while(rem)
    {
	if((*in & 0xc0) != 0x80)
	    throw DATA_CONVERSION();

	value <<= 6;
	value |= *in++ & 0x3f;
	rem--;
    }

    return value;
}

namespace PortableInterceptor
{

// ----------------------------------------------------------------------
// Add code set information to IOR. This is done with an
// IORInterceptor. This will be called to establish an IOR template
// for each POA.
// ----------------------------------------------------------------------

class CodeSetIORInterceptor_impl : public PortableInterceptor::IORInterceptor
{
public:

    CodeSetIORInterceptor_impl()
    {
    }

    //
    // IDL to C++ Mapping
    //
    virtual char*
    name()
    {
        return string_dup("");
    }

    virtual void
    establish_components(PortableInterceptor::IORInfo_ptr info)
    {
        try
        {
            CONV_FRAME::CodeSetComponentInfo_var codeSetInfo =
                CodeSetUtil::createCodeSetComponentInfo();

            IOP::TaggedComponent component;
            component.tag = IOP::TAG_CODE_SETS;

            //
            // The Codec could be used here -- but that means that the
            // Any insertion/extraction operators would have to be
            // generated unnecessarily
            //
	    OCI::BufferImpl buf;
	    OutputStreamImpl out(&buf);
            out._OB_writeEndian();
            codeSetInfo -> _OB_marshal(&out);

            component.component_data.length(out._OB_pos());
            memcpy(component.component_data.get_buffer(),
		   buf.data(), buf.length());
            info -> add_ior_component(component);
        }
        catch(const INV_POLICY&)
        {
            return;
        }
    }
};

} // End namespace PortableInterceptor

void
CodeSetUtil::registerIORInterceptor(PIManager_ptr manager)
{
    PortableInterceptor::IORInterceptor_var interceptor =
        new PortableInterceptor::CodeSetIORInterceptor_impl;

    try
    {
	manager -> addIORInterceptor(interceptor);
    }
    catch(const PortableInterceptor::ORBInitInfo::DuplicateName&)
    {
	assert(false);
    }
}
