// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/OAInterface.h>
#include <OB/Except.h>
#include <OB/OCI.h>
#include <OB/ORBInstance.h>
#include <OB/Downcall.h>
#include <OB/RefCounted.h>
#include <OB/Stream.h>

#include <CollocatedServer.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::CollocatedServer_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::CollocatedServer_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// CollocatedServer constructor and destructor
// ----------------------------------------------------------------------

CollocatedServer::CollocatedServer(OAInterface_ptr oaInterface,
				   Server::ConcModel concModel)
    : Server(concModel),
      nextRequestId_(0),
      callMap_(13),
      destroy_(false),
      hold_(true),
      oaInterface_(OAInterface::_duplicate(oaInterface))
{
}

CollocatedServer::~CollocatedServer()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// CollocatedServer public member implementation
// ----------------------------------------------------------------------

void
CollocatedServer::destroy()
{
    OB_SYNCHRONIZED

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;

    //
    // Set the destroy flag
    //
    destroy_ = true;

    //
    // Set the status of all downcalls, and empty the call map
    //
    CallMap::Enumerator e = callMap_.keys();
    for(ULong i = 0 ; i < e.length() ; i++)
    {
	Downcall_var down;
	bool getResult = callMap_.get(e[i], down);
	assert(getResult);
	assert(down -> pending());
	down -> setFailureException(INITIALIZE(MinorORBDestroyed,
					       COMPLETED_NO));
    }
    callMap_.clear();

#ifdef HAVE_JTC
    //
    // Notify everyone of the state transition
    //
    notifyAll();
#endif
}

void
CollocatedServer::hold()
{
    OB_SYNCHRONIZED

    assert(!destroy_);
    hold_ = true;

#ifdef HAVE_JTC
    //
    // Notify everyone of the state transition
    //
    notifyAll();
#endif
}

void
CollocatedServer::activate()
{
    OB_SYNCHRONIZED

    assert(!destroy_);
    hold_ = false;

#ifdef HAVE_JTC
    //
    // Notify everyone of the state transition
    //
    notifyAll();
#endif
}

bool
CollocatedServer::send(Downcall_ptr down, bool)
    throw()
{
#ifdef HAVE_JTC
    //
    // We need a state monitor if the request is dispatched in a
    // separate thread.
    //
    // TODO: If we had a method to query for whether a separate
    // dispatch thread is used, we could avoid initializing the state
    // monitor when it is not needed.
    //
    down -> initStateMonitor();
#endif

    Upcall_var up;

    {
        OB_SYNCHRONIZED

        //
        // First check whether we're destroyed or on hold
        //
#ifdef HAVE_JTC
	
        while(hold_ && !destroy_)
        {
	    try
	    {
	        wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
        }
	
        if(destroy_)
        {
	    down -> setFailureException(TRANSIENT(0, COMPLETED_NO));
	    return true;
        }

#else

        if(hold_)
        {
	    down -> setFailureException(BAD_INV_ORDER());
	    return true;
        }

#endif
    
        //
        // Collect the Upcall data
        //
        const OCI::ProfileInfo& profileInfo =
	    down -> profileInfo_nodup() -> value;
        ULong reqId = down -> requestId();
        const char* op = down -> operation();
        OutputStream_ptr out = down -> output_nodup();
        OCI::Buffer_var buf = new OCI::Buffer;
        buf -> consume(out -> _OB_buffer_nodup());
        const IOP::ServiceContextList& requestSCL = down -> getRequestSCL();
    
        if(down -> responseExpected())
        {
	    //
	    // Put the Downcall in the call map
	    //
	    callMap_.put(reqId, Downcall::_duplicate(down));

	    //
	    // From here on, we consider the Downcall as pending
	    //
	    down -> setPending();
	
	    //
	    // Note: For efficiency reasons, the Upcall constructor
	    // assumes ownership for all parameters marked with "/**/"
	    //
	    up = oaInterface_ -> createUpcall(
	        this,
	        /**/ new OCI::ProfileInfo(profileInfo),
	        /**/ OCI::TransportInfo::_nil(),
	        reqId,
	        /**/ string_dup(op),
	        /**/ new InputStream(buf, 0, false),
	        /**/ new IOP::ServiceContextList(requestSCL));
        }
        else
        {
	    //
	    // This is a oneway call, and if there was no exception so
	    // far, everything is fine
	    //
	    down -> setNoException(InputStream::_nil());
	
	    //
	    // Note: For efficiency reasons, the Upcall constructor
	    // assumes ownership for all parameters marked with "/**/"
	    //
	    up = oaInterface_ -> createUpcall(
	        UpcallReturn::_nil(),
	        /**/ new OCI::ProfileInfo(profileInfo),
	        /**/ OCI::TransportInfo::_nil(),
	        reqId,
	        /**/ string_dup(op),
	        /**/ new InputStream(buf, 0, false),
	        /**/ new IOP::ServiceContextList(requestSCL));
        }
    }

    //
    // Invoke the upcall
    //
    // Note: This must be done *outside* the synchronization, so that
    // nested method invocations are possible.
    //
    up -> invoke();

#ifdef HAVE_JTC

    if(down -> responseExpected())
    {
	try
	{
	    //
	    // Even though the downcall has definitely been "sent" at
	    // this point, we call waitUntilSent() in order to find
	    // out whether it's safe already to use this downcall
	    // object.
	    //
	    return down -> waitUntilSent(false);
	}
	catch(const SystemException&)
	{
	    assert(false);
	    return false;
	}
    }
    else
	return true;

#else

    //
    // If we don't have multiple threads, the call must have completed
    // by now
    //
    if(down -> responseExpected())
        assert(!down -> pending());

    return true;

#endif
}

bool
CollocatedServer::receive(Downcall_ptr down, bool block)
    throw()
{
#ifdef HAVE_JTC

    //
    // Try to receive the reply
    //
    try
    {
	//
	// TODO: If we had a method to query for whether a separate
	// dispatch thread is used, we could avoid calling
	// waitUntilCompleted() if no such separate thread is used,
	// and query the state directly instead.
	//
	return down -> waitUntilCompleted(block);
    }
    catch(const SystemException& ex)
    {
	JTCSynchronized synchronized(*this);
	callMap_.remove(down -> requestId());
	down -> setFailureException(ex);
	return true;
    }

#else

    assert(!down -> pending());
    return true;

#endif
}

bool
CollocatedServer::sendReceive(Downcall_ptr down)
    throw()
{
    send(down, true);
    return receive(down, true);
}

ULong
CollocatedServer::requestId()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(nextRequestIdMutex_);
#endif

    return nextRequestId_++;
}

OCI::ProfileInfoSeq*
CollocatedServer::getUsableProfiles(const OCI::IOR& ior,
				    const PolicyList& policies)
{
    return oaInterface_ -> getUsableProfiles(ior, policies);
}

void
CollocatedServer::upcallBeginReply(Upcall_ptr upcall,
				   const IOP::ServiceContextList& replySCL)
{
    upcall -> createOutputStream(0);
    
    if(replySCL.length() > 0)
    {
	OB_SYNCHRONIZED

	Downcall_var down;
	bool ok = callMap_.get(upcall -> requestId(), down);
	if(ok) // Might be false if the request timed out or destroyed
	    down -> setReplySCL(replySCL);
    }
}

void
CollocatedServer::upcallEndReply(Upcall_ptr upcall)
{
    OB_SYNCHRONIZED

    Downcall_var down;
    bool ok = callMap_.get(upcall -> requestId(), down);
    if(ok) // Might be false if the request timed out or destroyed
    {
	OutputStream_ptr out = upcall -> output_nodup();
	OCI::Buffer_var buf = new OCI::Buffer;
	buf -> consume(out -> _OB_buffer_nodup());
	InputStream_var in = new InputStream(buf, 0, false);
	down -> setNoException(in);
	callMap_.remove(down -> requestId());
    }
}

void
CollocatedServer::upcallBeginUserException(Upcall_ptr upcall,
					   const IOP::ServiceContextList&
					   replySCL)
{
    upcall -> createOutputStream(0);
    
    if(replySCL.length() > 0)
    {
	OB_SYNCHRONIZED

	Downcall_var down;
	bool ok = callMap_.get(upcall -> requestId(), down);
	if(ok) // Might be false if the request timed out or destroyed
	    down -> setReplySCL(replySCL);
    }
}

void
CollocatedServer::upcallEndUserException(Upcall_ptr upcall)
{
    OB_SYNCHRONIZED

    Downcall_var down;
    bool ok = callMap_.get(upcall -> requestId(), down);
    if(ok) // Might be false if the request timed out or destroyed
    {
	OutputStream_ptr out = upcall -> output_nodup();
	OCI::Buffer_var buf = new OCI::Buffer;
	buf -> consume(out -> _OB_buffer_nodup());
	InputStream_var in = new InputStream(buf, 0, false);
	down -> setUserException(in);
	callMap_.remove(down -> requestId());
    }
}

void
CollocatedServer::upcallUserException(Upcall_ptr upcall,
				      const UserException& ex,
				      const IOP::ServiceContextList&
				      replySCL)
{
    //
    // We marshal to preserve 100% location transparency. If we would
    // set the exception in the Downcall directly as shown below, then
    // we wouldn't get an UNKNOWN exception if we're calling from the
    // DII without setting the exception TypeCode.
    //

/*
    OB_SYNCHRONIZED

    Downcall_var down;
    bool ok = callMap_.get(upcall -> requestId(), down);
    if(ok) // Might be false if the request timed out or destroyed
    {
        if(replySCL.length() > 0)
            down -> setReplySCL(replySCL);
        down -> setUserException(ex);
	callMap_.remove(down -> requestId());
    }
*/

    upcallBeginUserException(upcall, replySCL);
    OutputStream_ptr out = upcall -> output_nodup();
    try
    {
	ex._OB_marshal(out);
    }
    catch(CORBA::SystemException& ex)
    {
	upcall -> marshalEx(ex);
    }
    upcallEndUserException(upcall);
}

void
CollocatedServer::upcallSystemException(Upcall_ptr upcall,
					const SystemException& ex,
					const IOP::ServiceContextList&
					replySCL)
{
    OB_SYNCHRONIZED

    Downcall_var down;
    bool ok = callMap_.get(upcall -> requestId(), down);
    if(ok) // Might be false if the request timed out or destroyed
    {
	if(replySCL.length() > 0)
	    down -> setReplySCL(replySCL);
	down -> setSystemException(ex);
	callMap_.remove(down -> requestId());
    }
}

void
CollocatedServer::upcallForward(Upcall_ptr upcall,
				RefCountIOR_ptr ior,
				bool perm,
				const IOP::ServiceContextList& replySCL)
{
    OB_SYNCHRONIZED

    Downcall_var down;
    bool ok = callMap_.get(upcall -> requestId(), down);
    if(ok) // Might be false if the request timed out or destroyed
    {
	if(replySCL.length() > 0)
	    down -> setReplySCL(replySCL);
	down -> setLocationForward(ior, perm);
	callMap_.remove(down -> requestId());
    }
}
