// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_CORBA_URL_H
#define OB_CORBA_URL_H

#include <OB/Basic.h>
#include <OB/Object.h>
#include <OB/OCITypes.h>

namespace OB 
{

//
// Base structure for representing "protocol:address_info" in 
// corbaloc/corbaname URLs. Specific protocol classes are derived
// from this type (see the end of this header file).
//
class URLAddr 
{
    
public:
    virtual const char* getProtocolName()=0;
    virtual void setProfile(IOP::IOR& ior, int profileNumber,
                            const OCI::ObjectKey& key)=0;
    virtual ~URLAddr();
};

typedef URLAddr*  URLAddrPtr;

// 
// Helper class for sequence of URLAddr*. Specialized
// destructor for correct memory management
// 

class UniqueClass {};

class URLAddrSeq : public OB::FixSeq < URLAddrPtr, UniqueClass >
{
    // 
    // no copies or assignment for URLAddrSeq. 
    // 
    URLAddrSeq(const URLAddrSeq& );
    URLAddrSeq& operator=(const URLAddrSeq&);
    
public:
    URLAddrSeq() {};

    // 
    // Delete the sequence, deleting each element
    // 
    ~URLAddrSeq()
    {
        for(CORBA::ULong p=0; p<length(); ++p)
        {
            delete operator[](p);
        }
    }
};

// The CorbaURL class represents CORBA URL forms and implements
// string_to_object for these URLS. Supported schemes are:
//
// OMG schemes: 
//         IOR, corbaloc, corbaname
//
// ORBacus schemes: 
//         file, relfile
//
// 
class CorbaURL 
{
    // 
    // No copies or assignment for CorbaURL
    // 
    CorbaURL(const CorbaURL&);
    CorbaURL& operator=(const CorbaURL&);

    const char* urlStr_;        // R/O copy of original URL string
    const char* fragmentStart_; // end of URL proper,
                                // corbaname # fragment start
    
    CORBA::String_var scheme_;   // URL scheme
    URLAddrSeq  addresses_;     // addresses and protocols
    OCI::ObjectKey key_;         // object key
    
    bool rirURL_;               // true if URL uses rir: protocol
    
    // 
    // parse the ':' terminated scheme from the URL , in lowercase
    // 
    void parseScheme(); 

    // 
    // parse a single corbaloc iiop: or rir: address
    // 
    URLAddr* parseSingleProtocolAddr(const char* addrStart,
                                            const char* addrEnd);
    // 
    // parse a set of corbaloc protocol addresses. return a char* pointer
    // to the end of the address portion of the URL
    // 
    const char* parseCORBALOCAddresses(); 

    // 
    // parse a corbaloc URL key, return char pointer past key
    // or null if end of string found
    // 
    const char* parseKey(const char* keyStart);
    
    // 
    // Conversions for specific URL schemes
    // 
    CORBA::Object_ptr IORToObject(ORBInstance_ptr objectFactory);
    CORBA::Object_ptr CORBALOCToObject(ORBInstance_ptr objectFactory);
    CORBA::Object_ptr CORBANAMEToObject(ORBInstance_ptr ORBInstance);

    CORBA::Object_ptr OBIIOPToObject(ORBInstance_ptr orb);
    CORBA::Object_ptr OBFileToObject(ORBInstance_ptr orb, 
				     const char* dirPrefix);

public:

    // 
    // create a CorbaURL from a URLString
    // 
    CorbaURL(const char* urlStr);
    
    // 
    // return an object reference for this URL
    // 
    CORBA::Object_ptr toObject(ORBInstance_ptr ORBInstance); 

    // 
    // return the URL scheme (no ":" at the end)
    // 
    const char* getScheme();

    // 
    // Utilities:
    // 
    // Generic string / char
    // 
    static char*  lowerStr(char* start);
    static char*  copyStr(const char* start, const char* end);
    static int convertHex(char ch);
    
    // 
    // corbaloc URL specific
    // 
    // 
    // remove URL escape sequences from a string. On return trueEnd contains a
    // pointer to the true end of the escaped URL string. This prevents
    // strings that end in %00 escapes from being misinterpreted.
    // 
    static char*  unescapeURL(char* start, const char* end, 
                              const char*& trueEnd);
    
    // 
    // parse valid integers from a URL string, with range
    // checking if minVal and maxVal are not equal to each
    // other.
    // 
    static int urlStrToInt(const char* start, const char* end,
                           int minVal = 0, 
                           int maxVal = 0);
    // 
    // convert a string to an object key. Object key will *not* be
    // null terminated
    // 
    OCI::ObjectKey& strToKey(const char* start,
                              const char* end, OCI::ObjectKey& key);

    // 
    // convert a key to a string
    // An exception will be raised if the key contains nulls and cannot
    // be represented by a null terminated string
    // 
    char* keyToStr(const OCI::ObjectKey& key, const char* prefixStr);

    // 
    // Corba URL Scheme Names
    // 
    // OMG Standard
    // 
    static const char iorPrefix_[];
    static const char corbalocPrefix_[];
    static const char corbanamePrefix_[];

    // 
    // ORBacus Extensions
    // 
    static const char filePrefix_[];
    static const char relfilePrefix_[];

    // 
    // corbaloc protocol table
    // 
    static const char iiopProt_[];
    static const char iiopProtColon_[];
    static const char rirProt_[];
};

// 
// OMG Standard URL protocols: iiop and rir
//
// "rir:" resolve_initial_reference protocol is a special case and
// does not have any additional address info
//
class RIRAddr : public URLAddr
{
public:
    RIRAddr(const char* addrStr, const char* addrEnd);

    // 
    // return protocol name "rir"
    // 
    const char* getProtocolName();

    // 
    // an rir profile can not be added to an IOR
    // this will raise an exception and should not
    // be called during normal operation
    // 
    void setProfile(IOP::IOR& ior, int profileNumber,
                            const OCI::ObjectKey& key);
};

//
// "iiop:" requires [iiop version], host, [port]
//
class IIOPAddr : public URLAddr
{
    // 
    // iiop version, host and port
    // 
    CORBA::Octet major_;
    CORBA::Octet minor_;
    CORBA::String_var host_;
    CORBA::UShort port_;

public:

    // 
    // iiop: protocol default port 2809
    // 
    static const CORBA::UShort DEFAULT_PORT;

    // 
    // construct an IIOPAddr from iiop: string
    // 
    IIOPAddr(const char* addrStr, const char* addrEnd);

    // 
    // get protocol name "iiop:"
    // 
    const char* getProtocolName();

    // 
    // set an iiop profile in an IOR
    // 
    void setProfile(IOP::IOR& ior, int profileNumber,
                            const OCI::ObjectKey& key);
};

//
// Additional protocol address structures added here, for Example:
// 
// class ATMAddr: public URLAddr
// {
//   ....Atm Specific Addressing information
// }

}; // end namespace OB


#endif





