// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/StubImpl.h>
#include <OB/Downcall.h>
#include <OB/Client.h>
#include <OB/NamedValue.h>
#include <OB/DII.h>
#include <OB/PIClient.h>
#include <OB/PIORBInit.h>
#include <OB/PIManager.h>
#include <OB/ORBInstance.h>
#include <OB/Stream.h>
#include <OB/RefCounted.h>

#include <CodeConverter.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(Downcall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(Downcall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

/*
void
OBDuplicate(PIDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(PIArgsDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIArgsDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(PIDIIDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIDIIDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
*/

void
OBDuplicate(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Downcall private and protected members
// ----------------------------------------------------------------------

void
Downcall::checkForException()
    throw(ExceptionBase)
{
    switch(state_)
    {
    case DowncallStateUserException:
	if(ex_) // Only raise if a user exception has been set
	    ex_ -> _raise();
	break;

    case DowncallStateSystemException:
        assert(ex_);
	ex_ -> _raise();
	break;

    case DowncallStateFailureException:
	assert(ex_);
	//
	// COMPILERFIX HPUX 64
	//
	// was throw FailureException(*SystemException::_downcast(ex_));
	//
        throw FailureException(*(SystemException::_downcast(ex_)));

    case DowncallStateForward:
	assert(!is_nil(forwardIOR_));
	throw LocationForward(forwardIOR_, false);

    case DowncallStateForwardPerm:
	assert(!is_nil(forwardIOR_));
	throw LocationForward(forwardIOR_, true);

    default:
	break;
    }
}

// ----------------------------------------------------------------------
// Downcall constructor and destructor
// ----------------------------------------------------------------------

Downcall::Downcall(ORBInstance_ptr orbInstance,
		   Client_ptr client,
		   RefCountProfileInfo_ptr profileInfo,
		   RefCountPolicyList_ptr policies,
		   const char* op,
		   bool resp)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      client_(Client::_duplicate(client)),
      profileInfo_(RefCountProfileInfo::_duplicate(profileInfo)),
      policies_(RefCountPolicyList::_duplicate(policies)),
      op_(op),
      responseExpected_(resp),
      state_(DowncallStateUnsent),
#ifdef HAVE_JTC
      stateMonitor_(0),
#endif
      ex_(0)
{
    //
    // Get the next request ID
    //
    reqId_ = client_ -> requestId();
}

Downcall::~Downcall()
{
    delete ex_;

#ifdef HAVE_JTC
    delete stateMonitor_;
#endif
}

// ----------------------------------------------------------------------
// Downcall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
Downcall::preMarshal()
    throw(ExceptionBase)
{
    emitter_ = client_ -> startDowncall(this, out_);
    checkForException();
    return out_;
}

void
Downcall::marshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex);
    checkForException();
    assert(false);
}

void
Downcall::postMarshal()
    throw(ExceptionBase)
{
}

void
Downcall::request()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    //
    // We could also use send() and receive() separately. But using
    // sendReceive() is more efficient.
    //
/*
    bool finished = emitter_ -> send(this, true);
    if(finished)
	checkForException();

    finished = emitter_ -> receive(this, true);
    assert(finished);
    checkForException();
*/

    bool finished = emitter_ -> sendReceive(this);
    assert(finished);
    checkForException();
}

void
Downcall::oneway()
    throw(ExceptionBase)
{
    assert(!responseExpected_);

    if(policies_ -> syncScope == Messaging::SYNC_WITH_TRANSPORT)
    {
	bool finished = emitter_ -> send(this, true);
	assert(finished);
	checkForException();
    }
    else
    {
	bool finished = emitter_ -> send(this, false);
	if(finished)
	    checkForException();
    }
}

void
Downcall::deferred()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> send(this, true);
    if(finished)
	checkForException();
}

void
Downcall::response()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> receive(this, true);
    assert(finished);
    checkForException();
}

bool
Downcall::poll()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> receive(this, false);
    if(finished)
    {
	checkForException();
	return state_ != DowncallStatePending;
    }
    else
	return false;
}

InputStream_ptr
Downcall::preUnmarshal()
    throw(ExceptionBase)
{
    return in_;
}

void
Downcall::unmarshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex);
    checkForException();
    assert(false);
}

void
Downcall::postUnmarshal()
    throw(ExceptionBase)
{
    //
    // If the result of this downcall is a user exception, but no user
    // exception could be unmarshalled, then use the system exception
    // UNKNOWN
    //
    if(state_ == DowncallStateUserException && !ex_)
	setSystemException(UNKNOWN(MinorUnknownUserException, COMPLETED_YES));

    checkForException();
}

char*
Downcall::unmarshalExceptionId()
{
    assert(state_ == DowncallStateUserException);
    ULong pos = in_ -> _OB_pos();
    String_var id = in_ -> read_string();
    in_ -> _OB_pos(pos);
    return id._retn();
}

void
Downcall::setPending()
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(responseExpected_);
    state_ = DowncallStatePending;

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setNoException(InputStream_ptr in)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    state_ = DowncallStateNoException;
    if(is_nil(in))
    {
	assert(!responseExpected_);
    }
    else
    {
	assert(responseExpected_);
	in_ = InputStream::_duplicate(in);
	in_ -> _OB_ORBInstance(orbInstance_);
	const CodeConverters* codeConverters = client_ -> codeConverters();
	in_ -> _OB_codeConverters(codeConverters,
				  profileInfo_ -> value.major == 1 &&
				  profileInfo_ -> value.minor < 2);
    }

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setUserException(InputStream_ptr in)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(!is_nil(in));
    assert(responseExpected_);
    state_ = DowncallStateUserException;
    in_ = InputStream::_duplicate(in);
    in_ -> _OB_ORBInstance(orbInstance_);
    const CodeConverters* codeConverters = client_ -> codeConverters();
    in_ -> _OB_codeConverters(codeConverters,
                              profileInfo_ -> value.major == 1 &&
                              profileInfo_ -> value.minor < 2);

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setUserException(UserException* ex, const char*)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex;

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setUserException(const UserException& ex)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex._OB_clone();

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify(); // TODO: Is notify() necessary here?
	delete sync;
    }
#endif    

}

void
Downcall::setSystemException(const SystemException& ex)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateSystemException;
    ex_ = ex._OB_clone();

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setFailureException(const SystemException& ex)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(!ex_);
    state_ = DowncallStateFailureException;
    ex_ = ex._OB_clone();

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

void
Downcall::setLocationForward(RefCountIOR_ptr ior, bool perm)
    throw()
{
#ifdef HAVE_JTC
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);
#endif

    assert(responseExpected_);
    assert(is_nil(forwardIOR_));
    if(perm)
	state_ = DowncallStateForwardPerm;
    else
	state_ = DowncallStateForward;
    forwardIOR_ = RefCountIOR::_duplicate(ior);

#ifdef HAVE_JTC
    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
#endif    
}

#ifdef HAVE_JTC

void
Downcall::initStateMonitor()
{
    assert(!stateMonitor_);
    stateMonitor_ = new JTCMonitor;
}

bool
Downcall::waitUntilSent(bool block)
{
    //
    // Get the timeout
    //
    Long t = policies_ -> requestTimeout;

    //
    // Yield if non-blocking or blocking with zero timeout
    //
    if(!block || (block && t == 0))
	JTCThread::yield();

    //
    // Wait for the desired state, taking the timeout and blocking
    // flag into account
    //
    assert(stateMonitor_);
    JTCSynchronized synchronized(*stateMonitor_);

    while(state_ == DowncallStateUnsent)
    {
	if(!block)
	    return false;

	try
	{
	    if(t < 0)
	    {
		stateMonitor_ -> wait();
	    }
	    else
	    {
		DowncallState oldState = state_;

		stateMonitor_ -> wait(t);

		if(state_ == oldState)
		    throw NO_RESPONSE(0, COMPLETED_MAYBE);
	    }
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // Only return true if the downcall has fully completed
    //
    return state_ != DowncallStatePending;
}

bool
Downcall::waitUntilCompleted(bool block)
{
    //
    // Get the timeout
    //
    Long t = policies_ -> requestTimeout;

    //
    // Yield if non-blocking or blocking with zero timeout
    //
    if(!block || (block && t == 0))
	JTCThread::yield();

    //
    // Wait for the desired state, taking the timeout and blocking
    // flag into account
    //
    assert(stateMonitor_);
    JTCSynchronized synchronized(*stateMonitor_);

    while(state_ == DowncallStateUnsent || state_ == DowncallStatePending)
    {
	if(!block)
	    return false;

	try
	{
	    if(t < 0)
	    {
		stateMonitor_ -> wait();
	    }
	    else
	    {
		DowncallState oldState = state_;

		stateMonitor_ -> wait(t);

		if(state_ == oldState)
		    throw NO_RESPONSE(0, COMPLETED_MAYBE);
	    }
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // The downcall has completed
    //
    return true;
}

#endif

// ----------------------------------------------------------------------
// PIDowncall private and protected members
// ----------------------------------------------------------------------

void
PIDowncall::checkForException()
    throw(ExceptionBase)
{
    //
    // If ex_ is set, but exId_ is not, then set it now
    //
    if(ex_ && !exId_.in())
	exId_ = ex_ -> _rep_id();

    if(!is_nil(requestInfo_))
    {
	switch(state_)
	{
	case DowncallStateUserException:
	    //
	    // Only invoke interceptor if a user exception has been
	    // set
	    //
	    if(ex_)
		piManager_ -> clientReceiveException(requestInfo_, false,
						     *ex_, exId_);
	    break;
	    
	case DowncallStateSystemException:
	    assert(ex_);
	    piManager_ -> clientReceiveException(requestInfo_, true,
						 *ex_, exId_);
	    break;
	    
	case DowncallStateFailureException:
	    try
	    {
		assert(ex_);
		piManager_ -> clientReceiveException(requestInfo_, true,
						     *ex_, exId_);
	    }
	    catch(const SystemException&)
	    {
		//
		// Ignore any exception translations for failure
		// exceptions
		//
	    }
	    break;
	    
	case DowncallStateForward:
	    assert(!is_nil(forwardIOR_));
	    piManager_ -> clientReceiveLocationForward(requestInfo_,
						       forwardIOR_,
						       false);
	    break;
	    
	case DowncallStateForwardPerm:
	    assert(!is_nil(forwardIOR_));
	    piManager_ -> clientReceiveLocationForward(requestInfo_,
						       forwardIOR_,
						       true);
	    break;

	default:
	    break;
	}
    }

    Downcall::checkForException();
}

// ----------------------------------------------------------------------
// PIDowncall constructor and destructor
// ----------------------------------------------------------------------

PIDowncall::PIDowncall(ORBInstance_ptr orbInstance,
		       Client_ptr client,
		       RefCountProfileInfo_ptr profileInfo,
		       RefCountPolicyList_ptr policies,
		       const char* op,
		       bool resp,
		       RefCountIOR_ptr IOR,
		       RefCountIOR_ptr origIOR)
    : Downcall(orbInstance, client, profileInfo, policies, op, resp),
      IOR_(RefCountIOR::_duplicate(IOR)),
      origIOR_(RefCountIOR::_duplicate(origIOR))
{
    //
    // Get the PI manager
    //
    piManager_ = orbInstance_ -> getPIManager();
}

PIDowncall::~PIDowncall()
{
}

// ----------------------------------------------------------------------
// PIDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIDowncall::preMarshal()
    throw(ExceptionBase)
{
    if(piManager_ -> haveClientInterceptors())
    {
	requestInfo_ = piManager_ ->
	    clientSendRequest(op_,
			      responseExpected_,
			      IOR_,
			      origIOR_,
			      profileInfo_,
			      policies_,
			      requestSCL_,
			      replySCL_);
    }

    return Downcall::preMarshal();
}

void
PIDowncall::postUnmarshal()
    throw(ExceptionBase)
{
    //
    // If the result of this downcall is a user exception, but no user
    // exception could be unmarshalled, then use the system exception
    // UNKNOWN, but keep the original exception ID
    //
    if(state_ == DowncallStateUserException && !ex_)
    {
	String_var id = unmarshalExceptionId();
	setSystemException(UNKNOWN(MinorUnknownUserException, COMPLETED_YES));
	exId_ = id._retn();
    }
    
    Downcall::postUnmarshal();
    
    if(!is_nil(requestInfo_))
    {
        if(responseExpected_)
            assert(state_ == DowncallStateNoException);
        else
            assert(state_ == DowncallStateUnsent ||
		   state_ == DowncallStateNoException);
	piManager_ -> clientReceiveReply(requestInfo_);
    }
}

void
PIDowncall::setUserException(UserException* ex, const char* exId)
    throw()
{
    Downcall::setUserException(ex, exId);
    exId_ = exId;
}

// ----------------------------------------------------------------------
// PIArgsDowncall constructor and destructor
// ----------------------------------------------------------------------

PIArgsDowncall::PIArgsDowncall(ORBInstance_ptr orbInstance,
			       Client_ptr client,
			       RefCountProfileInfo_ptr profileInfo,
			       RefCountPolicyList_ptr policies,
			       const char* op,
			       bool resp,
			       RefCountIOR_ptr IOR,
			       RefCountIOR_ptr origIOR,
			       ParameterDesc argDesc[],
			       ULong nargDesc,
			       ParameterDesc* retDesc,
			       TypeCode_ptr* exceptionTC,
			       ULong nexceptionTC)
    : PIDowncall(orbInstance, client, profileInfo, policies, op, resp,
		 IOR, origIOR),
      argDesc_(argDesc),
      nargDesc_(nargDesc),
      retDesc_(retDesc),
      exceptionTC_(exceptionTC),
      nexceptionTC_(nexceptionTC)
{
}

PIArgsDowncall::~PIArgsDowncall()
{
}

// ----------------------------------------------------------------------
// PIArgsDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIArgsDowncall::preMarshal()
    throw(ExceptionBase)
{
    if(piManager_ -> haveClientInterceptors())
    {
	requestInfo_ = piManager_ ->
	    clientSendRequest(op_,
			      responseExpected_,
			      IOR_,
			      origIOR_,
			      profileInfo_,
			      policies_,
			      requestSCL_,
			      replySCL_,
                              argDesc_,
			      nargDesc_,
                              retDesc_,
                              exceptionTC_,
			      nexceptionTC_);
    }

    return Downcall::preMarshal();
}

// ----------------------------------------------------------------------
// PIDIIDowncall constructor and destructor
// ----------------------------------------------------------------------

PIDIIDowncall::PIDIIDowncall(ORBInstance_ptr orbInstance,
			     Client_ptr client,
			     RefCountProfileInfo_ptr profileInfo,
			     RefCountPolicyList_ptr policies,
			     const char* op,
			     bool resp,
			     RefCountIOR_ptr IOR,
			     RefCountIOR_ptr origIOR,
			     NVList_ptr args,
			     NamedValue_ptr result,
			     ExceptionList_ptr exceptions)
    : PIDowncall(orbInstance, client, profileInfo, policies, op, resp,
		 IOR, origIOR),
      args_(NVList::_duplicate(args)),
      result_(NamedValue::_duplicate(result)),
      exceptionList_(ExceptionList::_duplicate(exceptions))
{
}

PIDIIDowncall::~PIDIIDowncall()
{
}

// ----------------------------------------------------------------------
// PIDIIDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIDIIDowncall::preMarshal()
    throw(ExceptionBase)
{
    if(piManager_ -> haveClientInterceptors())
    {
	requestInfo_ = piManager_ ->
	    clientSendRequest(op_,
			      responseExpected_,
			      IOR_,
			      origIOR_,
			      profileInfo_,
			      policies_,
			      requestSCL_,
			      replySCL_,
			      args_,
			      result_,
			      exceptionList_);
    }

    return Downcall::preMarshal();
}
