// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/StubImpl.h>
#include <OB/Stream.h>
#include <OB/Logger.h>
#include <OB/OBPolicies.h>
#include <OB/OCI.h>
#include <OB/RefCounted.h>
#include <OB/ORBInstance.h>
#include <OB/CoreTraceLevels.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPClient private and protected member implementations
// ----------------------------------------------------------------------

GIOPClientWorker_ptr
GIOPClient::getWorker(bool create, Long t)
{
    OB_SYNCHRONIZED

    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);
    
    if(is_nil(worker_) && create)
    {
	//
	// Trace connection attempt
	//
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    OCI::ConnectorInfo_var info = connector_ -> get_info();
	    String_var msg = string_dup("trying to establish connection\n");
	    msg += "timeout: ";
	    if(t >= 0)
	    {
		msg += t;
		msg += "ms\n";
	    }
	    else
		msg += "none\n";
	    String_var desc = info -> describe();
	    msg += desc;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("outgoing", msg);
	}
	
	//
	// Create new transport, using the connector
	//
	// TODO: In reactive mode, this is not ideal. A non-blocking
	// connect() operation should be added to the OCI. Then a
	// GIOPClientStarterReactive class should be written, that
	// registers with the Reactor in order to listen to write
	// events on the filedescriptor of the non-blocking Connector
	// (handle() must be added to the Connector). If it receives a
	// write event, it gets the Transport from the Connector by
	// calling an operation complete_connect() (must also be
	// added). Then it creates an GIOPClientWorkerReactive with
	// the newly created Transport.
	//
	// For symetry reasons, GIOPClientStarterBlocking and
	// GIOPClientStarterThreaded should also be added, even though
	// these classes only have a trivial functionality. Or perhaps
	// the GIOPClientStarterThreaded tries to connect() in the
	// backgound? Just an idea...
	//
	
	OCI::Transport_var transport;
	
	if(t >= 0)
	{
	    transport = connector_ -> connect_timeout((ULong)t);
	    
	    //
	    // Was there a timeout?
	    //
	    if(is_nil(transport))
		throw NO_RESPONSE(0, COMPLETED_NO);
	}
	else
	{
	    transport = connector_ -> connect();
	    assert(!is_nil(transport));
	}
	
	//
	// Create new worker
	//
	switch(concModel_)
	{
	case Blocking:
	    worker_ = new GIOPClientWorkerBlocking(orbInstance_, this,
						   transport, acmTimeout_);
	    break;
	    
	case Reactive:
	    worker_ = new GIOPClientWorkerReactive(orbInstance_, this,
						   transport, acmTimeout_);
	    break;
	    
#ifdef HAVE_JTC
	case Threaded:
	    worker_ = new GIOPClientWorkerThreaded(orbInstance_, this,
						   transport, acmTimeout_);
	    break;
#endif
	}

	//
	// Lazy initialization of codeSetSC_. We don't want to
	// initialize codeSetSC_ in the constructor, in order to keep
	// creation of GIOPClients, which might later on not be used,
	// as light-weight as possible.
	//
	if(!codeSetSC_.ptr())
	{
	    //
	    // Create CONV_FRAME::CodeSetContext
	    //
	    CONV_FRAME::CodeSetContext ctx;
	    const CodeConverters* conv = codeConverters();
	    ctx.char_data =
                conv -> outputCharConverter -> getTo() -> rgy_value;
	    if(conv -> wcharConverter)
		ctx.wchar_data =
		    conv -> wcharConverter -> getTo() -> rgy_value;
	    else
		ctx.wchar_data = 0;
	    
	    //
	    // Create encapsulation for CONV_FRAME::CodeSetContext
	    //
	    OCI::BufferImpl buf;
	    OutputStreamImpl out(&buf);
	    out._OB_writeEndian();
	    ctx._OB_marshal(&out);
	    
	    //
	    // Create service context containing the
	    // CONV_FRAME::CodeSetContext encapsulation
	    //
	    codeSetSC_ = new IOP::ServiceContext;
	    codeSetSC_ -> context_id = IOP::CodeSets;
	    codeSetSC_ -> context_data.length(out._OB_pos());
	    Octet* oct = codeSetSC_ -> context_data.get_buffer();
	    memcpy(oct, buf.data(), buf.length());
	}
    }
    
    return GIOPClientWorker::_duplicate(worker_);
}

// ----------------------------------------------------------------------
// GIOPClient constructor and destructor
// ----------------------------------------------------------------------

GIOPClient::GIOPClient(ORBInstance_ptr orbInstance,
		       OCI::Connector_ptr connector,
		       Client::ConcModel concModel,
		       const CodeConverters& conv,
		       ULong acmTimeout)
    : Client(concModel, conv),
      orbInstance_(ORBInstance::_duplicate(orbInstance)),
      nextRequestId_(0),
      connector_(OCI::Connector::_duplicate(connector)),
      acmTimeout_(acmTimeout),
      destroy_(false)
{
}

GIOPClient::~GIOPClient()
{
    assert(destroy_);
    assert(is_nil(worker_));
}

// ----------------------------------------------------------------------
// GIOPClient public member implementation
// ----------------------------------------------------------------------

void
GIOPClient::destroy()
{
    GIOPClientWorker_var worker;
    
    {
	OB_SYNCHRONIZED
	    
	//
	// Don't destroy twice
	//
	if(destroy_)
	    return;

	//
	// Set the destroy flag
	//
	destroy_ = true;

	//
	// Use a copy of the worker, and destroy the worker outside
	// the synchronization, to avoid deadlocks
	//
	worker = worker_._retn();
    }

    if(!is_nil(worker))
	worker -> destroy();
}

void
GIOPClient::removeWorker(GIOPClientWorker_ptr worker)
{
    OB_SYNCHRONIZED

    if(worker_.in() == worker)
	worker_ = GIOPClientWorker::_nil();
}

ULong
GIOPClient::requestId()
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(nextRequestIdMutex_);
#endif

    return nextRequestId_++;
}

OCI::ProfileInfoSeq*
GIOPClient::getUsableProfiles(const OCI::IOR& ior, const PolicyList& policies)
{
    //
    // Get all profiles usable for the connector
    //
    OCI::ProfileInfoSeq_var all =
	connector_ -> get_usable_profiles(ior, policies);
    
    //
    // Filter out profiles which would require a different code converter
    //
    OCI::ProfileInfoSeq_var result = new OCI::ProfileInfoSeq;
    for(ULong i = 0 ; i < all -> length() ; i++)
    {
	CodeConverters conv;
	CodeSetUtil::getCodeConverters(all[i], conv);
	if(codeConverters() -> equal(conv))
        {
	    CORBA::ULong resultLength  = result -> length();
	    result -> length(resultLength + 1);
	    result[resultLength] = all[i];
        }
    }

    return result._retn();
}

OCI::ConnectorInfo_ptr
GIOPClient::connectorInfo()
{
    return connector_ -> get_info();
}

OCI::TransportInfo_ptr
GIOPClient::transportInfo()
{
    //
    // Get the worker, but do not create a new one if there is none
    // available
    //
    GIOPClientWorker_var worker = getWorker(false, -1);

    if(is_nil(worker))
	return OCI::TransportInfo::_nil();
    
    OCI::Transport_var transport = worker -> transport();
    return transport -> get_info();
}

DowncallEmitter_ptr
GIOPClient::startDowncall(Downcall_ptr down, OutputStream_out out)
    throw()
{
    GIOPClientWorker_var worker;
    try
    {
	//
	// Get the worker, creating a new one if there is none
	// available
	//
	worker = getWorker(true, down -> policies_nodup() -> connectTimeout);
    }
    catch(const SystemException& ex)
    {
	assert(ex.completed() == COMPLETED_NO);
	down -> setFailureException(ex);
	return DowncallEmitter::_nil();
    }

    try
    {
	//
	// We only need to add a code set context if we're GIOP
	// version 1.1 or higher, and if no messages have been sent so
	// far
	//
	Octet major = down -> profileInfo_nodup() -> value.major;
	Octet minor = down -> profileInfo_nodup() -> value.minor;
	if(!worker -> messageSent() && (major > 1 || minor >= 1))
	{
	    CoreTraceLevels_var coreTraceLevels =
		orbInstance_ -> getCoreTraceLevels();
	    if(coreTraceLevels -> traceConnections() >= 2)
	    {
		const CodeConverters* conv = codeConverters();
		String_var msg = string_dup("sending transmission code sets");
		msg += "\nchar code set: ";
		msg += conv -> outputCharConverter -> getTo() -> description;
		msg += "\nwchar code set: ";
		if(conv -> wcharConverter)
		    msg += conv -> wcharConverter -> getTo() -> description;
		else
		    msg += "none";
		Logger_var logger = orbInstance_ -> getLogger();
		logger -> trace("outgoing", msg);
	    }

	    assert(codeSetSC_.ptr());
	    down -> addToRequestSCL(codeSetSC_);
	}

	const OCI::ProfileInfo& profileInfo =
	    down -> profileInfo_nodup() -> value;
	OCI::Buffer_var buf = new OCI::Buffer(12);
	buf -> pos(12);
	out = new OutputStream(buf, codeConverters(),
			       profileInfo.major == 1 &&
			       profileInfo.minor < 2);
	
	//
	// Create GIOP outgoing message
	//
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	
	//
	// Write header
	//
	outgoing.writeRequestHeader(down -> requestId(),
				    down -> operation(),
				    down -> responseExpected(),
				    down -> getRequestSCL());

	return worker._retn();
    }
    catch(const SystemException& ex)
    {
	assert(ex.completed() == COMPLETED_NO);
	down -> setFailureException(ex);
	return DowncallEmitter::_nil();
    }
}

bool
GIOPClient::equal(Client_ptr cl)
{
    GIOPClient_ptr client = dynamic_cast<GIOPClient_ptr>(cl);
    if(is_nil(client))
	return false;
    
    if(!connector_ -> equal(client -> connector_))
	return false;
    
    if(!codeConverters() -> equal(*client -> codeConverters()))
	return false;
    
    if(acmTimeout_ != client -> acmTimeout_)
	return false;
    
    return true;
}
