// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/StubImpl.h>
#include <OB/Stream.h>
#include <OB/OCI.h>
#include <OB/RefCounted.h>
#include <OB/ORBInstance.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/Properties.h>
#include <OB/NativeTypes.h>
#include <OB/Reactor.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

#ifdef HAVE_SYS_TIMEB_H
#   include <sys/timeb.h>
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

/*
void
OBDuplicate(OB::GIOPClientWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
*/

#ifdef HAVE_JTC

void
OBDuplicate(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

#endif // HAVE_JTC

// ----------------------------------------------------------------------
// GIOPClientWorker::UnsentMessage public member implementations
// ----------------------------------------------------------------------

GIOPClientWorker::UnsentMessage::UnsentMessage(OCI::Buffer_ptr b)
{
    buf = OCI::Buffer::_duplicate(b);
    buf -> pos(0);
}

GIOPClientWorker::UnsentMessage::UnsentMessage(Downcall_ptr d)
{
    down = Downcall::_duplicate(d);
    buf = OCI::Buffer::_duplicate(d -> output_nodup() -> _OB_buffer_nodup());
    buf -> pos(0);
}

// ----------------------------------------------------------------------
// GIOPClientWorker private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorker::logCloseConnection()
{
    if(!closeLogged_)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    OCI::TransportInfo_var info = transport_ -> get_info();
	    String_var msg = string_dup("closing connection\n");
	    String_var desc = info -> describe();
	    msg += desc;
//          msg += '\n';
//  	    String_var exMsg = exception_ -> _to_string();
//  	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("outgoing", msg);
	}
	
	closeLogged_ = true;
    }
}

void
GIOPClientWorker::addCloseConnection()
{
    if(minor_ < 2) // GIOP 1.0 and 1.1 clients don't send CloseConnection
	return;

    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream_var out = new OutputStream(buf);
    
    OCI::ProfileInfo profileInfo;
    profileInfo.major = 1;
    profileInfo.minor = minor_;
    
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
    outgoing.writeMessageHeader(GIOP::CloseConnection, false, 0);
    
    addUnsent(buf);
}

void
GIOPClientWorker::addMessageError()
{
    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream_var out = new OutputStream(buf);

    OCI::ProfileInfo profileInfo;
    profileInfo.major = 1;
    profileInfo.minor = minor_;

    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
    outgoing.writeMessageHeader(GIOP::MessageError, false, 0);

    addUnsent(buf);
}

void
GIOPClientWorker::addUnsent(OCI::Buffer_ptr buf)
{
    //
    // Update the timestamp
    //
    updateTimestamp();

    //
    // If we are in StateError or StateClosed, don't add. In this case
    // its necessary to release the buffer since add() is expected to
    // consume the buffer.
    //
    if(state_ == StateError || state_ == StateClosed)
	return;

    //
    // Add new message to the message buffers
    //
    unsent_.push_back(UnsentMessage(buf));
}

void
GIOPClientWorker::addUnsent(Downcall_ptr down)
{
    //
    // Update the timestamp
    //
    updateTimestamp();

    //
    // If a failure exception is set, don't add new downcall
    // objects. Just set the exception in the downcall.
    //
    if(exception_)
    {
	down -> setFailureException(*exception_);
	return;
    }

    //
    // If we are in StateClosing, StateError, or StateClosed, then why
    // is no exception set?
    //
    assert(state_ != StateClosing && state_ != StateError &&
	   state_ != StateClosed);

    //
    // Add the message header
    //
    try
    {
	//
	// Save stream position and rewind
	//
	OutputStream_ptr out = down -> output_nodup();
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	
	//
	// Write message header
	//
	const OCI::ProfileInfo& profileInfo =
	    down -> profileInfo_nodup() -> value;
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	outgoing.writeMessageHeader(GIOP::Request, false, pos - 12);
	
	//
	// Restore stream position
	//
	out -> _OB_pos(pos);
    }
    catch(const SystemException& ex)
    {
	assert(ex.completed() == COMPLETED_NO);
	down -> setFailureException(ex);
	return;
    }

    //
    // Update the GIOP minor version
    //
    Octet minor = down -> profileInfo_nodup() -> value.minor;
    if(minor > minor_)
	minor_ = minor;

    //
    // Add new message to the message buffers
    //
    unsent_.push_back(UnsentMessage(down));
}

void
GIOPClientWorker::moveFirstUnsentToPending()
{
    //
    // If there are no unsent messages, then there must be a failure
    // exception, otherwise it's a bug.
    //
    if(unsent_.empty())
    {
	assert(exception_);
	return;
    }

    //
    // Remove first downcall
    //
    Downcall_var down = unsent_.front().down._retn();
    unsent_.pop_front();
    messageSent_ = true;
    
    //
    // Only add to pending if a response is expected
    //
    if(!is_nil(down))
    {
	if(down -> responseExpected())
	{
	    down -> setPending();
	    pending_.push_front(down);
	}
	else
	{
	    down -> setNoException(InputStream::_nil());
	}
    }
}

Downcall_ptr
GIOPClientWorker::findAndRemovePending(ULong reqId)
{
    OB_STL::list<Downcall_var>::iterator iter;
    for(iter = pending_.begin() ; iter != pending_.end() ; iter++)
    {
        if((*iter) -> requestId() == reqId)
            break;
    }

    if(iter == pending_.end())
        return Downcall::_nil();

    Downcall_var downcall = *iter;
    pending_.erase(iter);
    return downcall._retn();
}

bool
GIOPClientWorker::exception(State state,
			    const CORBA::SystemException& ex,
			    bool notCompleted)
{
    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
	return false;

    //
    // It is not possible to transition backwards
    //
    if(state < state_)
	return false;

    //
    // Set the failure exception
    //
    if(!exception_)
    {
	//
	// Note: The completion status of ex is not taken into account
	// here, since the completion status of an exception raised by
	// low-level operations (like OCI send and receive methods) is
	// not applicable as the completion status of the downcall as
	// a whole. Therefore the completion status is passed
	// explicitly with the notCompleted parameter.
	//
	
	//
	// Save the exception, with a completion status of COMPLETED_NO
	//
	exception_ = SystemException::_downcast(ex._OB_clone());
	exception_ -> completed(COMPLETED_NO);
	
	//
	// Always use a completion status of NO for unsent requests
	//
	while(!unsent_.empty())
	{
	    if(!is_nil(unsent_.front().down))
		unsent_.front().down -> setFailureException(*exception_);
	    
	    unsent_.pop_front();
	}
	
	//
	// If notCompleted is true, use a completion status of
	// COMPLETED_NO for pending requests, otherwise use
	// COMPLETED_MAYBE
	//
	SystemException* except =
	    SystemException::_downcast(exception_ -> _OB_clone());
	if(!notCompleted)
	    except -> completed(COMPLETED_MAYBE);
	while(!pending_.empty())
	{
	    pending_.front() -> setFailureException(*except);
	    pending_.pop_front();
	}
	delete except;
    }

    switch(state)
    {
    case StateActive:

	//
	// There is nothing special to do
	//
	break;

    case StateClosing:

        //
	// Add a CloseConnection message
	//
	addCloseConnection();

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

	break;

    case StateError:

        //
	// Add a MessageError message
	//
	addMessageError();

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

	break;

    case StateClosed:

	//
	// Close the transport
	//
	logCloseConnection();
	transport_ -> close();

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

	break;
    }

    //
    // Update the state
    //
    state_ = state;
    return true;
}

void
GIOPClientWorker::execute()
    throw()
{
    assert(state_ == StateActive || state_ == StateClosing);

    //
    // Update the timestamp
    //
    updateTimestamp();

    bool complete;
    try
    {
	//
	// Provide the buffer to the incoming message and reset buf_
	// to allow for reading the next message. A return value of
	// false indicates that a message is not available (i.e., a
	// fragment was received).
	//
        complete = incoming_.consumeBuffer(buf_._retn());
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return;
    }

    //
    // Keep track of the highest GIOP minor version seen
    //
    Octet minor = incoming_.version().minor;
    if(minor > minor_)
	minor_ = minor;

    //
    // If the message is not complete yet, there is nothing to do
    //
    if(!complete)
	return;

    //
    // OK, message is complete - let's execute it
    //
    switch(incoming_.type())
    {
    case GIOP::Reply:
    {
	ULong reqId;
	GIOP::ReplyStatusType_1_2 status;
	IOP::ServiceContextList scl;
	
	try
	{
	    incoming_.readReplyHeader(reqId, status, scl);
	}
	catch(const SystemException& ex)
	{
	    exception(StateError, ex);
	    return;
	}
	
	Downcall_var down = findAndRemovePending(reqId);
	if(is_nil(down))
	{ 
	    //
	    // Request id is unknown
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownReqId, COMPLETED_MAYBE));
	    return;
	}
	
	down -> setReplySCL(scl);
	InputStream_var in = incoming_.input();
	
	switch(status)
	{
	case GIOP::NO_EXCEPTION:
	    down -> setNoException(in);
	    break;
	    
	case GIOP::USER_EXCEPTION:
	    down -> setUserException(in);
	    break;
	    
	case GIOP::SYSTEM_EXCEPTION:
	{
	    SystemException* ex = SystemException::_OB_unmarshal(in);
	    ex -> remote(true);
	    down -> setSystemException(*ex);
	    delete ex;
	    break;
	}
	
	case GIOP::LOCATION_FORWARD:
	{
	    IOP::IOR ior;
	    IOP::IOR::_OB_unmarshal(ior, in);
	    RefCountIOR_var refIOR = new RefCountIOR(ior);
	    down -> setLocationForward(refIOR, false);
	    break;
	}
	
	case GIOP::LOCATION_FORWARD_PERM:
	{
	    IOP::IOR ior;
	    IOP::IOR::_OB_unmarshal(ior, in);
	    RefCountIOR_var refIOR = new RefCountIOR(ior);
	    down -> setLocationForward(refIOR, true);
	    break;
	}
	
	case GIOP::NEEDS_ADDRESSING_MODE:
	    // TODO: implement
	    assert(false);
	    break;
	    
	default:
	    assert(false); // TODO: Shouldn't just assert
	    break;
	}

	break;
    }
	
    case GIOP::Request:
    case GIOP::LocateRequest:
    case GIOP::CancelRequest:
	
	//
	// These messages may not be sent to a client
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	break;
	
    case GIOP::LocateReply:
	
	//
	// Since I am not sending LocateRequest messages, LocateReply
	// may not be sent to this client
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	break;
	    
    case GIOP::CloseConnection:

	//
	// If the peer closes the connection, all outstanding requests
	// can safely be reissued. Thus we send all of them a
	// TRANSIENT exception with a completion status of
	// COMPLETED_NO. This is done by calling exception() with the
	// notCompleted parameter set to true.
	//
	exception(StateClosed,
		  TRANSIENT(MinorCloseConnection, COMPLETED_MAYBE), true);
	break;
	
    case GIOP::MessageError:
	
	//
	// How should we react to MessageError messages? We just send
	// a MessageError back, and close the connection.
	//
	exception(StateError,
		  COMM_FAILURE(MinorMessageError, COMPLETED_MAYBE));
	break;
	
    case GIOP::Fragment:
	
	//
        // At this point there should be no fragments, only complete
        // messages
	//
	assert(false);
	break;
    }
}

void
GIOPClientWorker::updateTimestamp()
{
    //
    // Only update if an active connection management timeout is set
    //
    if(acmTimeout_ > 0)
    {
#ifdef WIN32
	struct _timeb tb;
	_ftime(&tb);
	timestamp_ = tb.time;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	timestamp_ = tv.tv_sec;
#endif
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorker::GIOPClientWorker(ORBInstance_ptr orbInstance,
				   GIOPClient_ptr client,
				   OCI::Transport_ptr transport,
				   ULong acmTimeout)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      client_(GIOPClient::_duplicate(client)),
      transport_(OCI::Transport::_duplicate(transport)),
      incoming_(orbInstance),
      messageSent_(false),
      minor_(0),
      closeLogged_(false),
      acmTimeout_(acmTimeout),
      state_(StateActive),
      exception_(0)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.client_shutdown_timeout");
    if(value)
	shutdownTimeout_ = atoi(value);
    else
	shutdownTimeout_ = 2;

    //
    // Trace new outgoing connection
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	OCI::TransportInfo_var info = transport_ -> get_info();
	String_var msg = string_dup("new connection\n");
	String_var desc = info -> describe();
	msg += desc;
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("outgoing", msg);
    }

    //
    // Update the active connection management timestamp
    //
    updateTimestamp();
}

GIOPClientWorker::~GIOPClientWorker()
{
    assert(state_ == StateClosed);
    delete exception_;
}

// ----------------------------------------------------------------------
// GIOPClientWorker public member implementations
// ----------------------------------------------------------------------

OCI::Transport_ptr
GIOPClientWorker::transport()
{
    return OCI::Transport::_duplicate(transport_);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerBlocking private and protected member implementations
// ----------------------------------------------------------------------

bool
GIOPClientWorkerBlocking::exception(State state,
				    const CORBA::SystemException& ex,
				    bool notCompleted)
{
    //
    // Let the base class change the state
    //
    if(!GIOPClientWorker::exception(state, ex, notCompleted))
	return false;

    //
    // If we are in StateClosing or StateError, we try to send all
    // remaining messages
    //
    if(state_ == StateClosing || state_ == StateError)
    {
	try
	{
	    while(!unsent_.empty())
	    {
		OCI::Buffer_ptr buf = unsent_.front().buf;
		
		transport_ -> send(buf, true);
		assert(buf -> is_full());
		
		moveFirstUnsentToPending();
	    }
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	}
    }

    //
    // If all messages have been sent and we are in StateClosing,
    // we gracefully shutdown the connection.
    //
    if(state_ == StateClosing)
    {
	assert(unsent_.empty());

	//
	// COMPILERFIX: witout this fix exceptions are not catched in
	// the following try/catch block.
	//
#if defined(__GNUC__) && defined(__OPTIMIZE__) 
	    volatile int dummy = 1;
	    dummy++;
#endif

	try
	{
	    logCloseConnection();
	    transport_ -> shutdown();
		
	    while(true)
	    {
		//
		// Setup the incoming message buffer
		//
		if(is_nil(buf_))
		    buf_ = new OCI::Buffer(12);
		
		if(buf_ -> pos() < 12)
		{
		    if(shutdownTimeout_ > 0)
		    {
			//
			// Receive header, with timeout
			//
			transport_ -> receive_timeout(buf_,
						      shutdownTimeout_ * 1000);
			if(!buf_ -> is_full())
			    throw TRANSIENT(MinorForcedShutdown,
					    COMPLETED_MAYBE);
		    }
		    else
		    {
			//
			// Receive header, blocking
			//
			transport_ -> receive(buf_, true);
			assert(buf_ -> is_full());
		    }
		    
		    //
		    // Header is complete
		    //
		    incoming_.extractHeader(buf_);
		    buf_ -> realloc(12 + incoming_.size());
		}
		
		if(!buf_ -> is_full())
		{
		    if(shutdownTimeout_ > 0)
		    {
			//
			// Receive body, with timeout
			//
			transport_ -> receive_timeout(buf_,
						      shutdownTimeout_ * 1000);
			if(!buf_ -> is_full())
			    throw TRANSIENT(MinorForcedShutdown,
					    COMPLETED_MAYBE);
		    }
		    else
		    {
			//
			// Receive body, blocking
			//
			transport_ -> receive(buf_, true);
			assert(buf_ -> is_full());
		    }
		}
		
		//
		// Execute the message
		//
		execute();
	    }
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	}
    }
    
    //
    // If all messages have been sent and we are in StateError, we
    // don't go through all the hula hoop with continuing to receive
    // messages until the peer closes. Instead, we just close the
    // connection, meaning that we can't be 100% sure that the peer
    // gets the last message.
    //
    if(state_ == StateError)
    {
	assert(unsent_.empty());

	exception(StateClosed,
		  TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
    }

    return true;
}

bool
GIOPClientWorkerBlocking::sendNoSync(Downcall_ptr down, bool block)
    throw()
{
    //
    // Add the downcall
    //
    addUnsent(down);

    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to send the request
    //
    while(down -> unsent())
    {
	OCI::Buffer_ptr buf = unsent_.front().buf;

	try
	{
	    if(block)
	    {
		if(t < 0)
		{
		    //
		    // Send buffer, blocking
		    //
		    transport_ -> send(buf, true);
		    assert(buf -> is_full());
		}
		else
		{
		    //
		    // Send buffer, with timeout
		    //
		    transport_ -> send_timeout(buf, (ULong)t);
		    if(!buf -> is_full())
			throw NO_RESPONSE(); // Timeout
		}
	    }
	    else
	    {
		//
		// Send buffer, non-blocking
		//
		transport_ -> send(buf, false);
		if(!buf -> is_full())
		    return true;
	    }
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	    return true;
	}

	//
	// Message was sent completely
	//
	moveFirstUnsentToPending();
    }

    return true;
}

bool
GIOPClientWorkerBlocking::receiveNoSync(Downcall_ptr down, bool block)
    throw()
{
    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to receive the reply
    //
    while(down -> unsent() || down -> pending())
    {
	//
	// Setup the incoming message buffer
	//
	if(is_nil(buf_))
	    buf_ = new OCI::Buffer(12);
	
	if(buf_ -> pos() < 12)
	{
	    try
	    {
		if(block)
		{
		    if(t < 0)
		    {
			//
			// Receive header, blocking
			//
			transport_ -> receive(buf_, true);
			assert(buf_ -> is_full());
		    }
		    else
		    {
			//
			// Receive header, with timeout
			//
			transport_ -> receive_timeout(buf_, (ULong)t);
			if(!buf_ -> is_full())
			    throw NO_RESPONSE(); // Timeout
		    }
		}
		else
		{
		    //
		    // Receive header, non-blocking
		    //
		    transport_ -> receive(buf_, false);
		    if(!buf_ -> is_full())
			return true;
		}
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateClosed, ex);
		return true;
	    }
	    
	    //
	    // Header is complete
	    //
	    try
	    {
		incoming_.extractHeader(buf_);
		buf_ -> realloc(12 + incoming_.size());
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateError, ex);
		return true;
	    }
	}

	if(buf_ -> pos() >= 12)
	{
	    if(!buf_ -> is_full())
	    {
		try
		{
		    if(block)
		    {
			if(t < 0)
			{
			    //
			    // Receive body, blocking
			    //
			    transport_ -> receive(buf_, true);
			    assert(buf_ -> is_full());
			}
			else
			{
			    //
			    // Receive body, with timeout
			    //
			    transport_ -> receive_timeout(buf_, (ULong)t);
			    if(!buf_ -> is_full())
				throw NO_RESPONSE(); // Timeout
			}
		    }
		    else
		    {
			//
			// Receive body, non-blocking
			//
			transport_ -> receive(buf_, false);
			if(!buf_ -> is_full())
			    return true;
		    }
		}
		catch(const SystemException& ex)
		{
		    exception(StateClosed, ex);
		    return true;
		}
	    }
	    
	    //
	    // Execute the message
	    //
	    execute();
	}
    }

    return true;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerBlocking constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerBlocking::GIOPClientWorkerBlocking(
    ORBInstance_ptr orbInstance,
    GIOPClient_ptr client,
    OCI::Transport_ptr transport,
    ULong acmTimeout)
    : GIOPClientWorker(orbInstance, client, transport, acmTimeout)
{
}

GIOPClientWorkerBlocking::~GIOPClientWorkerBlocking()
{
}

// ----------------------------------------------------------------------
// GIOPClientWorkerBlocking public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerBlocking::destroy()
{
    OB_SYNCHRONIZED

    exception(GIOPClientWorker::StateClosing,
	      INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

bool
GIOPClientWorkerBlocking::send(Downcall_ptr down, bool block)
    throw()
{
    OB_SYNCHRONIZED

    return sendNoSync(down, block);
}

bool
GIOPClientWorkerBlocking::receive(Downcall_ptr down, bool block)
    throw()
{
    OB_SYNCHRONIZED

    return receiveNoSync(down, block);
}

bool
GIOPClientWorkerBlocking::sendReceive(Downcall_ptr down)
    throw()
{
    OB_SYNCHRONIZED

    sendNoSync(down, true);
    return receiveNoSync(down, true);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::updateReactor()
{
    Mask mask = 0;

    if(state_ == StateActive || state_ == StateClosing)
	mask |= EventRead;

    if(state_ != StateClosed && !unsent_.empty())
	mask |= EventWrite;

    if(mask_ != mask)
    {
	mask_ = mask;

	if(mask_)
	{
	    OCI::Handle handle = transport_ -> handle();
	    assert(handle >= 0);
	    reactor_ -> register_handler(this, mask_, TypeClient, handle);
	}
	else
	{
	    reactor_ -> unregister_handler(this);
	}
    }
}

void
GIOPClientWorkerReactive::addUnsent(OCI::Buffer_ptr buf)
{
    GIOPClientWorker::addUnsent(buf);
    updateReactor();
    assert(unsent_.empty() || mask_ & EventWrite);
}

void
GIOPClientWorkerReactive::addUnsent(Downcall_ptr down)
{
    GIOPClientWorker::addUnsent(down);
    updateReactor();
    assert(unsent_.empty() || mask_ & EventWrite);
}

bool
GIOPClientWorkerReactive::exception(State state,
				    const CORBA::SystemException& ex,
				    bool notCompleted)
{
    //
    // Let the base class change the state
    //
    if(!GIOPClientWorker::exception(state, ex, notCompleted))
	return false;

    //
    // If we are in StateClosing and all messages have been sent, we
    // gracefully shutdown the connection.
    //
    if(state_ == StateClosing && unsent_.empty())
    {
	logCloseConnection();
	transport_ -> shutdown();
	
	//
	// TODO: The following code is temporary. This must be
	// changed when we have reactor timers back.
	//
	/* Begin temporary code */
	//
	// If we have a shutdown timeout, we receive dummy data, until
	// we get an exception.
	//
	if(shutdownTimeout_ > 0)
	{
	    try
	    {
		while(true)
		{
		    if(is_nil(buf_))
			buf_ = new OCI::Buffer(12);
		    
		    if(buf_ -> pos() < 12)
		    {
			transport_ -> receive_timeout(buf_,
						      shutdownTimeout_ * 1000);
			if(!buf_ -> is_full())
			    throw TRANSIENT(MinorForcedShutdown,
					    COMPLETED_MAYBE);
		    
			incoming_.extractHeader(buf_);
			buf_ -> realloc(12 + incoming_.size());
		    }
		    
		    if(!buf_ -> is_full())
		    {
			transport_ -> receive_timeout(buf_,
						      shutdownTimeout_ * 1000);
			if(!buf_ -> is_full())
			    throw TRANSIENT(MinorForcedShutdown,
					    COMPLETED_MAYBE);
		    }
		    
		    execute();
		}
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateClosed, ex);
	    }
	}
	/* End temporary code */
    }

    //
    // This cannot happen, in StateError we have always have a
    // MessageError message at this point
    //
    assert(!(unsent_.empty() && state_ == StateError));

    //
    // Update reactor registration
    //
    updateReactor();
    return true;
}

bool
GIOPClientWorkerReactive::receiveNoSync(Downcall_ptr down, bool block)
    throw()
{
    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to receive the reply
    //
    try
    {
	while(down -> unsent() || down -> pending())
	{
	    if(block)
	    {
		if(t < 0)
		{
		    //
		    // Dispatch one event, blocking
		    //
		    reactor_ -> dispatch_one_event(-1);
		}
		else
		{
		    //
		    // Dispatch one event, with timeout
		    //
		    if(!reactor_ -> dispatch_one_event(t))
			throw NO_RESPONSE(); // Timeout
		}
	    }
	    else
	    {
		//
		// Dispatch one event, with zero timeout
		//
		if(!reactor_ -> dispatch_one_event(0))
		    return false;
	    }
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
    }

    return true;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerReactive::GIOPClientWorkerReactive(
    ORBInstance_ptr orbInstance,
    GIOPClient_ptr client,
    OCI::Transport_ptr transport,
    ULong acmTimeout)
    : GIOPClientWorker(orbInstance, client, transport, acmTimeout),
      reactor_(orbInstance -> getReactor()),
      mask_(0)
{
    updateReactor();
}

GIOPClientWorkerReactive::~GIOPClientWorkerReactive()
{
    assert(mask_ == 0);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::destroy()
{
    OB_SYNCHRONIZED

    exception(GIOPClientWorker::StateClosing,
	      INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

bool
GIOPClientWorkerReactive::send(Downcall_ptr down, bool block)
    throw()
{
    OB_SYNCHRONIZED

    //
    // Add the downcall
    //
    addUnsent(down);
    
    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to send the request
    //
    try
    {
	while(down -> unsent())
	{
	    if(block)
	    {
		if(t < 0)
		{
		    //
		    // Dispatch one event, blocking
		    //
		    reactor_ -> dispatch_one_event(-1);
		}
		else
		{
		    //
		    // Dispatch one event, with timeout
		    //
		    if(!reactor_ -> dispatch_one_event(t))
			throw NO_RESPONSE(); // Timeout
		}
	    }
	    else
	    {
		//
		// Dispatch one event, with zero timeout
		//
		if(!reactor_ -> dispatch_one_event(0))
		    return false;
	    }
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
    }

    if(down -> pending())
	return false;
    else
	return true;
}

bool
GIOPClientWorkerReactive::receive(Downcall_ptr down, bool block)
    throw()
{
    OB_SYNCHRONIZED

    return receiveNoSync(down, block);
}

bool
GIOPClientWorkerReactive::sendReceive(Downcall_ptr down)
    throw()
{
    OB_SYNCHRONIZED

    //
    // We do not need to call send(). Instead, we just add the
    // downcall here. This is faster, and the result of the downcall
    // will be picked up by receiveNoSync() anyway (including any
    // failures).
    //
    addUnsent(down);

    return receiveNoSync(down, true);
}

Mask
GIOPClientWorkerReactive::handleEvent(Mask mask)
{
    OB_SYNCHRONIZED
    
    Mask retMask = 0;

    //
    // Handle write events only if we are not in StateClosed
    //
    if((mask & EventWrite) && state_ != StateClosed && !unsent_.empty())
    {
	OCI::Buffer_ptr buf = unsent_.front().buf;

	//
	// Send buffer, non-blocking
	//
	try
	{
	    retMask |= EventWrite;
	    transport_ -> send(buf, false);
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	    return retMask;
	}
	
	if(buf -> is_full())
	{
	    //
	    // Message was sent completely
	    //
	    moveFirstUnsentToPending();
	}
	
	if(unsent_.empty() && state_ == StateClosing)
	{
	    //
	    // If all messages have been sent and we are in
	    // StateClosing, we gracefully shutdown the connection.
	    //
	    logCloseConnection();
	    transport_ -> shutdown();

	    //
	    // TODO: The following code is temporary. This must be
	    // changed when we have reactor timers back.
	    //
	    /* Begin temporary code */
	    //
	    // If we have a shutdown timeout, we receive dummy data,
	    // until we get an exception.
	    //
	    if(shutdownTimeout_ > 0)
	    {
		try
		{
		    while(true)
		    {
			if(is_nil(buf_))
			    buf_ = new OCI::Buffer(12);
			
			if(buf_ -> pos() < 12)
			{
			    retMask |= EventRead;
			    transport_ -> receive_timeout(
				buf_, shutdownTimeout_ * 1000);
			    if(!buf_ -> is_full())
			    {
				exception(StateClosed,
					  TRANSIENT(MinorForcedShutdown,
						    COMPLETED_MAYBE));
				return retMask;
			    }
			    
			    incoming_.extractHeader(buf_);
			    buf_ -> realloc(12 + incoming_.size());
			}
			
			if(!buf_ -> is_full())
			{
			    retMask |= EventRead;
			    transport_ -> receive_timeout(
				buf_, shutdownTimeout_ * 1000);
			    if(!buf_ -> is_full())
			    {
				exception(StateClosed,
					  TRANSIENT(MinorForcedShutdown,
						    COMPLETED_MAYBE));
				return retMask;
			    }
			}
			
			execute();
		    }
		}
		catch(const SystemException& ex)
		{
		    exception(StateClosed, ex);
		    return retMask;
		}
	    }
	    /* End temporary code */
	}

	if(unsent_.empty() && state_ == StateError)
	{
	    //
	    // If all messages have been sent and we are in
	    // StateError, we don't go through all the hula hoop with
	    // continuing to receive messages until the peer
	    // closes. Instead, we just close the connection, meaning
	    // that we can't be 100% sure that the peer gets the last
	    // message.
	    //
	    exception(StateClosed,
		      TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
	}
    }

    //
    // Handle read events only if we are in StateActive or
    // StateClosing. We must continue to read if we are in
    // StateClosing to make sure that no messages can get lost upon
    // close, and to make sure that CloseConnection messages from the
    // peer are processed.
    //
    if((mask & EventRead) &&
       (state_ == StateActive || state_ == StateClosing))
    {
	//
	// Setup the incoming message buffer if not already done
	//
	if(is_nil(buf_))
	    buf_ = new OCI::Buffer(12);
	
	if(buf_ -> pos() < 12)
	{
	    //
	    // Receive header, non-blocking
	    //
	    try
	    {
		retMask |= EventRead;
		transport_ -> receive(buf_, false);
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateClosed, ex);
		return retMask;
	    }
	    
	    //
	    // Header complete?
	    //
	    if(buf_ -> is_full())
	    {
		try
		{
		    incoming_.extractHeader(buf_);
		    buf_ -> realloc(12 + incoming_.size());
		}
		catch(const SystemException& ex)
		{
		    exception(StateError, ex);
		    return retMask;
		}
	    }
	}
	
	if(buf_ -> pos() >= 12)
	{
	    if(!buf_ -> is_full())
	    {
		//
		// Receive body, non-blocking
		//
		try
		{
		    retMask |= EventRead;
		    transport_ -> receive(buf_, false);
		}
		catch(const SystemException& ex)
		{
		    exception(StateClosed, ex);
		    return retMask;
		}
	    }
	    
	    //
	    // Execute message if buffer is full
	    //
	    if(buf_ -> is_full())
		execute();
	}
    }

    //
    // Update reactor registration
    //
    updateReactor();

    return retMask;
}

#ifdef HAVE_JTC

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::addUnsent(OCI::Buffer_ptr buf)
{
    GIOPClientWorker::addUnsent(buf);
    notifyAll();
}

void
GIOPClientWorkerThreaded::addUnsent(Downcall_ptr down)
{
    down -> initStateMonitor();
    GIOPClientWorker::addUnsent(down);
    notifyAll();
}

Downcall_ptr
GIOPClientWorkerThreaded::findAndRemovePending(ULong reqId)
{
    //
    // Wait until the first unsent message has really been moved to pending
    //
    while(true)
    {
	if(unsent_.empty())
	    break;
	
	Downcall_ptr d = unsent_.front().down;
	if(is_nil(d) || d -> requestId() != reqId)
	    break;

	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    Downcall_ptr down = GIOPClientWorker::findAndRemovePending(reqId);

    if(acmTimeout_ > 0 && unsent_.empty() && pending_.empty())
	notifyAll();

    return down;
}

bool
GIOPClientWorkerThreaded::exception(State state,
				    const CORBA::SystemException& ex,
				    bool notCompleted)
{
    //
    // Let the base class change the state
    //
    if(!GIOPClientWorker::exception(state, ex, notCompleted))
	return false;

    //
    // Notify about the state change
    //
    notifyAll();
    return true;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerThreaded::GIOPClientWorkerThreaded(
    ORBInstance_ptr orbInstance, 
    GIOPClient_ptr client,
    OCI::Transport_ptr transport,
    ULong acmTimeout)
    : GIOPClientWorker(orbInstance, client, transport, acmTimeout)
{
    try
    {
	//
	// Retrieve the thread group for the clients
	//
	JTCThreadGroupHandle group =
	    orbInstance_ -> getClientWorkerGroup();

	//
	// Start sender and receiver thread
	//
	senderThread_ = new SenderThread(group, this);
	senderThread_ -> start();
	receiverThread_ = new ReceiverThread(group, this);
	receiverThread_ -> start();
    }
    catch(const JTCException&)
    {
	transport_ -> close();
	state_ = StateClosed;
	_OB_setRef(0);
	throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
    }
}

GIOPClientWorkerThreaded::~GIOPClientWorkerThreaded()
{
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::destroy()
{
    JTCSynchronized synchronized(*this);

    exception(GIOPClientWorker::StateClosing,
	      INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

bool
GIOPClientWorkerThreaded::send(Downcall_ptr down, bool block)
    throw()
{
    {
	JTCSynchronized synchronized(*this);
	
	//
	// Add the downcall
	//
	addUnsent(down);
    }

    //
    // Try to send the request
    //
    try
    {
	return down -> waitUntilSent(block);
    }
    catch(const SystemException& ex)
    {
	JTCSynchronized synchronized(*this);
	exception(StateClosed, ex);
	return true;
    }
}

bool
GIOPClientWorkerThreaded::receive(Downcall_ptr down, bool block)
    throw()
{
    //
    // Try to receive the reply
    //
    try
    {
	return down -> waitUntilCompleted(block);
    }
    catch(const SystemException& ex)
    {
	JTCSynchronized synchronized(*this);
	exception(StateClosed, ex);
	return true;
    }
}

bool
GIOPClientWorkerThreaded::sendReceive(Downcall_ptr down)
    throw()
{
    {
	JTCSynchronized synchronized(*this);
	
	//
	// We do not need to call send(). Instead, we just add the
	// downcall here. This is faster, and the result of the
	// downcall will be picked up by receive anyway (including any
	// failures).
	//
	addUnsent(down);
    }

    return receive(down, true);
}

void
GIOPClientWorkerThreaded::senderRun()
{
    bool first = true;

    while(true)
    {
	//
	// Get unsent message
	//
	OCI::Buffer_ptr buf;
	
	{
	    JTCSynchronized synchronized(*this);
	    
	    //
	    // For efficiency reasons (to avoid having two
	    // synchronized scopes) the following has been moved from
	    // the end of the while loop to the beginning. The
	    // if(!first) makes sure that this is not called upon the
	    // first run of the while loop.
	    //
	    if(!first)
	    {
		//
		// Message was sent completely
		//
		moveFirstUnsentToPending();
		notifyAll();
	    }
	    else
		first = false;
	    
	    //
	    // Wait for destruction or message
	    //
	    while(true)
	    {
		if(state_ == StateClosed)
		{
		    //
		    // If our state is StateClosed, stop this thread
		    //
		    return;
		}
		if(unsent_.empty() && state_ == StateClosing)
		{
		    //
		    // If all messages have been sent and we are in
		    // StateClosing and there are no outstanding
		    // upcalls, we gracefully shutdown the connection.
		    //
		    logCloseConnection();
		    transport_ -> shutdown();
		    return;
		}
		else if(unsent_.empty() && state_ == StateError)
		{
		    //
		    // If all messages have been sent and we are in
		    // StateError, we don't go through all the hula
		    // hoop with continuing to receive messages until
		    // the peer closes. Instead, we just close the
		    // connection, meaning that we can't be 100% sure
		    // that the peer gets the last message.
		    //
		    exception(StateClosed, TRANSIENT(MinorForcedShutdown,
						     COMPLETED_MAYBE));
		    return;
		}
		else if(!unsent_.empty())
		{
		    //
		    // Got a new message
		    //
		    buf = unsent_.front().buf.in();
		    break;
		}
		else if(acmTimeout_ > 0 &&
			unsent_.empty() && pending_.empty())
		{		    
		    //
		    // Shut down if the time limit has been exceeded
		    //
		    ULong current;
#ifdef WIN32
		    struct _timeb tb;
		    _ftime(&tb);
		    current = tb.time;
#else
		    timeval tv;
		    gettimeofday(&tv, 0);
		    current = tv.tv_sec;
#endif
		    if(current >= timestamp_ + acmTimeout_)
		    {
			exception(StateClosing,
				  TRANSIENT(MinorActiveConnectionManagement,
					    COMPLETED_MAYBE));
			continue;
		    }
		}

		//
		// Wait for an event
		//
		try
		{
		    if(acmTimeout_ > 0 &&
		       unsent_.empty() && pending_.empty())
			wait(acmTimeout_ * 1000);
		    else
			wait();
		}
		catch(const JTCInterruptedException&)
		{
		}
	    }
	}
	
	//
	// Send buffer, blocking
	//
	try
	{
            transport_ -> send(buf, true);
            assert(buf -> is_full());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
    }
}

void
GIOPClientWorkerThreaded::receiverRun()
{
    while(true)
    {
	//
	// Setup the incoming message buffer
	//
	assert(is_nil(buf_));
	buf_ = new OCI::Buffer(12);
	
	//
	// Receive header, blocking
	//
	try
	{
	    transport_ -> receive(buf_, true);
	    assert(buf_ -> is_full());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
	    
	//
	// Header is complete
	//
	try
	{
	    incoming_.extractHeader(buf_);
	    buf_ -> realloc(12 + incoming_.size());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateError, ex);
	    return;
	}
	
	if(!buf_ -> is_full())
	{
	    //
	    // Receive body, blocking
	    //
	    try
	    {
		transport_ -> receive(buf_, true);
		assert(buf_ -> is_full());
	    }
	    catch(const SystemException& ex)
	    {
		JTCSynchronized synchronized(*this);
		exception(StateClosed, ex);
		return;
	    }
	}
	    
	{
	    JTCSynchronized synchronized(*this);

	    //
	    // If we are not in StateActive or StateClosing, stop this
	    // thread. We do *not* stop this thread if we are in
	    // StateClosing, since we must continue to read data from
	    // the Transport to make sure that no messages can get
	    // lost upon close, and to make sure that CloseConnection
	    // messages from the peer are processed.
	    //
	    if(state_ != StateActive && state_ != StateClosing)
		return;

	    //
	    // Execute the message
	    //
	    execute();
	}
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::SenderThread::run()
{
    try
    {
	worker_ -> senderRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    while(true)
    {
	try
	{
	    if(worker_ -> shutdownTimeout_ > 0)
	    {
		//
		// Join with the receiver thread, with timeout
		//
		worker_ -> receiverThread_ ->
		    join(worker_ -> shutdownTimeout_ * 1000);
	    }
	    else
	    {
		//
		// Join with the receiver thread, no timeout
		//
		worker_ -> receiverThread_ -> join();
	    }
	    
	    break;
	}
	catch(const JTCInterruptedException&)
	{
	    continue;
	}
    }
    
    //
    // We now close the connection actively, since it may still be
    // open under certain circumstances. (For example, if the receiver
    // thread has not terminated yet, or if the receiver thread sets
    // the state to StateError before termination.)
    //
    {
	JTCSynchronized synchronized(*worker_.in());
	worker_ -> exception(GIOPClientWorker::StateClosed,
			     TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPClientWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::ReceiverThread::run()
{
    try
    {
	worker_ -> receiverRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPClientWorkerThreaded::_nil();
}

#endif // HAVE_JTC
