// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_WORKER_H
#define OB_GIOP_SERVER_WORKER_H

#include <OB/Server_fwd.h>
#include <OB/OAInterface_fwd.h>
#include <OB/OCI_fwd.h>
#include <OB/ORBInstance_fwd.h>

#include <GIOPServerStarter_fwd.h>
#include <GIOPServerWorker_fwd.h>

#include <OB/Upcall.h>
#include <OB/EventHandlerBase.h>
#include <OB/list.h>

#include <GIOPMessage.h>

namespace OB
{

struct CodeConverters;

//
// The GIOPServerWorker class
//
class GIOPServerWorker : public UpcallReturn
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorker(const GIOPServerWorker&);
    void operator=(const GIOPServerWorker&);

protected:

    ORBInstance_var orbInstance_; // The ORBInstance
    GIOPServerStarter_var starter_; // The server starter
    OCI::Transport_var transport_; // The transport
    OAInterface_var oaInterface_; // The object adapter interface
    OCI::Buffer_var buf_; // The incoming message buffer
    GIOPIncomingMessage incoming_; // The incoming message
    CodeConverters* codeConverters_; // The code converters
    OB_STL::list<OCI::Buffer_var> unsent_; // The outgoing message buffers
    CORBA::ULong upcallCount_; // Number of outstanding upcalls
    CORBA::Octet minor_; // The highest minor version seen in peer messages
    bool closeLogged_; // True if connection closure was already logged
    bool closeConnection_; // True if there was a CloseConnection message
    CORBA::ULong shutdownTimeout_; // The shutdown timeout
    CORBA::ULong acmTimeout_; // The active connection management timeout
    CORBA::ULong timestamp_; // Timestamp for the last connection activity

    //
    // The worker's state
    //
public:
    enum State
    {
	StateActive,
	StateHolding,
	StateClosing,
	StateError,
	StateClosed
    };
protected:
    State state_;

    //
    // Emit a trace message when closing the connection
    //
    void logCloseConnection();

    //
    // Add a CloseConnection message
    // Add a MessageError message
    //
    void addCloseConnection();
    void addMessageError();

    //
    // Add outgoing message. Note: The OCI::Buffer_ptr is
    // consumed. See the comments in the implementation file.
    //
    virtual void add(OCI::Buffer_ptr);

    //
    // Add an upcall
    //
    void addUpcall(Upcall_ptr);

    //
    // Get the code converters
    //
    void getCodeConverters(const IOP::ServiceContextList& scl);

    //
    // Change the state of the worker due to an exception
    //
    void exception(State, const CORBA::SystemException&);

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

    //
    // Execute incoming message
    //
    Upcall_ptr execute()
	throw();

    //
    // Execute Request message
    //
    Upcall_ptr executeRequest()
	throw();

    //
    // Execute LocateRequest message
    //
    void executeLocate()
	throw();

    //
    // Update the timestamp for the last connection activity
    //
    void updateTimestamp();

public:

    GIOPServerWorker(ORBInstance_ptr, GIOPServerStarter_ptr,
		     OCI::Transport_ptr, OAInterface_ptr);
    virtual ~GIOPServerWorker();

    static inline GIOPServerWorker_ptr _duplicate(GIOPServerWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorker_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State) = 0;

    //
    // Called upon return from an upcall
    //
    virtual void upcallBeginReply(Upcall_ptr,
				  const IOP::ServiceContextList&);
    virtual void upcallEndReply(Upcall_ptr) = 0;
    virtual void upcallBeginUserException(Upcall_ptr,
					  const IOP::ServiceContextList&);
    virtual void upcallEndUserException(Upcall_ptr);
    virtual void upcallUserException(Upcall_ptr,
                                     const CORBA::UserException&,
                                     const IOP::ServiceContextList&);
    virtual void upcallSystemException(Upcall_ptr,
				       const CORBA::SystemException&,
				       const IOP::ServiceContextList&);
    virtual void upcallForward(Upcall_ptr, RefCountIOR_ptr, bool,
			       const IOP::ServiceContextList&);
};

//
// The GIOPServerWorkerReactive class
//
class GIOPServerWorkerReactive : public GIOPServerWorker,
				 public EventHandlerBase
                                 OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorkerReactive(const GIOPServerWorkerReactive&);
    void operator=(const GIOPServerWorkerReactive&);

protected:

    Reactor_var reactor_; // The reactor
    Mask mask_; // The registered mask for the event handler

    //
    // Update the reactor registration
    //
    void updateReactor();

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

public:

    GIOPServerWorkerReactive(ORBInstance_ptr,
			     GIOPServerStarterReactive_ptr,
			     OCI::Transport_ptr,
                             OAInterface_ptr);
    virtual ~GIOPServerWorkerReactive();

    static inline GIOPServerWorkerReactive_ptr
    _duplicate(GIOPServerWorkerReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerReactive_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Called upon return from an upcall
    //
    virtual void upcallEndReply(Upcall_ptr);

    //
    // Handle event
    //
    virtual Mask handleEvent(Mask);
};

#ifdef HAVE_JTC

//
// The GIOPServerWorkerThreaded class
//
class GIOPServerWorkerThreaded : public GIOPServerWorker, public JTCMonitor
{
protected:

    //
    // The sender thread
    //
    class SenderThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	SenderThread(JTCThreadGroupHandle& group,
		     GIOPServerWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~SenderThread() { }
	virtual void run();
    };
    friend class SenderThread;
    JTCThreadHandle senderThread_;

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	ReceiverThread(JTCThreadGroupHandle& group,
		       GIOPServerWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // Add outgoing message
    //
    virtual void add(OCI::Buffer_ptr);

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

public:

    GIOPServerWorkerThreaded(ORBInstance_ptr,
			     GIOPServerStarterThreaded_ptr,
			     OCI::Transport_ptr,
                             OAInterface_ptr);
    ~GIOPServerWorkerThreaded();

    static inline GIOPServerWorkerThreaded_ptr
    _duplicate(GIOPServerWorkerThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerThreaded_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Called upon return from an upcall
    //
    virtual void upcallEndReply(Upcall_ptr);

    //
    // Run method for receiver and sender thread
    //
    void senderRun();
    void receiverRun();
};

#endif // HAVE_JTC

} // End of namespace OB


namespace CORBA
{

inline void
release(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorker_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerReactive_ptr p)
{
    return p == 0;
}

#ifdef HAVE_JTC

inline void
release(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerThreaded_ptr p)
{
    return p == 0;
}

#endif // HAVE_JTC

} // End of namespace CORBA

#endif
