// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/IIOPConfig.h>
#include <OB/Properties.h>
#include <OB/Logger.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// IIOPConfig private and protected member implementations
// ----------------------------------------------------------------------

OCI::Param*
IIOPConfig::findParam(const char* name)
{
    for(ULong i = 0 ; i < params_.length() ; i++)
    {
        if(strcmp(params_[i].name, name) == 0)
            return &params_[i];
    }

    ULong len = params_.length() + 1;
    params_.length(len);
    params_[len - 1].name = string_dup(name);
    return &params_[len - 1];
}

void
IIOPConfig::parseHost(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        //
        // Parse hosts, allowing a space or comma as the delimiters
        //
        StringSeq hosts;
        const char* cp = value;
        while(*cp != '\0')
        {
            while(*cp != '\0' && (*cp == ' ' || *cp == ','))
                cp++;
            const char* start = cp;
            while(*cp != '\0' && *cp != ' ' && *cp != ',')
                cp++;
            ULong len = cp - start;
            if(len > 0)
            {
                hosts.length(hosts.length() + 1);
                hosts[hosts.length() - 1] = string_alloc(len);
                strncpy((hosts[hosts.length() - 1].inout()), start, len);
                hosts[hosts.length() - 1][len] = '\0';
            }
        }
        if(hosts.length() == 0)
        {
            String_var err = prop;
            err += ": no hosts specified";
            logger_ -> error(err);
            throw INITIALIZE();
        }
        OCI::Param* param = findParam("host");
        param -> value <<= hosts;
    }
}

void
IIOPConfig::parsePort(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        int port = atoi(value);
        if(port < 1 || port > 65535)
        {
            String_var err = prop;
            err += ": port must be between 1 and 65535";
            logger_ -> error(err);
            throw INITIALIZE();
        }
        OCI::Param* param = findParam("port");
        param -> value <<= (UShort)port;
    }
}

void
IIOPConfig::parseBacklog(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        int backlog = atoi(value);
        if(backlog < 0 || backlog > 65535)
        {
            String_var err = prop;
            err += ": backlog must be between 0 and 65535";
            logger_ -> error(err);
            throw INITIALIZE();
        }
        OCI::Param* param = findParam("backlog");
        param -> value <<= (UShort)backlog;
    }
}

void
IIOPConfig::parseBind(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        if(strlen(value) == 0)
        {
            String_var err = prop;
            err += ": no address specified";
            logger_ -> error(err);
            throw INITIALIZE();
        }
        OCI::Param* param = findParam("bind");
        param -> value <<= value;
    }
}

void
IIOPConfig::parseNumeric(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        bool numeric = (strcmp(value, "true") == 0);
        OCI::Param* param = findParam("numeric");
        param -> value <<= numeric;
    }
}

void
IIOPConfig::parseMultiProfile(const char* prop)
{
    const char* value = props_ -> getProperty(prop);
    if(value != 0)
    {
        bool multiProfile = (strcmp(value, "true") == 0);
        OCI::Param* param = findParam("multi_profile");
        param -> value <<= multiProfile;
    }
}

// ----------------------------------------------------------------------
// IIOPConfig constructor
// ----------------------------------------------------------------------

IIOPConfig::IIOPConfig(Properties_ptr props, Logger_ptr logger)
   : props_(Properties::_duplicate(props)),
     logger_(Logger::_duplicate(logger))
{
}

// ----------------------------------------------------------------------
// IIOPConfig public member implementation
// ----------------------------------------------------------------------

//
// Get the parameters for all properties that have been parsed
//
OCI::ParamSeq*
IIOPConfig::getParams()
{
    return new OCI::ParamSeq(params_);
}

//
// Parse the following deprecated properties:
//
// ooc.orb.oa.host
// ooc.orb.oa.numeric
//
void
IIOPConfig::parseDeprecatedDefaults()
{
    parseHost("ooc.orb.oa.host");
    parseNumeric("ooc.orb.oa.numeric");
}

//
// Parse the following deprecated properties:
//
// ooc.orb.poamanager.<name>.host
// ooc.orb.poamanager.<name>.numeric
// ooc.orb.poamanager.<name>.port
//
void
IIOPConfig::parseDeprecatedForName(const char* name)
{
    String_var rootKey = string_dup("ooc.orb.poamanager.");
    rootKey += name;
    rootKey += ".";
    String_var prop;

    prop = rootKey;
    prop += "host";
    parseHost(prop);
    prop = rootKey;
    prop += "numeric";
    parseNumeric(prop);
    prop = rootKey;
    prop += "port";
    parsePort(prop);
}

//
// Parse the following deprecated property:
//
// ooc.orb.oa.port
//
// This property generally only applies to the RootPOAManager.
//
void
IIOPConfig::parseDeprecatedDefaultPort()
{
    parsePort("ooc.orb.oa.port");
}

//
// Parse the following property:
//
// ooc.iiop.port
//
// This property generally only applies to the RootPOAManager.
//
void
IIOPConfig::parseDefaultPort()
{
    parsePort("ooc.iiop.port");
}

//
// Parse the following property:
//
// ooc.iiop.bind
//
// This property generally only applies to the RootPOAManager.
//
void
IIOPConfig::parseDefaultBind()
{
    parseBind("ooc.iiop.bind");
}

//
// Parse the following properties:
//
// ooc.iiop.backlog
// ooc.iiop.host
// ooc.iiop.multi_profile
// ooc.iiop.numeric
//
void
IIOPConfig::parseDefaults()
{
    parseHost("ooc.iiop.host");
    parseNumeric("ooc.iiop.numeric");
    parseMultiProfile("ooc.iiop.multi_profile");
    parseBacklog("ooc.iiop.backlog");
}

//
// Parse the following properties:
//
// ooc.iiop.acceptor.<name>.backlog
// ooc.iiop.acceptor.<name>.bind
// ooc.iiop.acceptor.<name>.host
// ooc.iiop.acceptor.<name>.multi_profile
// ooc.iiop.acceptor.<name>.numeric
// ooc.iiop.acceptor.<name>.port
//
void
IIOPConfig::parseForName(const char* name)
{
    String_var rootKey = string_dup("ooc.iiop.acceptor.");
    rootKey += name;
    rootKey += ".";
    String_var prop;

    prop = rootKey;
    prop += "host";
    parseHost(prop);
    prop = rootKey;
    prop += "numeric";
    parseNumeric(prop);
    prop = rootKey;
    prop += "port";
    parsePort(prop);
    prop = rootKey;
    prop += "backlog";
    parseBacklog(prop);
    prop = rootKey;
    prop += "bind";
    parseBind(prop);
    prop = rootKey;
    prop += "multi_profile";
    parseMultiProfile(prop);
}

//
// Parse all properties in the standard order of precedence
//
void
IIOPConfig::parseAll(const char* name)
{
    parseDeprecatedDefaults();
    if(strcmp(name, "RootPOAManager") == 0)
        parseDeprecatedDefaultPort();
    parseDefaults();
    if(strcmp(name, "RootPOAManager") == 0)
    {
        parseDefaultPort();
        parseDefaultBind();
    }
    parseDeprecatedForName(name);
    parseForName(name);
}
