// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_OCI_IIOP_IMPL_H
#define OB_OCI_IIOP_IMPL_H

#include <OB/Object.h>
#include <OB/Except.h>
#include <OB/OCI_IIOP.h>
#include <OB/Net.h>

#include <OB/list.h>

namespace OCI
{

namespace IIOP
{

//
// Declare classes
//
class Transport_impl;
class TransportInfo_impl;
class Acceptor_impl;
class AcceptorInfo_impl;
class Connector_impl;
class ConnectorInfo_impl;
class AccFactory_impl;
class AccFactoryInfo_impl;
class ConFactory_impl;
class ConFactoryInfo_impl;

//
// The Transport_impl class
//
class Transport_impl : public Transport
{
    //
    // Hide copy-constructor and assignment operator
    //
    Transport_impl(const Transport_impl&);
    void operator=(const Transport_impl&);

protected:

    Handle fd_; // The handle
    CORBA::Boolean block_; // Set to true if blocking IO is selected
    TransportInfo_impl* info_; // Transport information
    
    void setBlock(CORBA::Boolean); // Set blocking/non-blocking behavior

    //
    // The corresponding info class is a friend
    //
    friend class TransportInfo_impl;

public:

    Transport_impl(Connector_impl*, Handle);
    Transport_impl(Acceptor_impl*, Handle);
    virtual ~Transport_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual void close()
	throw(CORBA::SystemException);
    
    virtual void shutdown()
	throw(CORBA::SystemException);
    
    virtual void receive(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean receive_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual void receive_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual void send(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean send_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual void send_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual OCI::TransportInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Connector_impl class
//
class Connector_impl : public Connector
{
    //
    // Hide copy-constructor and assignment operator
    //
    Connector_impl(const Connector_impl&);
    void operator=(const Connector_impl&);

protected:

    CORBA::String_var host_; // The host
    CORBA::UShort port_; // The port
    Handle fd_; // The handle
    ConnectorInfo_impl* info_; // Connector information

    void close(); // Close the connector

    //
    // The corresponding info class is a friend
    //
    friend class ConnectorInfo_impl;

public:

    Connector_impl(const char*, CORBA::UShort, const ConnectCBSeq& cb);
    virtual ~Connector_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual OCI::Transport_ptr connect()
        throw(CORBA::SystemException);

    virtual OCI::Transport_ptr connect_timeout(CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::ProfileInfoSeq* get_usable_profiles(const IOR&,
						     const CORBA::PolicyList&)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean equal(Connector_ptr)
        throw(CORBA::SystemException);

    virtual OCI::ConnectorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Acceptor_impl class
//
class Acceptor_impl : public Acceptor
{
    //
    // Hide copy-constructor and assignment operator
    //
    Acceptor_impl(const Acceptor_impl&);
    void operator=(const Acceptor_impl&);

protected:

    CORBA::StringSeq hosts_; // The hosts
    bool multiProfile_; // Use multiple profiles?
    CORBA::UShort port_; // The port
    CORBA::UShort backlog_; // The listen backlog
    Handle fd_; // The handle
    struct sockaddr_in localAddress_; // The local address
    AcceptorInfo_impl* info_; // Acceptor info

    //
    // The corresponding info class is a friend
    //
    friend class AcceptorInfo_impl;

public:

    Acceptor_impl(const CORBA::StringSeq&, bool, CORBA::UShort, CORBA::UShort);
    Acceptor_impl(const char*, const CORBA::StringSeq&, bool, CORBA::UShort,
                  CORBA::UShort);
    virtual ~Acceptor_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual void close()
	throw(CORBA::SystemException);
    
    virtual void listen()
        throw(CORBA::SystemException);
    
    virtual OCI::Transport_ptr accept(CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual OCI::Transport_ptr connect_self()
        throw(CORBA::SystemException);
    
    virtual void add_profiles(const ProfileInfo&, IOR&)
        throw(CORBA::SystemException);

    virtual ProfileInfoSeq* get_local_profiles(const IOR&)
        throw(CORBA::SystemException);
    
    virtual OCI::AcceptorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The AccFactory_impl class
//
class AccFactory_impl : public AccFactory
{
    //
    // Hide copy-constructor and assignment operator
    //
    AccFactory_impl(const AccFactory_impl&);
    void operator=(const AccFactory_impl&);

    AccFactoryInfo_impl* info_; // AccFactory info

public:

    AccFactory_impl();
    virtual ~AccFactory_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }

    virtual OCI::AccFactoryInfo_ptr get_info()
        throw(CORBA::SystemException);

    virtual OCI::Acceptor_ptr create_acceptor(const OCI::ParamSeq&)
        throw(OCI::InvalidParam,
              CORBA::SystemException);
};

//
// The ConFactory_impl class
//
class ConFactory_impl : public ConFactory
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactory_impl(const ConFactory_impl&);
    void operator=(const ConFactory_impl&);

    ConFactoryInfo_impl* info_; // ConFactory info

public:

    ConFactory_impl();
    virtual ~ConFactory_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual OCI::ConnectorSeq* create_connectors(const IOR&,
						 const CORBA::PolicyList&)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean equivalent(const IOR&, const IOR&)
        throw(CORBA::SystemException);
    
    virtual CORBA::ULong hash(const IOR&, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual OCI::ConFactoryInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The TransportInfo_impl class
//
class TransportInfo_impl : public TransportInfo OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    TransportInfo_impl(const TransportInfo_impl&);
    void operator=(const TransportInfo_impl&);

    OCI::ConnectorInfo_var connectorInfo_; // The connector info
    OCI::AcceptorInfo_var acceptorInfo_; // The acceptor info
    OB_STL::list<OCI::CloseCB_var> closeCBs_; // All close callbacks
    Transport_impl* transport_; // The associated transport
    struct sockaddr_in addr_; // Local address and port
    struct sockaddr_in remoteAddr_; // Remote address and port

    //
    // Initialize data members
    //
    void init();

public:

    TransportInfo_impl(Transport_impl* transport,
				Connector_impl* connector);
    TransportInfo_impl(Transport_impl* transport,
				Acceptor_impl* acceptor);
    virtual ~TransportInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual OCI::ConnectorInfo_ptr connector_info()
        throw(CORBA::SystemException);
    
    virtual OCI::AcceptorInfo_ptr acceptor_info()
        throw(CORBA::SystemException);
    
    virtual void add_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort port()
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callCloseCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The ConnectorInfo_impl class
//
class ConnectorInfo_impl : public ConnectorInfo OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConnectorInfo_impl(const ConnectorInfo_impl&);
    void operator=(const ConnectorInfo_impl&);

    OB_STL::list<OCI::ConnectCB_var> connectCBs_; // All connect callbacks
    Connector_impl* connector_; // The associated connector

public:

    ConnectorInfo_impl(Connector_impl*, const OCI::ConnectCBSeq&);
    virtual ~ConnectorInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callConnectCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The AcceptorInfo_impl class
//
class AcceptorInfo_impl : public AcceptorInfo OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    AcceptorInfo_impl(const AcceptorInfo_impl&);
    void operator=(const AcceptorInfo_impl&);

    OB_STL::list<OCI::AcceptCB_var> acceptCBs_; // All accept callbacks
    Acceptor_impl* acceptor_; // The associated acceptor

public:
    
    AcceptorInfo_impl(Acceptor_impl*);
    virtual ~AcceptorInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual CORBA::StringSeq* hosts()
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callAcceptCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The AccFactoryInfo_impl class
//
class AccFactoryInfo_impl : public AccFactoryInfo
{
    //
    // Hide copy-constructor and assignment operator
    //
    AccFactoryInfo_impl(const AccFactoryInfo_impl&);
    void operator=(const AccFactoryInfo_impl&);

public:

    AccFactoryInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual char* describe()
        throw(CORBA::SystemException);
};

//
// The ConFactoryInfo_impl class
//
class ConFactoryInfo_impl : public ConFactoryInfo OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactoryInfo_impl(const ConFactoryInfo_impl&);
    void operator=(const ConFactoryInfo_impl&);

    OB_STL::list<OCI::ConnectCB_var> connectCBs_; // All connect callbacks

public:

    ConFactoryInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProtocolId id()
        throw(CORBA::SystemException)
    {
        return TAG_IIOP;
    }

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    OCI::ConnectCBSeq* _OB_getConnectCBSeq();
};

} // End of namespace IIOP

} // End of namespace OCI

#endif
