// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBControl.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/Properties.h>
#include <OB/Reactor.h>
#include <OB/Logger.h>

#include <POA_impl.h>
#include <POACurrent_impl.h>
#include <POAManagerFactory_impl.h>
#include <Dispatch_impl.h>

using namespace OB;
using namespace CORBA;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::ORBControl_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::ORBControl_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ORBControl private and protected member implementation
// ----------------------------------------------------------------------

void
ORBControl::completeServerShutdown()
    throw()
{
    OB_SYNCHRONIZED

    //
    // If the shutdown_ is false, or the server side has already
    // shutdown then do nothing
    //
    if(!shutdown_ || state_ == StateServerShutdown)
	return;

    assert(state_ != StateClientShutdown && state_ != StateDestroyed);

#ifdef HAVE_JTC
    //
    // If run was called then only the main thread may complete the
    // shutdown
    //
    assert(state_ == StateNotRunning || mainThread_ == JTCThreadId::self());
#endif

    //
    // Get the POAManagerFactory implementation
    //
    OBPortableServer::POAManagerFactory_var pmFactory =
	orbInstance_ -> getPOAManagerFactory();
    
    OBPortableServer::POAManagerFactory_impl* factory =
	dynamic_cast<OBPortableServer::POAManagerFactory_impl*>
	(pmFactory.in());

    //
    // Deactivate all of the POAManagers
    //
    factory -> _OB_deactivate();
    
    //
    // Continue to dispatch events until all server type event
    // handlers have unregistered.
    //
    while(!reactor_ -> dispatch(TypeServer))
	;

    //
    // Wait for all the threads in the server worker group to
    // terminate
    //
#ifdef HAVE_JTC
    JTCThreadGroupHandle group = orbInstance_ -> getServerWorkerGroup();
    JTCSynchronized groupSync(*group);
    while(group -> activeCount() > 0)
    {
	try
	{
	    group -> wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
#endif

    //
    // Get the DispatchStrategyFactory implementation and destroy
    // it. It must be destroyed here so that the thread pools get
    // destroyed before OCI::Current_impl gets destroyed by the
    // destruction of the Root POA. Otherwise, thread specific data
    // for the thread pool threads will not get released.
    //
    OB::DispatchStrategyFactory_var dsFactory = 
        orbInstance_ -> getDispatchStrategyFactory();

    DispatchStrategyFactory_impl* dsFactoryImpl =
        dynamic_cast<DispatchStrategyFactory_impl*> (dsFactory.in());

    dsFactoryImpl -> _OB_destroy();

    //
    // Mark the server side state as shutdown and notify any waiting
    // threads
    //
    state_ = StateServerShutdown;

    //
    // Destroy the root POA
    //
    if(!is_nil(rootPOA_))
    {
        rootPOA_ -> destroy(true, true);
        rootPOA_ = PortableServer::POA::_nil();
    }

#ifdef HAVE_JTC
    notifyAll();
#endif
}

void
ORBControl::validateState()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(state_ == StateServerShutdown || state_ == StateClientShutdown)
	throw BAD_INV_ORDER("", MinorShutdownCalled, COMPLETED_NO);
	
    if(state_ == StateNotRunning)
    {
#ifdef HAVE_JTC
	//
	// Remember the main thread id
	//
	mainThread_ = JTCThreadId::self();
#endif

	//
	// Set the state to StateRunning
	//
	state_ = StateRunning;
    }
}

#ifdef HAVE_JTC
void
ORBControl::blockServerShutdownComplete()
    throw()
{
    OB_SYNCHRONIZED

    //
    // Wait for the server side to shutdown. Note that the client side
    // shutting down or the state being destroyed also implies that
    // the server side has shutdown
    //
    while(state_ == StateRunning)
    {
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
}
#endif

// ----------------------------------------------------------------------
// ORBControl constructor and destructor
// ----------------------------------------------------------------------

ORBControl::ORBControl()
    : state_(StateNotRunning),
      shutdown_(false)
{
}

ORBControl::~ORBControl()
{
}

// ----------------------------------------------------------------------
// ORBControl public member implementation
// ----------------------------------------------------------------------

void
ORBControl::destroy()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    //
    // destroy() may not be called unless the client side has been
    // shutdown
    //
    assert(state_ == StateClientShutdown);
    state_ = StateDestroyed;

    //
    // Set the ORBInstance object to nil
    //
    orbInstance_ = ORBInstance::_nil();
}

void
ORBControl::setORBInstance(ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);

    //
    // Take the Reactor from the ORBInstance
    //
    reactor_ = orbInstance_ -> getReactor();
}

bool
ORBControl::workPending()
    throw(SystemException)
{
    validateState();

#ifdef HAVE_JTC
    //
    // If this is not the main thread then do nothing
    //
    if(mainThread_ != JTCThreadId::self())
	return false;
#endif

    //
    // Validate state will throw an exception if state is
    // ServerShutdown, ClientShutdown or Destroyed. Therefore if
    // shutdown_ is true, then a server side shutdown is pending.
    //
    if(shutdown_)
	return true;

    return reactor_ -> event_ready();
}

void
ORBControl::performWork()
    throw(SystemException)
{
    validateState();

#ifdef HAVE_JTC
    //
    // If this is not the main thread then do nothing
    //
    if(mainThread_ != JTCThreadId::self())
	return;
#endif

    reactor_ -> dispatch_one_event(0);

    completeServerShutdown();
}

void
ORBControl::run()
    throw(SystemException)
{
    validateState();
	
#ifdef HAVE_JTC
    if(mainThread_ != JTCThreadId::self())
    {
	//
	// If this is not the main thread then block until the ORB is
	// shutdown
	//
	blockServerShutdownComplete();
	return;
    }
#endif

    //
    // Run the event loop until the ORB server side has shutdown. Note
    // that the client side shutting down or the state being destroyed
    // also implies that the server side has shutdown
    //
    do
    {
	reactor_ -> dispatch(0);

	//
	// After this call state is either ShutdownClient, or Running
	//
	completeServerShutdown();
    }
    while(state_ == StateRunning);
}

//
// ASYNC SAFE if waitForCompletion == false
//
void
ORBControl::shutdownServer(bool waitForCompletion)
    throw(SystemException)
{
    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(state_ == StateServerShutdown || state_ == StateClientShutdown)
	throw BAD_INV_ORDER("", MinorShutdownCalled, COMPLETED_NO);

    //
    // If waitForCompletion is true then find out whether we're inside
    // a method invocation -- if so throw a BAD_INV_ORDER exception
    //
    if(waitForCompletion)
    {
	bool inInvocation = false;
	try
	{
	    InitialServiceManager_var initialServiceManager =
                orbInstance_ -> getInitialServiceManager();
	    Object_var o = initialServiceManager ->
                resolveInitialReferences("POACurrent");
	    PortableServer::Current_impl* current =
                dynamic_cast<PortableServer::Current_impl*>(o.in());
	    inInvocation = (current != 0 && current -> _OB_inUpcall());
	}
	catch(const ORB::InvalidName&)
	{
	}
	
	if(inInvocation)
	    throw BAD_INV_ORDER(MinorDestroyWouldBlock, COMPLETED_NO);
    }

    shutdown_ = true;

    //
    // Interrupt the reactor. This should be done immediately before
    // the return since this can cause the main loop to wake and
    // complete the shutdown (thus, for instance, destroying the
    // POAManagerFactory).
    //
    reactor_ -> interrupt_dispatch();

    //
    // waitForCompletion false? We're done.
    //
    if(!waitForCompletion)
	return;

#ifdef HAVE_JTC
    //
    // If run was called and this is not the main thread and
    // waitForCompletion is true then wait for the shutdown to
    // complete.
    //
    if(state_ == StateRunning && mainThread_ != JTCThreadId::self())
    {
	blockServerShutdownComplete();
	return;
    }
#endif

    //
    // This is the main thread -- complete the shutdown process
    //
    completeServerShutdown();
}

void
ORBControl::shutdownServerClient()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // If the ORB client side is already shutdown, then we're done
    //
    if(state_ == StateClientShutdown)
	return;

    //
    // First shutdown the server side, if necessary.
    //
    if(state_ != StateServerShutdown)
	shutdownServer(true);

    //
    // The server shutdown must have completed
    //
    assert(state_ == StateServerShutdown);

    //
    // Shutdown the client side. Continue to dispatch events until all
    // client type event handlers have unregistered.
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();
    clientManager -> destroy();

    while(!reactor_ -> dispatch(TypeClient))
	;

    //
    // Wait for all the threads in the client worker group to
    // terminate
    //
#ifdef HAVE_JTC
    JTCThreadGroupHandle group = orbInstance_ -> getClientWorkerGroup();
    JTCSynchronized groupSync(*group);
    while(group -> activeCount() > 0)
    {
	try
	{
	    group -> wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
#endif

    //
    // Mark the ORB is client side as shutdown and notify any waiters
    //
    state_ = StateClientShutdown;
#ifdef HAVE_JTC
    notifyAll();
#endif
}
    
void
ORBControl::initializeRootPOA(const char* serverName)
    throw(SystemException)
{
    String_var rootPOAName;
    if(serverName == 0)
    {
	rootPOAName = string_dup("_RootPOA");
    }
    else
    {
	assert(serverName[0] != '\0');
	rootPOAName = serverName;
    }

    //
    // Get the initial service manager
    //
    InitialServiceManager_var ism = orbInstance_ -> getInitialServiceManager();

    //
    // Create the Root POAManager
    //
    OBPortableServer::POAManagerFactory_var factory =
	OBPortableServer::POAManagerFactory::_narrow(
	    Object_var(ism -> resolveInitialReferences("POAManagerFactory")));

    //
    // First attempt to locate the root POAManager
    //
    OBPortableServer::POAManager_var manager;

    OBPortableServer::POAManagerSeq_var managers =
	factory -> get_poa_managers();
    for(ULong i = 0 ; i < managers -> length() ; i++)
    {
	String_var name = managers[i] -> name();
	if(strcmp(name, "RootPOAManager") == 0)
	{
	    manager = managers[i];
	    break;
	}
    }

    //
    // If the root POAManager doesn't exist then create it
    //
    if(is_nil(manager))
    {
        try
        {
            manager = factory -> create_poa_manager("RootPOAManager");
        }
        catch(const OBPortableServer::POAManagerFactory::
              POAManagerAlreadyExists&)
        {
            assert(false);
        }
        catch(const OCI::InvalidParam& ex)
        {
            Logger_var logger = orbInstance_ -> getLogger();
            String_var err = string_dup("invalid configuration parameter `");
            err += ex.p.name;
            err += "' for RootPOAManager: ";
            err += ex.reason;
            logger -> error(err);
            throw INITIALIZE();
        }
    }

    //
    // Create the Root POA
    //
    OBPortableServer::POA_impl* root =
	new OBPortableServer::POA_impl(orbInstance_, rootPOAName,
				       manager);
    root -> _OB_addPolicyFactory();
    rootPOA_ = root;

    ism -> addInitialReference("RootPOA", root, true);

    //
    // Ask the POAManagerFactory to initialize the connection to the
    // IMR
    //
    // TODO-B3: Is there some other point that can be used to do this?
    //
    OBPortableServer::POAManagerFactory_impl* factoryImpl =
        dynamic_cast<OBPortableServer::POAManagerFactory_impl*>(factory.in());
    factoryImpl -> _OB_initializeIMR(root, this);
}
