// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/ValueFactoryManager.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/Codec.h>
#include <OB/POACommon.h>
#include <OB/PIArgs.h>
#include <OB/PIORBInit.h>
#include <OB/PIManager.h>
#include <OB/OBPOAManager.h>
#include <OB/MultiRequestSender.h>
#include <OB/Properties.h>
#include <OB/DirectStubImpl.h>
#include <OB/Reactor.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/Client.h>
#include <OB/Net.h>

#include <Dispatch_impl.h>
#include <BootManager_impl.h>

using namespace CORBA;
using namespace OB;

#ifdef HAVE_JTC

JTCMutex* ORBInstance::globalStateMutex_ = new JTCMutex;
JTCInitialize* ORBInstance::globalStateJTC_ = 0;

namespace OB
{

class GlobalStateMutexDestroyer
{
public:

    ~GlobalStateMutexDestroyer()
    {
	delete ORBInstance::globalStateMutex_;
	ORBInstance::globalStateMutex_ = 0;
    }
};

static GlobalStateMutexDestroyer destroyer;

} // End of namespace OB

#endif

int ORBInstance::globalStateCounter_ = 0;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::ORBInstance_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::ORBInstance_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ORBInstance private member implementation
// ----------------------------------------------------------------------

void
ORBInstance::createGlobalState()
{
#ifdef HAVE_JTC
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> lock();
#endif


    //
    // Only create on first call
    //
    if(globalStateCounter_++ == 0)
    {
#ifdef HAVE_WINSOCK_H
	//
	// Winsock startup
	//
	WORD wVersionRequested = MAKEWORD(1, 1);
	WSADATA wsaData;
	if(WSAStartup(wVersionRequested, &wsaData) != 0)
	{
#ifdef HAVE_JTC
	    if(globalStateMutex_ != 0)
		globalStateMutex_ -> unlock();
#endif
	    throw COMM_FAILURE(LastError(), MinorWSAStartup, COMPLETED_NO);
	}
#endif
	
#ifndef WIN32
	//
	// Set SIGPIPE signal handler
	//
	struct sigaction actPIPE;
	actPIPE.sa_handler = SIG_IGN;
	sigemptyset(&actPIPE.sa_mask);
	actPIPE.sa_flags = 0;
	sigaction(SIGPIPE, &actPIPE, 0);
#endif
	
#ifdef HAVE_JTC
	//
	// Initialize JTC, if necessary
	//
	if(!JTCInitialize::initialized())
	    globalStateJTC_ = new JTCInitialize();
#endif
    }
    
#ifdef HAVE_JTC
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> unlock();
#endif
}

void
ORBInstance::destroyGlobalState()
{
#ifdef HAVE_JTC
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> lock();
#endif

    //
    // Only clean up on last call
    //
    assert(globalStateCounter_ > 0);
    if(--globalStateCounter_ == 0)
    {
#ifdef HAVE_WINSOCK_H
	//
	// Winsock shutdown
	//
	WSACleanup();
#endif
	
#ifndef WIN32
	//
	// Reset SIGPIPE signal handler
	//
	struct sigaction actPIPE;
	actPIPE.sa_handler = SIG_DFL;
	sigemptyset(&actPIPE.sa_mask);
	actPIPE.sa_flags = 0;
	sigaction(SIGPIPE, &actPIPE, 0);
#endif
	
#ifdef HAVE_JTC
	delete globalStateJTC_;
#endif
    }
    
#ifdef HAVE_JTC
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> unlock();
#endif
}

// ----------------------------------------------------------------------
// ORBInstance constructor and destructor
// ----------------------------------------------------------------------

ORBInstance::ORBInstance(ObjectFactory_ptr objectFactory,
			 DirectStubImplFactory_ptr directStubImplFactory,
			 ClientManager_ptr clientManager,
                         PolicyFactoryManager_ptr pfManager,
                         PIManager_ptr piManager,
                         InitialServiceManager_ptr initServiceManager,
                         ValueFactoryManager_ptr valueFactoryManager,
                         IOP::CodecFactory_ptr codecFactory,
			 OBPortableServer::POAManagerFactory_ptr pmFactory,
			 MultiRequestSender_ptr multiRequestSender,
			 Properties_ptr properties,
			 DispatchStrategyFactory_ptr dsf,
			 BootManager_ptr bootManager,
			 Reactor_ptr reactor,
			 Logger_ptr logger,
			 CoreTraceLevels_ptr coreTraceLevels,
                         OCI::ConFactoryRegistry_ptr conFactoryRegistry,
                         OCI::AccFactoryRegistry_ptr accFactoryRegistry)
    : destroy_(false),
      objectFactory_(ObjectFactory::_duplicate(objectFactory)),
      directStubImplFactory_(
	  DirectStubImplFactory::_duplicate(directStubImplFactory)),
      clientManager_(ClientManager::_duplicate(clientManager)),
      policyFactoryManager_(
          PolicyFactoryManager::_duplicate(pfManager)),
      interceptorManager_(PIManager::_duplicate(piManager)),
      initServiceManager_(InitialServiceManager::_duplicate(
          initServiceManager)),
      valueFactoryManager_(
          ValueFactoryManager::_duplicate(valueFactoryManager)),
      codecFactory_(IOP::CodecFactory::_duplicate(codecFactory)),
      pmFactory_(OBPortableServer::POAManagerFactory::_duplicate(pmFactory)),
      multiRequestSender_(MultiRequestSender::_duplicate(multiRequestSender)),
      properties_(Properties::_duplicate(properties)),
      dispatchStrategyFactory_(DispatchStrategyFactory::_duplicate(dsf)),
      bootManager_(BootManager::_duplicate(bootManager)),
      reactor_(Reactor::_duplicate(reactor)),
      logger_(Logger::_duplicate(logger)),
      coreTraceLevels_(CoreTraceLevels::_duplicate(coreTraceLevels)),
      conFactoryRegistry_(
          OCI::ConFactoryRegistry::_duplicate(conFactoryRegistry)),
      accFactoryRegistry_(
          OCI::AccFactoryRegistry::_duplicate(accFactoryRegistry))
{
    createGlobalState();

#ifdef HAVE_JTC
    //
    // Create the server and client worker groups
    //
    clientWorkerGroup_ = new JTCThreadGroup("ClientWorkers");
    serverWorkerGroup_ = new JTCThreadGroup("ServerWorkers");
#endif
}

ORBInstance::~ORBInstance()
{
    assert(destroy_);

    destroyGlobalState();
}

// ----------------------------------------------------------------------
// ORBInstance public member implementation
// ----------------------------------------------------------------------

void
ORBInstance::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Destroy the POAManagerFactory
    //
    pmFactory_ -> destroy();
    pmFactory_ = OBPortableServer::POAManagerFactory::_nil();

    //
    // Destroy the Initial Service manager
    //
    initServiceManager_ -> destroy();
    initServiceManager_ = InitialServiceManager::_nil();

    //
    // Destroy the Object factory
    //
    objectFactory_ -> destroy();
    objectFactory_ = ObjectFactory::_nil();

    //
    // Destroy the StubImpl factory
    //
    directStubImplFactory_ -> destroy();
    directStubImplFactory_ = DirectStubImplFactory::_nil();

    //
    // Destroy the ClientManager
    //
    // ORBControl destroys the ClientManager
    //clientManager_ -> destroy();
    clientManager_ = ClientManager::_nil();

    //
    // Destroy the PolicyFactoryManager
    //
    policyFactoryManager_ -> destroy();
    policyFactoryManager_ = PolicyFactoryManager::_nil();

    //
    // Destroy the PortableInterceptor manager
    //
    interceptorManager_ -> destroy(); // No destroy operation defined
    interceptorManager_ = PIManager::_nil();

    //
    // Destroy the ValueFactory manager
    //
    valueFactoryManager_ -> destroy();
    valueFactoryManager_ = ValueFactoryManager::_nil();

    //
    // Destroy the CodecFactory
    //
    //codecFactory_ -> destroy(); // No destroy operation defined
    codecFactory_ = IOP::CodecFactory::_nil();

    //
    // Destroy the MultiRequestSender
    //
    //multiRequestSender_ -> destroy(); // No destroy operation defined
    multiRequestSender_ = MultiRequestSender::_nil();

    //
    // Properties are not destroyed - they are indestructible
    //
    //properties_ -> destroy(); // No destroy operation defined
    //properties_ = Properties::_nil();

    //
    // Destroy the DispatchStrategyFactory
    //
    // NOTE: destruction is taken care of in ORBControl
    //
    //DispatchStrategyFactory_impl* dsfImpl =
    //	dynamic_cast<DispatchStrategyFactory_impl*>
    //	(dispatchStrategyFactory_.in());
    //dsfImpl -> _OB_destroy();
    dispatchStrategyFactory_ = DispatchStrategyFactory::_nil();

    //
    // Destroy the BootManager
    //
    //bootManager_ -> destroy(); // No destroy operation defined
    bootManager_ = BootManager::_nil();

    //
    // Destroy the reactor
    //
    //reactor_ -> destroy();
    reactor_ = Reactor::_nil();

    //
    // Logger is not destroyed -- it is indestructable
    //
    //logger_ -> destroy();
    //logger_ = Logger::_nil();

    //
    // CoreTraceLevels is not destroyed -- it is indestructible
    //
    //coreTraceLevels_ -> destroy();
    //coreTraceLevels_ = CoreTraceLevels::_nil();

#ifdef HAVE_JTC
    try
    {
	//
	// Destroy the client and server worker groups
	//
	serverWorkerGroup_ -> destroy();
    }
    catch(const JTCIllegalThreadStateException&)
    {
	assert(false);
    }

    try
    {
	clientWorkerGroup_ -> destroy();
    }
    catch(const JTCIllegalThreadStateException&)
    {
	assert(false);
    }
#endif

    //
    // Destroy the ConFactoryRegistry
    //
    //conFactoryRegistry_ -> destroy(); // No destroy operation defined
    conFactoryRegistry_ = OCI::ConFactoryRegistry::_nil();

    //
    // Destroy the AccFactoryRegistry
    //
    //accFactoryRegistry_ -> destroy(); // No destroy operation defined
    accFactoryRegistry_ = OCI::AccFactoryRegistry::_nil();
}

ObjectFactory_ptr
ORBInstance::getObjectFactory()
{
    return ObjectFactory::_duplicate(objectFactory_);
}

DirectStubImplFactory_ptr
ORBInstance::getDirectStubImplFactory()
{
    return DirectStubImplFactory::_duplicate(directStubImplFactory_);
}

ClientManager_ptr
ORBInstance::getClientManager()
{
    return ClientManager::_duplicate(clientManager_);
}

PolicyFactoryManager_ptr
ORBInstance::getPolicyFactoryManager()
{
    return PolicyFactoryManager::_duplicate(policyFactoryManager_);
}

PIManager_ptr
ORBInstance::getPIManager()
{
    return PIManager::_duplicate(interceptorManager_);
}

InitialServiceManager_ptr
ORBInstance::getInitialServiceManager()
{
    return InitialServiceManager::_duplicate(initServiceManager_);
}

ValueFactoryManager_ptr
ORBInstance::getValueFactoryManager()
{
    return ValueFactoryManager::_duplicate(valueFactoryManager_);
}

IOP::CodecFactory_ptr
ORBInstance::getCodecFactory()
{
    return IOP::CodecFactory::_duplicate(codecFactory_);
}

OBPortableServer::POAManagerFactory_ptr
ORBInstance::getPOAManagerFactory()
{
    return OBPortableServer::POAManagerFactory::_duplicate(pmFactory_);
}

MultiRequestSender_ptr
ORBInstance::getMultiRequestSender()
{
    return MultiRequestSender::_duplicate(multiRequestSender_);
}

Properties_ptr
ORBInstance::getProperties()
{
    return Properties::_duplicate(properties_);
}

DispatchStrategyFactory_ptr
ORBInstance::getDispatchStrategyFactory()
{
    return DispatchStrategyFactory::_duplicate(dispatchStrategyFactory_);
}

BootManager_ptr
ORBInstance::getBootManager()
{
    return BootManager::_duplicate(bootManager_);
}

Reactor_ptr
ORBInstance::getReactor()
{
    return Reactor::_duplicate(reactor_);
}

Logger_ptr
ORBInstance::getLogger()
{
    return Logger::_duplicate(logger_);
}

CoreTraceLevels_ptr
ORBInstance::getCoreTraceLevels()
{
    return CoreTraceLevels::_duplicate(coreTraceLevels_);
}

#ifdef HAVE_JTC
JTCRecursiveMutex&
ORBInstance::getORBSyncMutex()
{
    return orbSyncMutex_;
}

JTCThreadGroupHandle
ORBInstance::getServerWorkerGroup()
{
    return serverWorkerGroup_;
}

JTCThreadGroupHandle
ORBInstance::getClientWorkerGroup()
{
    return clientWorkerGroup_;
}
#endif

OCI::ConFactoryRegistry_ptr
ORBInstance::getConFactoryRegistry()
{
    return OCI::ConFactoryRegistry::_duplicate(conFactoryRegistry_);
}

OCI::AccFactoryRegistry_ptr
ORBInstance::getAccFactoryRegistry()
{
    return OCI::AccFactoryRegistry::_duplicate(accFactoryRegistry_);
}

// ----------------------------------------------------------------------
// Get an ORBInstance, given and ORB_ptr
// ----------------------------------------------------------------------

/* In ORB_impl.cpp */
