// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Except.h>
#include <OB/PIPolicy.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/PolicyFactoryManager.h>

#include <ORB_impl.h>
#include <ORBPolicyManager_impl.h>
#include <OBPolicies_impl.h>
#include <Messaging_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// The ORBPolicyFactory_impl
// ----------------------------------------------------------------------

namespace PortableInterceptor
{

class ORBPolicyFactory_impl : public PortableInterceptor::PolicyFactory
{
public:

    //
    // IDL to C++ Mapping
    //
    virtual CORBA::Policy_ptr create_policy(CORBA::PolicyType,
                                            const CORBA::Any&);
};

} // End namespace PortableInterceptor

// ----------------------------------------------------------------------
// ORB private member functions
// ----------------------------------------------------------------------

void
OBCORBA::ORB_impl::initializeDefaultPolicies()
{
    //
    // Add the default policies
    //
    ConnectionReusePolicy_impl* reusePolicy =
        new ConnectionReusePolicy_impl(true);

    CORBA::ULong policiesLength = policies_.length();
    policies_.length(policiesLength + 1);
    policies_[policiesLength] = reusePolicy;
    
    RetryPolicy_impl* retryPolicy = new RetryPolicy_impl(OB::RETRY_STRICT);

    policiesLength = policies_.length();
    policies_.length(policiesLength + 1);
    policies_[policiesLength] = retryPolicy;

    //
    // Create the ORBPolicyManager
    //
    PolicyManager_var pm = new ORBPolicyManager_impl(policies_);
    InitialServiceManager_var initServiceManager =
        orbInstance_ -> getInitialServiceManager();
    initServiceManager -> addInitialReference("ORBPolicyManager", pm);
    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    objectFactory -> setPolicyManager(pm);

    //
    // Register the default PolicyFactory policies for the ORB
    //
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();

    PortableInterceptor::PolicyFactory_var factory = new
        PortableInterceptor::ORBPolicyFactory_impl;
    pfm -> registerPolicyFactory(CONNECTION_REUSE_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(PROTOCOL_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(RETRY_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(LOCATION_TRANSPARENCY_POLICY_ID, factory,
                                 true);
    pfm -> registerPolicyFactory(Messaging::SYNC_SCOPE_POLICY_TYPE,
                                 factory, true);
    pfm -> registerPolicyFactory(INTERCEPTOR_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(CONNECT_TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(REQUEST_TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(ACM_TIMEOUT_POLICY_ID, factory, true);
}

// ----------------------------------------------------------------------
// ORBPolicyFactory public member functions
// ----------------------------------------------------------------------

Policy_ptr
PortableInterceptor::ORBPolicyFactory_impl::create_policy(
    PolicyType policy_type,
    const Any& val)
{
    //
    // A case stmt would be better here, but egcs complains: case
    // label `CONNECTION_REUSE_POLICY_ID' does not reduce to an
    // integer constant
    //
    if(policy_type == CONNECTION_REUSE_POLICY_ID)
    {
	Boolean b;
	if(val >>= Any::to_boolean(b))
        {
	    return new ConnectionReusePolicy_impl(b);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == PROTOCOL_POLICY_ID)
    {
        OCI::ProtocolId id;
        if(val >>= id)
        {
            return new ProtocolPolicy_impl(id);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == RETRY_POLICY_ID)
    {
        Short v;
        if(val >>= v)
        {
            if(v > RETRY_ALWAYS)
                throw PolicyError(BAD_POLICY_VALUE);
            return new RetryPolicy_impl(v);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new TimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == LOCATION_TRANSPARENCY_POLICY_ID)
    {
        Short v;
        if(val >>= v)
            return new LocationTransparencyPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::SYNC_SCOPE_POLICY_TYPE)
    {
        Messaging::SyncScope v;
        if(val >>= v)
            return new Messaging::SyncScopePolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == INTERCEPTOR_POLICY_ID)
    {
        bool v;
        if(val >>= v)
            return new InterceptorPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == CONNECT_TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new ConnectTimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == REQUEST_TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new RequestTimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == ACM_TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new ACMTimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    throw PolicyError(BAD_POLICY);
    return Policy::_nil(); // Some compilers need this
}
