// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/ORBInstance.h>
#include <OB/PIManager.h>
#include <OB/InitialServiceManager.h>
#include <OB/Client.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/ValueFactoryManager.h>
#include <OB/ORBControl.h>
#include <OB/MultiRequestSender.h>
#include <OB/Stream.h>
#include <OB/Context.h>
#include <OB/Environment.h>
#include <OB/RefCounted.h>
#include <OB/DirectStubImpl.h>
#include <OB/CoreTraceLevels.h>
#include <OB/Logger.h>  
#include <OB/Properties.h>
#include <OB/OCI_IIOP_init.h>

#include <ORB_impl.h>
#include <CodeSetUtil.h>
#include <TypeCodeFactory.h>
#include <OCI_impl.h>
#include <POAManagerFactory_impl.h>
#include <DynamicAny_impl.h>
#include <CodecFactory_impl.h>
#include <Dispatch_impl.h>
#include <BootManager_impl.h>
#include <OBORBInitInfo_impl.h>

using namespace CORBA;
using namespace OB;

//
// The ORB implementation map
//
OBCORBA::ORB_impl::ORBImplMap OBCORBA::ORB_impl::orbImplMap_(7);
OBCORBA::ORB_impl::ORBInitializerSeq OBCORBA::ORB_impl::orbInitializers_;

// ----------------------------------------------------------------------
// ORB_impl constructor and destructor
// ----------------------------------------------------------------------

OBCORBA::ORB_impl::ORB_impl(int ac, char** av, const char* id,
			    const char* serverName,
			    const char* serverInstance,
			    Client::ConcModel concModel,
			    CoreTraceLevels_ptr coreTraceLevels,
			    Properties_ptr properties,
			    Logger_ptr logger,
			    Reactor_ptr reactor)

    : id_(id), serverName_(serverName), destroy_(false)
{
    try
    {
	//
	// It's not necessary to lock the orbImplMap here since ORB_init
	// locks the map during ORB creation
	//
	//JTCSynchronized orbImplMapSync(orbImplMap_);
	//

	//
	// Create the ORBControl
	//
	orbControl_ = new ORBControl;

	//
	// Create the ORBInstance object
	//
	//
	InitialServiceManager_var initServiceManager =
	    new InitialServiceManager;
	ClientManager_var clientManager =
	new ClientManager((Client::ConcModel)concModel);
	ObjectFactory_var objectFactory = new ObjectFactory;
	DirectStubImplFactory_var directStubImplFactory =
	new DirectStubImplFactory;
	PolicyFactoryManager_var pfManager = new PolicyFactoryManager;
	PIManager_var piManager = new PIManager;
	ValueFactoryManager_var valueFactoryManager = new ValueFactoryManager;
	IOP::CodecFactory_impl* codecFactoryImpl = new IOP::CodecFactory_impl;
	IOP::CodecFactory_var codecFactory = codecFactoryImpl;
	OBPortableServer::POAManagerFactory_impl* pmFactoryImpl =
	new OBPortableServer::POAManagerFactory_impl(serverName,
						     serverInstance);
	OBPortableServer::POAManagerFactory_var pmFactory = pmFactoryImpl;
	MultiRequestSender_var multiRequestSender = new MultiRequestSender;
	DispatchStrategyFactory_impl* dsfImpl =
	    new DispatchStrategyFactory_impl;
	DispatchStrategyFactory_var dsf = dsfImpl;
	BootManager_impl* bootManagerImpl = new BootManager_impl;
	BootManager_var bootManager = bootManagerImpl;
	OCI::ConFactoryRegistry_var conFactoryRegistry =
            new OCI::ConFactoryRegistry_impl();
	OCI::AccFactoryRegistry_var accFactoryRegistry =
            new OCI::AccFactoryRegistry_impl();

	orbInstance_ = new ORBInstance(objectFactory,
				       directStubImplFactory,
				       clientManager,
				       pfManager,
				       piManager,
				       initServiceManager,
				       valueFactoryManager,
				       codecFactory,
				       pmFactory,
				       multiRequestSender,
				       properties,
				       dsf,
				       bootManager,
				       reactor,
				       logger,
				       coreTraceLevels,
				       conFactoryRegistry,
				       accFactoryRegistry);

	objectFactory -> setORBInstance(orbInstance_);
	initServiceManager -> setORBInstance(orbInstance_);
	pmFactoryImpl -> _OB_setORBInstance(orbInstance_);
	piManager -> setORBInstance(orbInstance_);
	codecFactoryImpl -> _OB_setORBInstance(orbInstance_);
	clientManager -> setORBInstance(orbInstance_);
	directStubImplFactory -> setORBInstance(orbInstance_);
	dsfImpl -> _OB_setORBInstance(orbInstance_);

	//
	// Set the ORBInstance on the ORBControl
	//
	orbControl_ -> setORBInstance(orbInstance_);

	//
	// Initialize default policies
	//
	initializeDefaultPolicies();

	//
	// Create the DynamicAny::DynAnyFactory object
	//
	DynamicAny::DynAnyFactory_var dynAnyFactory =
	    new DynamicAny::DynAnyFactory_impl(orbInstance_);

	//
	// Add initial references
	//
	initServiceManager -> addInitialReference("POAManagerFactory",
						  pmFactory);
	initServiceManager -> addInitialReference("DynAnyFactory",
						  dynAnyFactory);
	initServiceManager -> addInitialReference("CodecFactory",
						  codecFactory);
	initServiceManager -> addInitialReference("DispatchStrategyFactory",
						  dsf);
	initServiceManager -> addInitialReference("BootManager",
						  bootManager);
	initServiceManager -> addInitialReference("RootPOA",
						  Object::_nil()); // Dummy
	initServiceManager -> addInitialReference("OCIConFactoryRegistry",
						  conFactoryRegistry);
	initServiceManager -> addInitialReference("OCIAccFactoryRegistry",
						  accFactoryRegistry);

	//
	// Initialize the IIOP OCI plug-in, if necessary
	//
	const char* initIIOP = properties -> getProperty("ooc.orb.init_iiop");
	if(initIIOP == 0)
	    initIIOP = "both";
	if(strcmp(initIIOP, "none") != 0)
	{
	    if(strcmp(initIIOP, "client") == 0)
		OCI::IIOP::initClient(conFactoryRegistry);
	    else if(strcmp(initIIOP, "server") == 0)
		OCI::IIOP::initServer(accFactoryRegistry);
	    else
	    {
		OCI::IIOP::initClient(conFactoryRegistry);
		OCI::IIOP::initServer(accFactoryRegistry);
	    }
	}

	//
	// Initialize Portable Interceptors - this must be done after
	// installing the IIOP plug-in to allow an ORBInitializer or
	// interceptor to make a remote invocation
	//

	CodeSetUtil::registerIORInterceptor(piManager);

	//
	// Call each of the ORB initializers. If there are no ORB
	// initializers it's not necessary to setup the PIManager
	// since no interceptors will be called.
	//
	if(orbInitializers_.length() > 0)
	{
	    OBPortableInterceptor::ORBInitInfo_impl* infoImpl =
                new OBPortableInterceptor::ORBInitInfo_impl(
                    this,
                    ac, av, id, 
                    piManager,
                    initServiceManager,
                    codecFactory);

	    OBPortableInterceptor::ORBInitInfo_var info = infoImpl;

	    ULong i;
	    for(i = 0 ; i < orbInitializers_.length() ; i++)
		orbInitializers_[i] -> pre_init(info);

	    //TODO: change state
	    for(i = 0 ; i < orbInitializers_.length() ; i++)
		orbInitializers_[i] -> post_init(info);

	    infoImpl -> _OB_destroy();
	}
	piManager -> setupComplete();

	//
	// Add this ORB to the implementation map -- this must be done
	// last.
	//
	orbImplMap_.put(id_, ORB::_duplicate(this));
    }
    catch(...)
    {
	//
	// Here the same thing as ORB::destroy must be done since
	// although the ORB itself isn't fully initialized all of the
	// ORB components may be.
	//
	if(!CORBA::is_nil(orbControl_))
	{
	    orbControl_ -> shutdownServerClient();
	    orbControl_ -> destroy();
	}
	if(!CORBA::is_nil(orbInstance_))
	{
	    orbInstance_ -> destroy();
	    orbInstance_ = ORBInstance::_nil();
	}
	_OB_setRef(0);
	throw;
    }
}

OBCORBA::ORB_impl::~ORB_impl()
{
    //
    // Check whether the ORB has been properly destroyed
    //
    if(!destroy_)
    {
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> warning("ORB::destroy() was not called. "
			  "This may result in memory leaks.");
    }
}

// ----------------------------------------------------------------------
// ORB_impl public member implementation
// ----------------------------------------------------------------------

char*
OBCORBA::ORB_impl::object_to_string(Object_ptr p)
{
    {
	OB_SYNCHRONIZED
	if(destroy_)
	    throw OBJECT_NOT_EXIST();
    }

    RefCountIOR_var ior;
    
    if(is_nil(p))
    {
	IOP::IOR empty;
	ior = new RefCountIOR(empty);
    }
    else
    {
        if(p -> _OB_isLC())
            throw BAD_PARAM();
	
	ior = p -> _OB_origIOR(); 
    }

    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    ior -> value._OB_marshal(&out);
    String_var s = OctetsToAscii(buf.data(), buf.length());
    
    static const char* prefix = "IOR:";
    char* result = string_alloc(strlen(prefix) + strlen(s));
    strcpy(result, prefix);
    strcat(result, s);

    return result;
}

Object_ptr
OBCORBA::ORB_impl::string_to_object(const char* ior)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    return objectFactory -> stringToObject(ior);
}

/* DII related operations are in DII.cpp */

void
OBCORBA::ORB_impl::get_default_context(Context_out ctx)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    ctx = new Context("");
}

void
OBCORBA::ORB_impl::create_environment(Environment_out new_env)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    new_env = new Environment;
}

CORBA::ORB::ObjectIdList*
OBCORBA::ORB_impl::list_initial_services()
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    InitialServiceManager_var initServiceManager =
	orbInstance_ -> getInitialServiceManager();
    return initServiceManager -> listInitialServices();
}

Boolean
OBCORBA::ORB_impl::get_service_information(ServiceType,
					   ServiceInformation_out info)
{
    info = new ServiceInformation;
    return false;
}

Object_ptr
OBCORBA::ORB_impl::resolve_initial_references(const char* identifier)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    InitialServiceManager_var initServiceManager =
        orbInstance_ -> getInitialServiceManager();

    Object_var obj;

    try
    {
        obj = initServiceManager -> resolveInitialReferences(identifier);
    }
    catch(const InvalidName&)
    {
        //
        // If the service is the RootPOA and it hasn't yet been
        // initialized, create it. We could put in some automatic
        // method here for late binding of objects at some later
        // point.
        //
        if(strcmp(identifier, "RootPOA") == 0)
        {
            orbControl_ -> initializeRootPOA(serverName_);
            return resolve_initial_references(identifier);
        }
        else
            throw;
    }

    return obj._retn();
}

void
OBCORBA::ORB_impl::register_initial_reference(const char* name, Object_ptr obj)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    InitialServiceManager_var initServiceManager =
        orbInstance_ -> getInitialServiceManager();
    initServiceManager -> addInitialReference(name, obj);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_struct_tc(const char* id, const char* name,
			     const StructMemberSeq& members)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createStructTC(id, name, members);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_union_tc(const char* id, const char* name,
			  TypeCode_ptr discriminator_type,
			  const UnionMemberSeq& members)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createUnionTC(id, name, discriminator_type,
					  members);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_enum_tc(const char* id, const char* name,
			 const EnumMemberSeq& members)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createEnumTC(id, name, members);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_alias_tc(const char* id, const char* name,
			  TypeCode_ptr original_type)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createAliasTC(id, name, original_type);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_exception_tc(const char* id, const char* name,
			      const StructMemberSeq& members)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createExceptionTC(id, name, members);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_interface_tc(const char* id, const char* name)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createInterfaceTC(id, name);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_string_tc(ULong bound)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createStringTC(bound);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_wstring_tc(ULong bound)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createWstringTC(bound);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_fixed_tc(UShort digits, Short scale)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createFixedTC(digits, scale);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_sequence_tc(ULong bound, TypeCode_ptr element_type)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createSequenceTC(bound, element_type);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_recursive_sequence_tc(ULong bound, ULong offset)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createRecursiveSequenceTC(bound, offset);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_array_tc(ULong length, TypeCode_ptr element_type)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createArrayTC(length, element_type);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_value_tc(const char* id, const char* name,
			    ValueModifier type_modifier,
			    TypeCode_ptr concrete_base,
			    const ValueMemberSeq& members)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createValueTC(id, name, type_modifier,
					  concrete_base, members);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_value_box_tc(const char* id, const char* name,
			       TypeCode_ptr boxed_type)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createValueBoxTC(id, name, boxed_type);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_native_tc(const char* id, const char* name)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createNativeTC(id, name);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_recursive_tc(const char* id)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createRecursiveTC(id);
}

TypeCode_ptr
OBCORBA::ORB_impl::create_abstract_interface_tc(const char* id,
						const char* name)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return TypeCodeFactory::createAbstractInterfaceTC(id, name);
}

Boolean
OBCORBA::ORB_impl::work_pending()
{
    //
    // Ensure that the ORB mutex is not locked during the call to
    // ORBControl methods
    //
    if(destroy_)
	throw OBJECT_NOT_EXIST();
    return orbControl_ -> workPending();
}

void
OBCORBA::ORB_impl::perform_work()
{
    //
    // Ensure that the ORB mutex is not locked during the call to
    // ORBControl methods
    //
    if(destroy_)
	throw OBJECT_NOT_EXIST();
    orbControl_ -> performWork();
}

void
OBCORBA::ORB_impl::run()
{
    //
    // Ensure that the ORB mutex is not locked during the call to
    // ORBControl methods
    //
    if(destroy_)
	throw OBJECT_NOT_EXIST();
    orbControl_ -> run();
}

//
// ASYNC SAFE if called with wait_for_completion = false
//
void
OBCORBA::ORB_impl::shutdown(Boolean wait_for_completion)
{
    //
    // Ensure that the ORB mutex is not locked during the call to
    // ORBControl methods
    //
    if(destroy_)
	throw OBJECT_NOT_EXIST();
    orbControl_ -> shutdownServer(wait_for_completion);
}

void
OBCORBA::ORB_impl::destroy()
{
    //
    // From the specification:
    //
    // This operation destroys the ORB so that its resources can be
    // reclaimed by the application.  Any operation invoked on a
    // destroyed ORB reference will raise the OBJECT_NOT_EXIST
    // exception.  Once an ORB has been destroyed, another call to
    // ORB_init with the same ORBid will return a reference to a newly
    // constructed ORB.
    //
    // If destroy is called on an ORB that has not been shut down, it
    // will start the shut down process and block until the ORB has
    // shut down before it destroys the ORB.  If an application calls
    // destroy in a thread that is currently servicing an invocation,
    // the BAD_INV_ORDER system exception will be raised with the OMG
    // minor code 3, since blocking would result in a deadlock.
    //
    // For maximum portability and to avoid resource leaks, an
    // application should always call shutdown and destroy on all ORB
    // instances before exiting.
    //
    OB_SYNCHRONIZED

    //
    // Has the ORB been destroyed yet?
    //
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    //
    // Shutdown both the server & client side of the ORB
    //
    orbControl_ -> shutdownServerClient();

    //
    // Destroy the ORBControl. Don't set to _nil.
    //
    orbControl_ -> destroy();
    //orbControl_ = ORBControl::_nil();
    
    //
    // Destroy the ORBInstance object
    //
    orbInstance_ -> destroy();
    orbInstance_ = ORBInstance::_nil();

    //
    // Mark the ORB as destroyed
    //
    destroy_ = true;

    //
    // Remove the ORB from the implementation map
    //
    orbImplMap_.remove(id_);
}

Policy_ptr
OBCORBA::ORB_impl::create_policy(PolicyType type, const Any& any)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    PolicyFactoryManager_var manager =
	orbInstance_ -> getPolicyFactoryManager();
    return manager -> createPolicy(type, any);
}

ValueFactory
OBCORBA::ORB_impl::register_value_factory(const char* id, ValueFactory factory)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    ValueFactoryManager_var manager = orbInstance_ -> getValueFactoryManager();
    return manager -> registerValueFactory(id, factory);
}

void
OBCORBA::ORB_impl::unregister_value_factory(const char* id)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    ValueFactoryManager_var manager = orbInstance_ -> getValueFactoryManager();
    manager -> unregisterValueFactory(id);
}

ValueFactory
OBCORBA::ORB_impl::lookup_value_factory(const char* id)
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    ValueFactoryManager_var manager = orbInstance_ -> getValueFactoryManager();
    return manager -> lookupValueFactory(id);
}

OB::PropertySet
OBCORBA::ORB_impl::properties()
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return orbInstance_ -> getProperties();
}

OB::Logger_ptr
OBCORBA::ORB_impl::logger()
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return orbInstance_ -> getLogger();
}

OB::Reactor_ptr
OBCORBA::ORB_impl::reactor()
{
    OB_SYNCHRONIZED
    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return orbInstance_ -> getReactor();
}

// ----------------------------------------------------------------------
// Get an ORBInstance, given an ORB_ptr
// ----------------------------------------------------------------------

ORBInstance_ptr
OB::GetORBInstance(CORBA::ORB_ptr p)
{
    if(is_nil(p))
	return ORBInstance::_nil();
    
    OBCORBA::ORB_impl* impl = dynamic_cast<OBCORBA::ORB_impl*>(p);
    if(!impl)
	throw INITIALIZE();
    
    return ORBInstance::_duplicate(impl -> orbInstance_);
}
