// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/NativeTypes.h>
#include <OB/TypeCode.h>
#include <OB/DirectStubImpl.h>
#include <OB/MarshalStubImpl.h>
#include <OB/Policy.h>
#include <OB/Environment.h>
#include <OB/OBORB.h>
#include <OB/OBPolicies.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>
#include <OB/Client.h>
#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>

using namespace CORBA;
using namespace OB;

//
// The ids_ for Object
//
const char* Object::ids_[] =
{
    //
    // Don't put "IDL:omg.org/CORBA/Object:1.0" here, as this ID is
    // always implictly there.
    //
    0
};

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(Object_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(Object_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Object private and protected members
// ----------------------------------------------------------------------

void
Object::checkRetry(bool retry, const SystemException& ex)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif
    
    //
    // We remove the StubImpl, whether we retry or not
    //
    stubImpl_ = StubImplBase::_nil();

    //
    // Reset to the original IOR
    //
    IOR_ = origIOR_;

    //
    // Get the core trace levels
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();

    //
    // We only retry upon COMM_FAILURE, TRANSIENT, and NO_RESPONSE
    //
    if(COMM_FAILURE::_downcast(&ex) == 0 && TRANSIENT::_downcast(&ex) == 0 &&
       NO_RESPONSE::_downcast(&ex) == 0)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"retry only upon COMM_FAILURE, TRANSIENT and "
		"NO_RESPONSE exceptions\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }

    //
    // Only retry once
    //
    if(!retry)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because I tried again once already\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_NEVER is set
    //
    if(policies_ -> retry == RETRY_NEVER)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again because the RETRY_NEVER policy is set\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_STRICT is set and the completion status
    // is not COMPLETED_NO
    //
    if(policies_ -> retry == RETRY_STRICT && ex.completed() != COMPLETED_NO)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because the RETRY_STRICT policy is set\n"
		"and completion status is not COMPLETED_NO\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }
}

StubImplBase_ptr
Object::_OB_getStubImpl()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif
    
    if(is_nil(stubImpl_))
    {
	while(true)
	{
	    try
	    {
        	//
		// First try to create a DirectStubImpl
		//
		DirectStubImplFactory_var factory =
		    orbInstance_ -> getDirectStubImplFactory();
		stubImpl_ = factory ->
		    createStubImpl(IOR_, origIOR_, policies_);
	    }
	    catch(const LocationForward& ex)
	    {
		IOR_ = ex.ior;
		if(ex.perm)
		    origIOR_ = IOR_;
		//
		// stubImpl_ is already nil
		//
		continue;
	    }
	    break;
	}

	//
	// If that fails then setup a MarshalStubImpl
	//
	if(is_nil(stubImpl_))
	{
	    MarshalStubImpl_var marshalStubImpl =
		_OB_createMarshalStubImpl();
	    marshalStubImpl -> _OB_setup(
		orbInstance_, IOR_, origIOR_, policies_);
	    stubImpl_ = StubImplBase::_duplicate(marshalStubImpl);
	}
	assert(!is_nil(stubImpl_));
    }
	    
    return StubImplBase::_duplicate(stubImpl_);
}

MarshalStubImpl_ptr
Object::_OB_getMarshalStubImpl()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif

    //
    // Is the stubImpl_ already marshal?
    //
    MarshalStubImpl_var marshalStubImpl = MarshalStubImpl::_duplicate(
	dynamic_cast<MarshalStubImpl_ptr>(stubImpl_.in()));
    
    if(is_nil(marshalStubImpl))
    {
	marshalStubImpl = _OB_createMarshalStubImpl();
	marshalStubImpl -> _OB_setup(
	    orbInstance_, IOR_, origIOR_, policies_);
	stubImpl_ = StubImplBase::_duplicate(marshalStubImpl);
    }

    return marshalStubImpl._retn();
}

MarshalStubImpl_ptr
Object::_OB_createMarshalStubImpl()
{
    return new MarshalStubImpl();
}

// ----------------------------------------------------------------------
// Object constructor and destructor
// ----------------------------------------------------------------------

Object::Object()
{
}

Object::~Object()
{
}

// ----------------------------------------------------------------------
// Object public member implementation
// ----------------------------------------------------------------------

InterfaceDef_ptr
Object::_get_interface()
{
    if(_OB_isLC())
        throw NO_IMPLEMENT();
    
    bool _ob_retry = true;
    while(true)
    {
        StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
        try
        {
            return _ob_stubImpl -> _get_interface();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry);
        }
    }
}

Boolean
Object::_is_a(const char* _ob_id)
{
    assert_nca(_ob_id != 0, NCANullString);

    //
    // Check IDL:omg.org/CORBA/Object:1.0
    //
    if(strcmp(_ob_id, "IDL:omg.org/CORBA/Object:1.0") == 0)
        return true;
    
    //
    // Check all other ids
    //
    const char** ids = _OB_ids();
    assert(ids != 0);
    for(ULong i = 0; ids[i] != 0; i++)
        if(strcmp(_ob_id, ids[i]) == 0)
            return true;
    
    //
    // Check the type_id in the IOR and the original IOR
    //
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(mutex_);
#endif

        if(strcmp(_ob_id, IOR_ -> value.type_id) == 0 ||
           strcmp(_ob_id, origIOR_ -> value.type_id) == 0)
            return true;
    }

    //
    // If this is an l-c object, return false now
    //
    if(_OB_isLC())
        return false;

    //
    // Execute _is_a remotely
    //
    bool _ob_retry = true;
    while(true)
    {
        StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
        try
        {
            return _ob_stubImpl -> _is_a(_ob_id);
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry);
        }
    }
}

Boolean
Object::_non_existent()
{
    if(_OB_isLC())
        throw NO_IMPLEMENT();

    bool _ob_retry = true;
    while(true)
    {
        StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
        try
        {
            return _ob_stubImpl -> _non_existent();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry);
        }
    }
}

Boolean
Object::_is_equivalent(Object_ptr p)
{
    //
    // Check for nil reference
    //
    if(is_nil(p))
	return false;

    //
    // For l-c objects, equivalence means that the references
    // (pointers) point to the same object
    //
    if(_OB_isLC() || p -> _OB_isLC())
    {
        if(_OB_isLC() && p -> _OB_isLC())
            return this == p;
	else
	    return false;
    }
    
    //
    // Direct pointer comparison
    //
    if(p == this)
	return true;

    //
    // Ask the client manager
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif
    return clientManager -> equivalent(origIOR_ -> value,
				       p -> origIOR_ -> value);
}

ULong
Object::_hash(ULong maximum)
{
    //
    // For l-c objects return a hash based on the memory address
    //
    if(_OB_isLC())
        return ((ULong)(unsigned long)this / 256) % (maximum + 1);
    
    //
    // Ask the client manager
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif
    return clientManager -> hash(origIOR_ -> value, maximum);
}

Object_ptr
Object::_narrow(AbstractBase_ptr a)
{
    if(!is_nil(a))
        return a -> _to_object();
    return _nil();
}

/*
 *
 * _request() and _create_request() implementations in DII.cpp
 *
 */

Policy_ptr
Object::_get_policy(PolicyType policy_type)
{
    return _get_effective_policy_override(policy_type);
}

Policy_ptr
Object::_get_effective_policy_override(PolicyType policy_type)
{
    Policy_var policy = _OB_getPolicy(policy_type);
    if(is_nil(policy))
        throw INV_POLICY();
    return policy._retn();
}

Object_ptr
Object::_set_policy_overrides(const PolicyList& np, SetOverrideType set_add)
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    //
    // Create the new policy list
    //
    RefCountPolicyList_var newPolicies;

    if(set_add == SET_OVERRIDE)
    {
	newPolicies = new RefCountPolicyList(np);
    }
    else // ADD_OVERRIDE
    {
	PolicyList policies = policies_ -> value;
	
	for(ULong i = 0 ; i < np.length() ; i++)
	{
	    ULong j;
	    for(j = 0 ; j < policies.length() ; j++)
		if(policies[j] -> policy_type() == np[i] -> policy_type())
		    break;
	    
	    if(j < policies.length()) {
		policies[j] = np[i];
	    } else {
		CORBA::ULong policiesLength = policies.length();
		policies.length(policiesLength + 1);
		policies[policiesLength] = np[i];
            }
	}
	
	newPolicies = new RefCountPolicyList(policies);
    }

    //
    // Create a new reference
    //
    Object_var p = new Object;

    //
    // Copy the data from this object reference
    //
    p -> _OB_copyFrom(this);

    //
    // Setup the policies
    //
    p -> policies_ = newPolicies;

    return p._retn();
}

PolicyList*
Object::_get_policy_overrides(const PolicyTypeSeq& types)
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    ULong len = policies_ -> value.length();
    ULong typesLen = types.length();

    PolicyList_var policies = new PolicyList(typesLen);

    for(ULong i = 0 ; i < typesLen; ++i)
    {
        PolicyType currType = types[i];
        for(ULong j = 0 ; j < len ; ++j)
            if(policies_ -> value[j] -> policy_type() == currType) {
                CORBA::ULong policiesLength  = policies -> length();
                policies -> length(policiesLength + 1);
                policies[policiesLength] = policies_ -> value[j];
            }
    }

    return policies._retn();
}

Boolean
Object::_validate_policies(const PolicyList&)
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    //
    // For now all policy combinations are valid
    //
    return true;
}

OCI::TransportInfo_ptr
Object::_get_oci_transport_info()
{
    if(_OB_isLC())
        throw NO_IMPLEMENT();

    StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
    return _ob_stubImpl -> _get_oci_transport_info();
}

OCI::ConnectorInfo_ptr
Object::_get_oci_connector_info()
{
    if(_OB_isLC())
        throw NO_IMPLEMENT();

    StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
    return _ob_stubImpl -> _get_oci_connector_info();
}

const char**
Object::_OB_ids() const
{
    return ids_;
}

const char**
Object::_OB_staticIds()
{
    return ids_;
}

void
Object::_OB_setup(ORBInstance_ptr instance, RefCountIOR_ptr ior,
		  RefCountPolicyList_ptr policies)
{
    if(_OB_isLC())
        throw BAD_OPERATION();

    orbInstance_ = ORBInstance::_duplicate(instance);
    IOR_ = RefCountIOR::_duplicate(ior);
    origIOR_ = RefCountIOR::_duplicate(ior);
    policies_ = RefCountPolicyList::_duplicate(policies);
}

void
Object::_OB_copyFrom(Object_ptr p)
{
    if(_OB_isLC())
        throw BAD_OPERATION();

    //
    // Synchronization on this -> _ob_mutex is not necessary, since
    // _OB_copyFrom is called right after reference creation, before
    // any other thread gets a chance to use it. We must, however,
    // synchronize on p -> _ob_mutex, since the reference we're
    // copying from might be used by other threads.
    //

#ifdef HAVE_JTC
    JTCSynchronized sync(p -> mutex_);
#endif

    orbInstance_ = p -> orbInstance_;
    IOR_ = p -> IOR_;
    origIOR_ = p -> origIOR_;
    policies_ = p -> policies_;

    //
    // Don't copy stubImpl_, because it might be a StubImpl for a base
    // type, but not for the most derived type.
    //
}

ORBInstance_ptr
Object::_OB_ORBInstance() const
{
    return ORBInstance::_duplicate(orbInstance_);
}

RefCountIOR_ptr
Object::_OB_IOR() const
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif

    return RefCountIOR::_duplicate(IOR_);
}

RefCountIOR_ptr
Object::_OB_origIOR() const
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif

    return RefCountIOR::_duplicate(origIOR_);
}

void
Object::_OB_marshalOrigIOR(OutputStreamImpl* out) const
{
#ifdef HAVE_JTC
    JTCSynchronized sync(mutex_);
#endif

    origIOR_ -> value._OB_marshal(out);
}

RefCountPolicyList_ptr
Object::_OB_policies() const
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    return RefCountPolicyList::_duplicate(policies_);
}

Policy_ptr
Object::_OB_getPolicy(PolicyType policy_type) const
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    for(ULong i = 0 ; i < policies_ -> value.length() ; i++)
	if(policies_ -> value[i] -> policy_type() == policy_type)
	    return Policy::_duplicate(policies_ -> value[i]);
    
    return Policy::_nil();
}

void
Object::_OB_handleException(const ExceptionBase& ex, bool& retry)
{
    try
    {
        ex._raise();
    }
    catch(const LocationForward& ex)
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(mutex_);
#endif
	
	//
	// Can't location forward l-c objects
	//
	assert(!_OB_isLC());
	
	//
	// Change the IOR
	//
	IOR_ = RefCountIOR::_duplicate(ex.ior);
	if(ex.perm)
	    origIOR_ = RefCountIOR::_duplicate(ex.ior);
	
	//
	// We need to re-get the stub implementation
	//
        stubImpl_ = StubImplBase::_nil();

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry",
			    "trying again because of location forward");
	}
    }
    catch(const StubImplChanged&)
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(mutex_);
#endif

	//
	// l-c objects don't have a StubImpl
	//
	assert(!_OB_isLC());

	//
	// StubImpl has changed, so let's remove it
	//
	stubImpl_ = StubImplBase::_nil();

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace(
		"retry", "trying again because stub implementation changed");
	}
    }
    catch(const FailureException& ex)
    {
	//
	// If it's not safe to retry, throw the exception
	//
	checkRetry(retry, *ex.exception);

	//
	// Only retry once
	//
	retry = false;

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"trying again because of failure\n");
	    String_var exMsg = ex.exception -> _to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
    }
    //
    // We don't retry on non-locally raised exceptions
    //
/*
    catch(const TRANSIENT& ex)
    {
	//
	// If it's not safe to retry, throw the exception
	//
	checkRetry(retry, ex);

	//
	// Only retry once
	//
	retry = false;

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"trying again because server sent a TRANSIENT exception\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
    }
*/
}

AbstractBase_ptr
Object::_OB_toAbstractBase()
{
    return AbstractBase::_nil();
}

// ----------------------------------------------------------------------
// ObjectFactory constructor and destructor
// ----------------------------------------------------------------------

ObjectFactory::ObjectFactory()
    : destroy_(false)
{
}

ObjectFactory::~ObjectFactory()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// ObjectFactory public member implementation
// ----------------------------------------------------------------------

void
ObjectFactory::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Set the ORBInstance object to nil
    //
    orbInstance_ = ORBInstance::_nil();

    //
    // Set the PolicyManager object to nil
    //
    policyManager_ = PolicyManager::_nil();
}

void
ObjectFactory::setORBInstance(ORBInstance_ptr orbInstance)
{
    orbInstance_ = ORBInstance::_duplicate(orbInstance);
}

void
ObjectFactory::setPolicyManager(PolicyManager_ptr policyManager)
{
    policyManager_ = PolicyManager::_duplicate(policyManager);
}

Object_ptr
ObjectFactory::createObject(RefCountIOR_ptr ior)
{
    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // Check for nil object reference
    //
    if(strlen(ior -> value.type_id) == 0 &&
       ior -> value.profiles.length() == 0)
	return Object::_nil();

    //
    // Create a new CORBA::Object
    //
    Object_var p = new Object();
    RefCountPolicyList_var policyList = policies();
    p -> _OB_setup(orbInstance_, ior, policyList);

    return p._retn();
}

/* stringToObject operation is in INS.cpp */

RefCountPolicyList_ptr
ObjectFactory::policies()
{
    PolicyList_var policies = policyManager_ -> get_all_policy_overrides();
    return new RefCountPolicyList(policies);
}
