// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>

#include <POAControl.h>

// ----------------------------------------------------------------------
// POAControl constructor
// ----------------------------------------------------------------------

OBPortableServer::POAControl::POAControl()
    : state_(DestroyNotCalled),
      etherealize_(false),
      requests_(0)
{
}

// ----------------------------------------------------------------------
// POAControl public member implementation
// ----------------------------------------------------------------------

//
// Mark the state as destroy pending.  Return true if the destroy
// should complete, false otherwise.
//
bool
OBPortableServer::POAControl::markDestroyPending(
    bool etherealize, bool waitForCompletion)
{
    OB_SYNCHRONIZED;

    //
    // If the POA is in the middle of a destroy and waitForCompletion
    // is true then wait for the destroy to complete.
    //
    if(state_ != DestroyNotCalled)
    {
	while(waitForCompletion && state_ != DestroyCompleted)
	{
#ifndef HAVE_JTC        
	    //
	    // Should not occur if we're not multi-threaded
	    //
	    assert(false);
#else
	    //
	    // wait for the destroy to complete
	    //
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
#endif
	}
	return false;
    }

    //
    // Save the etherealize flag. Set the state to DestroyPending.
    //
    etherealize_ = etherealize;
    state_ = DestroyPending;

    return true;
}

//
// Increment the outstanding request count. Return true if the POA
// is not destroyed, false otherwise.
//
bool
OBPortableServer::POAControl::incrementRequestCount()
{
    OB_SYNCHRONIZED
	
    if(state_ != DestroyNotCalled)
    {
	while(state_ != DestroyCompleted)
	{
#ifndef HAVE_JTC        
	    //
	    // Should not occur if we're not multi-threaded
	    //
	    assert(false);
#else
	    //
	    // wait for the destroy to complete
	    //
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
#endif
	}
	return false;
    }

    //
    // Reference successfully acquired, return true
    //
    requests_++;

    return true;
}

//
// Decrement the outstanding request count. If the state is
// DestroyPending then wake any waiting threads and return true.
//
bool
OBPortableServer::POAControl::decrementRequestCount()
{
    OB_SYNCHRONIZED
	
    assert(requests_ > 0);
    requests_--;

    //
    // If we have no more outstanding requests notify anyone
    // waiting for this state
    //
    if(requests_ == 0)
    {
	if(state_ == DestroyPending)
	{
#ifdef HAVE_JTC
	    notifyAll();
#endif
	    return true;
	}
    }
    return false;
}

//
// Wait for any pending requests to terminate. Return true if
// the destroy should complete, false otherwise.
//
bool
OBPortableServer::POAControl::waitPendingRequests()
{
    OB_SYNCHRONIZED;

    while(requests_ > 0)
    {
#ifndef HAVE_JTC
	//
	// Should not occur if we're not multi-threaded
	//
	assert(false);
#else
	//
	// wait for the destroy to complete
	//
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
#endif
    }

    //
    // If the state is not DestroyPending then some other thread
    // is in the progress of completing destroying the POA
    //
    if(state_ != DestroyPending)
	return false;

    state_ = DestroyInProgress;
    return true;
}

//
// Return true if there are outstanding pending requests
//
bool
OBPortableServer::POAControl::hasPendingRequests()
{
    OB_SYNCHRONIZED;

    return requests_ > 0;
}

//
// Has the POA been destroyed?
//
bool
OBPortableServer::POAControl::getDestroyed()
{
    OB_SYNCHRONIZED;

    return state_ == DestroyCompleted;
}

//
// Mark the state as DestroyCompleted, wake any waiting threads
//
void
OBPortableServer::POAControl::markDestroyCompleted()
{
    OB_SYNCHRONIZED;
    state_ = DestroyCompleted;
#ifdef HAVE_JTC
    notifyAll();
#endif
}

//
// Should etherealize be called?
//
bool
OBPortableServer::POAControl::etherealize()
{
    OB_SYNCHRONIZED;
    return etherealize_;
}
