// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/POAServantBase.h>
#include <OB/POAInterface.h>

#include <POACurrent_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// POACurrentState class. This holds the internal state of the POA
// Current implementation object.
// ----------------------------------------------------------------------

namespace PortableServer
{

class POACurrentState
{
    void clean()
    {
        poa = POA::_nil();
        servant = 0;
	op = 0;
        oid = 0;
        cookie = 0;
    }

#ifdef HAVE_JTC
    static JTCMutex freeMutex_;
#endif
    static POACurrentState* freeList_;
    
public:
    
    POACurrentState() : cookie(0) { }

    static void freeState(POACurrentState* next)
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        next -> clean();
        next -> next = freeList_;
        freeList_ = next;
    }
    
    static POACurrentState* newState()
    {
#ifdef HAVE_JTC
        JTCSynchronized sync(freeMutex_);
#endif
        POACurrentState* next;
        if(freeList_ != 0)
        {
            next = freeList_;
            freeList_ = freeList_ -> next;
        }
        else
        {
            next = new POACurrentState;
        }
        return next;
    }

    static void destroy()
    {
        while(freeList_ != 0)
        {
            POACurrentState* tmp = freeList_;
            freeList_ = freeList_ -> next;
            delete tmp;
        }
    }

    POA_var poa;
    ServantBase_var servant;
    ServantLocator::Cookie cookie;

    //
    // We can make these fields pointers since they should be valid
    // for the duration of the invocation.
    //
    const ObjectId* oid;
    const char* op;

    POACurrentState* next;
};

} // End of namespace PortableServer

#ifdef HAVE_JTC
JTCMutex PortableServer::POACurrentState::freeMutex_;
#endif
PortableServer::POACurrentState*
    PortableServer::POACurrentState::freeList_ = 0;

// ----------------------------------------------------------------------
// Current_impl constructor and destructor
// ----------------------------------------------------------------------

PortableServer::Current_impl::Current_impl()
{
#ifdef HAVE_JTC
    stateKey_ = JTCTSS::allocate();
#else
    state_ = 0;
#endif
}

PortableServer::Current_impl::~Current_impl()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_ == 0);
    POACurrentState::destroy();
    
#ifdef HAVE_JTC
    JTCTSS::release(stateKey_);
#endif
}

// ----------------------------------------------------------------------
// Current_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
PortableServer::Current_impl::get_POA()
    throw(PortableServer::Current::NoContext,
          SystemException)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    if(state_ == 0 || is_nil(state_ -> poa))
        throw Current::NoContext();
    return POA::_duplicate(state_ -> poa);
}


PortableServer::ObjectId*
PortableServer::Current_impl::get_object_id()
    throw(PortableServer::Current::NoContext,
          SystemException)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    if(state_ == 0 || is_nil(state_ -> poa))
        throw Current::NoContext();
    return new ObjectId(*state_ -> oid);
}

void
PortableServer::Current_impl::_OB_preinvoke(
    PortableServer::POA_ptr poa,
    PortableServer::ServantBase* servant,
    const char* op,
    const PortableServer::ObjectId& oid,
    PortableServer::ServantLocator::Cookie cookie)
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    POACurrentState* ns = POACurrentState::newState();
    ns -> poa = POA::_duplicate(poa);

    //
    // These should be valid for the duration of the invocation
    //
    ns -> oid = &oid;
    ns -> cookie = cookie;
    ns -> servant = servant;
    ns -> op = op;
    if(servant != 0)
        servant -> _add_ref();

    ns -> next = state_;
    state_ = ns;

#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}

void
PortableServer::Current_impl::_OB_postinvoke()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    POACurrentState* save = state_;
    state_ = state_ -> next;
    POACurrentState::freeState(save);
    
#ifdef HAVE_JTC
    JTCTSS::set(stateKey_, state_);
#endif
}

bool
PortableServer::Current_impl::_OB_inUpcall() const
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    return state_ != 0;
}

PortableServer::ServantBase*
PortableServer::Current_impl::_OB_getServant()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif
    
    assert(state_);
    return state_ -> servant;
}

PortableServer::ServantLocator::Cookie
PortableServer::Current_impl::_OB_getCookie()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return state_ -> cookie;
}

const PortableServer::ObjectId&
PortableServer::Current_impl::_OB_getObjectId()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return *state_ -> oid;
}

const char*
PortableServer::Current_impl::_OB_getOp()
{
#ifdef HAVE_JTC
    POACurrentState* state_ = (POACurrentState*)JTCTSS::get(stateKey_);
#endif

    assert(state_);
    return state_ -> op;
}
