// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/MarshalStubImpl.h>
#include <OB/DirectStubImpl.h>
#include <OB/OAInterface.h>
#include <OB/Upcall.h>
#include <OB/Server.h>
#include <OB/IMRState.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/Timer.h>
#include <OB/RefCounted.h>
#include <OB/IIOPConfig.h>
#include <OB/OCI_IIOP.h>
#include <OB/Net.h> // For MAXHOSTNAMELEN

#include <POAManagerFactory_impl.h>
#include <POAManager_impl.h>
#include <POA_impl.h>
#include <OADProcessEndpoint_impl.h>
#include <IMRActiveStateHolder.h>
#include <POALocator.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// POAManagerFactory_impl private and protected member implementation
// ----------------------------------------------------------------------

char*
OBPortableServer::POAManagerFactory_impl::getUniqueName()
{
    timeval tv = TimeNow();

    String_var name = string_dup("POAManager-");
    name += tv.tv_sec;
    name += count_++;

    assert(!managers_ -> containsKey(name));

    return name._retn();
}

void
OBPortableServer::POAManagerFactory_impl::validateName(const char* name)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists)
{
    //
    // Does the POAManager exist?
    //
    POAManager_var manager;
    if(managers_ -> get(name, manager))
    {
        //
        // If the POAManager is INACTIVE then remove it from the list,
        // and allow the user to re-add the POAManager
        //
        if(manager -> get_state() == PortableServer::POAManager::INACTIVE)
            managers_ -> remove(name);
        else
            throw POAManagerAlreadyExists();
    }
}

// ----------------------------------------------------------------------
// POAManagerFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OBPortableServer::POAManagerFactory_impl::POAManagerFactory_impl(
    const char* serverName,
    const char* serverInstance)
    : managers_(new POAManagerTable(7)),
      serverName_(serverName),
      serverInstance_(CORBA::string_dup(serverInstance)),
      count_(0),
      imrHolder_(new OB::IMRActiveStateHolder(serverName_, serverInstance_)),
      poaLocator_(new POALocator)
{
    if(serverName == 0)
    {
	serverName_ = string_dup("_RootPOA");
        needIMR_ = false;
    }
    else
        needIMR_ = true;
}

OBPortableServer::POAManagerFactory_impl::~POAManagerFactory_impl()
{
    assert(managers_ -> size() == 0);

    delete managers_;
    delete imrHolder_;
    delete poaLocator_;
}

// ----------------------------------------------------------------------
// POAManagerFactory_impl public member implementation
// ----------------------------------------------------------------------

OBPortableServer::POAManager_ptr
OBPortableServer::POAManagerFactory_impl::create_poa_manager(const char* name)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists,
          OCI::InvalidParam,
          SystemException)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(*managers_);
#endif

    String_var nameDummy;
    if(strlen(name) == 0)
    {
        nameDummy = getUniqueName();
        name = nameDummy.in();
    }
    else
        validateName(name);

    //
    // Get the IIOP acceptor configuration parameters and
    // create an IIOP acceptor
    //
    Properties_var props = orbInstance_ -> getProperties();
    Logger_var logger = orbInstance_ -> getLogger();
    IIOPConfig cfg(props, logger);
    cfg.parseAll(name);
    OCI::ParamSeq_var params = cfg.getParams();

    OCI::Acceptor_var acceptor;
    try
    {
        OCI::AccFactoryRegistry_var registry =
            orbInstance_ -> getAccFactoryRegistry();
        OCI::AccFactory_var factory =
            registry -> get_factory(OCI::IIOP::TAG_IIOP);
        acceptor = factory -> create_acceptor(params);
    }
    catch(const OCI::NoSuchFactory&)
    {
        logger -> error("unable to create IIOP acceptor");
        throw INITIALIZE();
    }

    //
    // Create the new POAManager_impl and add to the table
    //
    OCI::AcceptorSeq acceptors;
    acceptors.length(1);
    acceptors[0] = acceptor;
    POAManager_var manager = new POAManager_impl(orbInstance_, imrHolder_,
                                                 poaLocator_, name,
                                                 serverName_, acceptors);
    managers_ -> put(name, manager);

    return manager._retn();
}

OBPortableServer::POAManager_ptr
OBPortableServer::POAManagerFactory_impl::create_poa_manager_with_config(
    const char* name,
    const AcceptorConfigSeq& config)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists,
          OCI::NoSuchFactory,
          OCI::InvalidParam,
	  CORBA::SystemException)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(*managers_);
#endif

    String_var nameDummy;
    if(strlen(name) == 0)
    {
        nameDummy = getUniqueName();
        name = nameDummy.in();
    }
    else
        validateName(name);

    //
    // Create acceptors
    //
    OCI::AccFactoryRegistry_var registry =
        orbInstance_ -> getAccFactoryRegistry();
    OCI::AcceptorSeq acceptors;
    acceptors.length(config.length());
    for(ULong i = 0 ; i < config.length() ; i++)
    {
        OCI::AccFactory_var factory = registry -> get_factory(config[i].id);
        acceptors[i] = factory -> create_acceptor(config[i].params);
    }

    //
    // Create the new POAManager_impl and add to the table
    //
    POAManager_var manager = new POAManager_impl(orbInstance_, imrHolder_,
                                                 poaLocator_, name,
                                                 serverName_, acceptors);
    managers_ -> put(name, manager);

    return manager._retn();
}

OBPortableServer::POAManagerSeq*
OBPortableServer::POAManagerFactory_impl::get_poa_managers()
    throw(SystemException)
{
    POAManagerTable::Enumerator keys = managers_ -> keys();
    POAManagerSeq_var result = new POAManagerSeq(keys.length());

    POAManager_var manager;
    for(ULong i = 0 ; i < keys.length() ; i++)
    	if(managers_ -> get(keys[i], manager)) {
	    CORBA::ULong resultLength  = result -> length();
	    result -> length(resultLength + 1);
	    result[resultLength] = manager;
        }

    return result._retn();
}

void
OBPortableServer::POAManagerFactory_impl::destroy()
    throw(SystemException)
{
    //
    // Remove the references to the orbInstance_
    //
    orbInstance_ = ORBInstance::_nil();

    //
    // Shutdown, if necessary
    //
    if(managers_ -> size() != 0)
	_OB_deactivate();
}

void
OBPortableServer::POAManagerFactory_impl::_OB_deactivate()
    throw(SystemException)
{
    //
    // Deactivate each of the POAManagers
    //
    POAManagerTable::Enumerator keys = managers_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	POAManager_var manager;
	if(managers_ -> get(keys[i], manager))
	{
	    try
	    {
		manager -> deactivate(true, true);
	    }
	    catch(const PortableServer::POAManager::AdapterInactive&)
	    {
		// Ignore
	    }
	}
    }
    managers_ -> clear();

    //
    // Tell the IMR that the ORB is STOPPING
    //
    IMR::ActiveState_var as = imrHolder_ -> getActiveState();
    if(!is_nil(as))
    {
	Logger_var logger = orbInstance_ -> getLogger();
	try
	{
	    as -> set_status(serverInstance_, IMR::STOPPING);
	}
	catch(...)
	{
	    String_var msg = string_dup(serverName_);
	    msg += ": Cannot contact IMR on shutdown";
	    logger -> warning(msg);
	}

	//
	// Clear the IMR::Server record
	//
	imrHolder_ -> setActiveState(IMR::ActiveState::_nil());
    }
}

DirectStubImpl_ptr
OBPortableServer::POAManagerFactory_impl::_OB_createDirectStubImpl(
    RefCountIOR_ptr ior, 
    RefCountPolicyList_ptr policies)
    throw(SystemException, LocationForward)
{
    POAManagerTable::Enumerator keys = managers_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	try
	{
	    POAManager_var manager;
	    if(managers_ -> get(keys[i], manager))
	    {
		OCI::AcceptorSeq_var acceptors = manager -> get_acceptors();
                for(ULong j = 0 ; j < acceptors -> length() ; j++)
                {
                    OCI::ProfileInfoSeq_var profileInfos =
                        acceptors[j] -> get_local_profiles(ior -> value);

                    //
                    // If the IOR is local then at least one ProfileInfo
                    // will be returned
                    //
                    if(profileInfos -> length() > 0)
                    {
                        POAManager_impl* managerImpl =
                            dynamic_cast<POAManager_impl*>(manager.in());
                        
                        //
                        // In the case that the servant cannot support a
                        // DirectStubImpl a null-pointer will be returned
                        //
                        return managerImpl ->
                            _OB_createDirectStubImpl(profileInfos[0].key,
			    			     policies);
                    }
                }
	    }
	}
	catch(const PortableServer::POAManager::AdapterInactive&)
	{
	    //
	    // Ignore -- the POAManager isn't valid anymore
	    //
	}
    }

    return DirectStubImpl::_nil();
}

void
OBPortableServer::POAManagerFactory_impl::_OB_setORBInstance(
    ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);
}

void
OBPortableServer::POAManagerFactory_impl::_OB_initializeIMR(
    OBPortableServer::POA_impl* root,
    OB::ORBControl_ptr orbControl)
{
    if(!needIMR_)
        return;

    //
    // Create the OAD::ProcessMonitor servant
    //
    OAD::ProcessEndpoint_impl* processEndpointImpl =
	new OAD::ProcessEndpoint_impl(serverName_, serverInstance_, root,
				      orbControl);

    //
    // We assign this to a _var type to ensure that it's destroyed
    // appropriately
    //
    processEndpoint_ = processEndpointImpl;

    IMR::Domain_var imrDomain;
    try
    {
        InitialServiceManager_var initServiceManager =
            orbInstance_ -> getInitialServiceManager();
        Object_var imrObj = initServiceManager ->
            resolveInitialReferences("IMR");
        imrDomain = IMR::Domain::_narrow(imrObj);
    }
    catch(const ORB::InvalidName&)
    {
	// Ignore -- this will be handled later
    }

    Logger_var logger = orbInstance_ -> getLogger();
    Properties_var properties = orbInstance_ -> getProperties();

    //
    // IMR::IMRDomain not reachable?
    //
    if(is_nil(imrDomain))
    {
	String_var msg = string_dup(serverName_);
	msg += ": IMRDomain not reachable";
	logger -> error(msg);
	throw INITIALIZE();
    }

    //
    // Check if we should try to register with the IMR.
    //
    const char* exec = properties -> getProperty("ooc.orb.imr.register");
    if(exec != 0)
    {
        char host[MAXHOSTNAMELEN + 1];
	gethostname(host, MAXHOSTNAMELEN);

	try
	{
	    imrDomain -> registerServer(serverName_, exec, host);
	}
	catch(const IMR::ServerAlreadyRegistered&)
	{
	    String_var msg = string_dup(serverName_);
	    msg += ": (IMR) Server already registered";
	    logger -> error(msg);
	    throw INITIALIZE();
	}
	catch(const NO_PERMISSION&)
	{
	    String_var msg = string_dup(serverName_);
	    msg += ": (IMR) Not running in administrative mode";
	    logger -> error(msg);
	    throw INITIALIZE();
	}
    }

    //
    // Tell the IMR that we're starting up
    //
    try
    {
	//
	// Retrieve the ServerFactory
	//
	IMR::ActiveStateFactory_var factory =
	    imrDomain -> get_active_state_factory();

	//
	// Retrieve our server record
	//
	IMR::ActiveState_var as = factory -> get_active_state(serverName_);
	imrHolder_ -> setActiveState(as);
	
	//
	// This is the ProcessEndpointManager
	//
        OAD::ProcessEndpointManager_var endpoint;

	PortableServer::ObjectId empty;
	RefCountIOR_var ior = root -> _OB_createIOR(empty);

	//
	// Tell the IMR that we are STARTING.
	//
	as -> startup(serverInstance_, ior -> value, endpoint);

	//
	// Link with the OAD ProcessEndpoint
	//
	processEndpointImpl -> _OB_establish_link(endpoint);
    }
    catch(const BAD_PARAM&)
    {
	String_var msg = string_dup(serverName_);
	msg += ": (IMR) Server already running";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::NoSuchServer&)
    {
	String_var msg = string_dup(serverName_);
	msg += ": (IMR) Not registered with IMR";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::NoSuchOAD&)
    {
	String_var msg = string_dup(serverName_);
	msg += ": (IMR) No OAD for host";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::OADNotRunning&)
    {
	String_var msg = string_dup(serverName_);
	msg += ": (IMR) OAD not registered";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const OAD::AlreadyLinked&)
    {
	String_var msg = string_dup(serverName_);
	msg += ": (IMR) Process registered with OAD";
	logger -> error(msg);
	throw INITIALIZE();
    }
}
