// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/MarshalStubImpl.h>
#include <OB/DirectStubImpl.h>
#include <OB/Upcall.h>
#include <OB/Server.h>
#include <OB/IMRState.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/RefCounted.h>
#include <OB/OAInterface.h>

#include <POAManager_impl.h>
#include <POA_impl.h>
#include <OCI_impl.h>
#include <OCI_IIOP_impl.h>
#include <IMRActiveStateHolder.h>
#include <POAOAInterface_impl.h>
#include <POACurrent_impl.h>
#include <BootManager_impl.h>
#include <POALocator.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// POAManager_impl private member implementation
// ----------------------------------------------------------------------

//
// If we're in the context of a method invocation then raise a
// BAD_INV_ORDER exception, otherwise wait for all pending requests to
// complete
//
void
OBPortableServer::POAManager_impl::waitPendingRequests()
    throw(SystemException)
{
    //
    // Find out whether we're inside a method invocation
    //
    bool inInvocation = false;
    try
    {
	InitialServiceManager_var initialServiceManager =
	    orbInstance_ -> getInitialServiceManager();
	CORBA::Object_var o = initialServiceManager ->
	    resolveInitialReferences("POACurrent");
	PortableServer::Current_impl* current =
	    dynamic_cast<PortableServer::Current_impl*>(o.in());
	inInvocation = (current != 0 && current -> _OB_inUpcall());
    }
    catch(const ORB::InvalidName&)
    {
    }

    //
    // TODO: No minor code?
    //
    if(inInvocation)
	throw BAD_INV_ORDER(0, COMPLETED_NO);

    //
    // Wait for all pending requests from all POAs to complete
    //
    POAHashTable::Enumerator keys = poas_ -> keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
	PortableServer::POA_var poa;
	if(poas_ -> get(keys[i], poa))
	{
	    POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
	    poaImpl -> _OB_waitPendingRequests();
	}
    }
}

// ----------------------------------------------------------------------
// POAManager_impl constructor and destructor
// ----------------------------------------------------------------------

OBPortableServer::POAManager_impl::POAManager_impl(
    ORBInstance_ptr orbInstance,
    IMRActiveStateHolder* imrHolder,
    POALocator* poaLocator,
    const char* name,
    const char* serverName,
    const OCI::AcceptorSeq& acceptors)
    throw(SystemException)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      poas_(new POAHashTable(63)),
      state_(PortableServer::POAManager::HOLDING),
      acceptors_(acceptors),
      name_(name),
      serverName_(serverName),
      imrHolder_(imrHolder),
      bootManager_(orbInstance -> getBootManager()),
      poaLocator_(poaLocator)
{
    try
    {
        //
        // Create the OAInterface
        //
        oaInterface_ = new POAOAInterface_impl(this, serverName_,
					       orbInstance_);

        //
        // Construct the root name of the property key for this POAManager
        // instance
        //
        String_var rootKey = string_dup("ooc.orb.poamanager.");
        rootKey += name_;
        rootKey += ".";
        ULong rootKeyLen = strlen(rootKey);

        //
        // Get the Logger
        //
        Logger_var logger = orbInstance_ -> getLogger();

        //
        // Get the ORB properties
        //
        Properties_var properties = orbInstance -> getProperties();

        //
        // The set of properties that the POAManager supports
        //
        const char* props[] =
        {
            "conc_model",
            "host", // deprecated
            "numeric", // deprecated
            "port", // deprecated
            "version"
        };
        int numProps = sizeof(props)/sizeof(props[0]);

        //
        // Check over the POAManager properties and find out whether an
        // unknown property is present
        //
        Properties::KeySeq keys = properties -> getKeys();
        for(ULong j = 0 ; j < keys.length() ; j++)
        {
            const char* key = keys[j];

            //
            // Ignore ooc.orb.oa.thread_pool
            //
            if(strcmp(key, "ooc.orb.oa.thread_pool") == 0)
                continue;

	    //
    	    // Remove the property prefix
	    //
            const char* prop = key;
            if(strncmp(key, rootKey, rootKeyLen) == 0)
                prop += rootKeyLen;
            else if(strncmp(key, "ooc.orb.oa.", 11) == 0)
                prop += 11;
            else
                continue;

            //
            // Check for a match among the supported properties
            //
            int i;
            for(i = 0 ; i < numProps ; i++)
                if(strcmp(prop, props[i]) == 0)
                    break;

	    //
	    // Was the property present?
	    //
	    if(i >= numProps)
	    {
                String_var err = key;
	        err += ": unknown property";
                logger -> warning(err);
	    }
        }

        //
        // Configuration parameters
        //
        version_.major = 1;
        version_.minor = 2;
        Server::ConcModel concModel = Server::Reactive;

        //
        // Parse the individual properties
        //
        for(int i = 0 ; i < numProps ; i++)
        {
	    const char* key = props[i];

	    //
	    // First check the specific POAManager key
	    //
	    String_var fullkey = rootKey;
	    fullkey += key;
	    const char* value = properties -> getProperty(fullkey);

	    //
	    // If the specific POAManager key doesn't have a value then
	    // check the default value "ooc.orb.oa.*"
	    //
	    if(value == 0)
	    {
	        fullkey = string_dup("ooc.orb.oa.");
	        fullkey += key;
	        value = properties -> getProperty(fullkey);
	    }

	    //
	    // No value
	    //
	    if(value == 0)
	        continue;

	    if(strcmp(key, "conc_model") == 0)
	    {
	        if(strcmp(value, "reactive") == 0)
		    concModel = Server::Reactive;
#ifdef HAVE_JTC
	        else if(strcmp(value, "threaded") == 0)
		    concModel = Server::Threaded;
	        //
	        // Technically the only valid values for
	        // ooc.orb.poamanager.*.conc_model are "reactive" and
	        // "threaded" since this is the communications conc model.
	        // However, we'll also accept the following values since
	        // we might be parsing "ooc.orb.oa.conc_model" (which
	        // represents the default value for both the comm conc
	        // model *and* the method dispatch model).
	        //
	        else if(strcmp(value, "thread_per_client") == 0)
		    concModel = Server::Threaded;
	        else if(strcmp(value, "thread_per_request") == 0)
		    concModel = Server::Threaded;
	        else if(strcmp(value, "thread_pool") == 0)
		    concModel = Server::Threaded;
#endif
	        else
	        {
		    String_var err = fullkey;
		    err += ": unknown value";
		    logger -> warning(err);
	        }
    	    }
            else if(strcmp(key, "version") == 0)
            {
                if(strcmp(value, "1.0") == 0)
                {
                    version_.major = 1;
                    version_.minor = 0;
                }
                else if(strcmp(value, "1.1") == 0)
                {
                    version_.major = 1;
                    version_.minor = 1;
                }
                else if(strcmp(value, "1.2") == 0)
                {
                    version_.major = 1;
                    version_.minor = 2;
                }
                else
                {
                    String_var err = fullkey;
                    err += ": expected `1.0', `1.1', or `1.2'";
                    logger -> error(err);
                    throw INITIALIZE();
                }
            }
        }

        //
        // Create the server manager
        //
        serverManager_ = new ServerManager(orbInstance_,
					   acceptors_, 
					   oaInterface_,
					   concModel);

	//
	// Get the boot manager implementation
	//
	bootManagerImpl_ = dynamic_cast<BootManager_impl*>(bootManager_.in());
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }
}

OBPortableServer::POAManager_impl::~POAManager_impl()
{
    //
    // Delete the POA hashtable
    //
    delete poas_;
}

// ----------------------------------------------------------------------
// POAManager_impl public member implementation
// ----------------------------------------------------------------------

//
// Mapping for PortableServer::POAManager
//
void
OBPortableServer::POAManager_impl::activate()
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    OB_SYNCHRONIZED;

    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == PortableServer::POAManager::INACTIVE)
        throw PortableServer::POAManager::AdapterInactive();

    if(state_ != PortableServer::POAManager::ACTIVE)
    {
        state_ = PortableServer::POAManager::ACTIVE;

#ifdef HAVE_JTC
	//
	// Notify all of a state transition
	//
	notifyAll();
#endif

	//
	// Activate the server manager
	//
	serverManager_ -> activate();

	//
	// Tell the OAInterface to accept requests
	//
	oaInterface_ -> activate();

	//
	// Inform IMR of state change
	//
	_OB_updateIMR();
    }

    imrHolder_ -> serverRunning(orbInstance_);
}

void
OBPortableServer::POAManager_impl::hold_requests(Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    OB_SYNCHRONIZED

    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == PortableServer::POAManager::INACTIVE)
        throw PortableServer::POAManager::AdapterInactive();

    if(state_ != PortableServer::POAManager::HOLDING)
    {
        state_ = PortableServer::POAManager::HOLDING;

#ifdef HAVE_JTC
	//
	// Notify all of a state transition
	//
	notifyAll();
#endif
        
        //
        // Ask the server manager to hold
        //
	serverManager_ -> hold();

	//
	// Tell the OAInterface to accept requests
	//
	oaInterface_ -> activate();

	//
	// Wait for all pending requests to complete, if asked
	//
	if(waitCompletion)
	    waitPendingRequests();

	//
	// Inform IMR of state change
	//
	_OB_updateIMR();
    }
}

void
OBPortableServer::POAManager_impl::discard_requests(Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    OB_SYNCHRONIZED

    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception.
    //
    if(state_ == PortableServer::POAManager::INACTIVE)
        throw PortableServer::POAManager::AdapterInactive();

    if(state_ != PortableServer::POAManager::DISCARDING)
    {
        state_ = PortableServer::POAManager::DISCARDING;

#ifdef HAVE_JTC
	//
	// Notify all of a state transition
	//
	notifyAll();
#endif

	//
	// Tell the OAInterface to discard requests
	//
	oaInterface_ -> discard();

	//
	// Wait for all pending requests to complete, if asked
	//
	if(waitCompletion)
	    waitPendingRequests();

	//
	// Inform IMR of state change
	//
	_OB_updateIMR();
    }
}

//
// TODO: waitCompletion?
//
void
OBPortableServer::POAManager_impl::deactivate(Boolean etherealize,
					      Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    OB_SYNCHRONIZED
    
    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception.
    //
    if(state_ == PortableServer::POAManager::INACTIVE)
	throw PortableServer::POAManager::AdapterInactive();

    if(state_ != PortableServer::POAManager::INACTIVE)
    {
	//
	// Destroy the server manager
	//
	serverManager_ -> destroy();

	//
	// Set the state to INACTIVE *after* the serverManager_ has
	// been destroyed, to avoid a race condition.
	//
	state_ = PortableServer::POAManager::INACTIVE;

#ifdef HAVE_JTC
	//
	// Notify all of a state transition
	//
	notifyAll();
#endif
	
	//
	// Etherealize each of the servants associated with each POA
	//
	if(etherealize)
	{
	    POAHashTable::Enumerator keys = poas_ -> keys();
	    for(ULong i = 0; i < keys.length(); ++i)
	    {
		PortableServer::POA_var poa;
		if(poas_ -> get(keys[i], poa))
		{
		    POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
		    poaImpl -> _OB_etherealize();
		}
	    }
	}

	//
	// Inform IMR of state change
	//
	_OB_updateIMR();
    }
}

PortableServer::POAManager::State
OBPortableServer::POAManager_impl::get_state()
    throw(SystemException)
{
    //
    // We'll remove the synchronization at present since this is a
    // simple state variable
    //
    //OB_SYNCHRONIZED

    return state_;
}

//
// Mapping for OBPortableServer::POAManager
//

char*
OBPortableServer::POAManager_impl::name()
    throw(SystemException)
{
    return string_dup(name_.in());
}

OCI::AcceptorSeq*
OBPortableServer::POAManager_impl::get_acceptors()
     throw(PortableServer::POAManager::AdapterInactive,
	   SystemException)
{
    OB_SYNCHRONIZED

    if(state_ == PortableServer::POAManager::INACTIVE)
	throw PortableServer::POAManager::AdapterInactive();

    return new OCI::AcceptorSeq(acceptors_);
}

//
// Internal methods.
//

//
// Register a POA with this POAManager
//
void
OBPortableServer::POAManager_impl::_OB_addPOA(PortableServer::POA_ptr poa,
					      const IMR::POAName& id)
    throw()
{
    OB_SYNCHRONIZED

    assert(!poas_ -> containsKey(id));
    poas_ -> put(id, PortableServer::POA::_duplicate(poa));

    poaLocator_ -> add(poa, id);
}

//
// Un-register a POA with this POAManager
//
void
OBPortableServer::POAManager_impl::_OB_removePOA(const IMR::POAName& id)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    assert(poas_ -> containsKey(id));
    poas_ -> remove(id);

    poaLocator_ -> remove(id);
}

IMR::ActiveState_ptr
OBPortableServer::POAManager_impl::_OB_getIMRActiveState()
    throw(SystemException)
{
    return imrHolder_ -> getActiveState();
}

void
OBPortableServer::POAManager_impl::_OB_updateIMR()
    throw(SystemException)
{
    IMR::ActiveState_var as = imrHolder_ -> getActiveState();
    if(is_nil(as))
        return;

    //
    // Get list of persistent POAs
    //
    IMR::POANameSeq poas(poas_ -> size());
    POAHashTable::Enumerator keys = poas_ -> keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
        PortableServer::POA_var poa;
        if(poas_ -> get(keys[i], poa))
        {
            POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
	    poaImpl -> _OB_isPersistent(poas);
        }
    }

    //
    // Inform the IMR of the state change of the persistent POAs
    //
    if(poas.length() != 0)
    {
        try
	{
	    IMR::POAStatus state;
	    switch(state_)
	    {
	    case PortableServer::POAManager::INACTIVE:
	        state = IMR::INACTIVE;
	        break;
	    case PortableServer::POAManager::ACTIVE:
	        state = IMR::ACTIVE;
	        break;
	    case PortableServer::POAManager::HOLDING:
	        state = IMR::HOLDING;
	        break;
	    case PortableServer::POAManager::DISCARDING:
	        state = IMR::DISCARDING;
	        break;
	    }
	    as -> poa_status_update(poas, state);
	}
	catch(const CORBA::SystemException&)
	{
	    //
	    // TODO: Is any action required here.
	    //
	}
    }
}

DirectStubImpl_ptr
OBPortableServer::POAManager_impl::_OB_createDirectStubImpl(
    const OCI::ObjectKey& key,
    RefCountPolicyList_ptr policies)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException, LocationForward)
{
    OB_SYNCHRONIZED

    if(state_ == PortableServer::POAManager::INACTIVE)
	throw PortableServer::POAManager::AdapterInactive();
    
    ObjectKeyData data;
    if(ParseObjectKey(key, data))
    {
        PortableServer::POA_var poa = _OB_locatePOA(data);
        if(!is_nil(poa))
        {
            POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
            return poaImpl -> _OB_createDirectStubImpl(data.oid, policies);
        }
    }

    //
    // Check to see if the BootManager knows of a reference for the
    // ObjectKey. If so, forward the request.
    //
    RefCountIOR_var ior = bootManagerImpl_ -> _OB_locate(key);
    if(!is_nil(ior))
	throw LocationForward(ior, false);

    //
    // In this case there is no POA for a local servant. This is an
    // OBJECT_NOT_EXIST exception.
    //
    throw OBJECT_NOT_EXIST();
    return DirectStubImpl::_nil(); // Some compilers need this
}

PortableServer::POA_ptr
OBPortableServer::POAManager_impl::_OB_locatePOA(const ObjectKeyData& data)
    throw(LocationForward,
	  SystemException)
{
    assert(data.poaID.length() != 0);

    //
    // If the GIOP engine sends a request while the POAManager is in
    // INACTIVE state, then something is wrong.
    //
    assert(get_state() != PortableServer::POAManager::INACTIVE);

    PortableServer::POA_var poa;
    if(!poas_ -> get(data.poaID, poa))
    {
	//
	// The POA isn't contained in our local POA table. Ask the
	// POALocator to locate the POA.
	//
	poa = poaLocator_ -> locate(data);

        //
        // If the POA is connected to some other POAManager (and hence
        // some other end-point) then location forward
        //
        if(!is_nil(poa))
        {
            PortableServer::POAManager_var manager = poa -> the_POAManager();
            if(manager.in() != this)
            {
                POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
                RefCountIOR_var ior = poaImpl -> _OB_createIOR(data.oid);
                throw LocationForward(ior, false);
            }
        }
    }

    //
    // If the POA doesn't match the ObjectKeyData then this POA isn't
    // present.
    //
    if(!is_nil(poa))
    {
        POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
        if(!poaImpl -> _OB_poaMatches(data))
            poa = POA::_nil();
    }

    return poa._retn();
}

CollocatedServer_ptr
OBPortableServer::POAManager_impl::_OB_getCollocatedServer()
    throw()
{
    return serverManager_ -> getCollocatedServer();
}

void
OBPortableServer::POAManager_impl::_OB_validateState()
    throw(SystemException)
{
    OB_SYNCHRONIZED

    while(true)
    {
	//
	// If POAManager::activate() has been called then we're done
	//
	if(state_ == PortableServer::POAManager::ACTIVE)
	    break;

	//
	// If the POAManager is INACTIVE or DISCARDING then throw a
	// TRANSIENT exception
	//
	if(state_ == PortableServer::POAManager::INACTIVE ||
	  state_ == PortableServer::POAManager::DISCARDING)
	    throw TRANSIENT(0, COMPLETED_NO);
	
#ifdef HAVE_JTC	    
	//
	// Wait for a state transition
	//
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
#else
	//
	// This will cause a deadlock
	//
	throw BAD_INV_ORDER();
#endif
    }
}

const GIOP::Version&
OBPortableServer::POAManager_impl::_OB_getGIOPVersion() const
    throw()
{
    return version_;
}
