// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/DirectStubImpl.h>
#include <OB/POAServantActivator.h>
#include <OB/POAServantBase.h>
#include <OB/PIManager.h>
#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/OBPolicies.h>
#include <OB/DSI.h>
#include <OB/ObjectIdHasher.h>
#include <OB/list.h>

#include <POAStrategy.h>
#include <POAPolicies_impl.h>
#include <POACurrent_impl.h>

using namespace OB;
using namespace CORBA;

// ----------------------------------------------------------------------
// PortableServer servant manager strategies
// ----------------------------------------------------------------------


namespace OBPortableServer
{

//
// Strategy for ServantLocators
//
class ServantLocatorStrategy : public ServantManagerStrategy OB_COMMA_MUTEX
{
    //
    // The servant locator
    //
    PortableServer::ServantLocator_var servantLocator_;

public:

    virtual void
    destroy();

    virtual void
    setServantManager(PortableServer::ServantManager_ptr)
	throw(SystemException);

    virtual PortableServer::ServantManager_ptr
    getServantManager();

    PortableServer::ServantBase*
    preinvoke(const PortableServer::ObjectId&,
	      PortableServer::POA_ptr,
	      const char*,
	      PortableServer::ServantLocator::Cookie&)
	throw(SystemException,
	      LocationForward);

    void
    postinvoke(const PortableServer::ObjectId&,
	       PortableServer::POA_ptr,
	       const char*,
	       PortableServer::ServantLocator::Cookie,
	       PortableServer::ServantBase*);
};

//
// Strategy for ServantActivators
//
class ServantActivatorStrategy : public ServantManagerStrategy OB_COMMA_MUTEX
{
    //
    // The servant activator
    //
    PortableServer::ServantActivator_var servantActivator_;

public:

    virtual void
    destroy();

    virtual void
    setServantManager(PortableServer::ServantManager_ptr manager)
	throw(SystemException);

    virtual PortableServer::ServantManager_ptr
    getServantManager();

    PortableServer::ServantBase*
    incarnate(const PortableServer::ObjectId&, PortableServer::POA_ptr)
	throw(SystemException,
	      LocationForward);

    void
    etherealize(const PortableServer::ObjectId&,
		PortableServer::POA_ptr,
		PortableServer::ServantBase*,
		bool, bool);
};

//
// Definitions for TableEntry
//
class TableEntry;
typedef TableEntry* TableEntry_ptr;

} // end namespace OBPortableServer

void OBDuplicate(OBPortableServer::TableEntry_ptr);
void OBRelease(OBPortableServer::TableEntry_ptr);

namespace OBPortableServer
{

typedef ObjVar< TableEntry > TableEntry_var;

//
// An entry in the active object table
//
class TableEntry : public RefCount OB_COMMA_MONITOR
{
    //
    // The servant
    //
    PortableServer::ServantBase_var servant_;

    //
    // The state of the TableEntry
    //
    enum
    {
	ACTIVE,
	DEACTIVATE_PENDING,
	DEACTIVATED
    }
    state_;

    //
    // The number of outstanding requests
    //
    int outstandingRequests_;

public:

    static inline TableEntry_ptr _duplicate(TableEntry_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline TableEntry_ptr _nil()
    { return 0; }

    TableEntry(PortableServer::ServantBase* s)
	: servant_(s),
	  state_(ACTIVE),
	  outstandingRequests_(0)
    {
	servant_ -> _add_ref();
    }

    void
    setServant(PortableServer::ServantBase* s)
    {
	assert(s != 0);
	servant_ = s;
	servant_ -> _add_ref();
    }

    void
    clearServant()
    {
	servant_ = 0;
    }

    PortableServer::ServantBase*
    getServant()
    {
	servant_ -> _add_ref();
	return servant_;
    }

    //
    // Is the table entry active?
    //
    bool
    active()
    {
	return state_ == ACTIVE;
    }

    //
    // Mark the table entry as deactivated
    //
    void
    setDeactivatePending()
    {
	state_ = DEACTIVATE_PENDING;
    }

    //
    // Mark the deactivate of the table entry as complete
    //
    void
    setDeactivated()
    {
	//
	// This method is not synchronized for efficiencies sake
	//
	state_ = DEACTIVATED;
#ifdef HAVE_JTC
	notifyAll();
#endif
    }

#ifdef HAVE_JTC
    void
    waitDeactivated()
    {
	//
	// This method is not synchronized for efficiencies sake
	//
	assert(state_ != ACTIVE);
	while(state_ != DEACTIVATED)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }
#endif

    int
    getOutstandingRequests()
    {
	return outstandingRequests_;
    }

    void
    incOutstandingRequest()
    {
	outstandingRequests_++;
    }

    int
    decOutstandingRequest()
    {
	return --outstandingRequests_;
    }
};

} // End namespace OBPortableServer

namespace CORBA
{

inline void
release(OBPortableServer::TableEntry_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OBPortableServer::TableEntry_ptr p)
{
    return p == 0;
}

} // End namespace CORBA

namespace OBPortableServer
{

//
// All connected implementations
//
typedef Hashtable<PortableServer::ObjectId,
                  TableEntry_var,
    		  OBPortableServer::ObjectIdHasher,
		  OB_LOCK_STRATEGY_SYNCHRONIZED> ActiveObjectTable;

//
// Definitions for DirectSeqEntry
//
class DirectSeqEntry;
typedef DirectSeqEntry* DirectSeqEntry_ptr;

} // End namespace OBPortableServer

void OBDuplicate(OBPortableServer::DirectSeqEntry_ptr);
void OBRelease(OBPortableServer::DirectSeqEntry_ptr);

namespace OBPortableServer
{

typedef ObjVar< DirectSeqEntry > DirectSeqEntry_var;

//
// Mapping for ObjectId to a sequence of DirectStubImpl
//
class DirectSeqEntry : public RefCount
{
    typedef OB_STL::list<DirectStubImpl_ptr> DirectList;
    DirectList stubs_;

    PortableServer::ObjectId oid_; //TODO: tmp

    void
    traceoid()
    {
	const Octet* oct = oid_.get_buffer();
	String_var o = OctetsToAscii(oct, oid_.length());
	OB_COUT << o << OB_ENDL;
    }

public:

    static inline DirectSeqEntry_ptr _duplicate(DirectSeqEntry_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline DirectSeqEntry_ptr _nil()
    { return 0; }

    DirectSeqEntry(const PortableServer::ObjectId& oid)
	: oid_(oid)
    {
    }

    ~DirectSeqEntry()
    {
	deactivate();
    }

    void
    deactivate()
    {
	//traceoid();
	//cout << "deactivate: " << endl;
	for(DirectList::iterator iter = stubs_.begin();
	    iter != stubs_.end();
	    iter++)
	{
	    (*iter) -> _OB_deactivate();
	}

	stubs_.clear();
    }

    void
    add(DirectStubImpl_ptr stubImpl)
    {
	//traceoid();
	//cout << "add: " << stubImpl << endl;

	stubs_.push_back(stubImpl);
    }

    bool
    remove(DirectStubImpl_ptr stubImpl)
    {
	//traceoid();
	//cout << "remove: " << stubImpl << endl;

	for(DirectList::iterator iter = stubs_.begin();
	    iter != stubs_.end();
	    iter++)
	{
	    if(*iter == stubImpl)
	    {
		stubs_.erase(iter);
		return stubs_.empty();
	    }
	}

	return false;
    }
};

typedef Hashtable<PortableServer::ObjectId,
		  DirectSeqEntry_var,
    		  OBPortableServer::ObjectIdHasher,
		  OB_LOCK_STRATEGY_SYNCHRONIZED> DirectSeqTable;

} // End namespace OBPortableServer

namespace CORBA
{

inline void
release(OBPortableServer::DirectSeqEntry_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OBPortableServer::DirectSeqEntry_ptr p)
{
    return p == 0;
}

} // End namespace CORBA

namespace OBPortableServer
{

//
// If using UNIQUE_ID policy then this will map a servant to its
// id
//
struct ServantBasePtrHasher
{
    static ULong
    hash(const PortableServer::ServantBase* k)
    {
	return ((ULong)(long)k);
    }
    
    static bool
    comp(const PortableServer::ServantBase* k1,
	 const PortableServer::ServantBase* k2)
    {
	return k1 == k2;
    }
};
typedef Hashtable<PortableServer::ServantBase*, PortableServer::ObjectId,
    ServantBasePtrHasher, OB_LOCK_STRATEGY_SYNCHRONIZED> ServantIdTable;

//
// If USE_ACTIVE_OBJECT_MAP_ONLY this strategy is used
//
class ActiveObjectOnlyStrategy : public ServantLocationStrategy
{
protected:

    //
    // The AOM
    //
    ActiveObjectTable* activeObjectTable_;

    //
    // Reverse map from servant to id
    //
    ServantIdTable* servantIdTable_;

    //
    // Mapping for ObjectId's to DirectStubImpl
    //
    DirectSeqTable* directSeqTable_;

    //
    // The ORBInstance
    //
    ORBInstance_var orbInstance_;

    void
    completeActivation(const PortableServer::ObjectId&,
		       PortableServer::ServantBase*);

    virtual void
    completeDeactivate(PortableServer::POA_ptr,
		       const PortableServer::ObjectId&, TableEntry_ptr);

    virtual DirectStubImpl_ptr
    completeDirectStubImpl(PortableServer::POA_ptr,
			   const PortableServer::ObjectId&,
			   PortableServer::ServantBase*,
			   RefCountPolicyList_ptr policies);

public:

    ActiveObjectOnlyStrategy(POAPolicies*, ORBInstance_ptr);
    
    ~ActiveObjectOnlyStrategy();
    
    virtual void
    destroy(PortableServer::POA_ptr, bool)
	throw(SystemException);

    virtual void
    etherealize(PortableServer::POA_ptr)
	throw(SystemException);

    virtual void
    activate(const PortableServer::ObjectId&, PortableServer::ServantBase*)
	throw(PortableServer::POA::ServantAlreadyActive,
	      PortableServer::POA::WrongPolicy,
	      PortableServer::POA::ObjectAlreadyActive,
	      SystemException);
	
    virtual void
    deactivate(PortableServer::POA_ptr, const PortableServer::ObjectId&)
	throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      SystemException);

    virtual PortableServer::ObjectId*
    servantToId(PortableServer::ServantBase*, PortableServer::Current_impl*)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    idToServant(const PortableServer::ObjectId&, bool)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    locate(const PortableServer::ObjectId&, PortableServer::POA_ptr,
	   const char*, PortableServer::ServantLocator::Cookie&)
	throw(SystemException, LocationForward);

    virtual void
    preinvoke(const PortableServer::ObjectId&)
	throw(SystemException);

    virtual void
    postinvoke(const PortableServer::ObjectId&, PortableServer::POA_ptr,
	       const char*, PortableServer::ServantLocator::Cookie,
	       PortableServer::ServantBase*)
	throw(SystemException);

    virtual DirectStubImpl_ptr
    createDirectStubImpl(PortableServer::POA_ptr,
			 const PortableServer::ObjectId&,
			 RefCountPolicyList_ptr)
	throw(SystemException, LocationForward);

    virtual void
    removeDirectStubImpl(const PortableServer::ObjectId&, DirectStubImpl_ptr)
	throw(SystemException);

    virtual ServantManagerStrategy*
    getServantManagerStrategy();

    virtual DefaultServantHolder*
    getDefaultServantHolder();
};

//
// Strategy for RETAIN and USE_SERVANT_MANAGER or USE_DEFAULT_SERVANT
//
class RetainStrategy : public ActiveObjectOnlyStrategy
{
    typedef ActiveObjectOnlyStrategy Parent;

    ServantActivatorStrategy* servantManager_;

    DefaultServantHolder* defaultServant_;

    void
    etherealize(const PortableServer::ObjectId&, PortableServer::POA_ptr,
		PortableServer::ServantBase*, bool);

protected:

    virtual void
    completeDeactivate(PortableServer::POA_ptr,
		       const PortableServer::ObjectId&, TableEntry_ptr);

public:

    RetainStrategy(POAPolicies*, ORBInstance_ptr, ServantActivatorStrategy*,
		   DefaultServantHolder*);
    ~RetainStrategy();

    virtual void
    destroy(PortableServer::POA_ptr, bool)
	throw(SystemException);

    virtual void
    etherealize(PortableServer::POA_ptr)
	throw(SystemException);

    virtual PortableServer::ObjectId*
    servantToId(PortableServer::ServantBase*, PortableServer::Current_impl*)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    idToServant(const PortableServer::ObjectId&, bool)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    locate(const PortableServer::ObjectId&, PortableServer::POA_ptr,
	   const char*,
	   PortableServer::ServantLocator::Cookie&)
	throw(SystemException, LocationForward);

    virtual ServantManagerStrategy*
    getServantManagerStrategy();

    virtual DefaultServantHolder*
    getDefaultServantHolder();
};

//
// Strategy for NON_RETAIN and USE_SERVANT_MANAGER or
// USE_DEFAULT_SERVANT
//
class NonRetainStrategy : public ServantLocationStrategy
{
    ServantLocatorStrategy* servantManager_;

    DefaultServantHolder* defaultServant_;

public:

    NonRetainStrategy(ServantLocatorStrategy*,
		      DefaultServantHolder*);
    ~NonRetainStrategy();

    virtual void
    destroy(PortableServer::POA_ptr, bool)
	throw(SystemException);

    virtual void
    etherealize(PortableServer::POA_ptr)
	throw(SystemException);

    virtual void
    activate(const PortableServer::ObjectId&, PortableServer::ServantBase*)
	throw(PortableServer::POA::ServantAlreadyActive,
	      PortableServer::POA::WrongPolicy,
	      PortableServer::POA::ObjectAlreadyActive,
	      SystemException);

    virtual void
    deactivate(PortableServer::POA_ptr, const PortableServer::ObjectId&)
	throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      SystemException);

    virtual PortableServer::ObjectId*
    servantToId(PortableServer::ServantBase*, PortableServer::Current_impl*)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    idToServant(const PortableServer::ObjectId&, bool)
	throw(SystemException);

    virtual PortableServer::ServantBase*
    locate(const PortableServer::ObjectId&, PortableServer::POA_ptr,
	   const char*, PortableServer::ServantLocator::Cookie&)
	throw(SystemException, LocationForward);

    virtual void
    preinvoke(const PortableServer::ObjectId&)
	throw(SystemException);

    virtual void
    postinvoke(const PortableServer::ObjectId&, PortableServer::POA_ptr,
	       const char*,
	       PortableServer::ServantLocator::Cookie,
	       PortableServer::ServantBase*)
	throw(SystemException);

    virtual DirectStubImpl_ptr
    createDirectStubImpl(PortableServer::POA_ptr,
			 const PortableServer::ObjectId&,
			 RefCountPolicyList_ptr)
	throw(SystemException, LocationForward);

    virtual void
    removeDirectStubImpl(const PortableServer::ObjectId&, DirectStubImpl_ptr)
	throw(SystemException);

    virtual ServantManagerStrategy*
    getServantManagerStrategy();

    virtual DefaultServantHolder*
    getDefaultServantHolder();
};

} // End namespace OBPortableServer

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OBPortableServer::TableEntry_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OBPortableServer::TableEntry_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OBPortableServer::DirectSeqEntry_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OBPortableServer::DirectSeqEntry_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
// ----------------------------------------------------------------------
// ServantLocatorStrategy public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::ServantLocatorStrategy::destroy()
{
    OB_SYNCHRONIZED

    servantLocator_ = PortableServer::ServantLocator::_nil();
}

void
OBPortableServer::ServantLocatorStrategy::setServantManager(
    PortableServer::ServantManager_ptr manager)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    //
    // Attempting to set the servant manager after one has already
    // been set will result in the BAD_INV_ORDER exception being
    // raised.
    //
    if(!is_nil(servantLocator_))
	throw BAD_INV_ORDER();
    
    servantLocator_ = PortableServer::ServantLocator::_narrow(manager);
    if(is_nil(servantLocator_))
	throw OBJ_ADAPTER();
}

PortableServer::ServantManager_ptr
OBPortableServer::ServantLocatorStrategy::getServantManager()
{
    OB_SYNCHRONIZED

    return PortableServer::ServantManager::_duplicate(servantLocator_);
}

PortableServer::ServantBase*
OBPortableServer::ServantLocatorStrategy::preinvoke(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    const char* op,
    PortableServer::ServantLocator::Cookie& cookie)
    throw(SystemException, LocationForward)
{
    PortableServer::ServantLocator_var locator;
    {
        OB_SYNCHRONIZED

        //
        // If no servant manager has been associated with the POA,
        // OBJ_ADAPTER is raised
        //
        if(is_nil(servantLocator_))
	    throw OBJ_ADAPTER();

	locator = servantLocator_;
    }

    PortableServer::ServantBase* servant;
    try
    {
	servant = locator -> preinvoke(oid, poa, op, cookie);
    }
    catch(const PortableServer::ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward_reference -> _OB_IOR();
	throw LocationForward(ior, false);
    }
	
    //
    // 11-23:
    //
    // If a ServantManager returns a null Servant (or the
    // equivalent in a language mapping) as the result of an
    // incarnate() or preinvoke() operation, the POA will
    // return the OBJ_ADAPTER system exception as the result
    // of the request.
    //
    if(servant == 0)
	throw OBJ_ADAPTER();
    
    servant -> _add_ref();
    return servant;
}

void
OBPortableServer::ServantLocatorStrategy::postinvoke(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    const char* op,
    PortableServer::ServantLocator::Cookie cookie,
    PortableServer::ServantBase* servant)
{
    PortableServer::ServantLocator_var locator;
    {
        OB_SYNCHRONIZED

        if(is_nil(servantLocator_))
	    return;

	locator = servantLocator_;
    }

    locator -> postinvoke(oid, poa, op, cookie, servant);
}

// ----------------------------------------------------------------------
// ServantActivatorStrategy public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::ServantActivatorStrategy::destroy()
{
    OB_SYNCHRONIZED

    servantActivator_ = PortableServer::ServantActivator::_nil();
}

void
OBPortableServer::ServantActivatorStrategy::setServantManager(
    PortableServer::ServantManager_ptr manager)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    //
    // Attempting to set the servant manager after one has already
    // been set will result in the BAD_INV_ORDER exception being
    // raised.
    //
    if(!is_nil(servantActivator_))
	throw BAD_INV_ORDER();

    servantActivator_ = PortableServer::ServantActivator::_narrow(manager);
    if(is_nil(servantActivator_))
	throw OBJ_ADAPTER();
}

PortableServer::ServantManager_ptr
OBPortableServer::ServantActivatorStrategy::getServantManager()
{
    OB_SYNCHRONIZED

    return PortableServer::ServantManager::_duplicate(servantActivator_);
}

//
// Note that the synchronization in these methods ensures that
// concurrent calls to the incarnate and etherealize do not occur
//
PortableServer::ServantBase*
OBPortableServer::ServantActivatorStrategy::incarnate(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa)
    throw(SystemException,
	  LocationForward)
{
    OB_SYNCHRONIZED

    //
    // If no servant manager has been associated with the POA,
    // OBJ_ADAPTER is raised
    //
    if(is_nil(servantActivator_))
	throw OBJ_ADAPTER();

    PortableServer::ServantBase* servant;

    try
    {
	servant = servantActivator_ -> incarnate(oid, poa);
    }
    catch(const PortableServer::ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward_reference -> _OB_IOR();
	throw LocationForward(ior, false);
    }
	
    //
    // 11-23:
    //
    // If a ServantManager returns a null Servant (or the
    // equivalent in a language mapping) as the result of an
    // incarnate() or preinvoke() operation, the POA will return
    // the OBJ_ADAPTER system exception as the result of the
    // request.
    //
    if(servant == 0)
	throw OBJ_ADAPTER();

    servant -> _add_ref();

    return servant;
}

void
OBPortableServer::ServantActivatorStrategy::etherealize(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    PortableServer::ServantBase* servant,
    bool cleanup,
    bool remaining)
{
    OB_SYNCHRONIZED

    if(!is_nil(servantActivator_))
    {
	try
	{
	    servantActivator_ -> etherealize(oid, poa, servant, cleanup,
					     remaining);
	}
	catch(const SystemException&)
	{
	    // Ignore
	}
    }
}

// ----------------------------------------------------------------------
// ActiveObjectOnlyStrategy protected member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::ActiveObjectOnlyStrategy::completeActivation(
    const PortableServer::ObjectId& oid,
    PortableServer::ServantBase* servant)
{
    //
    // If there is a DirectStubImpl that refers to a default servant
    // under this oid then deactivate each
    //
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*directSeqTable_);
#endif
	DirectSeqEntry_var table;
	if(directSeqTable_ -> get(oid, table))
	{
	    table -> deactivate();
	    directSeqTable_ -> remove(oid);
	}
    }
    
    //
    // If using UNIQUE_ID add the servant to the servantIdTable
    //
    if(servantIdTable_ != 0)
    {
	assert(!servantIdTable_ -> containsKey(servant));
	servantIdTable_ -> put(servant, oid);
    }
}

//
// This method is synchronized on the TableEntry
///
void
OBPortableServer::ActiveObjectOnlyStrategy::completeDeactivate(
    PortableServer::POA_ptr poa, const PortableServer::ObjectId& oid,
    TableEntry_ptr entry)
{
    assert(entry -> getOutstandingRequests() == 0);
    assert(!entry -> active());
	
    //
    // Mark each DirectStubImpl associated with this oid as
    // deactivated
    //
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*directSeqTable_);
#endif
	DirectSeqEntry_var table;
	if(directSeqTable_ -> get(oid, table))
	{
	    table -> deactivate();
	    directSeqTable_ -> remove(oid);
	}
    }
	
    //
    // If we're using UNIQUE_ID then remove the servant from the
    // servantIdTable
    //
    if(servantIdTable_ != 0)
    {
	PortableServer::ServantBase_var servant = entry -> getServant();
	assert(servantIdTable_ -> containsKey(servant));
	servantIdTable_ -> remove(servant);
    }
	
    //
    // Release the servant reference before calling etherealize
    //
    entry -> clearServant();
	
    //
    // Inform the entry that the servant has been deactivated
    //
    entry -> setDeactivated();

    //
    // Remove the entry from the active object map
    //
    activeObjectTable_ -> remove(oid);
}

DirectStubImpl_ptr
OBPortableServer::ActiveObjectOnlyStrategy::completeDirectStubImpl(
    PortableServer::POA_ptr poa, const PortableServer::ObjectId& oid,
    PortableServer::ServantBase* servant, RefCountPolicyList_ptr policies)
{
    DirectStubImpl_ptr directStubImpl;

    //
    // No direct invocations for DSI servants
    //
    PortableServer::DynamicImplementation* dyn =
	dynamic_cast<PortableServer::DynamicImplementation*>(servant);
    if(dyn != 0)
	return DirectStubImpl::_nil();
    
    //
    // We must have direct invocations if the servant has native
    // types. Always use direct invocations, if possible, if there are
    // no interceptors installed.
    //
    // TODO: Check the POA interceptor policy
    //
    PIManager_var piManager = orbInstance_ -> getPIManager();
    if(!servant -> _OB_hasNativeTypes() &&
       (policies -> locationTransparency == LOCATION_TRANSPARENCY_STRICT ||
        piManager -> haveClientInterceptors() ||
	piManager -> haveServerInterceptors()))
	return DirectStubImpl::_nil();
    
    //
    // Create a DirectStubImpl
    //
    directStubImpl = servant -> _OB_createDirectStubImpl(poa, oid);

    //
    // Add the DirectStubImpl to the table
    //
    if(!is_nil(directStubImpl))
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*directSeqTable_);
#endif
	DirectSeqEntry_var table;
	if(!directSeqTable_ -> get(oid, table))
	{
	    table = new DirectSeqEntry(oid);
	    directSeqTable_ -> put(oid, table);
	}
	table -> add(directStubImpl);
    }

    return directStubImpl;
}

// ----------------------------------------------------------------------
// ActiveObjectOnlyStrategy constructor/destructor
// ----------------------------------------------------------------------

OBPortableServer::ActiveObjectOnlyStrategy::ActiveObjectOnlyStrategy(
    POAPolicies* policies,
    ORBInstance_ptr orbInstance)
    : activeObjectTable_(new ActiveObjectTable(1023)),
      directSeqTable_(new DirectSeqTable(1023)),
      orbInstance_(ORBInstance::_duplicate(orbInstance))
{
    if(policies -> idUniquenessPolicy() == PortableServer::UNIQUE_ID)
	servantIdTable_ = new ServantIdTable(1023);
    else
	servantIdTable_ = 0;
}

OBPortableServer::ActiveObjectOnlyStrategy::~ActiveObjectOnlyStrategy()
{
    delete activeObjectTable_;
    delete servantIdTable_;
    delete directSeqTable_;
}

// ----------------------------------------------------------------------
// ActiveObjectOnlyStrategy public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::ActiveObjectOnlyStrategy::destroy(
    PortableServer::POA_ptr, bool)
    throw(SystemException)
{
    activeObjectTable_ -> clear();

    directSeqTable_ -> clear();

    if(servantIdTable_ != 0)
	servantIdTable_ -> clear();
}

void
OBPortableServer::ActiveObjectOnlyStrategy::etherealize(
    PortableServer::POA_ptr)
    throw(SystemException)
{
    // Do nothing
}

void
OBPortableServer::ActiveObjectOnlyStrategy::activate(
    const PortableServer::ObjectId& oid,
    PortableServer::ServantBase* servant)
    throw(PortableServer::POA::ServantAlreadyActive,
	  PortableServer::POA::WrongPolicy,
	  PortableServer::POA::ObjectAlreadyActive,
	  SystemException)
{
    //
    // Find out whether a servant is already bound under this id.  If
    // there is an entry and the servant is deactivated then wait
    // until the servant deactivation completes.
    //
    while(true)
    {
	TableEntry_var entry;
	if(activeObjectTable_ -> get(oid, entry))
	{
#ifdef HAVE_JTC
	    JTCSynchronized sync(*entry.in());
	    if(!entry -> active())
	    {
		entry -> waitDeactivated();
		continue;
	    }
#endif
	    throw PortableServer::POA::ObjectAlreadyActive();
	}

#ifdef HAVE_JTC
	JTCSynchronized sync(*activeObjectTable_);
#endif

        //
        // If using UNIQUE_ID, then verify that the servant isn't
        // already activated. Synchronization is necessary to
        // avoid a race condition.
        //
	if(servantIdTable_ != 0 && servantIdTable_ -> containsKey(servant))
	    throw PortableServer::POA::ServantAlreadyActive();

        //
        // Insert the servant in the active object table with the
        // provided id. If there is already an entry in the table
        // for this object-id then some other thread has already
        // added the entry.
        //
	if(!activeObjectTable_ -> containsKey(oid))
	{
            entry = new TableEntry(servant);
	    activeObjectTable_ -> put(oid, entry);
	    completeActivation(oid, servant);
	    return;
	}
    }
}

void
OBPortableServer::ActiveObjectOnlyStrategy::deactivate(
    PortableServer::POA_ptr poa,
    const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
	  PortableServer::POA::WrongPolicy,
	  SystemException)
{
    //
    // If no object in the active object table associated with this
    // key then raise an ObjectNotActive exception.
    //
    TableEntry_var entry;
    if(activeObjectTable_ -> get(oid, entry))
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*entry.in());
#endif
	if(entry -> active())
	{
	    entry -> setDeactivatePending();
	    
	    //
	    // If there are no futher outstanding requests then complete the
	    // deactivation
	    //
	    if(entry -> getOutstandingRequests() == 0)
		completeDeactivate(poa, oid, entry);
	}
    }
    else
	throw PortableServer::POA::ObjectNotActive();
}

PortableServer::ObjectId*
OBPortableServer::ActiveObjectOnlyStrategy::servantToId(
    PortableServer::ServantBase* servant,
    PortableServer::Current_impl*)
    throw(SystemException)
{
    PortableServer::ObjectId id;
    if(servantIdTable_ != 0 && servantIdTable_ -> get(servant, id))
	return new PortableServer::ObjectId(id);
    return 0;
}

PortableServer::ServantBase*
OBPortableServer::ActiveObjectOnlyStrategy::idToServant(
    const PortableServer::ObjectId& oid,
    bool)
    throw(SystemException)
{
    while(true)
    {
	TableEntry_var entry;
	if(activeObjectTable_ -> get(oid, entry))
	{
#ifdef HAVE_JTC
	    JTCSynchronized sync(*entry.in());
	    if(!entry -> active())
	    {
		entry -> waitDeactivated();
		continue;
	    }
#else
	    if(!entry -> active())
		return 0;
#endif
	    return entry -> getServant();
	}
	break;
    }
    return 0;
}

PortableServer::ServantBase*
OBPortableServer::ActiveObjectOnlyStrategy::locate(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr,
    const char*,
    PortableServer::ServantLocator::Cookie&)
    throw(SystemException,
	  LocationForward)
{
    while(true)
    {
	TableEntry_var entry;
	if(activeObjectTable_ -> get(oid, entry))
	{
#ifdef HAVE_JTC
	    JTCSynchronized sync(*entry.in());
	    if(!entry -> active())
	    {
		entry -> waitDeactivated();
		continue;
	    }
#else
	    if(!entry -> active())
		throw OBJECT_NOT_EXIST();
#endif
	    return entry -> getServant();
	}
	break;
    }

    //
    // If the servant isn't in the table then this is an
    // OBJECT_NOT_EXIST exception
    //
    throw OBJECT_NOT_EXIST();
    return 0; // Some compilers need this
}

void
OBPortableServer::ActiveObjectOnlyStrategy::preinvoke(
    const PortableServer::ObjectId& oid)
    throw(SystemException)
{
    TableEntry_var entry;
    if(activeObjectTable_ -> get(oid, entry))
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*entry.in());
#endif
	entry -> incOutstandingRequest();
    }
}

void
OBPortableServer::ActiveObjectOnlyStrategy::postinvoke(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    const char* op,
    PortableServer::ServantLocator::Cookie cookie,
    PortableServer::ServantBase* servant)
    throw(SystemException)
{
    TableEntry_var entry;
    if(activeObjectTable_ -> get(oid, entry))
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*entry.in());
#endif
	//
	// If the number of outstanding requests is now 0 and the
	// entry has been deactivated then complete the deactivation
	//
	if(entry -> decOutstandingRequest() == 0 &&
	   !entry -> active())
	    completeDeactivate(poa, oid, entry);
    }
}

DirectStubImpl_ptr
OBPortableServer::ActiveObjectOnlyStrategy::createDirectStubImpl(
    PortableServer::POA_ptr poa,
    const PortableServer::ObjectId& oid,
    RefCountPolicyList_ptr policies)
    throw(SystemException,
	  LocationForward)
{
    try
    {
	PortableServer::ServantLocator::Cookie cookie;
	PortableServer::ServantBase_var servant = locate(oid, poa, "", cookie);
	return completeDirectStubImpl(poa, oid, servant, policies);
    }
    catch(const SystemException&)
    {
    }
    return DirectStubImpl::_nil();
}

void
OBPortableServer::ActiveObjectOnlyStrategy::removeDirectStubImpl(
    const PortableServer::ObjectId& oid,
    DirectStubImpl_ptr directStubImpl)
    throw(SystemException)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(*directSeqTable_);
#endif
    DirectSeqEntry_var table;
    if(directSeqTable_ -> get(oid, table))
    {
	if(table -> remove(directStubImpl))
	    directSeqTable_ -> remove(oid);
    }
}

OBPortableServer::ServantManagerStrategy*
OBPortableServer::ActiveObjectOnlyStrategy::getServantManagerStrategy()
{
    return 0;
}

OBPortableServer::DefaultServantHolder*
OBPortableServer::ActiveObjectOnlyStrategy::getDefaultServantHolder()
{
    return 0;
}

// ----------------------------------------------------------------------
// RetainStrategy private member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::RetainStrategy::etherealize(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    PortableServer::ServantBase* servant,
    bool cleanup)
{
    //
    // If we have an ActivationStrategy then we have to
    // etherealize the object
    //
    if(servantManager_ != 0)
    {
	bool remaining = false;
	    
	if(servantIdTable_ == 0)
	{
	    //
	    // TODO: optimize!  If servant is still in the active
	    // object map, then we still have remaining
	    // activations.
	    //
	    ActiveObjectTable::Enumerator keys =
		activeObjectTable_ -> keys();
	    for(ULong i = 0; i < keys.length(); ++i)
	    {
		TableEntry_var entry2;
		if(activeObjectTable_ -> get(keys[i], entry2))
		{
#ifdef HAVE_JTC
		    JTCSynchronized sync(*entry2.in());
#endif
		    PortableServer::ServantBase_var servant2 = entry2 -> getServant();
		    if(servant2.in() == servant)
		    {
			remaining = true;
			break;
		    }
		}
	    }
	}

	servantManager_ -> etherealize(oid, poa, servant, cleanup, remaining);
    }
}

void
OBPortableServer::RetainStrategy::completeDeactivate(
    PortableServer::POA_ptr poa,
    const PortableServer::ObjectId& oid,
    TableEntry_ptr entry)
{
    PortableServer::ServantBase_var servant = entry -> getServant();

    Parent::completeDeactivate(poa, oid, entry);

    PortableServer::ServantBase* s = servant;
    servant = 0;

    etherealize(oid, poa, s, false);
}

// ----------------------------------------------------------------------
// RetainStrategy constructor/destructor
// ----------------------------------------------------------------------

OBPortableServer::RetainStrategy::RetainStrategy(
    POAPolicies* policies,
    ORBInstance_ptr orbInstance,
    ServantActivatorStrategy* servantManager,
    DefaultServantHolder* defaultServant)
    : ActiveObjectOnlyStrategy(policies, orbInstance),
      servantManager_(servantManager),
      defaultServant_(defaultServant)
{
}

OBPortableServer::RetainStrategy::~RetainStrategy()
{
    delete servantManager_;
    delete defaultServant_;
}

// ----------------------------------------------------------------------
// RetainStrategy public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::RetainStrategy::destroy(PortableServer::POA_ptr poa, bool e)
    throw(SystemException)
{
    if(servantManager_ != 0 && e)
    {
	ActiveObjectTable::Enumerator keys = activeObjectTable_ -> keys();
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    TableEntry_var entry;
	    if(activeObjectTable_ -> get(keys[i], entry))
	    {
#ifdef HAVE_JTC
		JTCSynchronized sync(*entry.in());
#endif
		activeObjectTable_ -> remove(keys[i]);

		PortableServer::ServantBase_var servant =
		    entry -> getServant();
		entry -> clearServant();

		PortableServer::ServantBase* s = servant;
		servant = 0;
		etherealize(keys[i], poa, s, true);
	    }
	}
    }
    Parent::destroy(poa, e);

    if(servantManager_ != 0)
	servantManager_ -> destroy();
    if(defaultServant_ != 0)
	defaultServant_ -> destroy();
}

void
OBPortableServer::RetainStrategy::etherealize(PortableServer::POA_ptr poa)
    throw(SystemException)
{
    destroy(poa, true);
}

PortableServer::ObjectId*
OBPortableServer::RetainStrategy::servantToId(
    PortableServer::ServantBase* servant,
    PortableServer::Current_impl* poaCurrent)
    throw(SystemException)
{
    PortableServer::ObjectId* oid = Parent::servantToId(servant, poaCurrent);
    if(oid == 0 && defaultServant_ != 0)
	return defaultServant_ -> servantToId(servant, poaCurrent);

    return oid;
}

PortableServer::ServantBase*
OBPortableServer::RetainStrategy::idToServant(
    const PortableServer::ObjectId& oid,
    bool useDefaultServant)
    throw(SystemException)
{
    PortableServer::ServantBase_var servant = Parent::idToServant(oid, useDefaultServant);
    if(servant.in() == 0 && useDefaultServant && defaultServant_ != 0)
	servant = defaultServant_ -> getDefaultServant();
    return servant;
}

PortableServer::ServantBase*
OBPortableServer::RetainStrategy::locate(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    const char* op,
    PortableServer::ServantLocator::Cookie& cookie)
    throw(SystemException,
	  LocationForward)
{
    while(true)
    {
	TableEntry_var entry;
	if(activeObjectTable_ -> get(oid, entry))
	{
#ifdef HAVE_JTC
	    JTCSynchronized sync(*entry.in());
	    if(!entry -> active())
	    {
		entry -> waitDeactivated();
		continue;
	    }
#else
	    if(!entry -> active())
		return 0;
#endif
	    return entry -> getServant();
	}
	
	if(defaultServant_ != 0)
	{
	    PortableServer::ServantBase_var servant =
		defaultServant_ -> getDefaultServant();
	    if(!servant)
		throw OBJ_ADAPTER();
	    return servant._retn();
	}
	
	if(servantManager_ == 0)
	    throw OBJECT_NOT_EXIST();
	
#ifdef HAVE_JTC
	JTCSynchronized sync(*activeObjectTable_);
#endif
	if(!activeObjectTable_ -> containsKey(oid))
	{
	    PortableServer::ServantBase_var servant =
		servantManager_ -> incarnate(oid, poa);

	    //
	    // 11-25:
	    //
	    // "If the incarnate operation returns a servant that is
	    // already active for a different Object Id and if the POA
	    // also has the UNIQUE_ID policy, the incarnate has violated
	    // the POA policy and is considered to be in error. The POA
	    // will raise an OBJ_ADAPTER system exception for the
	    // request."
	    //
	    // Note: The activator should be allowed to explicitly
	    // activate the servant for the given PortableServer::ObjectId.
	    //
	    if(servantIdTable_ != 0 && // TODO: initialize anyway?
	       servantIdTable_ -> containsKey(servant))
	    {
		PortableServer::ObjectId oid2;
		servantIdTable_ -> get(servant, oid2);
		if(!OBPortableServer::ObjectIdHasher::comp(oid, oid2))
		    throw OBJ_ADAPTER();
	    }

	    //
	    // First verify that the user didn't register the
	    // PortableServer::ObjectId with the POA in the incarnate
	    // call. If the PortableServer::ObjectId was registered
	    // with the POA and returned servant was different then
	    // raise an OBJ_ADAPTER exception.
	    //
	    if(activeObjectTable_ -> get(oid, entry))
	    {
		PortableServer::ServantBase_var es = entry -> getServant();
		if(es.in() != servant)
		    throw OBJ_ADAPTER();
	    }
	    else
	    {
		entry = new TableEntry(servant);
		activeObjectTable_ -> put(oid, entry);
		completeActivation(oid, servant);
	    }
	    return servant._retn();
	}
    }
    return 0; // Some compilers need this
}

OBPortableServer::ServantManagerStrategy*
OBPortableServer::RetainStrategy::getServantManagerStrategy()
{
    return servantManager_;
}

OBPortableServer::DefaultServantHolder*
OBPortableServer::RetainStrategy::getDefaultServantHolder()
{
    return defaultServant_;
}

// ----------------------------------------------------------------------
// NonRetainStrategy constructor/destructor
// ----------------------------------------------------------------------

OBPortableServer::NonRetainStrategy::NonRetainStrategy(
    ServantLocatorStrategy* servantManager,
    DefaultServantHolder* defaultServant)
    : servantManager_(servantManager),
      defaultServant_(defaultServant)
{
}

OBPortableServer::NonRetainStrategy::~NonRetainStrategy()
{
    delete servantManager_;
    delete defaultServant_;
}

// ----------------------------------------------------------------------
// NonRetainStrategy public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::NonRetainStrategy::destroy(PortableServer::POA_ptr, bool)
    throw(SystemException)
{
    if(servantManager_ != 0)
	servantManager_ -> destroy();
    if(defaultServant_ != 0)
	defaultServant_ -> destroy();
}

void
OBPortableServer::NonRetainStrategy::etherealize(PortableServer::POA_ptr)
    throw(SystemException)
{
    // Do nothing
}

void
OBPortableServer::NonRetainStrategy::activate(
    const PortableServer::ObjectId&,
    PortableServer::ServantBase*)
    throw(PortableServer::POA::ServantAlreadyActive,
	  PortableServer::POA::WrongPolicy,
	  PortableServer::POA::ObjectAlreadyActive,
	  SystemException)
{
    //
    // Requires the RETAIN policy.
    //
    throw PortableServer::POA::WrongPolicy();
}

void
OBPortableServer::NonRetainStrategy::deactivate(
    PortableServer::POA_ptr,
    const PortableServer::ObjectId&)
    throw(PortableServer::POA::ObjectNotActive,
	  PortableServer::POA::WrongPolicy,
	  SystemException)
{
    //
    // Requires the RETAIN policy.
    //
    throw PortableServer::POA::WrongPolicy();
}

PortableServer::ObjectId*
OBPortableServer::NonRetainStrategy::servantToId(
    PortableServer::ServantBase* servant,
    PortableServer::Current_impl* poaCurrent)
    throw(SystemException)
{
    if(defaultServant_ != 0)
	return defaultServant_ -> servantToId(servant, poaCurrent);
    return 0;
}

PortableServer::ServantBase*
OBPortableServer::NonRetainStrategy::idToServant(
    const PortableServer::ObjectId& oid,
    bool)
    throw(SystemException)
{
    if(defaultServant_ != 0)
	return defaultServant_ -> getDefaultServant();
    return 0;
}

PortableServer::ServantBase*
OBPortableServer::NonRetainStrategy::locate(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa, const char* op,
    PortableServer::ServantLocator::Cookie& cookie)
    throw(SystemException,
	  LocationForward)
{
    if(defaultServant_ != 0)
    {
	PortableServer::ServantBase_var servant =
	    defaultServant_ -> getDefaultServant();
	if(!servant)
	    throw OBJ_ADAPTER();
	return servant._retn();
    }

    if(servantManager_ == 0)
	throw OBJECT_NOT_EXIST();
    
    return servantManager_ -> preinvoke(oid, poa, op, cookie);
}

void
OBPortableServer::NonRetainStrategy::preinvoke(
    const PortableServer::ObjectId&)
    throw(SystemException)
{
}

void
OBPortableServer::NonRetainStrategy::postinvoke(
    const PortableServer::ObjectId& oid,
    PortableServer::POA_ptr poa,
    const char* op,
    PortableServer::ServantLocator::Cookie cookie,
    PortableServer::ServantBase* servant)
    throw(SystemException)
{
    if(servantManager_ != 0)
	servantManager_ -> postinvoke(oid, poa, op, cookie, servant);
}

DirectStubImpl_ptr
OBPortableServer::NonRetainStrategy::createDirectStubImpl(
    PortableServer::POA_ptr,
    const PortableServer::ObjectId&,
    RefCountPolicyList_ptr policies)
    throw(SystemException,
	  LocationForward)
{
    return DirectStubImpl::_nil();
}

void
OBPortableServer::NonRetainStrategy::removeDirectStubImpl(
    const PortableServer::ObjectId&,
    DirectStubImpl_ptr)
    throw(SystemException)
{
}

OBPortableServer::ServantManagerStrategy*
OBPortableServer::NonRetainStrategy::getServantManagerStrategy()
{
    return servantManager_;
}

OBPortableServer::DefaultServantHolder*
OBPortableServer::NonRetainStrategy::getDefaultServantHolder()
{
    return defaultServant_;
}

// ----------------------------------------------------------------------
// ServantLocationStrategyFactory public member implementation
// ----------------------------------------------------------------------

OBPortableServer::ServantLocationStrategy*
OBPortableServer::ServantLocationStrategyFactory::
createServantLocationStrategy(
    POAPolicies* policies,
    ORBInstance_ptr orbInstance)
{
    if(policies -> requestProcessingPolicy() ==
       PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY)
	return new ActiveObjectOnlyStrategy(policies, orbInstance);
    
    ServantLocatorStrategy* servantLocator = 0;
    ServantActivatorStrategy* servantActivator = 0;
    DefaultServantHolder* defaultServant = 0;
    
    if(policies -> requestProcessingPolicy() ==
       PortableServer::USE_DEFAULT_SERVANT)
	defaultServant = new DefaultServantHolder;
    
    if(policies -> requestProcessingPolicy() ==
       PortableServer::USE_SERVANT_MANAGER)
    {
	if(policies -> servantRetentionPolicy() == PortableServer::RETAIN)
	{
	    servantActivator = new ServantActivatorStrategy;
	}
	else
	{
	    servantLocator = new ServantLocatorStrategy;
	}
    }
    
    if(policies -> servantRetentionPolicy() == PortableServer::RETAIN)
	return new RetainStrategy(policies, orbInstance,
				  servantActivator, defaultServant);
    // NON_RETAIN
    return new NonRetainStrategy(servantLocator, defaultServant);
}

// ----------------------------------------------------------------------
// DefaultServantHolder constructor/destructor
// ----------------------------------------------------------------------

OBPortableServer::DefaultServantHolder::DefaultServantHolder()
{
}

OBPortableServer::DefaultServantHolder::~DefaultServantHolder()
{
}

// ----------------------------------------------------------------------
// DefaultServantHolder public member implementation
// ----------------------------------------------------------------------

void
OBPortableServer::DefaultServantHolder::destroy()
{
    OB_SYNCHRONIZED

    servant_ = 0;
}

void
OBPortableServer::DefaultServantHolder::setDefaultServant(
    PortableServer::ServantBase* servant)
{
    OB_SYNCHRONIZED

    servant_ = servant;
    servant -> _add_ref();
}

PortableServer::ServantBase*
OBPortableServer::DefaultServantHolder::getDefaultServant()
{
    OB_SYNCHRONIZED

    if(servant_)
	servant_ -> _add_ref();
    return servant_;
}

PortableServer::ObjectId*
OBPortableServer::DefaultServantHolder::servantToId(
    PortableServer::ServantBase* servant,
    PortableServer::Current_impl* poaCurrent)
    throw(SystemException)
{
    OB_SYNCHRONIZED

    if(servant != servant_.in())
	return 0;

    if(poaCurrent -> _OB_inUpcall() &&
       poaCurrent -> _OB_getServant() == servant)
    {
	try
	{
	    return poaCurrent -> get_object_id();
	}
	catch(const PortableServer::Current::NoContext&)
	{
	    assert(false);
	}
    }
    return 0; // Some compilers need this
}
