// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/POAServantBase.h>
#include <OB/IMRState.h>
#include <OB/Logger.h>
#include <OB/Stream.h>
#include <OB/Timer.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/PIManager.h>
#include <OB/RefCounted.h>
#include <OB/Upcall.h>

#include <OB/IIOP.h> // TODO: for fix to reference_to_id IMR problem

#include <POAStrategy.h>
#include <IDStrategy.h>
#include <IIOPUtil.h>
#include <POA_impl.h>
#include <OCI_impl.h>
#include <Dispatch_impl.h>
#include <POAPolicies_impl.h>
#include <POAManager_impl.h>
#include <POACurrent_impl.h>
#include <OCI_impl.h>
#include <POAControl.h>
#include <DispatchRequest_impl.h>

using namespace CORBA;
using namespace OB;

namespace OBPortableServer
{

//
// This class is used to hold the adapter activator
//
class AdapterActivatorHolder OB_COLON_MUTEX
{
    PortableServer::AdapterActivator_var activator_;

public:

    void
    destroy()
    {
	OB_SYNCHRONIZED
	activator_ = PortableServer::AdapterActivator::_nil();
    }

    PortableServer::AdapterActivator_ptr
    getAdapterActivator()
    {
	OB_SYNCHRONIZED
        return PortableServer::AdapterActivator::_duplicate(activator_);
    }

    void
    setAdapterActivator(PortableServer::AdapterActivator_ptr activator)
    {
	OB_SYNCHRONIZED
        activator_ = PortableServer::AdapterActivator::_duplicate(activator);
    }
};

} // End of namespace OBPortableServer

// ----------------------------------------------------------------------
// POA_impl private member implementation
// ----------------------------------------------------------------------

OCI::ObjectKey*
OBPortableServer::POA_impl::createObjectKey(const PortableServer::ObjectId& id)
{
    return ::OB::CreateObjectKey(ObjectKeyData(
	poaId_, id,
	policies_ -> lifespanPolicy() == PortableServer::PERSISTENT,
	poaCreateTime_));
}

bool
OBPortableServer::POA_impl::findPolicyIndex(const PolicyList& p,
					    CORBA::PolicyType type,
					    ULong& i)
{
    for(i = 0 ; i < p.length() ; i++)
	if(p[i] -> policy_type() == type)
	    return true;
    return false;
}

//
// Validate a specific set of POA policies
//
bool
OBPortableServer::POA_impl::validatePolicies(const POAPolicies& policies,
					     const PolicyList& policyList,
					     ULong& i)
{
    //
    // NON_RETAIN requires either USE_DEFAULT_SERVANT or
    // USE_SERVANT_MANAGER
    //
    if(policies.servantRetentionPolicy() == PortableServer::NON_RETAIN &&
       policies.requestProcessingPolicy()
       != PortableServer::USE_SERVANT_MANAGER &&
       policies.requestProcessingPolicy()
       != PortableServer::USE_DEFAULT_SERVANT)
    {
	//
	// Since the default value is RETAIN then the policy must be
	// present
	//
	bool ok = findPolicyIndex(policyList,
				  PortableServer::SERVANT_RETENTION_POLICY_ID,
				  i);
	assert(ok);
        return false;
    }
    
    //
    // USE_ACTIVE_OBJECT_MAP_ONLY requires RETAIN
    //
    if(policies.requestProcessingPolicy() ==
       PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY &&
       policies.servantRetentionPolicy() !=  PortableServer::RETAIN)
    {
	//
	// One of the two must be present
	//
	bool ok = findPolicyIndex(policyList,
				  PortableServer::SERVANT_RETENTION_POLICY_ID,
				  i);
	if(!ok)
	    ok = findPolicyIndex(policyList,
				 PortableServer::REQUEST_PROCESSING_POLICY_ID,
				 i);
	assert(ok);
        return false;
    }
    
    //
    // USE_DEFAULT_SERVANT requires MULTIPLE_ID
    //
    if(policies.requestProcessingPolicy() ==
       PortableServer::USE_DEFAULT_SERVANT &&
       policies.idUniquenessPolicy() !=  PortableServer::MULTIPLE_ID)
    {
	//
	// Since USE_DEFAULT_SERVANT, this must be present
	//
	bool ok = findPolicyIndex(policyList,
				  PortableServer::REQUEST_PROCESSING_POLICY_ID,
				  i);
	assert(ok);
        return false;
    }
    
    //
    // IMPLICIT_ACTIVATION requires SYSTEM_ID and RETAIN
    //
    if(policies.implicitActivationPolicy() ==
       PortableServer::IMPLICIT_ACTIVATION &&
       policies.servantRetentionPolicy() !=  PortableServer::RETAIN &&
       policies.idAssignmentPolicy() !=  PortableServer::SYSTEM_ID)
    {
	//
	// Since IMPLICIT_ACTIVATION , this must be present
	//
	bool ok =findPolicyIndex(policyList,
				 PortableServer::IMPLICIT_ACTIVATION_POLICY_ID,
				 i);
	assert(ok);
        return false;
    }

    return true;
}

//
// Handle unknown exceptions
//
void
OBPortableServer::POA_impl::handleUnknownException(Upcall_ptr upcall)
{
    String_var msg = string_dup("Servant method raised a non-CORBA exception");
            
    if(upcall -> responseExpected())
        msg += "\nClient receives this exception as CORBA::UNKNOWN";
            
    msg += "\noperation name: \"";
    msg += upcall -> operation();
    msg += '"';

    OCI::TransportInfo_ptr transportInfo = upcall -> transportInfo_nodup();
    if(!is_nil(transportInfo))
    {
        String_var desc = transportInfo -> describe();
        msg += '\n';
        msg += desc;
    }
    else
    {
        msg += "\nCollocated method call";
    }
            
    Logger_var logger = orbInstance_ -> getLogger();
    logger -> warning(msg);
            
    upcall -> setSystemException(UNKNOWN());
}

void
OBPortableServer::POA_impl::completeDestroy()
{
    //
    // Wait for all pending requests to terminate. If the POA has
    // destruction has already completed waitPendingRequests returns
    // false.
    //
    if(poaControl_ -> waitPendingRequests())
    {
	//
	// TODO: Move to POA reaping
	//
	// Remove this POA from our parent
	//
	if(parent_ != 0)
	{
#ifdef HAVE_JTC
	    JTCSynchronized parentSync(*parent_ -> children_);
#endif
	    assert(parent_ -> children_ -> containsKey(name_));
	    parent_ -> children_ -> remove(name_);
	}

	//
	// Unregister from the POAManager and tell the IMR that the
	// POA is going away. If the manager_ is nil then the POA is
	// being destroyed on startup.
	//
	if(!is_nil(manager_))
	{
	    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
	    if(policies_ -> lifespanPolicy() == PortableServer::PERSISTENT)
	    {
		IMR::ActiveState_var as = m -> _OB_getIMRActiveState();
		if(!is_nil(as))
		{
		    try
		    {
			as -> poa_destroy(poaId_);
		    }
		    catch(const IMR::NoSuchPOA& e)
		    {
			String_var msg =
			string_dup("IMR: poa_destroy: NoSuchPOA: ");
			for(ULong i = 0; i < e.poa.length(); i++)
			{
			    msg += e.poa[i];
			    if(i != e.poa.length()-1)
				msg += "/";
			}
			Logger_var logger = orbInstance_ -> getLogger();
			logger -> warning(msg);
		    }
		    catch(const SystemException& ex)
		    {
			String_var msg =
			string_dup("IMR: poa_destroy: ");
			msg += ex._to_string();
			Logger_var logger = orbInstance_ -> getLogger();
			logger -> warning(msg);
		    }
		}
	    }
	    m -> _OB_removePOA(poaId_);
	    manager_ = POAManager::_nil();
	}

	//
	// Destroy the servant location strategy
	//
	servantLocationStrategy_ -> destroy(this,
					    poaControl_ -> etherealize());

	//
	// Clear internal state variables
	//
	adapterActivator_ -> destroy();

	//
	// Mark the destroy as complete
	//
	poaControl_ -> markDestroyCompleted();
    }
}

//
// Register the POA with the IMR. Note that this only affects
// PERSISTENT POAs.
//
void
OBPortableServer::POA_impl::registerWithIMR()
{
    //
    // This method does nothing if this is not a persistent POA
    //
    if(policies_ -> lifespanPolicy() !=  PortableServer::PERSISTENT)
        return;

    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    IMR::ActiveState_var as = m -> _OB_getIMRActiveState();
    if(is_nil(as))
        return;

    Logger_var logger = orbInstance_ -> getLogger();
    try
    {
    	//
	// Get the state from the POA Manager
	//
	IMR::POAStatus imrState;
	POAManager::State state = manager_ -> get_state();
	switch(state)
	{
	case POAManager::INACTIVE:
            imrState = IMR::INACTIVE;
	    break;
	case POAManager::ACTIVE:
            imrState = IMR::ACTIVE;
	    break;
	case POAManager::HOLDING:
            imrState = IMR::HOLDING;
	    break;
	case POAManager::DISCARDING:
            imrState = IMR::DISCARDING;
	    break;
	}

        //
        // Create dummy `empty' template IOR. This IOR is provided to
        // the IMR for use when serving location forward requests.
        //
        OCI::ObjectKey emptykey;
        RefCountIOR_var ior = createIOR(emptykey);
        
        //
        // Add the information to the IMR. The IMR returns a skeleton
        // IOR that is used to provide object references to servants
        // that reside in this POA.
        //
        as -> poa_create(poaId_, imrState, ior -> value, imrIOR_);
        hasIMR_ = true;
        OCI::ProfileInfo_var imrProfileInfo;
        bool b = IIOPExtractProfileInfo(imrIOR_, imrProfileInfo);
        assert(b); // TODO: Internal error
        bool ok = ParseObjectKey(imrProfileInfo -> key, imrKey_);
        assert(ok); // TODO: Internal error
    }
    catch(const IMR::NoSuchPOA& e)
    {
        String_var msg = string_dup("IMR: POA not registered: ");
        for(ULong i = 0; i < e.poa.length(); i++)
        {
            msg += e.poa[i];
            if(i != e.poa.length()-1)
                msg += "/";
        }
        logger -> error(msg);
        throw INITIALIZE();
    }
    catch(SystemException& ex)
    {
	String_var msg = string_dup("IMR: Cannot contact: ");
	msg += ex._to_string();
        logger -> error(msg);
        throw INITIALIZE(); // TODO: - some exception
    }
}

RefCountIOR_ptr
OBPortableServer::POA_impl::createIOR(const OCI::ObjectKey& key,
				      const char* intf)
{
    IOP::IOR ior;
    ior.type_id = intf;
    OCI::AcceptorSeq_var acceptors = manager_ -> get_acceptors();
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    const GIOP::Version& version = m -> _OB_getGIOPVersion();
    OCI::ProfileInfo profileInfo;
    profileInfo.major = version.major;
    profileInfo.minor = version.minor;
    profileInfo.key = key;

    for(ULong i = 0 ; i < acceptors -> length() ; i++)
    {
        //
        // Collect the components to be supplied to add_profiles().
        // We start with the generic components (allComponents_),
        // and may have to append components specific to this
        // acceptor's profile tag.
        //
        IOP::TaggedComponentSeq* components;
        componentTable_ -> get(acceptors[i] -> tag(), components);
        if(components -> length() > 0)
        {
            profileInfo.components.length(allComponents_.length() +
                                          components -> length());
            ULong j;
            for(j = 0 ; j < allComponents_.length() ; j++)
                profileInfo.components[j] = allComponents_[j];
            for(j = 0 ; j < components -> length() ; j++)
                profileInfo.components[allComponents_.length() + j] =
                    (*components)[j];
        }
        else
            profileInfo.components = allComponents_;

        acceptors[i] -> add_profiles(profileInfo, ior);
    }

    return new RefCountIOR(ior);
}

void
OBPortableServer::POA_impl::init()
    throw()
{
    //
    // Setup the component table.
    //
    componentTable_ = new ComponentTable(7);

    //
    // Create the id generation strategy
    //
    idGenerationStrategy_ =
	IdGenerationStrategyFactory::createIdGenerationStrategy(policies_);

    //
    // The ORB doesn't have an IMR, by default
    //
    hasIMR_ = false;

    //
    // Create the adapter activator holder
    //
    adapterActivator_ = new AdapterActivatorHolder;

    //
    // Create the POA control
    //
    poaControl_ = new POAControl;

    //
    // Initialize the ServantLocationStategy
    //
    servantLocationStrategy_ =
	ServantLocationStrategyFactory::createServantLocationStrategy(
	    policies_, orbInstance_);

    //
    // Initialize the POA child hashtable
    //
    children_ = new POAHashTable(31);

    //
    // Initialize the dispatch strategy
    //
    dispatchStrategy_ = policies_ -> dispatchStrategyPolicy();

    //
    // Set the POA create time
    //
    struct timeval tv = TimeNow();
    poaCreateTime_ = tv.tv_sec;

    //
    // Setup the adapter id. This ID is unique for the lifespan of
    // this POA.
    //
    PortableServer::ObjectId oid;
    adapterId_ = createObjectKey(oid);

    //
    // Register with the POAManager
    //
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    m -> _OB_addPOA(this, poaId_);

    //
    // Add an entry to the componentTable_ for each acceptor
    //
    // TODO: It may not be possible to add components for this
    // acceptor -- perhaps there should be flag to indicate this?
    //
    OCI::AcceptorSeq_var acceptors = manager_ -> get_acceptors();
    for(ULong i = 0 ; i < acceptors -> length() ; i++)
        componentTable_ -> put(acceptors[i] -> tag(),
			       new IOP::TaggedComponentSeq);

    //
    // Establish the components to place in IORs generated by this POA
    //
    PIManager_var piManager = orbInstance_ -> getPIManager();
    piManager -> establishComponents(*componentTable_, allComponents_,
                                     externalPolicies_);
}

// ----------------------------------------------------------------------
// POA_impl constructor and destructor
// ----------------------------------------------------------------------

//
// Constructor for any POA other than the root
//
OBPortableServer::POA_impl::POA_impl(ORBInstance_ptr orbInstance,
				     const char* name,
				     OBPortableServer::POA_impl* parent,
				     POAManager_ptr manager,
				     PortableServer::Current_impl* poaCurrent,
				     OCI::Current_impl* ociCurrent,
				     const POAPolicies& policies,
				     const PolicyList& external)
    throw()
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      name_(name),
      parent_(parent),
      manager_(POAManager::_duplicate(manager)),
      policies_(new POAPolicies(policies)),
      poaCurrent_(poaCurrent),
      ociCurrent_(ociCurrent),
      externalPolicies_(new RefCountPolicyList(external))
{
    //
    // Set the root POA name
    //
    POA_impl* p = this;
    while(p -> parent_ != 0)
	p = p -> parent_;
    rootName_ = p -> name_;
        
    //
    // Set my POA id
    //
    assert(parent != 0);
    poaId_ = parent -> poaId_;

    CORBA::ULong poaIdLength = poaId_.length();
    poaId_.length(poaIdLength + 1);
    poaId_[poaIdLength] = name;

    init();
}

//
// Constructor for the root POA
//
OBPortableServer::POA_impl::POA_impl(ORBInstance_ptr orbInstance,
				     const char* name,
				     POAManager_ptr manager)
    throw()
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      name_(name),
      rootName_(name),
      parent_(0),
      manager_(POAManager::_duplicate(manager)),
      policies_(0),
      poaCurrent_(0),
      ociCurrent_(0),
      hasIMR_(false)
{
    PolicyList pl(1);
    pl.length(1);
    pl[0] = new PortableServer::ImplicitActivationPolicy_impl(
	PortableServer::IMPLICIT_ACTIVATION);

    //
    // Create the policies_ member
    //
    policies_ = new POAPolicies(orbInstance_, pl);

    //
    // There are no external policies for the root POA
    //
    CORBA::PolicyList external;
    externalPolicies_ = new RefCountPolicyList(external);

    //
    // Create the POA and OCI Current implementation object
    //
    poaCurrent_ = new PortableServer::Current_impl;
    ociCurrent_ = new OCI::Current_impl;

    //
    // Register the POA::Current, and the OCI::Current classes with
    // the ORB
    //
    InitialServiceManager_var ism =
	orbInstance_ -> getInitialServiceManager();

    ism -> addInitialReference("POACurrent", poaCurrent_);
    ism -> addInitialReference("OCICurrent", ociCurrent_);

    //
    // Set the POAID
    //
    poaId_.length(1);
    poaId_[0] = name_;

    //
    // Initialize the POA
    //
    init();
}

OBPortableServer::POA_impl::~POA_impl()
{
    assert(poaControl_ -> getDestroyed());

    delete policies_;

    delete children_;

    delete idGenerationStrategy_;

    delete servantLocationStrategy_;

    delete adapterActivator_;

    delete poaControl_;

    ComponentTable::Enumerator keys = componentTable_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
        IOP::TaggedComponentSeq* components;
        componentTable_ -> get(keys[i], components);
        delete components;
    }
    delete componentTable_;

    //
    // If this is the root POA then release the
    // PortableServer::Current and OCI::Current
    //
    if(parent_ == 0)
    {
        release(poaCurrent_);
        release(ociCurrent_);
    }
}

// ----------------------------------------------------------------------
// POA_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
OBPortableServer::POA_impl::create_POA(
    const char* adapter,
    PortableServer::POAManager_ptr manager,
    const PolicyList& policyList)
    throw(PortableServer::POA::AdapterAlreadyExists,
          PortableServer::POA::InvalidPolicy,
          SystemException)
{
    assert_nca(adapter != 0, NCANullString);

    //
    // Has the POA been destroyed?
    //
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Seperate the provided policy list into POA policies and
    // externally provided policies
    //
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();

    PolicyList remainder = policyList;
    PolicyList external;
    pfm -> filterPolicyList(remainder, external);

    //
    // Are the requested policies valid?
    //
    POAPolicies policies(orbInstance_, remainder);
    ULong idx;
    if(!validatePolicies(policies, policyList, idx))
	throw POA::InvalidPolicy((CORBA::UShort)idx);

    POA_impl* childImpl = 0;
    POA_var child;

    //
    // It's not possible to create a POA that already exists. If the
    // POA is being destroyed an AdapterAlreadyExists exception is
    // still thrown -- the application has to be capable of dealing
    // with this (by calling create_POA again)
    //
    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*children_);
#endif
	if(children_ -> containsKey(adapter))
	    throw POA::AdapterAlreadyExists();

	//
	// If necessary create a new POAManager for this POA
	//
	POAManager_var obmanager;
	if(is_nil(manager))
	{
	    InitialServiceManager_var ism =
		orbInstance_ -> getInitialServiceManager();
	    POAManagerFactory_var factory =
		OBPortableServer::POAManagerFactory::_narrow(
		    Object_var(ism -> resolveInitialReferences(
			"POAManagerFactory")));
	    obmanager = factory -> create_poa_manager("");
	}
	else
	{
	    obmanager = POAManager::_narrow(manager);
	    assert(!is_nil(obmanager));
	}
	
	//
	// Create the new POA object
	//
	childImpl = new POA_impl(orbInstance_, adapter, this, obmanager,
				 poaCurrent_, ociCurrent_, policies, external);
	child = childImpl;

	//
	// Add this to the child list
	//
	children_ -> put(adapter, POA::_duplicate(child));
    }

    //
    // Initialize the IMR, if necessary
    //
    try
    {
        childImpl -> registerWithIMR();
    }
    catch(...)
    {
        child -> destroy(true, false);
        throw;
    }
    
    return child._retn();
}

PortableServer::POA_ptr
OBPortableServer::POA_impl::find_POA(const char* adapter, Boolean activate)
    throw(PortableServer::POA::AdapterNonExistent,
          SystemException)
{
    assert_nca(adapter != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    bool check = true;

    //
    // If the child isn't registered, and we've been asked to activate
    // the POA and we have an AdapterActivator and the POAManager's
    // state is active then we activate the POA (see below
    // requirement)
    //
    // From 11-18:
    //
    // In addition, when a POA manager is in the holding state, the
    // adapter activators registered with the associated POAs will not
    // get called.
    //
    if(!children_ -> containsKey(adapter) && activate)
    {
	PortableServer::AdapterActivator_var adapterActivator =
	    adapterActivator_ -> getAdapterActivator();
	if(!is_nil(adapterActivator))
	{
	    //
	    // From the spec:
	    //
	    // Creating a POA using a POA manager that is in the
	    // active state can lead to race conditions if the POA
	    // supports preexisting objects, because the new POA may
	    // receive a request before its adapter activator, servant
	    // manager, or default servant have been
	    // initialized. These problems do not occur if the POA is
	    // created by an adapter activator registered with a
	    // parent of the new POA, because requests are queued
	    // until the adapter activator returns.  To avoid these
	    // problems when a POA must be explicitly initialized, the
	    // application can initialize the POA by invoking find_POA
	    // with a TRUE activate parameter.
	    //
	    // TODO: This requirement is not met
	    //
	    check = adapterActivator -> unknown_adapter(this, adapter);
	}
    }

    //
    // Lookup the POA in our set of children.
    //
    PortableServer::POA_var poa;
    if(check)
	children_ -> get(adapter, poa);

    if(is_nil(poa))
        throw POA::AdapterNonExistent();

    return poa._retn();
}

void
OBPortableServer::POA_impl::destroy(
    Boolean etherealize,
    Boolean waitForCompletion)
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // If waitForCompletion is TRUE and the current thread is in an
    // invocation context dispatched from some POA belonging to the
    // same ORB as this POA, the BAD_INV_ORDER exception is raised and
    // POA destruction does not occur.
    //
    if(waitForCompletion && poaCurrent_ -> _OB_inUpcall())
	throw BAD_INV_ORDER();

    //
    // Mark the POA as destroy pending. Only one thread will continue.
    // All others will wait (if waitForCompletion is true) until the
    // POA destruction is complete.
    //
    if(poaControl_ -> markDestroyPending(etherealize, waitForCompletion))
    {
	//
	// Recursively destroy all children
	//
	POAHashTable::Enumerator keys = children_ -> keys();
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    PortableServer::POA_var child;
	    if(children_ -> get(keys[i], child))
		child -> destroy(etherealize, waitForCompletion);
	}
	
	//
	// If waitForCompletion is FALSE, the destroy operation
	// destroys the POA and its children but waits neither for
	// active requests to complete nor for etherealization to
	// occur. If destroy is called multiple times before
	// destruction is complete (because there are active
	// requests), the etherealize_objects parameter applies only
	// to the first call of destroy. Subsequent calls with
	// conflicting etherealize_objects settings use the value of
	// etherealize_objects from the first call. The
	// waitForCompletion parameter is handled as defined above for
	// each individual call (some callers may choose to block,
	// while others may not).
	//
	// We only return if there are outstanding requests. In this
	// case when the last outstanding request is completed then
	// destruction of the the POA will complete.
	//
	if(!waitForCompletion && poaControl_ -> hasPendingRequests())
	    return;
	
	completeDestroy();
    }
}

//
// Factory methods for policy objects don't bother to check if the POA
// has already been destroyed
//

PortableServer::ThreadPolicy_ptr
OBPortableServer::POA_impl::create_thread_policy(
    PortableServer::ThreadPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ThreadPolicy_impl(value);
}

PortableServer::LifespanPolicy_ptr
OBPortableServer::POA_impl::create_lifespan_policy(
    PortableServer::LifespanPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::LifespanPolicy_impl(value);
}

PortableServer::IdUniquenessPolicy_ptr
OBPortableServer::POA_impl::create_id_uniqueness_policy(
    PortableServer::IdUniquenessPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::IdUniquenessPolicy_impl(value);
}

PortableServer::IdAssignmentPolicy_ptr
OBPortableServer::POA_impl::create_id_assignment_policy(
    PortableServer::IdAssignmentPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::IdAssignmentPolicy_impl(value);
}

PortableServer::ImplicitActivationPolicy_ptr
OBPortableServer::POA_impl::create_implicit_activation_policy(
    PortableServer::ImplicitActivationPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ImplicitActivationPolicy_impl(value);
}

PortableServer::ServantRetentionPolicy_ptr
OBPortableServer::POA_impl::create_servant_retention_policy(
    PortableServer::ServantRetentionPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ServantRetentionPolicy_impl(value);
}

PortableServer::RequestProcessingPolicy_ptr
OBPortableServer::POA_impl::create_request_processing_policy(
    PortableServer::RequestProcessingPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::RequestProcessingPolicy_impl(value);
}

OBPortableServer::SynchronizationPolicy_ptr
OBPortableServer::POA_impl::create_synchronization_policy(
    SynchronizationPolicyValue value)
    throw(SystemException)
{
    return new SynchronizationPolicy_impl(value);
}

OBPortableServer::DispatchStrategyPolicy_ptr
OBPortableServer::POA_impl::create_dispatch_strategy_policy(
    DispatchStrategy_ptr value)
    throw(SystemException)
{
    return new DispatchStrategyPolicy_impl(value);
}

char*
OBPortableServer::POA_impl::the_name()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return string_dup(name_);
}

PortableServer::POA_ptr
OBPortableServer::POA_impl::the_parent()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return POA::_duplicate(parent_);
}

PortableServer::POAList*
OBPortableServer::POA_impl::the_children()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    POAHashTable::Enumerator keys = children_ -> keys();

    //
    // Since its possible that the children list changes while this
    // method call is in progress its necessary to check the return
    // value of Hashtable::get.
    //
    PortableServer::POAList_var children =
	 new PortableServer::POAList(keys.length());
    for(ULong i = 0; i < keys.length(); ++i)
    {
        PortableServer::POA_var child;
        if(children_ -> get(keys[i], child)) {
	    CORBA::ULong childrenLength  = children -> length();
	    children -> length(childrenLength + 1);
	    children[childrenLength] = child;
        }
    }

    return children._retn();
}

PortableServer::POAManager_ptr
OBPortableServer::POA_impl::the_POAManager()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return PortableServer::POAManager::_duplicate(manager_);
}

PortableServer::AdapterActivator_ptr
OBPortableServer::POA_impl::the_activator()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return adapterActivator_ -> getAdapterActivator();
}

void
OBPortableServer::POA_impl::the_activator(
    PortableServer::AdapterActivator_ptr activator)
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    adapterActivator_ -> setAdapterActivator(activator);
}

CORBA::PolicyList*
OBPortableServer::POA_impl::the_policies()
    throw(CORBA::SystemException)
{
    return policies_ -> recreate();
}

OB::DispatchStrategy_ptr
OBPortableServer::POA_impl::the_dispatch_strategy()
    throw(CORBA::SystemException)
{
    return policies_ -> dispatchStrategyPolicy();
}

PortableServer::ServantManager_ptr
OBPortableServer::POA_impl::get_servant_manager()
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    ServantManagerStrategy* servantManagerStrategy =
	servantLocationStrategy_ -> getServantManagerStrategy();

    if(servantManagerStrategy == 0)
	throw POA::WrongPolicy();

    return servantManagerStrategy -> getServantManager();
}

void
OBPortableServer::POA_impl::set_servant_manager(
    PortableServer::ServantManager_ptr mgr)
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(mgr != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    ServantManagerStrategy* servantManagerStrategy =
	servantLocationStrategy_ -> getServantManagerStrategy();

    if(servantManagerStrategy == 0)
	throw POA::WrongPolicy();

    servantManagerStrategy -> setServantManager(mgr);
}

PortableServer::Servant
OBPortableServer::POA_impl::get_servant()
    throw(PortableServer::POA::NoServant,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    DefaultServantHolder* defaultServantHolder =
	servantLocationStrategy_ -> getDefaultServantHolder();

    if(defaultServantHolder == 0)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	 defaultServantHolder -> getDefaultServant();

    if(!servant)
	throw POA::NoServant();
    return servant._retn();
}

void
OBPortableServer::POA_impl::set_servant(PortableServer::Servant servant)
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    DefaultServantHolder* defaultServantHolder =
	servantLocationStrategy_ -> getDefaultServantHolder();
    if(defaultServantHolder == 0)
        throw POA::WrongPolicy();

    defaultServantHolder -> setDefaultServant(servant);
}

PortableServer::ObjectId*
OBPortableServer::POA_impl::activate_object(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantAlreadyActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    PortableServer::ObjectId_var oid = idGenerationStrategy_ -> createId();
    
    try
    {
	servantLocationStrategy_ -> activate(oid, servant);
    }
    catch(const POA::ObjectAlreadyActive&)
    {
	assert(false); // Should not occur
    }

    return oid._retn();
}

void
OBPortableServer::POA_impl::activate_object_with_id(
    const PortableServer::ObjectId& oid,
    PortableServer::Servant servant)
    throw(PortableServer::POA::ServantAlreadyActive,
          PortableServer::POA::ObjectAlreadyActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Validate that the ObjectId is valid
    //
    if(!idGenerationStrategy_ -> isValid(oid))
	throw BAD_PARAM();

    servantLocationStrategy_ -> activate(oid, servant);
}

void
OBPortableServer::POA_impl::deactivate_object(
    const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    servantLocationStrategy_ -> deactivate(this, oid);
}

Object_ptr
OBPortableServer::POA_impl::create_reference(const char* intf)
    throw(PortableServer::POA::WrongPolicy,
	  SystemException)
{
    assert_nca(intf != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    PortableServer::ObjectId_var oid = idGenerationStrategy_ -> createId();
    return _OB_createReference(oid, intf);
}

Object_ptr
OBPortableServer::POA_impl::create_reference_with_id(
    const PortableServer::ObjectId& oid,
    const char* intf)
    throw(PortableServer::POA::WrongPolicy,
	  SystemException)
{
    assert_nca(intf != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Validate that the ObjectId is valid
    //
    if(!idGenerationStrategy_ -> isValid(oid))
	throw BAD_PARAM();

    return _OB_createReference(oid, intf);
}

PortableServer::ObjectId*
OBPortableServer::POA_impl::servant_to_id(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires USE_DEFAULT_SERVANT policy or RETAIN policy and either
    // the UNIQUE_ID policy or the IMPLICIT_ACTIVATION (w/ SYSTEM_ID)
    // policies.
    //
    if(policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT &&
       (policies_ -> servantRetentionPolicy() != PortableServer::RETAIN ||
	(policies_ -> idUniquenessPolicy() != PortableServer::UNIQUE_ID &&
	 (policies_ -> implicitActivationPolicy() !=
	  PortableServer::IMPLICIT_ACTIVATION ||
	  policies_ -> idAssignmentPolicy() != PortableServer::SYSTEM_ID))))
        throw POA::WrongPolicy();

    PortableServer::ObjectId* oid =
	servantLocationStrategy_ -> servantToId(servant, poaCurrent_);

    if(oid == 0)
    {
	//
	// If the POA doesn't have the IMPLICIT_ACTIVATION with
	// SYSTEM_ID policy or the POA doesn't have the MULTIPLE_ID
	// policy then a ServantNotActive exception
	//
	if(policies_ -> implicitActivationPolicy() !=
	   PortableServer::IMPLICIT_ACTIVATION &&
	   policies_ -> idUniquenessPolicy() != PortableServer::MULTIPLE_ID )
	    throw POA::ServantNotActive();

	try
	{
	    oid = activate_object(servant);
	}
	catch(const POA::ServantAlreadyActive&)
	{
	    assert(false); // Should not occur
	}
    }

    return oid;
}

Object_ptr
OBPortableServer::POA_impl::servant_to_reference(
    PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // If the operation was invoked in the context of executing
    // request on the specified servant, the reference associated with
    // the current invocation is returned.
    //
    if(poaCurrent_ -> _OB_inUpcall() &&
       poaCurrent_ -> _OB_getServant() == servant)
    {
        try
        {
            PortableServer::ObjectId_var oid = poaCurrent_ -> get_object_id();
            String_var intf = servant -> _primary_interface(oid, this);
            return _OB_createReference(oid, intf);
        }
        catch(const PortableServer::Current::NoContext&)
        {
            assert(false);
        }
    }
    
    PortableServer::ObjectId_var oid = servant_to_id(servant);
    String_var intf = servant -> _primary_interface(oid, this);
    return create_reference_with_id(oid, intf);
}

PortableServer::Servant
OBPortableServer::POA_impl::reference_to_servant(Object_ptr reference)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongAdapter,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy.
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN &&
       policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT)
        throw POA::WrongPolicy();

    PortableServer::ObjectId_var oid = reference_to_id(reference);

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, true);

    if(!servant)
	throw POA::ObjectNotActive();

    return servant._retn();
}

//
// TODO: This is not nice and should be fixed.
//
static bool
IIOPExtractProfileBody(const IOP::IOR& ior, ::IIOP::ProfileBody_1_0& body)
{
    //
    // Extract the IIOP profile information from the provided IOR
    //
    ULong profile;
    for(profile = 0 ; profile < ior.profiles.length() ; profile++)
	if(ior.profiles[profile].tag == IOP::TAG_INTERNET_IOP)
	    break;
	
    if(profile >= ior.profiles.length())
	return false;
	
    const Octet* coct =	ior.profiles[profile].profile_data.get_buffer();
    ULong len = ior.profiles[profile].profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

    return true;
}

PortableServer::ObjectId*
OBPortableServer::POA_impl::reference_to_id(Object_ptr reference)
    throw(PortableServer::POA::WrongAdapter,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    RefCountIOR_var ior = reference -> _OB_IOR();

    //
    // If this POA has an IMR then the object key must be extracted
    // from the object id field of the object key contained in the
    // reference
    //
    if(hasIMR_)
    {
	//
	// First verify that the reference actually refers to the IMR
	//
	::IIOP::ProfileBody_1_0 iorBody;
	if(IIOPExtractProfileBody(ior -> value, iorBody))
	{
	    ::IIOP::ProfileBody_1_0 imrBody;
	    bool b = IIOPExtractProfileBody(imrIOR_, imrBody);
	    assert(b);

	    //
	    // The host, port and object key must be the same
	    //
	    if(strcmp(imrBody.host, iorBody.host) == 0 &&
	       imrBody.port == iorBody.port)
	    {
		OCI::ObjectKey key;
		key.length(iorBody.object_key.length());
		memcpy(key.get_buffer(), iorBody.object_key.get_buffer(),
		       key.length());

		//
		// Extract the object key data from the IMR object key
		//
		ObjectKeyData keyData;
		if(ParseObjectKey(key, keyData))
		{
		    //
		    // The object id is the actual object key, extract
		    // information from that
		    //
		    key.length(keyData.oid.length());
		    memcpy(key.get_buffer(), keyData.oid.get_buffer(),
			   key.length());

		    //
		    // Verify the poaIds are the same, have the
		    // persistency values and the same create time if
		    // transient
		    //
		    if(ParseObjectKey(key, keyData) && _OB_poaMatches(keyData))
			return new PortableServer::ObjectId(keyData.oid);
		}
	    }
	}
    }
    else
    {
	//
	// Extract the object key from the IOR of the object reference.
	//
	OCI::AcceptorSeq_var acceptors = manager_ -> get_acceptors();    
        bool local = false;
        ULong i;

	OCI::ProfileInfoSeq_var profileInfos;
        for(i = 0 ; i < acceptors -> length() ; i++)
        {
            profileInfos = acceptors[i] -> get_local_profiles(ior -> value);
            if(profileInfos -> length() > 0)
                local = true;
        }

	//
	// Is the key local?
	//
	if(local)
	{
	    //
	    // Verify the poaIds are the same, have the persistency
	    // values and the same create time if transient
	    //
	    ObjectKeyData keyData;
	    if(ParseObjectKey(profileInfos[0].key, keyData) &&
	       _OB_poaMatches(keyData))
            {
		return new PortableServer::ObjectId(keyData.oid);
            }
	}
    }

    throw POA::WrongAdapter();
    return 0; // Some compilers need this
}

PortableServer::Servant
OBPortableServer::POA_impl::id_to_servant(const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN &&
       policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, true);

    if(!servant)
        throw POA::ObjectNotActive();

    return servant._retn();
}

Object_ptr
OBPortableServer::POA_impl::id_to_reference(
    const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy.
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, false);
    if(!servant)
        throw POA::ObjectNotActive();

    String_var intf = servant -> _primary_interface(oid, this);
    return _OB_createReference(oid, intf);
}

void
OBPortableServer::POA_impl::_OB_preinvoke(
    const char* op,
    const PortableServer::ObjectId& oid,
    PortableServer::ServantBase* servant,
    PortableServer::ServantLocator::Cookie cookie,
    OCI::TransportInfo_ptr info)
{
    //
    // Depending on the synchronization policy we have to lock a
    // variety of mutexes. This has to be done before the preinvoke:
    //
    //  No serialization of invocations of preinvoke or postinvoke may
    //  be assumed; there may be multiple concurrent invocations of
    //  preinvoke for the same ObjectId. (However, if the
    //  SINGLE_THREAD_MODEL policy is being used, that policy will
    //  serialize these calls.)
    //

#ifdef HAVE_JTC
    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.lock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbInstance_ -> getORBSyncMutex().lock();
	break;
    }
#endif

    //
    // preinvoke the servant location strategy
    //
    servantLocationStrategy_ -> preinvoke(oid);

    //
    // Setup the OCI::Current context
    //
    ociCurrent_ -> _OB_preinvoke(info);

    //
    // Setup the PortableServer::Current context
    //
    poaCurrent_ -> _OB_preinvoke(this, servant, op, oid, cookie);
}

void
OBPortableServer::POA_impl::_OB_postinvoke()
{
    const PortableServer::ObjectId& oid = poaCurrent_ -> _OB_getObjectId();
    const char* op = poaCurrent_ -> _OB_getOp();
    PortableServer::ServantBase* servant = poaCurrent_ -> _OB_getServant();
    PortableServer::ServantLocator::Cookie cookie =
	 poaCurrent_ -> _OB_getCookie();

    //
    // Clean up the PortableServant::Current
    //
    poaCurrent_ -> _OB_postinvoke();

    //
    // Clean up the OCI:Current
    //
    ociCurrent_ -> _OB_postinvoke();

    //
    // postinvoke the servant location strategy
    //
    servantLocationStrategy_ -> postinvoke(oid, this, op, cookie, servant);

    //
    // Depending on the synchronization policy we have to unlock a
    // variety of mutexes
    //
#ifdef HAVE_JTC
    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.unlock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbInstance_ -> getORBSyncMutex().unlock();
	break;
    }
#endif
}

bool
OBPortableServer::POA_impl::_OB_locateServant(
    const PortableServer::ObjectId& oid)
    throw(SystemException, LocationForward)
{
    PortableServer::ServantLocator::Cookie cookie;
    const char op[] = "_locate";
    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> locate(oid, this, op, cookie);

    if(servant)
    {
	servantLocationStrategy_ -> preinvoke(oid);
	servantLocationStrategy_ -> postinvoke(oid, this, op, cookie,
					       servant);
	return true;
    }

    return false;
}

//
// Create the upcall object for a method invocation
//
Upcall_ptr
OBPortableServer::POA_impl::_OB_createUpcall(
    const PortableServer::ObjectId& oid,
    UpcallReturn_ptr upcallReturn,
    OCI::ProfileInfo* profileInfo,
    OCI::TransportInfo_ptr transportInfo,
    CORBA::ULong requestId,
    char* op,
    InputStream_ptr in,
    IOP::ServiceContextList* requestSCL)

    throw(SystemException, LocationForward)
{
    Upcall_var upcall;

    //
    // Increment the outstanding request count
    //
    if(poaControl_ -> incrementRequestCount())
    {
	try
	{
	    PIUpcall_var piUpcall;

	    //
	    // Create the upcall object
	    //
	    if(policies_ -> interceptorCallPolicy())
	    {
		piUpcall = new PIUpcall(orbInstance_, upcallReturn,
					profileInfo, transportInfo,
					requestId, op, in, requestSCL);
		upcall = Upcall::_duplicate(piUpcall);
		
		//
		// Call the receive_request_service_contexts
		// interception point
		//
		OctetSeq oidSeq(oid.length(), oid.length(),
				(Octet*)oid.get_buffer(),
				false);
		piUpcall -> receiveRequestServiceContexts(externalPolicies_,
							  adapterId_, oidSeq);
		piUpcall -> contextSwitch();
	    }
	    else
	    {
		upcall = new Upcall(orbInstance_, upcallReturn,
				    profileInfo, transportInfo,
				    requestId, op, in, requestSCL);
	    }

	    DispatchRequest_impl* dispatchRequestImpl =
		new DispatchRequest_impl(this, oid, upcall);

	    upcall -> setDispatchInfo(dispatchRequestImpl, dispatchStrategy_);
	}
	catch(const SystemException& ex)
	{
	    upcall -> setSystemException(ex);
	    _OB_decrementRequestCount();
	}
	catch(const LocationForward& ex)
	{
	    upcall -> setLocationForward(ex.ior, ex.perm);
	    _OB_decrementRequestCount();
	}
    }

    return upcall._retn();
}

void
OBPortableServer::POA_impl::_OB_dispatch(
    const PortableServer::ObjectId& oid,
    Upcall_ptr upcall)
    throw()
{
    const char* op = upcall -> operation();

    //
    // Notify the PIUpcall about a potential change in thread
    // context. This has to be done before the servant is located
    // since PICurrent information has to be available in servant
    // locators.
    //
    PIUpcall_ptr piUpcall = dynamic_cast<PIUpcall_ptr>(upcall);
    if(!is_nil(piUpcall))
	piUpcall -> contextSwitch();

    try
    {
        //
        // Locating a servant can throw LocationForward and
        // SystemException errors
        //
        PortableServer::ServantLocator::Cookie cookie;
        PortableServer::ServantBase* servant =
            servantLocationStrategy_ -> locate(oid, this, op, cookie);
    
        //
        // If there is a servant then dispatch the request
        //
        if(servant != 0)
        {
            OCI::TransportInfo_ptr transportInfo =
                upcall -> transportInfo_nodup();

            //
            // Do any pre invocation stuff
            //
            _OB_preinvoke(op, oid, servant, cookie, transportInfo);

            //
            // The Upcall may call _OB_postinvoke
            //
            upcall -> setServantAndPOA(servant, this);

            try
            {
                //
                // Dispatch the request
                //
                servant -> _OB_dispatch(upcall);
            }
            catch(...)
            {
                //
                // If postinvoke() hasn't been called, then do it now,
                // since it may raise an exception that must be returned
                // to the client
                //
                if(!upcall -> postinvokeCalled())
                {
                    servant -> _remove_ref(); // Must be called before
                                              // _OB_postinvoke
                    _OB_postinvoke(); // May raise SystemException
                }

                throw;
            }

            //
            // If postinvoke() hasn't been called, then do it now,
            // since it may raise an exception that must be returned
            // to the client
            //
            if(!upcall -> postinvokeCalled())
            {
                servant -> _remove_ref(); // Must be called before
                                          // _OB_postinvoke
                _OB_postinvoke(); // May raise SystemException
            }
        }
        else if(strcmp(op, "_non_existent") == 0 ||
                strcmp(op, "_not_existent") == 0)
        {
            upcall -> preUnmarshal();
            upcall -> postUnmarshal();
            OutputStream_ptr out = upcall -> preMarshal();
            out -> write_boolean(true);
            upcall -> postMarshal();
        }
        else
        {
            upcall -> setSystemException(OBJECT_NOT_EXIST());
        }
    }
    catch(const LocationForward& ex)
    {
	upcall -> setLocationForward(ex.ior, ex.perm);
    }
    catch(const UserException& ex)
    {
	upcall -> setUserException(ex);
    }
    catch(const SystemException& ex)
    {
	upcall -> setSystemException(ex);
    }
    catch(...)
    {
        handleUnknownException(upcall);
    }

    //
    // Decrement the outstanding request count
    //
    _OB_decrementRequestCount();
}

Object_ptr
OBPortableServer::POA_impl::_OB_createReference(
    const PortableServer::ObjectId& oid,
    const char* intf)
{
    //
    // Create the object key for this object id
    //
    OCI::ObjectKey_var key = createObjectKey(oid);

    RefCountIOR_var ior;

    if(hasIMR_)
    {
	//
	// Create an IOR that refers to the IMR. The generated object
	// key is placed in the object-id field of the IMR object key.
        //
	imrKey_.oid.length(key -> length());
        memcpy(imrKey_.oid.get_buffer(), key -> get_buffer(), key -> length());
	OCI::ObjectKey_var key = OB::CreateObjectKey(imrKey_);
	IOP::IOR_var imrIOR = IIOPCreateIOR(imrIOR_, intf, key);
	ior = new RefCountIOR(imrIOR);	
    }
    else
    {
	//
	// Create local IOR
	//
	ior = createIOR(key, intf);
    }

    //
    // Create and return reference
    //
    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    return objectFactory -> createObject(ior);
}

DirectStubImpl_ptr
OBPortableServer::POA_impl::_OB_createDirectStubImpl(
    const PortableServer::ObjectId& oid,
    RefCountPolicyList_ptr policies)
    throw(SystemException, LocationForward)
{
    return servantLocationStrategy_ ->
        createDirectStubImpl(this, oid, policies);
}

PortableServer::Current_impl*   
OBPortableServer::POA_impl::_OB_POACurrent()
{
    return poaCurrent_;
}

void
OBPortableServer::POA_impl::_OB_removeDirectStubImpl(
    const PortableServer::ObjectId& oid,
    DirectStubImpl_ptr directStubImpl)
{
    servantLocationStrategy_ -> removeDirectStubImpl(oid, directStubImpl);
}

bool
OBPortableServer::POA_impl::_OB_poaMatches(const ObjectKeyData& data)
{
    //
    // If the length is incorrect, return false
    //
    if(data.poaID.length() != poaId_.length())
	return false;

    //
    // Check each name
    //
    for(ULong i = 0; i < data.poaID.length(); ++i)
	if(strcmp(data.poaID[i], poaId_[i]) != 0)
	    return false;

    //
    // If the persistent flags are different then false
    //
    if(policies_ -> lifespanPolicy() ==
       PortableServer::PERSISTENT != data.persistent)
        return false;
    
    //
    // If persistent then true
    //
    if(policies_ -> lifespanPolicy() == PortableServer::PERSISTENT)
    	return true;

    //
    // Otherwise, the time has to be the same
    //
    return data.createTime == poaCreateTime_;
}

bool
OBPortableServer::POA_impl::_OB_incrementRequestCount()
{
    return poaControl_ -> incrementRequestCount();
}

void
OBPortableServer::POA_impl::_OB_decrementRequestCount()
{
    if(poaControl_ -> decrementRequestCount())
	completeDestroy();
}

void
OBPortableServer::POA_impl::_OB_waitPendingRequests()
{
    poaControl_ -> waitPendingRequests();
}

ORBInstance_ptr
OBPortableServer::POA_impl::_OB_ORBInstance()
{
    return ORBInstance::_duplicate(orbInstance_);
}

//
// Add the policy factory for the POA Policies
//
void
OBPortableServer::POA_impl::_OB_addPolicyFactory()
{
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();

    PortableInterceptor::PolicyFactory_var factory =
        new PortableInterceptor::POAPolicyFactory_impl;

    pfm -> registerPolicyFactory(
        PortableServer::THREAD_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::LIFESPAN_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::ID_UNIQUENESS_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::ID_ASSIGNMENT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::IMPLICIT_ACTIVATION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::SERVANT_RETENTION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::REQUEST_PROCESSING_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        SYNCHRONIZATION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        DISPATCH_STRATEGY_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
	INTERCEPTOR_CALL_POLICY_ID, factory, true);
}

void
OBPortableServer::POA_impl::_OB_validateManagerState()
    throw(SystemException)
{
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    m -> _OB_validateState();
}

//
// Create an IOR that refers to this POA using the object-id and
// upcall information
//
RefCountIOR_ptr
OBPortableServer::POA_impl::_OB_createIOR(const PortableServer::ObjectId& oid)
{
    OCI::ObjectKey_var key = createObjectKey(oid);
    return createIOR(key);
}

void
OBPortableServer::POA_impl::_OB_etherealize()
{
    servantLocationStrategy_ -> etherealize(this);
}

void
OBPortableServer::POA_impl::_OB_isPersistent(IMR::POANameSeq& persistentPOAs)
{
    if(policies_ -> lifespanPolicy() == PortableServer::PERSISTENT)
    {
        ULong length = persistentPOAs.length();
	persistentPOAs.length(length + 1);
	persistentPOAs[length] = poaId_;
    }
}
