// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_POA_IMPL_H
#define OB_POA_IMPL_H

#include <OB/Upcall_fwd.h>
#include <OB/DirectStubImpl_fwd.h>

#include <OB/Upcall_fwd.h>

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/OBPOAInterface.h>
#include <OB/POAServantLocator.h>
#include <OB/IMRTypes.h>
#include <OB/ObjectKey.h>
#include <OB/Hashtable.h>

namespace OCI
{

class Current_impl;

} // End of namespace OCI

namespace OB
{

//
// The component table
//
typedef OB::Hashtable< CORBA::ULong, IOP::TaggedComponentSeq*,
    OB::ULongHasher, OB::LockStrategyNull > ComponentTable;

}

namespace PortableServer
{

class Current_impl; // Implementation of PortableServer::Current

} // end of namespace PortableServer

namespace OBPortableServer
{

//
// Forward declarations
//
class POAControl;
class AdapterActivatorHolder;
class ServantLocationStrategy;
class IdGenerationStrategy;
class POAPolicies;

//
// The POA_impl class
//
class POA_impl : public POA
{
    //
    // The ORBInstance object
    //
    OB::ORBInstance_var orbInstance_;

    //
    // The name of this POA and the name of the root POA
    //
    CORBA::String_var name_;
    CORBA::String_var rootName_;

    //
    // The adapter id
    //
    CORBA::OctetSeq_var adapterId_;

    //
    // The POA name. This is the full path to the POA.
    //
    IMR::POAName poaId_;

    //
    // My parent POA
    //
    POA_impl* parent_;

    //
    // My POA manager
    //
    OBPortableServer::POAManager_var manager_;

    //
    // My servant location strategy
    //
    ServantLocationStrategy* servantLocationStrategy_;

    //
    // My POAControl object
    //
    POAControl* poaControl_;

    //
    // The associated AdapterActivator
    //
    AdapterActivatorHolder* adapterActivator_;
    
    //
    // All children of this POA
    //
    typedef OB::Hashtable<CORBA::String_var, PortableServer::POA_var,
			  OB::StringHasher, 
	OB::OB_LOCK_STRATEGY_SYNCHRONIZED> POAHashTable;
    POAHashTable* children_;

    //
    // My policies
    //
    POAPolicies* policies_;
    
    //
    // Used for generating object ids
    //
    IdGenerationStrategy* idGenerationStrategy_;
    
    //
    // The POA creation time, if not persistent
    //
    time_t poaCreateTime_;

#ifdef HAVE_JTC
    //
    // If this POA has the single thread policy set then this mutex is
    // used ensure this policy is met
    //
    JTCRecursiveMutex poaSyncMutex_;
    static JTCRecursiveMutex orbSyncMutex_;
#endif

    OB::DispatchStrategy_var dispatchStrategy_;

    //                                                    
    // The Current implementation. The RootPOA is responsible for the
    // creation and deletion of the one and only PortableServer::Current.
    //                                               
    PortableServer::Current_impl* poaCurrent_;

    //
    // The OCI::Current implementation. The RootPOA is responsible for
    // the creation and deletion of the one and only OCI::Current.
    //
    OCI::Current_impl* ociCurrent_;

    //
    // This is the set of policies that are not POA policies and are
    // registered with a valid PolicyFactory
    //
    OB::RefCountPolicyList_var externalPolicies_;

    //
    // The set of components to add to each IOR created by this POA.
    // The PortableInterceptors has the ability to add components to
    // specific profiles (hence the hashtable) and the ability to add
    // components to ALL profiles (hence the sequence).
    //
    IOP::TaggedComponentSeq allComponents_;
    OB::ComponentTable* componentTable_;

    //
    // IMR Data
    //

    //
    // Does this POA have an IMR?
    //
    bool hasIMR_;
    
    //
    // The IMR's IOR
    //
    IOP::IOR_var imrIOR_;
    
    //
    // The IMR ObjectKeyData
    //
    OB::ObjectKeyData imrKey_;

    //
    // Given an object id create an object key
    //
    OCI::ObjectKey* createObjectKey(const PortableServer::ObjectId&);

    //
    // Find the index of the given policy
    //
    bool findPolicyIndex(const CORBA::PolicyList&, CORBA::PolicyType,
			 CORBA::ULong&);

    //
    // Validate the set of provided POA policies
    //
    bool validatePolicies(const POAPolicies&, const CORBA::PolicyList&,
			  CORBA::ULong&);

    //
    // Handle unknown exceptions
    //
    void handleUnknownException(OB::Upcall_ptr);

    //
    // Complete the destroy of the POA
    //
    void completeDestroy();

    //
    // Register with the IMR, if necessary
    //
    void registerWithIMR();

    //
    // Create IOR
    //
    OB::RefCountIOR_ptr createIOR(const OCI::ObjectKey&, const char* = "");
    
    //
    // Common initialization code
    //
    void init()
	throw();

    //
    // Private, internal constructor, for non-root POAs
    //
    POA_impl(OB::ORBInstance_ptr,
	     const char*,
	     POA_impl*,
	     POAManager_ptr,
             PortableServer::Current_impl*,
             OCI::Current_impl*,
	     const POAPolicies&,
             const CORBA::PolicyList&)
	throw();

public:

    //
    // Public constructor for the root POA
    //
    POA_impl(OB::ORBInstance_ptr, const char*, POAManager_ptr)
	throw();
    ~POA_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PortableServer::POA_ptr
    create_POA(const char* adapter_name,
	       PortableServer::POAManager_ptr a_POAManager,
               const CORBA::PolicyList& policies)
        throw(PortableServer::POA::AdapterAlreadyExists,
	      PortableServer::POA::InvalidPolicy,
              CORBA::SystemException);

    virtual PortableServer::POA_ptr
    find_POA(const char* adapter_name, CORBA::Boolean activate_it)
        throw(PortableServer::POA::AdapterNonExistent,
	      CORBA::SystemException);

    virtual void
    destroy(CORBA::Boolean etherealize_objects,
	    CORBA::Boolean wait_for_completion)
        throw(CORBA::SystemException);

    virtual PortableServer::ThreadPolicy_ptr
    create_thread_policy(PortableServer::ThreadPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::LifespanPolicy_ptr
    create_lifespan_policy(PortableServer::LifespanPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::IdUniquenessPolicy_ptr
    create_id_uniqueness_policy(PortableServer::IdUniquenessPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::IdAssignmentPolicy_ptr
    create_id_assignment_policy(PortableServer::IdAssignmentPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::ImplicitActivationPolicy_ptr
    create_implicit_activation_policy(
	PortableServer::ImplicitActivationPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::ServantRetentionPolicy_ptr
    create_servant_retention_policy(
	PortableServer::ServantRetentionPolicyValue value)
        throw(CORBA::SystemException);

    virtual PortableServer::RequestProcessingPolicy_ptr
    create_request_processing_policy(
	PortableServer::RequestProcessingPolicyValue value)
        throw(CORBA::SystemException);

    virtual SynchronizationPolicy_ptr
    create_synchronization_policy(
	SynchronizationPolicyValue value)
        throw(CORBA::SystemException);

    virtual DispatchStrategyPolicy_ptr
    create_dispatch_strategy_policy(OB::DispatchStrategy_ptr value)
        throw(CORBA::SystemException);

    virtual char* the_name()
        throw(CORBA::SystemException);

    virtual PortableServer::POA_ptr the_parent()
        throw(CORBA::SystemException);

    virtual PortableServer::POAList* the_children()
        throw(CORBA::SystemException);
    
    virtual PortableServer::POAManager_ptr the_POAManager()
        throw(CORBA::SystemException);

    virtual PortableServer::AdapterActivator_ptr the_activator()
        throw(CORBA::SystemException);

    virtual void the_activator(PortableServer::AdapterActivator_ptr)
        throw(CORBA::SystemException);

    virtual CORBA::PolicyList* the_policies()
        throw(CORBA::SystemException);

    virtual OB::DispatchStrategy_ptr the_dispatch_strategy()
        throw(CORBA::SystemException);

    virtual PortableServer::ServantManager_ptr get_servant_manager()
        throw(PortableServer::POA::WrongPolicy, CORBA::SystemException);

    virtual void set_servant_manager(PortableServer::ServantManager_ptr imgr)
        throw(PortableServer::POA::WrongPolicy, CORBA::SystemException);

    virtual PortableServer::Servant get_servant()
        throw(PortableServer::POA::NoServant,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual void set_servant(PortableServer::Servant p_servant)
        throw(PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual PortableServer::ObjectId*
    activate_object(PortableServer::Servant p_servant)
        throw(PortableServer::POA::ServantAlreadyActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual void activate_object_with_id(const PortableServer::ObjectId& id,
					 PortableServer::Servant p_servant)
        throw(PortableServer::POA::ServantAlreadyActive,
	      PortableServer::POA::ObjectAlreadyActive,
              PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual void deactivate_object(const PortableServer::ObjectId& oid)
        throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual CORBA::Object_ptr create_reference(const char* intf)
        throw(PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual CORBA::Object_ptr
    create_reference_with_id(const PortableServer::ObjectId& oid,
			     const char* intf)
        throw(PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual PortableServer::ObjectId*
    servant_to_id(PortableServer::Servant p_servant)
        throw(PortableServer::POA::ServantNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual CORBA::Object_ptr
    servant_to_reference(PortableServer::Servant p_servant)
        throw(PortableServer::POA::ServantNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual PortableServer::Servant
    reference_to_servant(CORBA::Object_ptr reference)
        throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongAdapter,
	      PortableServer::POA::WrongPolicy,
              CORBA::SystemException);

    virtual PortableServer::ObjectId*
    reference_to_id(CORBA::Object_ptr reference)
        throw(PortableServer::POA::WrongAdapter,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual PortableServer::Servant
    id_to_servant(const PortableServer::ObjectId& oid)
        throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    virtual CORBA::Object_ptr
    id_to_reference(const PortableServer::ObjectId& oid)
        throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    
    //
    // Called before a method invocation on a servant
    //
    void _OB_preinvoke(const char*, const PortableServer::ObjectId&,
		       PortableServer::ServantBase*,
		       PortableServer::ServantLocator::Cookie,
                       OCI::TransportInfo_ptr);

    //
    // Called after a method invocation on a servant
    //
    void _OB_postinvoke();

    //
    // Determine if a servant is present
    //
    bool _OB_locateServant(const PortableServer::ObjectId&)
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Create an Upcall object. Return a _nil Upcall if the POA is in
    // the process of being destroyed.
    //
    // Note: For efficiency reasons, the Upcall constructor assumes
    // ownership for all parameter marked with "/**/"
    //
    OB::Upcall_ptr _OB_createUpcall(
	const PortableServer::ObjectId&,
	OB::UpcallReturn_ptr,
	/**/ OCI::ProfileInfo*,
	/**/ OCI::TransportInfo_ptr,
	CORBA::ULong,
	/**/ char*,
	/**/ OB::InputStream_ptr,
	/**/ IOP::ServiceContextList*)
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Dispatch a method invocation
    //
    void _OB_dispatch(const PortableServer::ObjectId&,
		      OB::Upcall_ptr)
	throw();

    //
    // Create an object reference given an object key and interface name
    //
    CORBA::Object_ptr _OB_createReference(const PortableServer::ObjectId&,
					  const char*);
        
    //
    // Ask a servant to create the StubImplBase, if possible.
    //
    OB::DirectStubImpl_ptr _OB_createDirectStubImpl(
	const PortableServer::ObjectId&,
	OB::RefCountPolicyList_ptr)
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Get the POA current implementation object
    //
    PortableServer::Current_impl* _OB_POACurrent();

    //
    // Remove a DirectStubImpl
    //
    void _OB_removeDirectStubImpl(const PortableServer::ObjectId&,
				  OB::DirectStubImpl*);

    //
    // Determine if this POA matches the provided object key
    //
    bool _OB_poaMatches(const OB::ObjectKeyData&);

    //
    // Increment the outstanding number of requests
    //
    bool _OB_incrementRequestCount();

    //
    // Decrement the outstanding number of requests
    //
    void _OB_decrementRequestCount();

    //
    // Wait for all pending requests to terminate
    //
    void _OB_waitPendingRequests();

    //
    // Get the ORBInstance object
    //
    OB::ORBInstance_ptr _OB_ORBInstance();

    void _OB_addPolicyFactory();

    void _OB_validateManagerState()
	throw(CORBA::SystemException);

    OB::RefCountIOR_ptr _OB_createIOR(const PortableServer::ObjectId&);

    void _OB_etherealize();

    void _OB_isPersistent(IMR::POANameSeq& persistentPOAs);
};

} // End of namespace OBPortableServer

#endif
