// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>

#include <TypeCodeFactory.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// TypeCodeFactory public member implementation
// ----------------------------------------------------------------------

TypeCode_ptr
TypeCodeFactory::createPrimitiveTC(TCKind kind)
{
    assert(kind == tk_null ||
	   kind == tk_void ||
	   kind == tk_short ||
	   kind == tk_long ||
	   kind == tk_longlong ||
	   kind == tk_ushort ||
	   kind == tk_ulong ||
	   kind == tk_ulonglong ||
	   kind == tk_float ||
	   kind == tk_double ||
	   kind == tk_longdouble ||
	   kind == tk_boolean ||
	   kind == tk_char ||
	   kind == tk_wchar ||
	   kind == tk_octet ||
	   kind == tk_any ||
	   kind == tk_TypeCode ||
	   kind == tk_Principal);

    TypeCode_var tc = new TypeCode;
    tc -> kind_ = kind;
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createStructTC(const char* id,
				const char* name,
				const StructMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_struct;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createUnionTC(const char* id,
			       const char* name,
			       TypeCode_ptr discriminator_type,
			       const UnionMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_union;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> discriminatorType_ = TypeCode::_duplicate(discriminator_type);

#ifndef NDEBUG
    TypeCode_var tc1 = GetOrigType(discriminator_type);
    TCKind k = tc1 -> kind();

    assert_nca(k == tk_short || k == tk_ushort ||
                k == tk_long || k == tk_ulong ||
                k == tk_longlong || k == tk_ulonglong ||
                k == tk_char || k == tk_boolean || k == tk_wchar ||
                k == tk_enum, NCADiscType);
#endif

    tc -> labels_.length(members.length());
    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
#ifndef NDEBUG
        TypeCode_var tc2 = members[i].label.type();
         if(tc2 -> kind() != tk_octet)
             assert_nca(tc1 -> equivalent(tc2), NCADiscType);
#endif

        tc -> labels_[i] = members[i].label;
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createEnumTC(const char* id,
			      const char* name,
			      const EnumMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_enum;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i];
    }

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createAliasTC(const char* id,
			       const char* name,
			       TypeCode_ptr original_type)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_alias;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> contentType_ = TypeCode::_duplicate(original_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createExceptionTC(const char* id,
				   const char* name,
				   const StructMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_except;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_objref;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createStringTC(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_string;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createWstringTC(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_wstring;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createFixedTC(UShort digits, Short scale)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_fixed;
    tc -> fixedDigits_ = digits;
    tc -> fixedScale_ = scale;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createSequenceTC(ULong bound, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_sequence;
    tc -> length_ = bound;
    tc -> contentType_ = TypeCode::_duplicate(element_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createRecursiveSequenceTC(ULong /*bound*/, ULong /*offset*/)
{
    assert_nca(false, NCADeprecated);
    return TypeCode::_nil();
}

TypeCode_ptr
TypeCodeFactory::createArrayTC(ULong length, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_array;
    tc -> length_ = length;
    tc -> contentType_ = TypeCode::_duplicate(element_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createValueTC(const char* id,
                                 const char* name,
                                 ValueModifier type_modifier,
                                 TypeCode_ptr concrete_base,
                                 const ValueMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_value;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> typeModifier_ = type_modifier;
    tc -> concreteBaseType_ = TypeCode::_duplicate(concrete_base);

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());
    tc -> memberVisibility_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
        tc -> memberVisibility_[i] = members[i].access;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createValueBoxTC(const char* id,
				  const char* name,
				  TypeCode_ptr boxed_type)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_value_box;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> contentType_ = TypeCode::_duplicate(boxed_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createNativeTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_native;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createRecursiveTC(const char* id)
{
    assert_nca(id != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> recId_ = id;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createAbstractInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_abstract_interface;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}
