// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IIOP.h>
#include <OB/CONV_FRAME.h>
#include <../library/CodeSetDatabase.h>

#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;

static void
PrintOctets(OB_STD(ostream&) out, const Octet* oct, ULong count)
{
    static const ULong inc = 8;
    
    ULong j;

    for(j = 0 ; j < count ; j += inc)
    {
	ULong k;

	for(k = j ; k - j < inc ; k++)
	{
	    if(k < count)
	    {
		out.width(3);
		out << (int)oct[k] << ' ';
	    }
	    else
		out << "    ";
	}

	out << '"';
	
	for(k = j ; k < count && k - j < inc ; k++)
	{
	    if(isprint((char)oct[k]))
		out << (char)oct[k];
	    else
		out << '.';
	}
	
	out << '"' << OB_ENDL;
    }
}

static void
PrintTagCodeSets(OB_STD(ostream&) out, const IOP::TaggedComponent& component)
{
    const Octet* coct = component.component_data.get_buffer();
    ULong len = component.component_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    CONV_FRAME::CodeSetComponentInfo info;
    CONV_FRAME::CodeSetComponentInfo::_OB_unmarshal(info, &in);

    CodeSetDatabase* db = CodeSetDatabase::instance();
    ULong i;

    //
    // Print char codeset information
    //
    const CodeSetDatabase::CodeSetInfo* charInfo;
    out << "Native char codeset: " << OB_ENDL;
    charInfo = db -> getCodeSetInfo(info.ForCharData.native_code_set);
    if(charInfo)
	out << "  \"" << charInfo -> description << "\"" << OB_ENDL;
    else
	out << "  [No native char codeset information]" << OB_ENDL;

    out << "Char conversion codesets:" << OB_ENDL;
    if(info.ForCharData.conversion_code_sets.length())
    {
	for(i = 0 ; i < info.ForCharData.conversion_code_sets.length() ; i++)
	{
	    charInfo =
		db -> getCodeSetInfo(info.ForCharData.conversion_code_sets[i]);
	    if(charInfo)
		out << "  \"" << charInfo -> description << "\"" << OB_ENDL;
	    else
		out << "  [Unknown codeset id: "
		    << info.ForCharData.conversion_code_sets[i]
		    << "]" << OB_ENDL;
	}
    }
    else
	out << "  [No char conversion codeset information]" << OB_ENDL;

    //
    // Print wchar codeset information
    //
    const CodeSetDatabase::CodeSetInfo* wCharInfo;
    out << "Native wchar codeset: " << OB_ENDL;
    wCharInfo = db -> getCodeSetInfo(info.ForWcharData.native_code_set);
    if(wCharInfo)
	out << "  \"" << wCharInfo -> description << "\"" << OB_ENDL;
    else
	out << "  [No native wchar codeset information]" << OB_ENDL;

    out << "Wchar conversion codesets:" << OB_ENDL;
    if(info.ForWcharData.conversion_code_sets.length())
    {
	for(i = 0 ; i < info.ForWcharData.conversion_code_sets.length() ; i++)
	{
	    wCharInfo =
		db -> getCodeSetInfo(info.ForWcharData.conversion_code_sets[i]);
	    if(wCharInfo)
		out << "  \"" << wCharInfo -> description << "\"" << OB_ENDL;
	    else
		out << "  [Unknown codeset id: "
		    << info.ForWcharData.conversion_code_sets[i]
		    << "]" << OB_ENDL;
	}
    }
    else
	out << "  [No wchar conversion codeset information]" << OB_ENDL;
}

static void
PrintIIOPProfile(OB_STD(ostream&) out, const IOP::TaggedProfile& profile)
{
    //
    // Get the IIOP profile body
    //
    const Octet* coct = profile.profile_data.get_buffer();
    ULong len = profile.profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_0 body;
    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

    //
    // Print general info
    //
    out << "iiop_version: " << int(body.iiop_version.major) << '.'
	<< int(body.iiop_version.minor) << '\n';
    out << "host: " << body.host << '\n';
    out << "port: " << body.port << '\n';
    out << "object_key: (" << body.object_key.length() << ")\n";
    PrintOctets(out, body.object_key.get_buffer(), body.object_key.length());

    //
    // Print IIOP 1.1 information
    //
    if(body.iiop_version.major > 1 || body.iiop_version.minor >= 1)
    {
	ULong l = in.read_ulong();

        ULong countAlternate = 1;
	for(ULong i = 0 ; i < l ; i++)
	{
	    IOP::TaggedComponent component;
	    IOP::TaggedComponent::_OB_unmarshal(component, &in);

            if(component.tag == IOP::TAG_ALTERNATE_IIOP_ADDRESS)
            {
                const Octet* coct = component.component_data.get_buffer();
                ULong len = component.component_data.length();
                OCI::BufferImpl buf((Octet*)coct, len);
                InputStreamImpl in(&buf, 0, false);
                in._OB_readEndian();
                String_var host = in.read_string();
                UShort port = in.read_ushort();
                out << "Alternate IIOP address #" << countAlternate << ':'
                    << OB_ENDL;
                out << "  host: " << host << OB_ENDL;
                out << "  port: " << port << OB_ENDL;
                countAlternate++;
            }
            else if(component.tag == IOP::TAG_CODE_SETS)
                PrintTagCodeSets(out, component);
	}
    }
}

static void
PrintObjref(OB_STD(ostream&) out, const IOP::IOR& ior)
{
    ULong countIOP = 1;
    ULong countMulti = 1;

    out << "type_id: " << ior.type_id << '\n';

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    out << "IIOP profile #" << countIOP++ << ":\n";
	    PrintIIOPProfile(out, ior.profiles[i]);
	}
	else if(ior.profiles[i].tag == IOP::TAG_MULTIPLE_COMPONENTS)
	{
	    out << "Multiple component profile #" << countMulti++ << ": ";

	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
	    ULong len = ior.profiles[i].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
	    in._OB_readEndian();

	    ULong cnt = in.read_ulong();
	    if(!cnt)
		out << OB_ENDL;
	    else
	    {
		for(ULong j = 0 ; j < cnt ; j++)
		{
		    IOP::TaggedComponent comp;
		    IOP::TaggedComponent::_OB_unmarshal(comp, &in);

		    if(comp.tag == IOP::TAG_ORB_TYPE)
		    {
			out << "IOP::TAG_ORB_TYPE" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_CODE_SETS)
		    {
			out << "IOP::TAG_CODE_SETS" << OB_ENDL;
			PrintTagCodeSets(out, comp);
		    }
		    else if(comp.tag == IOP::TAG_POLICIES)
		    {
			out << "IOP::TAG_POLICIES" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_ALTERNATE_IIOP_ADDRESS)
		    {
			out << "IOP::TAG_ALTERNATE_IIOP_ADDRESS"
			    << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_ASSOCIATION_OPTIONS)
		    {
			out << "IOP::TAG_ASSOCIATION_OPTION" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_SEC_NAME)
		    {
			out << "IOP::TAG_SEC_NAM" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_SPKM_1_SEC_MECH)
		    {
			out << "IOP::TAG_SPKM_1_SEC_MECH" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_SPKM_2_SEC_MECH)
		    {
			out << "IOP::TAG_SPKM_2_SEC_MECH" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_KerberosV5_SEC_MECH)
		    {
			out << "IOP::TAG_KerberosV5_SEC_MECH" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_CSI_ECMA_Secret_SEC_MECH)
		    {
			out << "IOP::TAG_CSI_ECMA_Secret_SEC_MECH"
			    << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_CSI_ECMA_Hybrid_SEC_MECH)
		    {
			out << "IOP::TAG_CSI_ECMA_Hybrid_SEC_MECH"
			    << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_SSL_SEC_TRANS)
		    {
			out << "IOP::TAG_SSL_SEC_TRANS" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_CSI_ECMA_Public_SEC_MECH)
		    {
			out << "IOP::TAG_CSI_ECMA_Public_SEC_MECH"
			    << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_GENERIC_SEC_MECH)
		    {
			out << "IOP::TAG_GENERIC_SEC_MECH" << OB_ENDL;
		    }
		    else if(comp.tag == IOP::TAG_JAVA_CODEBASE)
		    {
			out << "IOP::TAG_JAVA_CODEBASE" << OB_ENDL;
		    }
		    else
		    {
			out << "Unknown tag: " << comp.tag << OB_ENDL;
		    }
		}
	    }
	} 
        else 
        {
            out << "Unknown profile tag: " << ior.profiles[i].tag
		<< OB_ENDL;
        }
    }

    out << OB_FLUSH;
}

static void
PrintObjref(OB_STD(ostream&) out, const char* s, bool hasEndian)
{
    assert_nca(s != 0, NCANullString);
    
    if(strncmp(s, "IOR:", 4) != 0)
	throw INV_OBJREF();
    
    s += 4;
    
    Octet* poct = AsciiToOctets(s);
    OCI::BufferImpl buf(poct, strlen(s) / 2);
    InputStreamImpl in(&buf, 0, false);

    Boolean endian = in.read_boolean();
    in._OB_swap(endian != Endian);

    IOP::IOR ior;
    IOP::IOR::_OB_unmarshal(ior, &in);
    
    out << "byteorder: ";

    if(hasEndian)
	out << (endian ? "little" : "big") << " endian\n";
    else
	out << "n/a\n";

    PrintObjref(out, ior);
}

static void
DumpIOR(const char* ior, bool hasEndian)
{
    try
    {
	PrintObjref(OB_COUT, ior, hasEndian);
    }
    catch(const INV_OBJREF&)
    {
	OB_COUT << "IOR is invalid" << OB_ENDL;
    }

    OB_COUT << OB_FLUSH;
}

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] [-f FILE ... | IOR ...]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
        "-f                     Read IORs from files instead from the\n"
	"                       command line.\n"
	;
}

static bool
startsWith(const char* str, const char* prefix)
{
    const char* s = str;
    const char* p = prefix;

    while(*s != '\0' && *p != '\0')
    {
        if(tolower(*s) != tolower(*p))
            return false;
        s++;
        p++;
    }

    return true;
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];
	
    //
    // Get options
    //
    bool files = false;
    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 ||
           strcmp(argv[i], "-h") == 0)
        {
	    usage(progName);
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
        {
	    OB_CERR << "ORBacus " << OB::Version << OB_ENDL;
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
	    files = true;
        }
        else
        {
	    OB_CERR << progName << ": unknown option `" << argv[i]
			 << "'\n" << OB_ENDL;
	    usage(progName);
	    return EXIT_FAILURE;
        }
    }
	
    if(i == argc)
    {
        OB_CERR << progName << ": no IORs\n" << OB_ENDL;
        usage(progName);
        return EXIT_FAILURE;
    }

    if(!files)
    {
        if(!startsWith(argv[i], "IOR:") &&
           !startsWith(argv[i], "corbaloc:") &&
           !startsWith(argv[i], "corbaname:") &&
           !startsWith(argv[i], "file:") &&
           !startsWith(argv[i], "relfile:"))
	{
	    OB_CERR << "[No valid IOR in the command line, assuming -f]"
		    << OB_ENDL;

	    files = true;
	}
    }

    if(!files)
    {
        //
        // Dump all IORs given as arguments
        //
        int count = 0;
        for(; i < argc ; i++)
        {
	    if(count > 0)
	        OB_COUT << '\n';
	    OB_COUT << "IOR #" << ++count << ':' << OB_ENDL;

	    try
	    {
                //
                // The byte order can only be preserved for IOR: URLs
                //
                if(startsWith(argv[i], "IOR:"))
                    DumpIOR(argv[i], true);
                else
                {
                    //
                    // Let string_to_object do the dirty work
                    //
                    Object_var obj = orb -> string_to_object(argv[i]);
                    String_var s = orb -> object_to_string(obj);
                    DumpIOR(s, false);
                }
	    }
	    catch(const BAD_PARAM&)
	    {
		OB_CERR << "IOR is invalid" << OB_ENDL;
	    }
        }
    }
    else
    {
        //
        // Dump all IORs from the specified files
        //
        int count = 0;
        for(; i < argc ; i++)
        {
	    OB_IFSTREAM in;
	    // Must use open(name), not ifstream in(name) (VC++ bug)
	    in.open(argv[i]);
	    if(in.fail())
	    {
		    OB_CERR << progName << ": can't open `" << argv[i]
				 << "': " << strerror(errno) << OB_ENDL;
		    return EXIT_FAILURE;
	    }
		
	    String_var str = string_dup("");
		
	    while(!in.eof())
	    {
		char c;
		in.read(&c, 1);
		if(in.eof() || c == '\n' || c == '\r' || c == '\0')
		{
		    if(strlen(str) > 0)
		    {
			if(count > 0)
			    OB_COUT << '\n';
			OB_COUT << "IOR #" << ++count << ':' << OB_ENDL;

                        //
                        // The byte order can only be preserved for IOR: URLs
                        //
                        if(startsWith(str, "IOR:"))
                            DumpIOR(str, true);
                        else
                        {
                            //
                            // Let string_to_object do the dirty work
                            //
                            Object_var obj = orb -> string_to_object(str);
                            String_var s = orb -> object_to_string(obj);
                            DumpIOR(s, false);
                        }

			str = string_dup("");
	    	    }
	        }
	        else
	    	    str += c;
	    }
        }
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
