// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestMacro.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestUtil.h>

using namespace CORBA;
using namespace PortableServer;

#ifdef HAVE_JTC

class Test_impl2 : virtual public POA_Test, public JTCMonitor
{
    POA_var poa_;
    bool called_;
    bool finished_;

public:
    
    Test_impl2(POA_ptr poa)
        : poa_(POA::_duplicate(poa)),
          called_(false),
          finished_(false)
    {
    }
    
    virtual void aMethod()
	throw(CORBA::SystemException)
    {
        JTCSynchronized sync(*this);
        
        called_ = true;
        notify();

        try
        {
            wait(1000);
        }
        catch(const JTCInterruptedException&)
        {
        }
        finished_ = true;
    }

    virtual POA_ptr _default_POA()
    {
        return POA::_duplicate(poa_);
    }

    void blockUntilCalled()
    {
        JTCSynchronized sync(*this);
        while(!called_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }
    
    bool callComplete()
    {
        JTCSynchronized sync(*this);
        return finished_;
    }
};

class LongCaller : public JTCThread
{
    Test_var t_;

public:

    LongCaller(Test_ptr t)
        : t_(Test::_duplicate(t))
    {
    }

    virtual void run()
    {
        try
        {
            t_ -> aMethod();
        }
        catch(SystemException& se)
        {
            OB_CERR << se << OB_ENDL;
        }
    }
};

//
// In this test we want to spawn a thread to call a method on the Test
// interface. This method call should take some time. While the thread
// is calling the method we attempt to deactivate the object. This
// should not complete for some time, since it should wait for all outstanding
// method calls to complete.
//
void
TestDeactivateThreaded(ORB_ptr orb, POA_ptr root)
{
    Test_impl2* impl = new Test_impl2(root);

    Test_var t = impl -> _this();

    ObjectId_var oid = root -> servant_to_id(impl);

    JTCThreadHandle thr = new LongCaller(t);

    thr -> start();
    impl -> blockUntilCalled();

    //
    // Test: deactivate_object while method call is active
    //
    root -> deactivate_object(oid);

    //
    // Once we've deactivated the object the re-activation shouldn't
    // complete until the method call completes
    //
    root -> activate_object_with_id(oid, impl);

    //
    // The destroy call shouldn't return until the aMethod call is
    // complete
    //
    TEST(impl -> callComplete());

    //
    // Wait for the thread to terminate
    //
    while(thr -> isAlive())
    {
	try
	{
	    thr -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    root -> deactivate_object(oid);

    delete impl;
}
#endif

void
TestDeactivate(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl = new Test_impl(orb, "", false);

    Test_var t = impl -> _this();

    t -> aMethod();

    ObjectId_var oid = root -> servant_to_id(impl);

    root -> deactivate_object(oid);

    try
    {
        t -> aMethod();
        TEST(false); // expected OBJECT_NOT_EXIST
    }
    catch(const OBJECT_NOT_EXIST&)
    {
        // Expected
    }
    delete impl;
}

void
runtests(ORB_ptr orb, POA_ptr root)
{
    TestDeactivate(orb, root);
#ifdef HAVE_JTC
    TestDeactivateThreaded(orb, root);
#endif
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;

    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        //
        // Activate the RootPOA manager
        //
        POAManager_var rootMgr = root -> the_POAManager();
        TEST(!is_nil(rootMgr));
        
        rootMgr -> activate();

        OB_COUT << "Testing deactivate... " << OB_FLUSH;

        //
        // Run the tests using the root POA
        //
        runtests(orb, root);

        OB_COUT << "OK!" << OB_ENDL;
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
