// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestServer_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

//
// Implementation to test same thread dispatch strategy
//

class TestSameThread_impl : public POA_Test
{
#ifdef HAVE_JTC
    JTCThreadId tid_;
#endif
    bool failed_;
    bool first_;

public:

    TestSameThread_impl()
        : failed_(false),
	  first_(true)
    {
    }

    virtual void aMethod()
        throw(CORBA::SystemException)
    {
#ifdef HAVE_JTC
    	//
	// Test to ensure that all requests handled by the same thread
	//
    	if(first_)
	{
	    tid_ = JTCThreadId::self();
	    first_ = false;
	}
	else if(tid_ != JTCThreadId::self())
	{
	    failed_ = true;
	}
	JTCThread::sleep (100);
#endif
    }

    bool failed()
    {
        return failed_;
    }
};

#ifdef HAVE_JTC
//
// Implementation to test thread per request dispatch strategy
//

class TestThreadPerReq_impl : public POA_Test
{
    JTCMutex mut_;
    JTCThreadId tids_[10];
    bool failed_;

public:

    TestThreadPerReq_impl()
    	: failed_(false)
    {
    }

    virtual void aMethod()
        throw(CORBA::SystemException)
    {
    	unsigned int idx;
	JTCThreadId nullId;
	{
	    JTCSynchronized sync(mut_);

	    //
	    // Test to ensure that each request is being handled
	    // by a different thread.
	    //
	    for (idx = 0; idx < sizeof(tids_); ++idx)
	    {
	        if(tids_[idx] == nullId)
	    	    break;

	        if(tids_[idx] == JTCThreadId::self())
	            failed_ = true;
	    }

	    if(idx < sizeof(tids_))
	        tids_[idx] = JTCThreadId::self();
	}
	JTCThread::sleep (100);
    }

    bool failed()
    {
        return failed_;
    }
};

//
// Implementation to test thread pool dispatch strategy
//

class TestThreadPool_impl : public POA_Test
{
    JTCMutex mut_;
    JTCThreadId tids_[2];
    bool failed_;

public:

    TestThreadPool_impl()
    	: failed_(false)
    {
    }

    virtual void aMethod()
        throw(CORBA::SystemException)
    {
	JTCThreadId nullId;
	{
	    JTCSynchronized sync(mut_);

	    //
	    // Test to ensure that all requests are handled only by 
	    // the two threads in the thread pool.
	    //
	    if(tids_[0] == nullId)
	    {
	    	tids_[0] = JTCThreadId::self();
	    }
	    else if(tids_[0] != JTCThreadId::self())
	    {
	    	if(tids_[1] == nullId)
		    tids_[1] = JTCThreadId::self();
		else if(tids_[1] != JTCThreadId::self())
		    failed_ = true;
	    }
	}
	JTCThread::sleep (100);
    }

    bool failed()
    {
        return failed_;
    }
};
#endif

//
// Simple custom dispatch strategy (uses same thread)
//
class CustomDispatchStrategy :  public OB::DispatchStrategy
{
public:

    virtual OB::DispatchStrategyId
    id()
    {
	return 4; // Some unique number
    }

    virtual CORBA::Any*
    info()
    {
	return new CORBA::Any();
    }

    virtual void
    dispatch(OB::DispatchRequest_ptr request)
    {
        request -> invoke();
    }
};

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Set the communications concurrency model
	//
	OB::Properties_var props = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, props);
#ifdef HAVE_JTC
	props -> setProperty("ooc.orb.conc_model", "threaded");
	props -> setProperty("ooc.orb.oa.conc_model", "threaded");
#else
	props -> setProperty("ooc.orb.conc_model", "blocking");
	props -> setProperty("ooc.orb.oa.conc_model", "reactive");
#endif

	//
	// Create ORB
	//
	orb = CORBA::ORB_init(argc, argv);

        //
        // Resolve Root POA and POA Manager
        //
        CORBA::Object_var poaObj =
	    orb -> resolve_initial_references("RootPOA");
        OBPortableServer::POA_var rootPOA = 
	    OBPortableServer::POA::_narrow(poaObj);

        PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

	//
	// Resolve Dispatch Strategy Factory
	//
        CORBA::Object_var dsfObj =
	    orb -> resolve_initial_references("DispatchStrategyFactory");
	OB::DispatchStrategyFactory_var dsf = 
	    OB::DispatchStrategyFactory::_narrow(dsfObj);

	//
	// Create Dispatch Strategy objects
	//
	OB::DispatchStrategy_var ds_st =
	    dsf -> create_same_thread_strategy();

	int failed_creates = 0;

	OB::DispatchStrategy_var ds_tpr;
	try
	{
	    ds_tpr = dsf -> create_thread_per_request_strategy();
	}
	catch(CORBA::NO_IMPLEMENT&)
	{
	    ++failed_creates;
	}

	OB::DispatchStrategy_var ds_tp;
        OB::ThreadPoolId tpid;
	try
	{
	    tpid = dsf -> create_thread_pool(2);
	    ds_tp = dsf -> create_thread_pool_strategy(tpid);
	}
	catch(CORBA::NO_IMPLEMENT&)
	{
	    ++failed_creates;
	}

	OB::DispatchStrategy_var ds_cus = new CustomDispatchStrategy;

#ifndef HAVE_JTC
	//
	// If not using JThreads/C++ creation of thread per request
	// and thread pool strategies should fail.
	//
	if (failed_creates != 2)
	{
	    return EXIT_FAILURE;
	}
#endif

	//
	// Create POAs with threaded Dispatch Strategy Policies
	//     - same thread
	//     - thread per request
	//     - thread pool (same pool used for two POAs)
	//     - custom strategy
	//
	CORBA::PolicyList policies(1);
	policies.length(1);

	policies[0] = rootPOA -> create_dispatch_strategy_policy(ds_st);
	PortableServer::POA_var stPOA = 
	    rootPOA -> create_POA("stPOA", manager, policies);

#ifdef HAVE_JTC
	policies[0] = rootPOA -> create_dispatch_strategy_policy(ds_tpr);
	PortableServer::POA_var tprPOA = 
	    rootPOA -> create_POA("tprPOA", manager, policies);

	policies[0] = rootPOA -> create_dispatch_strategy_policy(ds_tp);
	PortableServer::POA_var tpPOA1 = 
	    rootPOA -> create_POA("tpPOA1", manager, policies);
	PortableServer::POA_var tpPOA2 = 
	    rootPOA -> create_POA("tpPOA2", manager, policies);
#endif

	policies[0] = rootPOA -> create_dispatch_strategy_policy(ds_cus);
	PortableServer::POA_var cusPOA = 
	    rootPOA -> create_POA("cusPOA", manager, policies);

	//
	// Create test implementation object in each POA
	//
	TestSameThread_impl* stTest = new TestSameThread_impl;
	PortableServer::ObjectId_var stObjId =
	    stPOA -> activate_object(stTest);
	CORBA::Object_var stObjRef = stPOA -> id_to_reference(stObjId);

#ifdef HAVE_JTC
	TestThreadPerReq_impl* tprTest = new TestThreadPerReq_impl;
	PortableServer::ObjectId_var tprObjId =
	    tprPOA -> activate_object(tprTest);
	CORBA::Object_var tprObjRef = tprPOA -> id_to_reference(tprObjId);

	TestThreadPool_impl* tpTest = new TestThreadPool_impl;
	PortableServer::ObjectId_var tpObjId1 =
	    tpPOA1 -> activate_object(tpTest);
	CORBA::Object_var tpObjRef1 = tpPOA1 -> id_to_reference(tpObjId1);

	PortableServer::ObjectId_var tpObjId2 =
	    tpPOA2 -> activate_object(tpTest);
	CORBA::Object_var tpObjRef2 = tpPOA2 -> id_to_reference(tpObjId2);
#endif

	TestSameThread_impl* cusTest = new TestSameThread_impl;
	PortableServer::ObjectId_var cusObjId =
	    cusPOA -> activate_object(cusTest);
	CORBA::Object_var cusObjRef = cusPOA -> id_to_reference(cusObjId);

	//
	// Create Test Server 
	//
#ifdef HAVE_JTC
	TestInfoSeq info(5);
	info.length(5);
	info[0].obj = Test::_narrow(stObjRef);
	info[0].except_id = CORBA::string_dup("");
	info[1].obj = Test::_narrow(tprObjRef);
	info[1].except_id = CORBA::string_dup("");
	info[2].obj = Test::_narrow(tpObjRef1);
	info[2].except_id = CORBA::string_dup("");
	info[3].obj = Test::_narrow(tpObjRef2);
	info[3].except_id = CORBA::string_dup("");
	info[4].obj = Test::_narrow(cusObjRef);
	info[4].except_id = CORBA::string_dup("");
#else
	TestInfoSeq info(2);
	info.length(2);
	info[0].obj = Test::_narrow(stObjRef);
	info[0].except_id = CORBA::string_dup("");
	info[1].obj = Test::_narrow(cusObjRef);
	info[1].except_id = CORBA::string_dup("");
#endif
	TestServer_impl* serverImpl = new TestServer_impl(orb, info);
	CORBA::Object_var svrObj = serverImpl -> _this();
	PortableServer::ServantBase_var server = serverImpl;

	//
	// Save reference
	//
	CORBA::String_var s = orb -> object_to_string(svrObj);

	const char* refFile = "Test.ref";
	OB_OFSTREAM out(refFile);
	if(out.fail())
	{
	    OB_CERR << "can't open `" << refFile << "': "
	 	 << strerror(errno) << OB_ENDL;
	    return EXIT_FAILURE;
	}
	out << s << OB_ENDL;
	out.close();

	//
	// Run implementation
	//
	manager -> activate();
	orb -> run();

	//
	// Check for failure
	//
	if(stTest -> failed() || cusTest -> failed()
#ifdef HAVE_JTC
	   || tpTest -> failed() || tprTest -> failed()
#endif
	   )
	    return EXIT_FAILURE;

	delete stTest;
	delete cusTest;
#ifdef HAVE_JTC
	delete tprTest;
	delete tpTest;
#endif

#ifdef HAVE__UNLINK
        _unlink(refFile);
#else
        unlink(refFile);
#endif

    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
