// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************


#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestServer_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_JTC

//
// Implementation to test multiple orbs
//

class TestOrb_impl : public Test_impl
{
    int count_;
    CORBA::ORB_var orb_;

public:

    TestOrb_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa)
        : Test_impl(orb, poa),
	  count_(0),
	  orb_(CORBA::ORB::_duplicate(orb))
    {
    }

    virtual void aMethod()
        throw(CORBA::SystemException)
    {
        if (++count_ == 3)
	{
	    //
	    // Shutdown th orb after three method calls
	    //
	    orb_ -> shutdown(false);
	}
    }
};

class OrbTestThread : public JTCThread
{
    CORBA::ORB_var orb_;
    PortableServer::POAManager_var manager_;
    TestOrb_impl* test_;
    Test_var obj_;

public:

    OrbTestThread(const char* name)
    {
	int argcDummy = 0;
	char* argvDummy[1];
	argvDummy[0] = 0;
        orb_ = CORBA::ORB_init(argcDummy, argvDummy, name);
        
	CORBA::Object_var obj = orb_ -> resolve_initial_references("RootPOA");
	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj);
	manager_ = rootPOA -> the_POAManager();

	test_ = new TestOrb_impl(orb_, rootPOA);
	obj_ = test_ -> _this();
    }

    ~OrbTestThread()
    {
        delete test_;
	try
	{
	    orb_ -> destroy();
	}
	catch(const CORBA::Exception&)
	{
	    // Ignore
	}
    }

    Test_ptr getObjectReference()
    {
    	return Test::_duplicate(obj_);
    }

    virtual void run()
    {
        try
	{
	    manager_ -> activate();
	    orb_ -> run();
	}
	catch(CORBA::Exception&)
	{
	}
    }

};

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;
    try
    {
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	properties -> setProperty("ooc.orb.oa.conc_model", "threaded");
	properties -> setProperty("ooc.orb.conc_model", "threaded");

	//
	// Create ORB
	//
	int ac = 0;
	orb = CORBA::ORB_init(ac, 0);

        //
        // Resolve Root POA and POA Manager
        //
        CORBA::Object_var poaObj =
	    orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var rootPOA = 
	    PortableServer::POA::_narrow(poaObj);
        PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

	//
	// Create threads to run other orb instances
	//
	JTCHandleT<OrbTestThread> thread1 = new OrbTestThread("orb1");
	JTCHandleT<OrbTestThread> thread2 = new OrbTestThread("orb2");
	JTCHandleT<OrbTestThread> thread3 = new OrbTestThread("orb3");
	JTCHandleT<OrbTestThread> thread4 = new OrbTestThread("orb4");
	JTCHandleT<OrbTestThread> thread5 = new OrbTestThread("orb5");

	thread1 -> start();
	thread2 -> start();
	thread3 -> start();
	thread4 -> start();
	thread5 -> start();

	//
	// Create Test Server
	//
	TestInfoSeq info(5);
	info.length(5);
	info[0].obj = thread1 -> getObjectReference();
	info[0].except_id = CORBA::string_dup("");
	info[1].obj = thread2 -> getObjectReference();
	info[1].except_id = CORBA::string_dup("");
	info[2].obj = thread3 -> getObjectReference();
	info[2].except_id = CORBA::string_dup("");
	info[3].obj = thread4 -> getObjectReference();
	info[3].except_id = CORBA::string_dup("");
	info[4].obj = thread5 -> getObjectReference();
	info[4].except_id = CORBA::string_dup("");

	TestServer_impl* serverImpl = new TestServer_impl(orb, info);
	PortableServer::ObjectId_var svrId =
	    rootPOA -> activate_object(serverImpl);
	CORBA::Object_var svrObj = rootPOA -> id_to_reference(svrId);
	PortableServer::ServantBase_var servant = serverImpl;

	//
	// Save references
	//
	const char* refFile = "Test.ref";
	
	CORBA::String_var s = orb -> object_to_string(svrObj);

	OB_OFSTREAM out(refFile);
	if(out.fail())
	{
	    OB_CERR << "can't open `" << refFile << "': "
	 	 << strerror(errno) << OB_ENDL;
	    return EXIT_FAILURE;
	}
	out << s << OB_ENDL;
	out.close();

	//
	// Run implementation
	//
	manager -> activate();
	orb -> run();

	//
	// Wait for all other threads to be finished
	//
	do
	{
	    try
	    {
	        thread1 -> join();
	        thread2 -> join();
	        thread3 -> join();
	        thread4 -> join();
	        thread5 -> join();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
	while(thread1 -> isAlive() || thread2 -> isAlive() || 
	      thread3 -> isAlive() || thread4 -> isAlive() ||
	      thread5 -> isAlive());

#ifdef HAVE__UNLINK
        _unlink(refFile);
#else
        unlink(refFile);
#endif

    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}

#else
int main() {return 0;}	// Not applicable without JThreads/C++
#endif
